//------------------------------------------------------------------------
// Copyright 2004-2005, Avocent Corporation
//------------------------------------------------------------------------
// FILE        :  adsap2.c
//
// DESCRIPTION :  Functions to support ADSAP2: Avocent DS
//                Authentication 2 Protocol
//
//                Ported from CPS_DSV3 code
//------------------------------------------------------------------------
// AUTHOR      :  Jamie Collins
//
// DATE        :  April 2004
//
//
// Notes:
//------------------------------------------------------------------------
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/stat.h>

#include "avoprint.h"
#include "list-item.h"

#include <openssl/ssl.h>
#include <openssl/rsa.h>       /* SSLeay stuff */
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/err.h>

#include "snmputil.h"
#include "ds2mib.h"
#include "adsap2.h"
#include "ssl.h"
#include "threadwrap.h"
#include "netwrap.h"
#include "eid_util.h"

#define ADSAP_WARNING     (AVODBG_WARNING | AVODBG_ADSAP)
#define ADSAP_FAILURE     (AVODBG_FAILURE | AVODBG_ADSAP)
#define ADSAP_INFO        (AVODBG_INFO    | AVODBG_ADSAP)
#define ADSAP_ADHOC       (AVODBG_ADHOC   | AVODBG_ADSAP)

extern int		get_ssl_data				(SSL *p_ssl, unsigned char *buf, unsigned long num, unsigned long timeout);
extern void		get_port_range				(int port_num, int max, int *start, int *end);
extern void		close_socket				(SOCKET socket);

int						send_ADSAP2_message			(void *p_ssl, P_ADSAP2_HEADER p_Header);
int						receive_ADSAP2_message		(void *p_ssl, P_ADSAP2_HEADER p_Header, int buf_size);
void					build_ADSAP2_result			(P_ADSAP2_HEADER p_header, unsigned char field_id, unsigned long result);
int						get_ADSAP2_auth_reply		(P_ADSAP2_HEADER p_header, ADSAP2_AUTHORIZE_REP * authorizeReply);

//*****************************************************************************
#define ADSAP2_LISTEN_PORT 3871

#define EXTENSION_OID_LEN		  9
#define EXTENSION_OID_ITEMS		10

unsigned short extension_oid[EXTENSION_OID_ITEMS][EXTENSION_OID_LEN] = {
	{ 0x2b, 0x6, 0x1, 0x4, 0x1, 0x28B2, 0x6, 0x1, 0x81 },	// .129 Casc. Chan.
	{ 0x2b, 0x6, 0x1, 0x4, 0x1, 0x28B2, 0x6, 0x1, 0x80 },	// .128 RIP ID
	{ 0x2b, 0x6, 0x1, 0x4, 0x1, 0x28B2, 0x6, 0x1, 0x6 },	// .6 video, media
	{ 0x2b, 0x6, 0x1, 0x4, 0x1, 0x28B2, 0x6, 0x1, 0x5 },	// .5 Role
	{ 0x2b, 0x6, 0x1, 0x4, 0x1, 0x28B2, 0x6, 0x1, 0x4 },	// .4 Expiration
	{ 0x2b, 0x6, 0x1, 0x4, 0x1, 0x28B2, 0x6, 0x1, 0x3 },	// .3 Mode
	{ 0x2b, 0x6, 0x1, 0x4, 0x1, 0x28B2, 0x6, 0x1, 0x2 },	// .2 Preemption
	{ 0x2b, 0x6, 0x1, 0x4, 0x1, 0x28B2, 0x6, 0x1, 0x1 },	// .1 Password
	{ 0x2b, 0x6, 0x1, 0x4, 0x1, 0x28B2, 0x6, 0x1, 0x82 },	// .130 Log DLA
	{ 0x2b, 0x6, 0x1, 0x4, 0x1, 0x28B2, 0x6, 0x1, 0x83 },	// .131 command
};

// Indexes into the above table
#define OID_INDEX_CHANNEL    0
#define OID_INDEX_RIP_ID     1
#define OID_INDEX_TYPE       2
#define OID_INDEX_ROLE       3
#define OID_INDEX_EXPIRATION 4
#define OID_INDEX_MODE       5
#define OID_INDEX_PREEMPTION 6
#define OID_INDEX_PASSWORD   7
#define OID_INDEX_LOGDLA     8
#define OID_INDEX_COMMAND    9

void *g_ssl_ctx_cert_client	= NULL;
void *g_ssl_ctx_cert_server	= NULL;

PGLOBAL_LIST preauth_cert_list = NULL;

#define DS_AUTH_RETRIES		1
#define DS_AUTH_TIMEOUT		10

static char sessionModeTbl[ADSAP2_MODE_MAX][ADSAP2_USERNAME_STR_LEN];
static char preauthTypeTbl[ADSAP2_PREAUTH_TYPE_MAX][ADSAP2_PREAUTH_TYPE_STR_LEN];
static char userRoleTbl   [ADSAP2_USER_ROLE_MAX][ADSAP2_USER_ROLE_STR_LEN];

// Used to keep us from storing the pre-auth cert for unencrypted
// VM sessions.
#define ADSAP2_NOERROR_VM_UNENCRYPTED 0x0900
//*****************************************************************************

/*
** Global credentials, received from the DSView Server
*/
typedef struct credential_pointers
{
	unsigned long	len;
	unsigned char	*p_data;
} __attribute__ ((packed)) CREDENTIAL_POINTERS, *P_CREDENTIAL_POINTERS;

CREDENTIAL_POINTERS ap_cert;
CREDENTIAL_POINTERS ap_key;
CREDENTIAL_POINTERS ds_cert;

// These global certs are copied to rootfs.d as part of the build process
#define GLOBAL_DSVIEW_CERTIFICATE_FILE "/etc/CA/global_dsview_cert.txt"
#define GLOBAL_APPLIANCE_CERTIFICATE_FILE "/etc/CA/global_appliance_cert.txt"
#define GLOBAL_APPLIANCE_RSA_PRIVATEKEY_FILE "/etc/CA/global_appliance_key.txt"

// These files will contain the certs received from DSView Server
#define DSVIEW_CERTIFICATE_FILE "/etc/CA/adsap2/dsview_cert"
#define APPLIANCE_CERTIFICATE_FILE "/etc/CA/adsap2/appliance_cert"
#define APPLIANCE_RSA_PRIVATEKEY_FILE "/etc/CA/adsap2/appliance_key"

//
// Session Mode Negotation Table which is defined in the ADSAP2 spec
//
typedef enum
{
  TARGET_AVAILABLE = 0,
  TARGET_IN_USE,
  TARGET_CONTENTION,

  TARGET_MAX

} TARGET_STATE;

#define CHECK_PREEMPT_LEVEL  (int)-1
#define CHECK_SHAREABLE      (int)-2
#define CHECK_STEALTHABLE    (int)-3

static const int sessionModeStatusTbl[ADSAP2_MODE_MAX][TARGET_MAX] = {
//                Available         In Use               Contention
//              =============================================================
/*Normal*/      {ADSAP2_NOERROR, ADSAP2_RESOURCE_BUSY, ADSAP2_RESOURCE_BLOCKED},
/*Exclusive*/   {ADSAP2_NOERROR, ADSAP2_RESOURCE_BUSY, ADSAP2_RESOURCE_BLOCKED},
/*Preempt*/     {ADSAP2_ERROR,   CHECK_PREEMPT_LEVEL,  CHECK_PREEMPT_LEVEL},
/*Preempt-Excl*/{ADSAP2_ERROR,   CHECK_PREEMPT_LEVEL,  CHECK_PREEMPT_LEVEL},
/*Share*/       {ADSAP2_ERROR,   CHECK_SHAREABLE,      ADSAP2_ERROR},
/*ShareForced*/ {ADSAP2_ERROR,   CHECK_SHAREABLE,      ADSAP2_ERROR},
/*SharePassive*/{ADSAP2_ERROR,   CHECK_SHAREABLE,      ADSAP2_ERROR},
/*Scan*/        {ADSAP2_NOERROR, ADSAP2_RESOURCE_BUSY, ADSAP2_RESOURCE_BLOCKED},
/*Stealth*/     {ADSAP2_ERROR,   CHECK_STEALTHABLE,    ADSAP2_ERROR},
/*SharPsvForce*/{ADSAP2_ERROR,   CHECK_SHAREABLE,      ADSAP2_ERROR}

};

extern void adsap2_system(char *cmd);

//---------------------------------------------------------------------
// FUNCTION     : sessionMode2Enum
//
// DESCRIPTION  : Converts session mode string to enum
//
// PARAMETERS   :
//
// RETURNS      : session mode enum
//
// NOTES        :
//
//---------------------------------------------------------------------
byte sessionMode2Enum (char * modeStr)
{
  int i;
  byte mode = ADSAP2_NORMAL;

  for (i = 0; i < ADSAP2_MODE_MAX; i++)
  {
    if ( strcmp( sessionModeTbl[i], modeStr ) == 0 )
    {
      mode = i;
      break;
    }
  }
  return mode;
}

//---------------------------------------------------------------------
// FUNCTION     : preauthType2Enum
//
// DESCRIPTION  : Converts preauth type string to enum
//
// PARAMETERS   :
//
// RETURNS      : preauth type enum
//
// NOTES        :
//
//---------------------------------------------------------------------
byte preauthType2Enum (char * typeStr)
{
  int i;
  byte type = ADSAP2_PREAUTH_TYPE_VIDEO;

  for (i = 0; i < ADSAP2_PREAUTH_TYPE_MAX; i++)
  {
    if ( strcmp( preauthTypeTbl[i], typeStr ) == 0 )
    {
      type = i;
      break;
    }
  }
  //avoPrintf(3, "preauthType2Enum enum = %d\n", type);
  return type;
}

//---------------------------------------------------------------------
// FUNCTION     : userRole2Enum
//
// DESCRIPTION  : Converts user role string to enum
//
// PARAMETERS   :
//
// RETURNS      : user role enum
//
// NOTES        :
//
//---------------------------------------------------------------------
byte userRole2Enum (char * roleStr)
{
  int i;
  byte userRole = ADSAP2_USER_ROLE_NONE;

  for (i = 0; i < ADSAP2_USER_ROLE_MAX; i++)
  {
    if ( strcmp( userRoleTbl[i], roleStr ) == 0 )
    {
      userRole = i;
      break;
    }
  }
  return userRole;
}

//---------------------------------------------------------------------
// FUNCTION     : extract_certificate_data
//
// DESCRIPTION  : Extract data from the certificate extensions
//
// PARAMETERS   : cert - certificate received from server
//                ext_data - structure to hold extracted data
//
// RETURNS      : TRUE or FALSE
//
// NOTES        :
//
//---------------------------------------------------------------------
int extract_certificate_data ( X509 * cert, ADSAP2_CERT_EXT_DATA * ext_data )
{
  X509_EXTENSION * extension = NULL;
  ASN1_OBJECT * object = NULL;
  ASN1_OCTET_STRING	* octet_string	= NULL;
  int ext_count = 0;
  int i,j,k;
  char * p_begin;
  char * p_end;
  int len;
  char *CN = "CN=";
  char *SLASH = "/";
  BOOLEAN good_string = FALSE;
  struct snval snvalue;

  avoPrintf(ADSAP_ADHOC, "extract_certificate_data\n");

  /*
  ** Set defaults.  Port must be zero to force us to look at the RIP ID.
  ** Assume this is a video preauth request by default.
  */
  memset(ext_data, 0, sizeof(ADSAP2_CERT_EXT_DATA));
  strcpy(ext_data->type, "VIDEO");
  ext_data->logdla    = 1;

  /*
  ** Retrieve the username
  */
  avoPrintf(ADSAP_ADHOC, "extract_certificate_data retrieve username\n");
  if((p_begin = strstr(cert->name, CN)) == NULL)
  {
    avoPrintf(ADSAP_WARNING, "extract_certificate_data: no CN found\n");
    return FALSE;
  }

  p_begin += strlen(CN);
  p_end = strstr(p_begin, SLASH);

  if(p_end != NULL)
    len = p_end - p_begin;
  else
    len = strlen(p_begin);

  if(len == 0)
  {
    avoPrintf(ADSAP_WARNING, "extract_certificate_data: name len 0\n");
    return FALSE;
  }

  if(len > ADSAP2_USERNAME_STR_LEN-1)
    len = ADSAP2_USERNAME_STR_LEN-1;

  strncpy(ext_data->username, p_begin, len);
  ext_data->username[len] = '\0';
  avoPrintf(ADSAP_INFO, "extract_certificate_data: username = %s\n", ext_data->username);

  /*
  ** Search for extensions
  */
  ext_count = X509_get_ext_count(cert);
  avoPrintf(ADSAP_ADHOC, "extract_certificate_data ext_count = %d\n", ext_count);

  for (i = 0; i < ext_count; i++, good_string = TRUE)
  {
    good_string = FALSE;

    if( (extension = X509_get_ext(cert, i)) == NULL ) {
      avoPrintf(ADSAP_FAILURE, 
        "Search extension failed: X509_get_ext %s %d\n", cert->name, i);
      break;
    }

    if( (object = X509_EXTENSION_get_object(extension)) == NULL ) {
      avoPrintf(ADSAP_FAILURE, 
        "Search extension failed: X509_EXTENSION_get_object\n");
      break;
	}

    if( (a1readoid(object->data, object->length, &(snvalue.sv_val.sv_oid))) != 
         SNMP_OK ) {
      avoPrintf(ADSAP_FAILURE, 
        "Search extension failed: a1readoid: %d %4x... \n", 
        object->length, *(unsigned int *)object->data);
//      break;
        continue;    //Ignore unknown extensions
    }

    if( snvalue.sv_val.sv_oid.len != EXTENSION_OID_LEN ) {
      avoPrintf(ADSAP_FAILURE, 
        "Search extension failed: snvalue.sv_val.sv_oid.len %d != %d\n", 
        snvalue.sv_val.sv_oid.len, EXTENSION_OID_LEN);
//      break;
      continue;    //Ignore unknown extensions
    }

    for(j = 0; j < EXTENSION_OID_ITEMS; j++)
    {
      for(k = 0; k < EXTENSION_OID_LEN; k++)
      {
        if(snvalue.sv_val.sv_oid.id[k] != extension_oid[j][k]) {
          //avoPrintf(ADSAP_FAILURE, "Search extension failed: snvalue.sv_val.sv_oid.id[%d] %d!=%d\n", k, snvalue.sv_val.sv_oid.id[k], extension_oid[j][k]);
          break;
        }
      }

      if(k == EXTENSION_OID_LEN)
        break;
    }

    if(k != EXTENSION_OID_LEN) {
      avoPrintf(ADSAP_FAILURE, "Search extension failed: EXTENSION_OID: %d", snvalue.sv_val.sv_oid.id[0]);
	  for(k=1; k < EXTENSION_OID_LEN; k++)
		avoPrintf(ADSAP_FAILURE, ".%d", snvalue.sv_val.sv_oid.id[k]);
	  avoPrintf(ADSAP_FAILURE, "\n");
      continue;    //Ignore unknown extensions
    }

    if((octet_string = X509_EXTENSION_get_data(extension)) == NULL) {
      avoPrintf(ADSAP_FAILURE, 
        "Search extension failed: X509_EXTENSION_get_data\n");
//      break;
      continue;    //Ignore unknown extensions
	}

    if((len = octet_string->length) == 0) {
      avoPrintf(ADSAP_FAILURE, 
        "Search extension failed: octet_string length = 0\n");
//      break;
      continue;    //Ignore unknown extensions
    }

    switch( j )
    {
      case OID_INDEX_COMMAND:
      {
        // Command
        if(len > ADSAP2_RIP_ID_STR_LEN-1)
        {
          len = ADSAP2_RIP_ID_STR_LEN-1;
        }
        strncpy(ext_data->command, octet_string->data, len);
        ext_data->command[len]	= '\0';
        avoPrintf(ADSAP_INFO, "extract_certificate_data: found Commandi[%d] %s\n", len, ext_data->command);
        break;
      }

	  case OID_INDEX_RIP_ID:
      {
        // RIP ID
        if(len > ADSAP2_RIP_ID_STR_LEN-1)
        {
          len = ADSAP2_RIP_ID_STR_LEN-1;
        }
        strncpy(ext_data->RIP_ID, octet_string->data, len);
        ext_data->RIP_ID[len]	= '\0';
        avoPrintf(ADSAP_INFO, "extract_certificate_data: found RIPID[%d] %s\n", len, ext_data->RIP_ID);
        break;
      }

      case OID_INDEX_CHANNEL:
      {
        // Channel
        ext_data->channel = atoi(octet_string->data);
        avoPrintf(ADSAP_INFO, "extract_certificate_data: found Channel %d\n", ext_data->channel);
        break;
      }

      case OID_INDEX_TYPE:
      {
        // Type (video, virtual_media, serial)
        if(len > ADSAP2_SESSION_MODE_STR_LEN-1)
        {
          len = ADSAP2_SESSION_MODE_STR_LEN-1;
        }
        strncpy(ext_data->type, octet_string->data, len);
        ext_data->type[len]	= '\0';
        avoPrintf(ADSAP_INFO, "extract_certificate_data: found type[%d] %s\n", len, ext_data->type);
        break;
      }

      case OID_INDEX_ROLE:
      {
        // Appliance Role
        if(len > ADSAP2_USER_ROLE_STR_LEN-1)
        {
          len = ADSAP2_USER_ROLE_STR_LEN-1;
        }
        strncpy(ext_data->role, octet_string->data, len);
        ext_data->role[len]	= '\0';
        avoPrintf(ADSAP_INFO, "extract_certificate_data: found role[%d] %s\n", len, ext_data->role);
        break;
      }

      case OID_INDEX_PREEMPTION:
      {
        // Preemption level
        ext_data->preempt = atoi(octet_string->data);
        avoPrintf(ADSAP_INFO, "extract_certificate_data found: preemption level %d\n", ext_data->preempt);
        break;
      }

      case OID_INDEX_PASSWORD:
      {
        // Password
        if(len > ADSAP2_PASSWORD_STR_LEN-1)
        {
          len = ADSAP2_PASSWORD_STR_LEN-1;
        }
        strncpy(ext_data->password, octet_string->data, len);
        ext_data->password[len]	= '\0';
        avoPrintf(ADSAP_INFO, "extract_certificate_data found: password[%d] %s\n", len, ext_data->password);
        break;
      }

      case OID_INDEX_MODE:
      {
        // Mode
        if(len > ADSAP2_SESSION_MODE_STR_LEN-1)
        {
          len = ADSAP2_SESSION_MODE_STR_LEN-1;
        }
        strncpy(ext_data->mode, octet_string->data, len);
        ext_data->mode[len]	= '\0';
        avoPrintf(ADSAP_INFO, "extract_certificate_data found: mode[%d] %s\n", len, ext_data->mode);
        break;
      }

      case OID_INDEX_EXPIRATION:
      {
        // Expiration
        ext_data->expiration = atoi(octet_string->data);
        avoPrintf(ADSAP_INFO, "extract_certificate_data found: expiration %d\n", ext_data->expiration);
        break;
      }

      case OID_INDEX_LOGDLA:
      {
        // Appliance Log DLA enable/disable
        ext_data->logdla = atoi(octet_string->data);
        avoPrintf(ADSAP_INFO, "extract_certificate_data found: logdla %d\n", ext_data->logdla);
        break;
      }

      default:
        avoPrintf(ADSAP_INFO, "extract_certificate_data:unrecognized extension\n");
        break;

    }
  } // search for extensions

  avoPrintf(ADSAP_ADHOC,"extract_certificate_data done extracting extensions\n");
  return good_string;
}

//---------------------------------------------------------------------
// FUNCTION     : ADSAP2_preauth_cert_valid
//
// DESCRIPTION  : Determine if the session certificate is still
//                valid - not expired.
//
// PARAMETERS   : pointer to certificate
//
// RETURNS      : FALSE if certificate has expired, TRUE otherwise
//
// NOTES        :
//
//---------------------------------------------------------------------
BOOLEAN ADSAP2_preauth_cert_valid(void *p_data)
{
	P_ADSAP2_PREAUTH_CERT	p_cert = (P_ADSAP2_PREAUTH_CERT) p_data;
	BOOLEAN					ret_val = TRUE;
  time_t current_time, delta;

  current_time = time( NULL );

	delta = current_time - (time_t)p_cert->cert_preauth_time;

	if( delta > p_cert->cert_valid_duration )
		ret_val = FALSE;

	return(ret_val);
}

//*****************************************************************************
EVP_PKEY *get_pubkey_from_DER_cert(unsigned char *p_data, unsigned short len, X509 **pp_x509)
{
	unsigned char	*p_cert_data = p_data;
  EVP_PKEY		*p_pubkey	= NULL;
	X509			*p_x509		= NULL;

	if((p_x509 = d2i_X509(NULL, (unsigned char **) &p_cert_data, len)) == NULL)
		goto get_pubkey_from_DER_cert_error;

	if((p_pubkey = X509_get_pubkey(p_x509)) != NULL)
		goto get_pubkey_from_DER_cert_exit;

get_pubkey_from_DER_cert_error:
	if(p_x509 != NULL)
	{
		X509_free(p_x509);
		p_x509 = NULL;
	}

get_pubkey_from_DER_cert_exit:
	if(pp_x509 != NULL)
		*pp_x509 = p_x509;
  else // Adam
    X509_free(p_x509); // Adam

	return(p_pubkey);
}


//*****************************************************************************

BOOLEAN ADSAP2_preauth_cert_compare(void *p_list, void *p_test, int data_type)
{
	P_ADSAP2_PREAUTH_CERT	p_list_cert = (P_ADSAP2_PREAUTH_CERT) p_list;
	P_ADSAP2_PREAUTH_CERT	p_test_cert = (P_ADSAP2_PREAUTH_CERT) p_test;
	unsigned long			i;
	unsigned char			*p_list_data;
	unsigned long			list_data_len;
	unsigned char			*p_test_data;
	unsigned long			test_data_len;
	BOOLEAN					ret_val		= TRUE;

	avoPrintf(ADSAP_ADHOC, "ADSAP2_preauth_cert_compare\n");

  if(p_list_cert->type != p_test_cert->type) {
    avoPrintf(ADSAP_INFO, "ADSAP2_preauth_cert_compare type mismatch\n");
    goto ADSAP2_preauth_cert_compare_error;
  }

  if(strlen(p_list_cert->username) != strlen(p_test_cert->username)) {
    avoPrintf(ADSAP_INFO, "ADSAP2_preauth_cert_compare name-len mismatch\n");
    goto ADSAP2_preauth_cert_compare_error;
  }
	if(strcmp(p_list_cert->username, p_test_cert->username)) {
    avoPrintf(ADSAP_INFO, "ADSAP2_preauth_cert_compare name mismatch\n");
    goto ADSAP2_preauth_cert_compare_error;
  }

	// DSRs compare the RIP ID
  if (p_test_cert->RIP_ID[0]) {
  	if(strcmp(p_list_cert->RIP_ID, p_test_cert->RIP_ID)) {
      avoPrintf(ADSAP_INFO, "ADSAP2_preauth_cert_compare RIP_ID mismatch\n");
      goto ADSAP2_preauth_cert_compare_error;
    }
  }

	// DSRs compare the channel ID
  if(p_test_cert->channel != 0) {
	if(p_list_cert->channel != p_test_cert->channel) {
      avoPrintf(ADSAP_INFO, "ADSAP2_preauth_cert_compare channel mismatch\n");
      goto ADSAP2_preauth_cert_compare_error;
    }
  }

  if(p_test_cert->req_port != 0) {
	if(p_list_cert->req_port != p_test_cert->req_port) {
      avoPrintf(ADSAP_INFO, "ADSAP2_preauth_cert_compare req_port mismatch\n");
      goto ADSAP2_preauth_cert_compare_error;
    }
  }

	switch(data_type)
	{
		case ADSAP2_CERT_DATA:
		{
			p_list_data		= p_list_cert->der_cert_data;
			list_data_len	= p_list_cert->der_cert_len;
			p_test_data		= p_test_cert->der_cert_data;
			test_data_len	= p_test_cert->der_cert_len;
			break;
		}

		case ADSAP2_KEY_DATA:
		{
			p_list_data		= p_list_cert->der_key_data;
			list_data_len	= p_list_cert->der_key_len;
			p_test_data		= p_test_cert->der_key_data;
			test_data_len	= p_test_cert->der_key_len;
			break;
		}

		default:
		{
			goto ADSAP2_preauth_cert_compare_error;
		}
	}

	if((p_list_data == NULL) || (list_data_len == 0) || (p_test_data == NULL) || (test_data_len == 0)) {
    avoPrintf(ADSAP_INFO, "ADSAP2_preauth_cert_compare key incomplete\n");
    goto ADSAP2_preauth_cert_compare_error;
  }

	if(list_data_len != test_data_len) {
    avoPrintf(ADSAP_INFO, "ADSAP2_preauth_cert_compare key-len mismatch\n");
    goto ADSAP2_preauth_cert_compare_error;
  }

	for(i = 0; i < list_data_len; i++)
	{
		if(p_list_data[i] != p_test_data[i]) {
    		avoPrintf(ADSAP_INFO, "ADSAP2_preauth_cert_compare key-data mismatch\n");
    		goto ADSAP2_preauth_cert_compare_error;
  		}
	}

	goto ADSAP2_preauth_cert_compare_exit;

ADSAP2_preauth_cert_compare_error:
	ret_val = FALSE;

ADSAP2_preauth_cert_compare_exit:

	return(ret_val);
}

//*****************************************************************************

BOOLEAN ADSAP2_preauth_cert_free(void *p_data)
{
	P_ADSAP2_PREAUTH_CERT	p_cert = (P_ADSAP2_PREAUTH_CERT) p_data;

	avoPrintf(ADSAP_ADHOC,"ADSAP2_preauth_cert_free\n");
	free(p_cert->der_key_data);
	free(p_cert->der_cert_data);

	p_cert->der_key_data		= NULL;
	p_cert->der_cert_data		= NULL;

	p_cert->der_key_len		= 0;
  p_cert->der_cert_len		= 0;

  p_cert->cert_preauth_time	= 0;
	p_cert->cert_valid_duration	= 0;

	free(p_cert);

	return(TRUE);
}

//*****************************************************************************

unsigned char get_number_preauth_port(PGLOBAL_LIST p_list, char req_port, BOOLEAN list_locked)
{
	PLIST_ITEM				p_item;
	unsigned char			count = 0;
	P_ADSAP2_PREAUTH_CERT	p_preauth_cert;

	avoPrintf(ADSAP_ADHOC, "get_number_preauth_port\n");
	if(p_list)
	{
		if(!list_locked)
    {
			pthread_mutex_lock(& p_list->list_lock);
    }

		p_item = p_list->list_head;

		while(p_item != NULL)
		{
      p_preauth_cert = (P_ADSAP2_PREAUTH_CERT) p_item->p_data;

			if((req_port == (char)-1) || (req_port == (char) p_preauth_cert->req_port))
				count++;

			p_item = p_item->p_next;
		}

		if(!list_locked)
    {
			pthread_mutex_unlock(& p_list->list_lock);
    }
	}

	return(count);
}

//*****************************************************************************

unsigned char get_number_preauth_rip(PGLOBAL_LIST p_list, char *ripId, BOOLEAN list_locked)
{
	PLIST_ITEM				p_item;
	unsigned char			count = 0;
	P_ADSAP2_PREAUTH_CERT	p_preauth_cert;

	avoPrintf(ADSAP_ADHOC, "get_number_preauth_rip\n");
	if(p_list)
	{
		if(!list_locked)
    {
			pthread_mutex_lock(& p_list->list_lock);
    }

		p_item = p_list->list_head;

		while(p_item != NULL)
		{
      p_preauth_cert = (P_ADSAP2_PREAUTH_CERT) p_item->p_data;

            if ( 0 == strcmp(ripId, p_preauth_cert->RIP_ID) )
				count++;

			p_item = p_item->p_next;
		}

		if(!list_locked)
    {
			pthread_mutex_unlock(& p_list->list_lock);
    }
	}

	return(count);
}

//*****************************************************************************

//---------------------------------------------------------------------
// FUNCTION     : ADSAP2_port_avail
//
// DESCRIPTION  : Determine if the target requested by the
//                DSView Server is available
//
// PARAMETERS   :
//
//
// RETURNS      :
//
// NOTES        : This may be a request for either a KVM or a VM session.
//                Does not reserve the resource.
//---------------------------------------------------------------------
WORD ADSAP2_port_avail(P_ADSAP2_PREAUTH_CERT p_cert, char *existingUsername,
                       BYTE *pExistingPreempt, BYTE *pExistingExclusive)
{
	return ADSAP2_NOERROR; 
}

//---------------------------------------------------------------------
// FUNCTION     : ADSAP2_credentials_present
//
// DESCRIPTION  : Determine if all of the certificates are stored
//                in appliance memory.
//
// PARAMETERS   : none
//
// RETURNS      : TRUE if all credentials are present, FALSE if not
//
// NOTES        : Credentials are received from the DSView Server.
//
//---------------------------------------------------------------------
BOOLEAN ADSAP2_credentials_present(void)
{
	if(	( ap_cert.len != 0 ) &&
		  ( ap_key.len  != 0 ) &&
			( ds_cert.len != 0 ) )
  {
    //avoPrintf(ADSAP_INFO, "ADSAP2_credentials present\n");
		return(TRUE);
  }

  //avoPrintf(ADSAP_INFO, "ADSAP2_credentials not present\n");
	return(FALSE);
}

//*****************************************************************************

//---------------------------------------------------------------------
// FUNCTION     : clear_credential_files
//
// DESCRIPTION  : Clear credential files
//
// PARAMETERS   : none
//
// RETURNS      :
//
// NOTES        :
//---------------------------------------------------------------------
int clear_credential_files( void )
{
	avoPrintf(ADSAP_ADHOC, "clear_credential_files\n");
  if( remove (DSVIEW_CERTIFICATE_FILE) != 0 )
  {
		return(FALSE);
  }
	if( remove (APPLIANCE_CERTIFICATE_FILE) != 0 )
  {
		return(FALSE);
  }
  if( remove (APPLIANCE_RSA_PRIVATEKEY_FILE) != 0 )
  {
		return(FALSE);
  } else {
// Remove private key
    remove (APPLIANCE_RSA_PRIVATEKEY_FILE".pem");
  }

// Update adsap2 to remove certs and keys from flash.
  adsap2_system("/bin/adsap2_update");

  return TRUE;
}

//---------------------------------------------------------------------
// FUNCTION     : clear_credentials
//
// DESCRIPTION  : Clear global credentials
//
// PARAMETERS   : none
//
// RETURNS      :
//
// NOTES        :
//---------------------------------------------------------------------
void clear_credentials(void)
{
	avoPrintf(ADSAP_ADHOC, "clear_credentials\n");
	if(ap_cert.len != 0)
	{
		memset(ap_cert.p_data, 0, ap_cert.len);
		free(ap_cert.p_data);
		ap_cert.p_data	= NULL;
		ap_cert.len		= 0;
	}

	if( ap_key.len != 0 )
	{
		memset( ap_key.p_data, 0, ap_key.len );
		free( ap_key.p_data );
		ap_key.p_data	= NULL;
		ap_key.len = 0;
	}

	if( ds_cert.len != 0 )
	{
		memset( ds_cert.p_data, 0, ds_cert.len );
		free( ds_cert.p_data );
		ds_cert.p_data = NULL;
		ds_cert.len = 0;
	}
}

//*****************************************************************************


//---------------------------------------------------------------------
// FUNCTION     : get_ADSAP2_mode
//
// DESCRIPTION  : Returns the current state of the appliance as
//                pertains to certificate management.
//
// PARAMETERS   : None
//
// RETURNS      : The current state of the appliance (ADSAP2 mode).
//
// NOTES        :
//  -ADSAP2_MODE_UNASSIGNED - the appliance has neither a DSView system
//   certificate nor an assigned MA certificate.  In this mode the only
//   valid requests is a Set Certificates request.
//  -ADSAP2_MODE_ASSIGNED - the appliance has a DSView system
//   certificate, an assigned MA certificate, and an assigned MA RSA
//   private key.  In this state ADSAP2 will perform full SSL
//   authentication.
//  -ADSAP2_MODE_TRUSTALL - the appliance has only the Global Avocent
//   DSView System certificate, the Global Avocent DSView Appliance
//   certificate, and the Global Avocent Appliance RSA private key.  In
//   this state ADSAP2 will not perform any certificate validation, and
//   any DSView system may control the appliance.
//---------------------------------------------------------------------
int get_ADSAP2_mode( void )
{
  int adsap2_mode;

  if ( DsmibGetSecurityMode() == DSMIB_SECURITY_MODE_TRUSTALL )
  {
    avoPrintf(ADSAP_ADHOC, "get_ADSAP2_mode: mode is ADSAP2_MODE_TRUSTALL\n");
    adsap2_mode = ADSAP2_MODE_TRUSTALL;
  }
  else if( ADSAP2_credentials_present() )
  {
    avoPrintf(ADSAP_ADHOC, "get_ADSAP2_mode: mode is ADSAP2_MODE_ASSIGNED\n");
    adsap2_mode = ADSAP2_MODE_ASSIGNED;
  }
  else
  {
    avoPrintf(ADSAP_ADHOC, "get_ADSAP2_mode: mode is ADSAP2_MODE_UNASSIGNED\n");
    adsap2_mode = ADSAP2_MODE_UNASSIGNED;
  }

  return(adsap2_mode);
}

//*****************************************************************************

void build_ADSAP2_PREAUTH_reply(P_ADSAP2_HEADER	p_header,
								WORD			wStatus,
								unsigned char	*username,
								BYTE			byPreempt,
								BYTE			byState)
{
	PF_ADSAP2_STATUS	p_status;
	PF_ADSAP2_USERNAME	p_username = NULL;
	PF_ADSAP2_PREEMPT	p_preempt;
	PF_ADSAP2_STATE		p_state;

	P_ADSAP2_REC_TERM	p_recterm;
	char				*p_start, *p_end;
	unsigned short		username_len = 0;

    if (username)
    {
	  username_len = strlen(username);
    }

	p_start									= (unsigned char *) p_header + sizeof(ADSAP2_HEADER);
	p_end									= p_start;

	p_status								= (PF_ADSAP2_STATUS) p_start;
	p_status->ADSAP2FieldHdr.byFieldID		= ADSAP2_PREAUTH_REP_STATUS;
	p_status->ADSAP2FieldHdr.wFieldLen		= htons(sizeof(p_status->wStatus));
	p_status->wStatus						= htons(wStatus);
	p_end									+= sizeof(F_ADSAP2_STATUS);

	if((wStatus == ADSAP2_RESOURCE_BUSY) || (wStatus == ADSAP2_RESOURCE_BLOCKED))
	{
		p_username								= (PF_ADSAP2_USERNAME) p_end;
		p_username->ADSAP2FieldHdr.byFieldID	= ADSAP2_PREAUTH_REP_USERNAME;
		p_username->ADSAP2FieldHdr.wFieldLen	= htons(username_len);
		memcpy((void *) (&(p_username->byUsername)), username, username_len);
		p_end									+=(sizeof(ADSAP2FIELD) + username_len);

		p_preempt								= (PF_ADSAP2_PREEMPT) p_end;
		p_preempt->ADSAP2FieldHdr.byFieldID		= ADSAP2_PREAUTH_REP_PREEMPT;
		p_preempt->ADSAP2FieldHdr.wFieldLen		= htons(sizeof(p_preempt->byPreempt));
		p_preempt->byPreempt					= byPreempt;
		p_end									+= sizeof(F_ADSAP2_PREEMPT);
	}

	if(wStatus == ADSAP2_RESOURCE_BUSY)
	{
		p_state									= (PF_ADSAP2_STATE) p_end;
		p_state->ADSAP2FieldHdr.byFieldID		= ADSAP2_PREAUTH_REP_STATE;
		p_state->ADSAP2FieldHdr.wFieldLen		= htons(sizeof(p_state->byState));
		p_state->byState						= byState;
		p_end									+= sizeof(F_ADSAP2_STATE);
	}

	p_recterm								= (P_ADSAP2_REC_TERM) p_end;
	p_recterm->byRecTerm					= ADSAP2_REC_TERM_VALUE;
	p_end									+= sizeof(ADSAP2_REC_TERM);

	p_header->byMsgType						|= 0x80;
	p_header->dwMsgLen						= htonl(p_end - p_start);
}

//*****************************************************************************

//---------------------------------------------------------------------
// FUNCTION     : ADSAP2_initialize_list
//
// DESCRIPTION  : Initialize ADSAP2 linked list
//
// PARAMETERS   : None
//
// RETURNS      :
//
// NOTES        :
//
//---------------------------------------------------------------------
void * ADSAP2_initialize_list( void )
{
	PGLOBAL_LIST p_list;

  avoPrintf(ADSAP_INFO, "ADSAP2_initialize_list\n");
	p_list = new_list();

	if(p_list != NULL)
	{
		set_item_compare_callback	(p_list, ADSAP2_preauth_cert_compare);
		set_item_valid_callback		(p_list, ADSAP2_preauth_cert_valid);
		set_item_free_callback		(p_list, ADSAP2_preauth_cert_free);
	}

	return((void *) p_list);
}

void free_preauth_cert(P_ADSAP2_PREAUTH_CERT *pp_cert)
{
	if(pp_cert != NULL)
	{
		if(*pp_cert != NULL)
		{
			if((*pp_cert)->der_cert_data != NULL)
				free((*pp_cert)->der_cert_data);

			if((*pp_cert)->der_key_data != NULL)
				free((*pp_cert)->der_key_data);

			free(*pp_cert);
			*pp_cert = NULL;
		}
	}
}

//*****************************************************************************
//*****************************************************************************

P_ADSAP2_PREAUTH_CERT new_preauth_cert(unsigned char *p_data, unsigned short data_len, int data_type)
{
	P_ADSAP2_PREAUTH_CERT	p_preauth_cert		= NULL;
	EVP_PKEY				*p_pubkey			= NULL;
	unsigned char			temp_key_data[256];
	unsigned char			*p_temp_key_data	= temp_key_data;
	unsigned short			temp_key_len;

  avoPrintf(ADSAP_ADHOC, "new_preauth_cert\n");
	if((p_preauth_cert = malloc(sizeof(ADSAP2_PREAUTH_CERT))) == NULL)
		goto new_preauth_cert_exit;

	memset(p_preauth_cert, 0, sizeof(ADSAP2_PREAUTH_CERT));

	if(data_type == ADSAP2_CERT_DATA)
	{
		if((p_preauth_cert->der_cert_data = malloc(data_len)) == NULL)
			goto new_preauth_cert_error;

		memcpy(p_preauth_cert->der_cert_data, p_data, data_len);
		p_preauth_cert->der_cert_len = data_len;

		if((p_pubkey = get_pubkey_from_DER_cert(p_data, data_len, NULL)) == NULL)
			goto new_preauth_cert_error;

		if((temp_key_len = i2d_PublicKey(p_pubkey, &p_temp_key_data)) == 0)
			goto new_preauth_cert_error;

		p_temp_key_data = temp_key_data;
	}
	else if(data_type == ADSAP2_KEY_DATA)
	{
		p_temp_key_data = p_data;
		temp_key_len = data_len;
	}
	else
	{
		goto new_preauth_cert_error;
	}

	if((p_preauth_cert->der_key_data = malloc(temp_key_len)) == NULL)
		goto new_preauth_cert_error;

	memcpy(p_preauth_cert->der_key_data, p_temp_key_data, temp_key_len);
	p_preauth_cert->der_key_len = temp_key_len;

	goto new_preauth_cert_exit;

new_preauth_cert_error:
	if(p_preauth_cert != NULL)
		free_preauth_cert(&p_preauth_cert);

new_preauth_cert_exit:

  if(p_pubkey != NULL)
  {
    EVP_PKEY_free(p_pubkey);
  }

	return(p_preauth_cert);
}

//*****************************************************************************


//---------------------------------------------------------------------
// FUNCTION     : process_preauth_cert
//
// DESCRIPTION  : Extract the preauth certificate extensions
//
// PARAMETERS   :
//
//
// RETURNS      :
//
// NOTES        :
//---------------------------------------------------------------------
WORD process_preauth_cert(unsigned char *p_cert_data, WORD cert_len, P_ADSAP2_PREAUTH_CERT *pp_preauth_cert)
{
  WORD status = ADSAP2_ERROR;
  X509 *p_x509 = NULL;
  EVP_PKEY *p_pubkey = NULL;
  P_ADSAP2_PREAUTH_CERT p_preauth_cert = NULL;
  ADSAP2_CERT_EXT_DATA ext_data;
  int len;
  BOOLEAN good_string= FALSE;

  avoPrintf(ADSAP_ADHOC, "process_preauth_cert\n");
  if((p_cert_data == NULL) || (cert_len == 0))
    goto process_preauth_cert_error;

  if((p_preauth_cert = new_preauth_cert(p_cert_data, cert_len, ADSAP2_CERT_DATA)) == NULL)
    goto process_preauth_cert_error;
  else
    avoPrintf(ADSAP_ADHOC, "process_preauth_cert new_preauth_cert OK\n");

  if((p_x509 = d2i_X509(NULL, (unsigned char **) &p_cert_data, cert_len)) == NULL)
    goto process_preauth_cert_error;
  else
    avoPrintf(ADSAP_ADHOC, "process_preauth_cert d2i_X509 OK\n");

  if((p_pubkey = X509_get_pubkey(p_x509)) == NULL)
    goto process_preauth_cert_error;
  else 
    avoPrintf(ADSAP_ADHOC, "process_preauth_cert X509_get_pubkey OK\n");

  if((X509_verify(p_x509, p_pubkey)) <= 0)
    goto process_preauth_cert_error;
  else 
    avoPrintf(ADSAP_ADHOC, "process_preauth_cert X509_verify OK\n");

  good_string = extract_certificate_data ( p_x509, & ext_data );

  if(!good_string)
    goto process_preauth_cert_error;
  else 
    avoPrintf(ADSAP_ADHOC, "process_preauth_cert extract_certificate_data OK\n");

  // Copy extracted certificate data to pre-auth cert structure
  p_preauth_cert->cert_preauth_time = time(NULL);
  p_preauth_cert->req_port = ext_data.port;
  p_preauth_cert->cert_valid_duration = ext_data.expiration;
  p_preauth_cert->preempt = ext_data.preempt;
  p_preauth_cert->channel = ext_data.channel;
  p_preauth_cert->mode = sessionMode2Enum(ext_data.mode);
  p_preauth_cert->type = preauthType2Enum(ext_data.type);
  p_preauth_cert->logdla = ext_data.logdla;

  // Copy username to pre-auth cert structure
  if((len = strlen(ext_data.username)) > ADSAP2_USERNAME_STR_LEN-1)
  {
    len = ADSAP2_USERNAME_STR_LEN-1;
  }
  strncpy(p_preauth_cert->username, ext_data.username, len);
  p_preauth_cert->username[len] = '\0';

  // Copy role to pre-auth cert structure
  if((len = strlen(ext_data.role)) > ADSAP2_USER_ROLE_STR_LEN-1)
  {
    len = ADSAP2_USER_ROLE_STR_LEN-1;
  }
  strncpy(p_preauth_cert->role, ext_data.role, len);
  p_preauth_cert->role[len] = '\0';

  // Copy ripid to pre-auth cert structure
  if((len = strlen(ext_data.RIP_ID)) > ADSAP2_RIP_ID_STR_LEN-1)
  {
    len = ADSAP2_RIP_ID_STR_LEN-1;
  }
  strncpy(p_preauth_cert->RIP_ID, ext_data.RIP_ID, len);
  p_preauth_cert->RIP_ID[len] = '\0';

  if((len = strlen(ext_data.command)) > ADSAP2_RIP_ID_STR_LEN-1)
  {
    len = ADSAP2_RIP_ID_STR_LEN-1;
  }
  strncpy(p_preauth_cert->command, ext_data.command, len);
  p_preauth_cert->command[len] = '\0';

  *pp_preauth_cert = p_preauth_cert;
  status = ADSAP2_NOERROR;
  goto process_preauth_cert_exit;

process_preauth_cert_error:
  printf("process_preauth_cert_error\n");
  if(p_preauth_cert != NULL)
    free_preauth_cert(&p_preauth_cert);

process_preauth_cert_exit:
  avoPrintf(ADSAP_ADHOC, "process_preauth_cert_exit\n");

  if(p_pubkey != NULL)
    EVP_PKEY_free(p_pubkey);

  if(p_x509 != NULL)
    X509_free(p_x509);

  avoPrintf(ADSAP_INFO, "process_preauth_cert status = %d\n", status);
  return(status);
}

//---------------------------------------------------------------------
// FUNCTION     : process_preauth
//
// DESCRIPTION  : Extract the session certificate from the
//                Pre-Authorization message and insert the session
//                certificate into the list.  Sets the status to
//                indicate whether or not the resource is available
//
// PARAMETERS   :
//
//
// RETURNS      :
//
// NOTES        : Does not reserve the resource.
//---------------------------------------------------------------------
void process_preauth(P_ADSAP2_HEADER p_header, int *p_adsap2_mode)
{
	ADSAP2_PREAUTH_REQ		request_data;
	char					*p_bytes;
	PLIST_ITEM				p_item			= NULL;
	P_ADSAP2_PREAUTH_CERT	p_preauth_cert	= NULL;

	WORD					status			= ADSAP2_NOERROR;
	char					existingUsername[ADSAP2_USERNAME_STR_LEN];
	BYTE					existingPreempt	= 0;
	BYTE					existingExclusive = 0;
	BOOLEAN					locked			= FALSE;

  avoPrintf(ADSAP_ADHOC, "process_preauth\n");
	memset(&request_data, 0, sizeof(request_data));
	memset(&existingUsername, 0, sizeof(existingUsername));

	p_bytes = (((unsigned char *) p_header) + sizeof(ADSAP2_HEADER));

	while(	(((PADSAP2FIELD) p_bytes)->byFieldID != ADSAP2_REC_TERM_VALUE) &&
			(((PADSAP2FIELD) p_bytes)->byFieldID != ADSAP2_TRAILER_VALUE) &&
			(status == ADSAP2_NOERROR))
	{
		switch(((PADSAP2FIELD) p_bytes)->byFieldID)
		{
			case ADSAP2_PREAUTH_REQ_CERT:
			{
				request_data.pCert = (PF_ADSAP2_CERT) p_bytes;
				break;
			}
			default:
			{
				goto process_preauth_error;
			}
		}

		p_bytes += sizeof(ADSAP2FIELD) + ntohs(((PADSAP2FIELD) p_bytes)->wFieldLen);
	}

	if(request_data.pCert == NULL)
    {
		goto process_preauth_error;
    }

	switch(*p_adsap2_mode)			// preauth message
	{
		case ADSAP2_MODE_TRUSTALL:
			break;

		case ADSAP2_MODE_UNASSIGNED:
			goto process_preauth_error;

		case ADSAP2_MODE_ASSIGNED:
			break;

		default:	// unknown mode
			goto process_preauth_error;
	}

	if((status = process_preauth_cert(&(request_data.pCert->byCert), ntohs(request_data.pCert->ADSAP2FieldHdr.wFieldLen), &p_preauth_cert)) != ADSAP2_NOERROR)
		goto process_preauth_exit;

  avoPrintf(ADSAP_ADHOC, "process_preauth new_list_item\n");
	if((p_item = new_list_item(p_preauth_cert)) == NULL)
		goto process_preauth_error;

  avoPrintf(ADSAP_ADHOC,"process_preauth avail_list_item\n");
	status = ADSAP2_port_avail(p_preauth_cert, existingUsername, &existingPreempt, &existingExclusive);
  if ( status != ADSAP2_NOERROR )
  {
    /*
    ** Remove the cert from the list
    */
    free_list_item(preauth_cert_list, p_item, locked);
    if ( status == ADSAP2_NOERROR_VM_UNENCRYPTED )
    {
      /*
      ** This status code means that a unencrypted VM session will follow
      ** which will not pull the cert from the list, so we went ahead and
      ** freed it.  Set the status code back to noerror to return to DSView
      */
      avoPrintf(ADSAP_INFO,"process_preauth removed cert for unencrypted VM\n");
      status = ADSAP2_NOERROR;
    }
		goto process_preauth_exit;
  }

  avoPrintf(ADSAP_ADHOC,"process_preauth add_list_item\n");
	add_list_item(preauth_cert_list, p_item, locked);

	goto process_preauth_exit;

process_preauth_error:
	status = ADSAP2_ERROR;
  avoPrintf(ADSAP_ADHOC,"process_preauth ERROR\n");

process_preauth_exit:
  avoPrintf(ADSAP_ADHOC,"process_preauth_exit\n");

build_ADSAP2_PREAUTH_reply(	p_header, status, existingUsername, existingPreempt, existingExclusive);
}

//*****************************************************************************

void processADSAP2Connect(void  *p_ssl, int *p_adsap2_mode)
{
	SOCKET			sock = SSL_Get_fd(p_ssl);
	unsigned char	dbgName[64];
	int				recv;
	P_ADSAP2_HEADER	p_header;
	BOOLEAN			disconnect = FALSE;
	struct in_addr peerAddr;

	GetPeerName(sock, & peerAddr );
	sprintf(dbgName, "processADSAP2Connect %d.%d.%d.%d",
            (peerAddr.s_addr >>24) & 0xff,
						(peerAddr.s_addr >>16) & 0xff,
						(peerAddr.s_addr >>8) & 0xff,
						peerAddr.s_addr & 0xff);

	avoPrintf(ADSAP_INFO, "\n%s Started\n", dbgName);

	p_header = (P_ADSAP2_HEADER) malloc(ADSAP2_MESSAGE_SIZE);
	memset(p_header, 0, ADSAP2_MESSAGE_SIZE);

	while ( ( GetPeerName( sock, & peerAddr ) == 0 ) && !disconnect )
	{
    recv = receive_ADSAP2_message(p_ssl, p_header, ADSAP2_MESSAGE_SIZE);
		if(recv == 0)		// Select timeout
			continue;

		if(recv == -1)		// Received disconnect
			break;

		if(recv > 0)		// Received data
		{
			switch(p_header->byMsgType)
			{
				case ADSAP2_MSG_STORE_REQ:
				{
					avoPrintf(ADSAP_INFO,"processADSAP2Connect: received store request\n");
					process_store_credentials(p_header, p_adsap2_mode);
					break;
				}

				case ADSAP2_MSG_CLEAR_REQ:
				{
					avoPrintf(ADSAP_INFO,"processADSAP2Connect: received clear request\n");
					process_clear_credentials(p_header, p_adsap2_mode);
					break;
				}

				case ADSAP2_MSG_PREAUTH_REQ:
				{
					avoPrintf(ADSAP_INFO,"processADSAP2Connect: received preauth request\n");
					process_preauth(p_header, p_adsap2_mode);
					break;
				}

				default:
					break;
			}

			send_ADSAP2_message(p_ssl, p_header);			// Send Reply
		}
	}

	free(p_header);

	avoPrintf(ADSAP_INFO,"processADSAP2Connect exiting\n");
}

//*****************************************************************************

int send_ADSAP2_message(void *p_ssl, P_ADSAP2_HEADER p_hdr_struct)
{
	int		header_bytes	= sizeof(ADSAP2_HEADER);
	int		data_bytes		= ntohl(p_hdr_struct->dwMsgLen);
	int		trailer_bytes	= sizeof(ADSAP2_TRAILER);
	int		total_bytes_to_send, total_bytes_sent, bytes_sent;
	char	*p_header		= (unsigned char *) p_hdr_struct;
	char	*p_trailer		= p_header + header_bytes + data_bytes;

	avoPrintf(ADSAP_INFO, "send_ADSAP2_message\n");
	total_bytes_to_send = header_bytes + data_bytes + trailer_bytes;

	*p_trailer = ADSAP2_TRAILER_VALUE;

	for(total_bytes_sent = 0; total_bytes_to_send > total_bytes_sent; )
	{
		bytes_sent = SSL_Write(p_ssl, p_header + total_bytes_sent, total_bytes_to_send - total_bytes_sent);
    #if 0 // debug
    {
    int i;
    printf("send_ADSAP2_message:\n");
    for ( i = 0; i < bytes_sent; i++ )
    {
      printf("%02x ", p_header[i+total_bytes_sent]);
    }
    printf("\n");
    }
    #endif
		if(bytes_sent > 0)
		{
			total_bytes_sent += bytes_sent;
			avoPrintf(ADSAP_ADHOC,"Sent %d bytes = %d header + %d data + %d trailer\n", bytes_sent, header_bytes, data_bytes, trailer_bytes);
		}
		else
		{
			return(-1);
		}
	}

	return(0);
}

//*****************************************************************************

int receive_ADSAP2_message(void  * p_ssl, P_ADSAP2_HEADER p_hdr_struct, int buf_size)
{
	int				header_bytes	= sizeof(ADSAP2_HEADER);
	int				data_bytes;
	int				trailer_bytes	= sizeof(ADSAP2_TRAILER);
	int				bytes_read, total_bytes;

	unsigned char	*p_header	= (unsigned char *) p_hdr_struct;
	unsigned char	*p_data		= p_header + sizeof(ADSAP2_HEADER);
	unsigned char	*p_trailer;

	avoPrintf(ADSAP_ADHOC,"receive_ADSAP2_message\n");
	bytes_read		= SSL_Get_Data(p_ssl, p_header, header_bytes, 30);

	if(bytes_read == 0)
	{
		avoPrintf(ADSAP_ADHOC,"receive_ADSAP2_message: timeout, no data received\n");
		return(0);
	}

	if(bytes_read == -1)
	{
		avoPrintf(ADSAP_ADHOC,"receive_ADSAP2_message: received disconnect\n");
		return(-1);
	}

	if(bytes_read != header_bytes)
	{
		avoPrintf(ADSAP_FAILURE,"receive_ADSAP2_message: partial header received\n");
		return(0);
	}

/* Check the signature */
	if(strncmp(p_hdr_struct->szUniqueID, CPS_ADSAP2_SIGNATURE, CPS_ADSAP2_SIGNATURE_SIZE))
	{
		avoPrintf(ADSAP_FAILURE,"receive_ADSAP2_message: invalid header\n");
		return(0);
	}

/* Read the data and the trailer */
	data_bytes		= ntohl(p_hdr_struct->dwMsgLen);

	if(data_bytes > (buf_size - header_bytes))
	{
		avoPrintf(ADSAP_FAILURE,"receive_ADSAP2_message: message too large - buffer size = %d bytes\n", buf_size);
		return(-1);
	}

	bytes_read		= SSL_Get_Data(p_ssl, p_data, data_bytes + trailer_bytes, 30);

	if(bytes_read == 0)
	{
		avoPrintf(ADSAP_WARNING,"receive_ADSAP2_message: timeout, no data received\n");
		return(0);
	}

	if(bytes_read == -1)
	{
		avoPrintf(ADSAP_INFO,"receive_ADSAP2_message: received disconnect\n");
		return(-1);
	}

	if(bytes_read != data_bytes + trailer_bytes)
	{
		avoPrintf(ADSAP_FAILURE,"receive_ADSAP2_message: partial message received\n");
		return(0);
	}

/* Check the trailer */

	p_trailer = p_data + data_bytes;

	if(*(p_trailer) != ADSAP2_TRAILER_VALUE)
	{
		avoPrintf(ADSAP_FAILURE,"receive_ADSAP2_message: invalid trailer\n");
		return(0);
	}

	total_bytes = header_bytes + data_bytes + trailer_bytes;

	avoPrintf(ADSAP_ADHOC,"Received %d bytes = %d header + %d data + %d trailer\n", total_bytes, header_bytes, data_bytes, trailer_bytes);

	return(total_bytes);
}

//*****************************************************************************

//---------------------------------------------------------------------
// FUNCTION     : send_ADSAP2_auth_request
//
// DESCRIPTION  : Assemble and send an ADSAP2 Authorized request
//
// PARAMETERS   : p_ssl - SSL struct
//                authorizeRequest - struct containing the parameters
//                       of the request
//                authorizeReply - struct containing the parameters
//                  returned by adsap2
//
// RETURNS      : ADSAP2 status code
//
// NOTES        : An Authorization request can be used for:
//                - user authentication - structure will contain username
//                  and password only
//                - target authorization - structure will contain username
//                  and password in addition to the mode string.  Will
//                  also contain EITHER RIP/channel data or SPC
//                  Device/socket data, not both.
//
//---------------------------------------------------------------------
WORD send_ADSAP2_auth_request(	SSL *p_ssl,
								ADSAP2_AUTHORIZE_REQ * authorizeRequest,
								ADSAP2_AUTHORIZE_REP * authorizeReply)
{
	P_ADSAP2_HEADER		  p_header;
	PF_ADSAP2_USERNAME	p_username;
	PF_ADSAP2_PASSWORD	p_password;
	PF_ADSAP2_MODE		  p_mode;
	PF_ADSAP2_ID		    p_id;
	P_ADSAP2_REC_TERM	  p_recterm;

	char				      message[2048];
	unsigned short		username_len	= 0;
	unsigned short		password_len	= 0;
	unsigned short		mode_len	   	= 0;
  unsigned short    ripId_len     = 0;
  unsigned short    cascadeChan_len  = 0;
  unsigned short    spcDevice_len = 0;
  unsigned short    spcSocket_len = 0;

  char ripStr[32];
  char chanStr[5];
  char modeStr[20];
  char spcDeviceStr[5];
  char spcSocketStr[5];

	char				*p_start, *p_end;
	WORD				status = ADSAP2_ERROR;  // jamie changed

  avoPrintf(ADSAP_INFO, "send_ADSAP2_auth_request\n");

	/*
  ** Username and password should always be present.
  */
  if( authorizeRequest->username )
  {
		username_len	= strlen(authorizeRequest->username);
    avoPrintf(ADSAP_INFO, "username = %s\n", authorizeRequest->username) ;
  }
	if( authorizeRequest->password )
  {
		password_len	= strlen(authorizeRequest->password);
    //avoPrintf(ADSAP_INFO, "password = %s\n", authorizeRequest->password);
  }

	/*
  ** Mode string defaults to "NORMAL" unless it is set to a differnt
  ** value.
  */
  strcpy( modeStr, sessionModeTbl[ authorizeRequest->mode ] );
  mode_len = strlen(modeStr);
  avoPrintf(ADSAP_INFO, "mode = %s\n", modeStr);

	/*
  ** For authorize requests to access a KVM target, the RIP ID parameter
  ** and possibly the cascadeChan parameter will be populated.
  */
  if( authorizeRequest->ripID[0] )
  {
    eid2Str( ripStr, authorizeRequest->ripID );
		ripId_len	= strlen(ripStr);
    avoPrintf(ADSAP_INFO, "ripID = %s\n", ripStr);
  } else {
	if (authorizeRequest->port) {
		sprintf(ripStr, "ttyS%d", authorizeRequest->port);
		ripId_len	= strlen(ripStr);
    	avoPrintf(ADSAP_INFO, "ripID = %s\n", ripStr);
	}
  }

  if ( authorizeRequest->cascadeChan > 0 )
  {
    sprintf(chanStr, "%d", authorizeRequest->cascadeChan);
    cascadeChan_len = strlen(chanStr);
    avoPrintf(ADSAP_INFO, "cascadeChan = %s\n", chanStr);
  }

	/*
  ** For authorize requests to access a SPC socket, the SPC device
  ** parameter and the SPC socket parameter will be populated.
  */
  if ( authorizeRequest->spcDevice > 0 )
  {
    // The "3" is the serial type index
    sprintf(spcDeviceStr, "3.%d", authorizeRequest->spcDevice);
    spcDevice_len = strlen(spcDeviceStr);
    avoPrintf(ADSAP_INFO, "spcDevice = %s\n", spcDeviceStr);
  }

  if ( authorizeRequest->spcSocket > 0 )
  {
    sprintf(spcSocketStr, "%d", authorizeRequest->spcSocket);
    spcSocket_len = strlen(spcSocketStr);
    avoPrintf(ADSAP_INFO, "spcSocket = %s\n", spcSocketStr);
  }

	p_end										= message;
	p_header									= (P_ADSAP2_HEADER) p_end;
	p_header->bySOH								= SOH;
	strncpy(p_header->szUniqueID, CPS_ADSAP2_SIGNATURE, CPS_ADSAP2_SIGNATURE_SIZE);

	p_header->byMsgType							= ADSAP2_MSG_AUTH_REQ;
	p_end										+= sizeof(ADSAP2_HEADER);
	p_start										= p_end;

	p_username									= (PF_ADSAP2_USERNAME) p_end;
	p_username->ADSAP2FieldHdr.byFieldID		= ADSAP2_AUTH_REQ_USERNAME;
	p_username->ADSAP2FieldHdr.wFieldLen		= htons(username_len);
	memcpy((void *) (&(p_username->byUsername)), authorizeRequest->username, username_len);
	p_end										+=(sizeof(ADSAP2FIELD) + username_len);

	p_password									= (PF_ADSAP2_PASSWORD) p_end;
	p_password->ADSAP2FieldHdr.byFieldID		= ADSAP2_AUTH_REQ_PASSWORD;
	p_password->ADSAP2FieldHdr.wFieldLen		= htons(password_len);
	memcpy((void *) (&(p_password->byPassword)), authorizeRequest->password, password_len);
	p_end										+= (sizeof(ADSAP2FIELD) + password_len);

  if ( mode_len)
  {
    p_mode										= (PF_ADSAP2_MODE) p_end;
	  p_mode->ADSAP2FieldHdr.byFieldID			= ADSAP2_AUTH_REQ_MODE;
	  p_mode->ADSAP2FieldHdr.wFieldLen			= htons(mode_len);
	  memcpy((void *) (&(p_mode->byMode)), modeStr, mode_len);
	  p_end										+= (sizeof(ADSAP2FIELD) + mode_len);
  }

  if ( ripId_len )
  {
    p_id										= (PF_ADSAP2_ID) p_end;
	  p_id->ADSAP2FieldHdr.byFieldID				= ADSAP2_AUTH_REQ_ID;
	  p_id->ADSAP2FieldHdr.wFieldLen				= htons(ripId_len);
	  memcpy((void *) (&(p_id->byID)), ripStr, ripId_len);
	  p_end										+= (sizeof(ADSAP2FIELD) + ripId_len);
  }

  if ( cascadeChan_len )
  {
    p_id										= (PF_ADSAP2_ID) p_end;
	  p_id->ADSAP2FieldHdr.byFieldID				= ADSAP2_AUTH_REQ_ID+1;
	  p_id->ADSAP2FieldHdr.wFieldLen				= htons(cascadeChan_len);
	  memcpy((void *) (&(p_id->byID)), chanStr, cascadeChan_len);
	  p_end										+= (sizeof(ADSAP2FIELD) + cascadeChan_len);
  }

  if ( spcDevice_len )
  {
    p_id										= (PF_ADSAP2_ID) p_end;
	  p_id->ADSAP2FieldHdr.byFieldID				= ADSAP2_AUTH_REQ_ID;
	  p_id->ADSAP2FieldHdr.wFieldLen				= htons(spcDevice_len);
	  memcpy((void *) (&(p_id->byID)), spcDeviceStr, spcDevice_len);
	  p_end										+= (sizeof(ADSAP2FIELD) + spcDevice_len);
  }

  if ( spcSocket_len )
  {
    p_id										= (PF_ADSAP2_ID) p_end;
	  p_id->ADSAP2FieldHdr.byFieldID				= ADSAP2_AUTH_REQ_ID+1;
	  p_id->ADSAP2FieldHdr.wFieldLen				= htons(spcSocket_len);
	  memcpy((void *) (&(p_id->byID)), spcSocketStr, spcSocket_len);
	  p_end										+= (sizeof(ADSAP2FIELD) + spcSocket_len);
  }

	p_recterm									= (P_ADSAP2_REC_TERM) p_end;
	p_recterm->byRecTerm						= ADSAP2_REC_TERM_VALUE;
	p_end										+= sizeof(ADSAP2_REC_TERM);

	p_header->dwMsgLen							= htonl(p_end - p_start);

	if((status = send_ADSAP2_message(p_ssl, p_header)) != (WORD)-1)
	{
		if((receive_ADSAP2_message(p_ssl, p_header, ADSAP2_MESSAGE_SIZE) > 0))
    {
			status = get_ADSAP2_auth_reply(p_header, authorizeReply);
    }
    else
    {
      avoPrintf(ADSAP_FAILURE, "receive_ADSAP2_message failed\n");
      status = ADSAP2_ERROR;
    }
	}

	return(status);
}

//*****************************************************************************

void build_ADSAP2_reply(P_ADSAP2_HEADER p_header, BYTE field_id, WORD wStatus)
{
	PF_ADSAP2_STATUS	p_status;
	P_ADSAP2_REC_TERM	p_recterm;
	char				*p_start, *p_end;

	p_start								= (unsigned char *) p_header + sizeof(ADSAP2_HEADER);
	p_end								= p_start;

	p_status							= (PF_ADSAP2_STATUS) p_start;
	p_status->ADSAP2FieldHdr.byFieldID	= field_id;
	p_status->ADSAP2FieldHdr.wFieldLen	= htons(sizeof(p_status->wStatus));
	p_status->wStatus					= htons(wStatus);
	p_end								+= sizeof(F_ADSAP2_STATUS);

	p_recterm							= (P_ADSAP2_REC_TERM) p_end;
	p_recterm->byRecTerm				= ADSAP2_REC_TERM_VALUE;
	p_end								+= sizeof(ADSAP2_REC_TERM);

	p_header->byMsgType					|= 0x80;
	p_header->dwMsgLen					= htonl(p_end - p_start);
}

//*****************************************************************************

int get_ADSAP2_auth_reply(	P_ADSAP2_HEADER	p_header,
							ADSAP2_AUTHORIZE_REP * authorizeReply)
{
	ADSAP2_AUTH_REPLY	reply_data;
	char				*p_bytes;
	WORD				status = ADSAP2_NOERROR;
	char				role[ADSAP2_USER_ROLE_STR_LEN];
	int					role_len;

	avoPrintf(ADSAP_ADHOC, "get_ADSAP2_auth_reply\n");
	memset(&reply_data, 0, sizeof(reply_data));

	if(p_header->byMsgType != (0x80 | ADSAP2_MSG_AUTH_REQ))
	{
		status = ADSAP2_ERROR;
	}
	else
	{
		p_bytes = (((unsigned char *) p_header) + sizeof(ADSAP2_HEADER));

		while(	(((PADSAP2FIELD) p_bytes)->byFieldID != ADSAP2_REC_TERM_VALUE) &&
				(((PADSAP2FIELD) p_bytes)->byFieldID != ADSAP2_TRAILER_VALUE) &&
				(status == ADSAP2_NOERROR))
		{
			switch(((PADSAP2FIELD) p_bytes)->byFieldID)
			{
				case ADSAP2_AUTH_REP_STATUS:
				{
					reply_data.pStatus = (PF_ADSAP2_STATUS) p_bytes;
					break;
				}
				case ADSAP2_AUTH_REP_PREEMPT:
				{
					reply_data.pPreempt = (PF_ADSAP2_PREEMPT) p_bytes;
					break;
				}
				case ADSAP2_AUTH_REP_ROLE:
				{
					reply_data.pRole = (PF_ADSAP2_ROLE) p_bytes;
					break;
				}
				default:
				{
					status = ADSAP2_ERROR;
					break;
				}
			}

			p_bytes += sizeof(ADSAP2FIELD) + ntohs(((PADSAP2FIELD) p_bytes)->wFieldLen);
		}
	}

	if(status == ADSAP2_NOERROR)
	{
		if(reply_data.pStatus == NULL)
			status = ADSAP2_ERROR;
		else
		{
			status = (int) (ntohs(reply_data.pStatus->wStatus));

			if(status == ADSAP2_NOERROR)
			{
				if((reply_data.pPreempt == NULL) || (reply_data.pRole == NULL))
					status = ADSAP2_ERROR;
				else
				{
					authorizeReply->preempt = reply_data.pPreempt->byPreempt;
					role_len = ntohs((reply_data.pRole)->ADSAP2FieldHdr.wFieldLen);
					memcpy(role, &((reply_data.pRole)->byRole), role_len);
					role[role_len] = '\0';
					authorizeReply->role =userRole2Enum(role);
					authorizeReply->logdla = 1;
				}
			}
		}
	}

	return(status);
}

//---------------------------------------------------------------------
//    FUNCTION: DER_read_PKCS8_PrivateKey_nopass()
//
// DESCRIPTION: Extract and teturn the private key
//
//  PARAMETERS: Private key
//
//     RETURNS:
//
//       NOTES: DER = Distinguished Encoding Rules
//              PKCS = standard for storing private keys and certificates
//                     securely
//              EVP = high-level cryptographic functions
//
//---------------------------------------------------------------------
EVP_PKEY *DER_read_PKCS8_PrivateKey_nopass(unsigned char *buf, int len)
{
	EVP_PKEY			*p_pkey	= NULL;
	BIO					*p_bio	= NULL;
	PKCS8_PRIV_KEY_INFO	*p8inf	= NULL;

	/*
  ** Create a memory BIO using len bytes of data at buf
  */
  if((p_bio = BIO_new_mem_buf(buf, len)) == NULL)
		goto DER_read_PKCS8_PrivateKey_nopass_exit;

  /*
  ** Read the DER representation of the private key
  */
  if((p8inf = d2i_PKCS8_PRIV_KEY_INFO_bio(p_bio, NULL)) == NULL)
		goto DER_read_PKCS8_PrivateKey_nopass_exit;

  /*
  ** Extract the private key from the PKCS8 structure
  */
	p_pkey = EVP_PKCS82PKEY(p8inf);

DER_read_PKCS8_PrivateKey_nopass_exit:
	if(p8inf)
		PKCS8_PRIV_KEY_INFO_free(p8inf);
	if(p_bio)
		BIO_free(p_bio);

	return(p_pkey);
}


//---------------------------------------------------------------------
// FUNCTION     : write_file
//
// DESCRIPTION  : Write credential to a flash file
//
// PARAMETERS   : destination buffer, filename
//
// RETURNS      : length of file, 0 if file not found
//
// NOTES        :
//
//---------------------------------------------------------------------
int write_file( unsigned char * buffer, int len, char * filename )
{
  FILE * fp = NULL;
  int result = FALSE;
  int mask = umask(077);

	fp = fopen(filename, "w");
  if ( fp )
	{
		result = fwrite(buffer, 1, len, fp);
		fclose(fp);
		avoPrintf(ADSAP_ADHOC,"Wrote File %s\n", filename);
	}
  umask(mask);
  return result;

}
//---------------------------------------------------------------------
// FUNCTION     : store_credentials_to_file
//
// DESCRIPTION  : Write the credentials to files.
//
// PARAMETERS   : none
//
// RETURNS      : TRUE if all credentials were written, FALSE otherwise
//
// NOTES        : Creates files in the flash file system that contain
//                the credentials.
//              : Call while the credentialsMutex is locked.
//---------------------------------------------------------------------
int store_credentials_to_file(void)
{
	int	ds_cert_result, appliance_cert_result, appliance_key_result;
	int	result = FALSE;

  avoPrintf(ADSAP_INFO, "store_credentials_to_file\n");
  if( ADSAP2_credentials_present() )
	{
    ds_cert_result        = write_file( ds_cert.p_data, ds_cert.len,
                                        DSVIEW_CERTIFICATE_FILE );
    appliance_cert_result = write_file( ap_cert.p_data, ap_cert.len,
                                        APPLIANCE_CERTIFICATE_FILE );
    appliance_key_result  = write_file( ap_key.p_data, ap_key.len,
                                        APPLIANCE_RSA_PRIVATEKEY_FILE );

		result = ds_cert_result && appliance_cert_result && appliance_key_result;
// Update adsap2 to add certs and keys to flash.
		if (result) {
			adsap2_system("/bin/adsap2_update");
		}
	}
  else
  {
    avoPrintf(ADSAP_FAILURE, "Credentials don't exist!\n");
  }

	return(result);
}

//---------------------------------------------------------------------
// FUNCTION     : process_store_credentials
//
// DESCRIPTION  : Store the ADSAP2 credentials
//
// PARAMETERS   : ADSAP2 message from DSView Server that contains
//                credentials
//
// RETURNS      : none
//
// NOTES        : Stores the credentials in global pointers and
//                stores them as files in the flash file system that
//              : Call while the credentialsMutex is locked.
//---------------------------------------------------------------------
void process_store_credentials(P_ADSAP2_HEADER p_header, int *p_adsap2_mode)
{
  ADSAP2_STORE_REQ  request_data;
  char *p_bytes;
  WORD status      = ADSAP2_NOERROR;

  X509           *p_x509_ds       = NULL;
  EVP_PKEY       *p_pubkey_ds     = NULL;
  unsigned char  *p_cert_data_ds  = NULL;
  unsigned short  cert_len_ds;

  X509           *p_x509_ap       = NULL;
  EVP_PKEY       *p_pubkey_ap     = NULL;
  EVP_PKEY       *p_privkey_ap    = NULL;
  unsigned char  *p_cert_data_ap = NULL;
  unsigned char  *p_key_data_ap  = NULL;
  unsigned short  cert_len_ap, key_len_ap;

  avoPrintf(ADSAP_INFO,"process_store_credentials\n");
  memset(&request_data, 0, sizeof(request_data));

  p_bytes = (((unsigned char *) p_header) + sizeof(ADSAP2_HEADER));

  while(  (((PADSAP2FIELD) p_bytes)->byFieldID != ADSAP2_REC_TERM_VALUE) &&
      (((PADSAP2FIELD) p_bytes)->byFieldID != ADSAP2_TRAILER_VALUE) &&
      (status == ADSAP2_NOERROR))
  {
    switch(((PADSAP2FIELD) p_bytes)->byFieldID)
    {
      case ADSAP2_STORE_REQ_DS_CERT:
      {
        avoPrintf(ADSAP_ADHOC,"process_store_credentials: ADSAP2_STORE_REQ_DS_CERT\n");
        request_data.pDS_Cert = (PF_ADSAP2_CERT) p_bytes;
        break;
      }

      case ADSAP2_STORE_REQ_AP_CERT:
      {
        avoPrintf(ADSAP_ADHOC,"process_store_credentials: ADSAP2_STORE_REQ_AP_CERT\n");
        request_data.pAP_Cert = (PF_ADSAP2_CERT) p_bytes;
        break;
      }

      case ADSAP2_STORE_REQ_AP_KEY:
      {
        avoPrintf(ADSAP_ADHOC,"process_store_credentials: ADSAP2_STORE_REQ_AP_KEY\n");
        request_data.pAP_Priv_Key = (PF_ADSAP2_KEY) p_bytes;
        break;
      }
      default:
      {
        status = ADSAP2_ERROR;
        break;
      }
    }

    p_bytes += sizeof(ADSAP2FIELD) + ntohs(((PADSAP2FIELD) p_bytes)->wFieldLen);
  }

  if((request_data.pDS_Cert == NULL) || (request_data.pAP_Cert == NULL) || (request_data.pAP_Priv_Key == NULL))
    goto process_store_credentials_error;

	switch(*p_adsap2_mode)		// set credentials
	{
		case ADSAP2_MODE_TRUSTALL:
		{
      avoPrintf(ADSAP_INFO, "process_store_credentials: ADSAP2_MODE_TRUSTALL\n");
			status = ADSAP2_INTRUST_ALL;
			goto process_store_credentials_exit;
		}
		case ADSAP2_MODE_UNASSIGNED:
		case ADSAP2_MODE_ASSIGNED:
      avoPrintf(ADSAP_INFO, "process_store_credentials: ADSAP2_MODE_UNASSIGNED or ASSIGNED\n");
			break;

		default:	// unknown mode
      avoPrintf(ADSAP_INFO, "process_store_credentials: unknown mode\n");
			goto process_store_credentials_error;
	}

  /*
  ** Process DSView Certificate
  */
  avoPrintf(ADSAP_ADHOC,"process_store_credentials: process dsview cert\n");
  cert_len_ds = ntohs(request_data.pDS_Cert->ADSAP2FieldHdr.wFieldLen);
  p_cert_data_ds = &request_data.pDS_Cert->byCert;

  p_pubkey_ds = get_pubkey_from_DER_cert(p_cert_data_ds, cert_len_ds, &p_x509_ds);

  if((p_pubkey_ds == NULL) || (p_x509_ds == NULL))
    goto process_store_credentials_error;

  if( ( X509_verify(p_x509_ds, p_pubkey_ds) ) <= 0 )
    goto process_store_credentials_error;

  /*
  ** Process Appliance Certificate
  */
  avoPrintf(ADSAP_ADHOC,"process_store_credentials: process appliance cert\n");
  cert_len_ap = ntohs(request_data.pAP_Cert->ADSAP2FieldHdr.wFieldLen);
  p_cert_data_ap = &request_data.pAP_Cert->byCert;

  p_pubkey_ap = get_pubkey_from_DER_cert(p_cert_data_ap, cert_len_ap, &p_x509_ap);

  if((p_pubkey_ap == NULL) || (p_x509_ap == NULL))
    goto process_store_credentials_error;

  if((X509_verify(p_x509_ap, p_pubkey_ds)) <= 0)
    goto process_store_credentials_error;

  /*
  ** Process Appliance Key
  */
  avoPrintf(ADSAP_ADHOC,"process_store_credentials: process appliance key\n");
  key_len_ap = ntohs(request_data.pAP_Priv_Key->ADSAP2FieldHdr.wFieldLen);

  if((p_privkey_ap = DER_read_PKCS8_PrivateKey_nopass(&request_data.pAP_Priv_Key->byKey, key_len_ap)) == NULL)
    goto process_store_credentials_error;

  /*
  ** Store DSView Certificate
  */
  avoPrintf(ADSAP_ADHOC,"process_store_credentials: store dsview cert\n");
	ds_cert.p_data = malloc(cert_len_ds);
  memcpy(ds_cert.p_data, (void *) &request_data.pDS_Cert->byCert, cert_len_ds);
  ds_cert.len = cert_len_ds;

  /*
  ** Store Appliance Certificate
  */
  avoPrintf(ADSAP_ADHOC,"process_store_credentials: store appliance cert\n");
	ap_cert.p_data = malloc(cert_len_ap);
  memcpy(ap_cert.p_data, (void *) &request_data.pAP_Cert->byCert, cert_len_ap);
  ap_cert.len = cert_len_ap;

  /*
  ** Store Appliance Key
  */
  avoPrintf(ADSAP_ADHOC,"process_store_credentials: store appliance key\n");
	ap_key.p_data = malloc(key_len_ap);
  memcpy(ap_key.p_data, (void *) &request_data.pAP_Priv_Key->byKey, key_len_ap);
  ap_key.len = key_len_ap;

#if 1
// Extract/store private key in PEM form. 
//The file will be saved in flash by store_credentials_to_file() below.
  {
    int mask = umask(077);
    FILE *fpem = fopen(APPLIANCE_RSA_PRIVATEKEY_FILE".pem", "w+");
    int retp = PEM_write_PKCS8PrivateKey(fpem, p_privkey_ap, 
                                         NULL, NULL, 0, NULL, NULL);
    fclose(fpem);
    umask(mask);
  }
#endif
  /*
  ** Update global pointers
  */
  store_credentials_to_file();

  *p_adsap2_mode = ADSAP2_MODE_ASSIGNED;

  goto process_store_credentials_exit;

process_store_credentials_error:
  avoPrintf(ADSAP_FAILURE,"process_store_credentials_error\n");
  status = ADSAP2_ERROR;

  if(p_cert_data_ds)
    free(p_cert_data_ds);
  if(p_cert_data_ap)
    free(p_cert_data_ap);
  if(p_key_data_ap)
    free(p_key_data_ap);

process_store_credentials_exit:
  avoPrintf(ADSAP_ADHOC,"process_store_credentials_exit\n");
  if(p_x509_ds != NULL)
    X509_free(p_x509_ds);
  if(p_x509_ap != NULL)
    X509_free(p_x509_ap);
  if(p_pubkey_ds != NULL)
    EVP_PKEY_free(p_pubkey_ds);

  build_ADSAP2_reply(p_header, ADSAP2_STORE_REP_STATUS, status);
}

//*****************************************************************************

void process_clear_credentials(P_ADSAP2_HEADER p_header, int *p_adsap2_mode)
{
  char  *p_bytes;
  WORD  status = ADSAP2_NOERROR;

  avoPrintf(ADSAP_ADHOC,"process_clear_credentials\n");
  p_bytes = (((unsigned char *) p_header) + sizeof(ADSAP2_HEADER));

  while(  (((PADSAP2FIELD) p_bytes)->byFieldID != ADSAP2_REC_TERM_VALUE) &&
      (((PADSAP2FIELD) p_bytes)->byFieldID != ADSAP2_TRAILER_VALUE) &&
      (status == ADSAP2_NOERROR))
  {
    switch(((PADSAP2FIELD) p_bytes)->byFieldID)
    {
      default:
      {
        goto process_clear_credentials_error;
      }
    }

    p_bytes += sizeof(ADSAP2FIELD) + ntohs(((PADSAP2FIELD) p_bytes)->wFieldLen);
  }

	switch(*p_adsap2_mode)		// clear credentials
	{
		case ADSAP2_MODE_TRUSTALL:
		{
			status = ADSAP2_INTRUST_ALL;
			goto process_clear_credentials_exit;
		}

		case ADSAP2_MODE_UNASSIGNED:
			goto process_clear_credentials_error;

		case ADSAP2_MODE_ASSIGNED:
			break;

		default:	// unknown mode
			goto process_clear_credentials_error;
	}

  clear_credentials();
  clear_credential_files();
  *p_adsap2_mode = ADSAP2_MODE_TRUSTALL;
  DsmibInitSecurityMode(DSMIB_SECURITY_MODE_TRUSTALL);

  goto process_clear_credentials_exit;

process_clear_credentials_error:
  avoPrintf(ADSAP_ADHOC,"process_clear_credentials_error\n");
  status = ADSAP2_ERROR;

process_clear_credentials_exit:
  build_ADSAP2_reply(p_header, ADSAP2_CLEAR_REP_STATUS, status);
}

//---------------------------------------------------------------------
//    FUNCTION: init_adsap2_credentials()
//
// DESCRIPTION: Load the certificates and private keys into the
//              SSL object.  If credentials have not been received
//              from the DSView Server, then we use the global
//              default credentials.
//
//  PARAMETERS: pointer to SSL object
//
//     RETURNS: TRUE if successful, FALSE otherwise
//
//       NOTES:
//
//---------------------------------------------------------------------
BOOLEAN init_adsap2_credentials(void  * p_ssl)
{
  EVP_PKEY  *p_pkey;
  int      use_ret;

  if( ADSAP2_credentials_present() )
  {
    avoPrintf(ADSAP_ADHOC, "init_adsap2_credentials load stored certs\n");
    /*
    ** Loads the stored appliance ASN1 encoded certificate
    */
    if( !SSL_use_certificate_ASN1((SSL *)p_ssl, ap_cert.p_data,
                                         ap_cert.len ) )
    {
      return(FALSE);
    }

    /*
    ** Extract the private key from the stored credentials
    */
    if( (p_pkey = DER_read_PKCS8_PrivateKey_nopass( ap_key.p_data,
                                          ap_key.len )) == NULL )
    {
      return(FALSE);
    }

    /*
    ** Adds the private key to the SSL object.
    */
    use_ret = SSL_use_PrivateKey((SSL *)p_ssl, p_pkey);
    EVP_PKEY_free(p_pkey);

    if( !use_ret )
    {
      return(FALSE);
    }
  }
  else
  {
    /*
    ** No credentials have been received from the DSView Server, so
    ** loads the default certificate stored in file into the SSL object.
    */
    avoPrintf(ADSAP_ADHOC, "init_adsap2_credentials load global certs\n");
    if(!SSL_use_certificate_file((SSL *)p_ssl, GLOBAL_APPLIANCE_CERTIFICATE_FILE, SSL_FILETYPE_PEM)) {
      avoPrintf(ADSAP_FAILURE, "SSL_use_certificate_file failed: %s\n", GLOBAL_APPLIANCE_CERTIFICATE_FILE);
      return(FALSE);
    }

    /*
    ** Adds the private key found in file to ctx.
    */
    if(!SSL_use_PrivateKey_file((SSL *)p_ssl, GLOBAL_APPLIANCE_RSA_PRIVATEKEY_FILE, SSL_FILETYPE_PEM)) {
      avoPrintf(ADSAP_FAILURE, "SSL_use_PrivateKey_file failed: %s\n", GLOBAL_APPLIANCE_RSA_PRIVATEKEY_FILE);
      return(FALSE);
    }
  }

  /*
  ** Checks the consistency of a private key with the corresponding
  ** certificate loaded into SSL object.
  */
  if(SSL_check_private_key((SSL *)p_ssl))
    return(TRUE);
  else {
    avoPrintf(ADSAP_FAILURE, "SSL_check_private_key failed\n");
    return(FALSE);
  }
}


//*****************************************************************************

void *accept_ssl(int sock, int adsap2_mode)
{
  SSL          *p_ssl = NULL;

  if( !SSL_Init_ctx(&g_ssl_ctx_cert_server,
                     SSL_CONTEXT_CERT,
                     SSL_SERVER, "RC4-MD5"))
  {
    if(p_ssl)
    {
      SSL_free(p_ssl);
      p_ssl = NULL;
    }
  }

  SSL_SetDsCertCallback(g_ssl_ctx_cert_server, NULL);
  p_ssl = SSL_Server_Handshake( g_ssl_ctx_cert_server, sock,
                              SSL_CONTEXT_CERT );
  return( p_ssl );
}

//*****************************************************************************

SSL *connect_ssl(int sock, int adsap2_mode)
{
  SSL          *p_ssl = NULL;

	avoPrintf(ADSAP_INFO,"connect_ssl\n");
  if( !SSL_Init_ctx (& g_ssl_ctx_cert_client,
                       SSL_CONTEXT_CERT,
                       SSL_CLIENT, "RC4-MD5"))
  {
    if(p_ssl)
    {
      SSL_free(p_ssl);
      p_ssl = NULL;
    }
  }

  SSL_SetDsCertCallback(g_ssl_ctx_cert_client, NULL);
  p_ssl = SSL_Client_Handshake( g_ssl_ctx_cert_client, sock,
                              SSL_CONTEXT_CERT );
  return( p_ssl );
}


//*****************************************************************************


X509 *PEM_read_file_X509(const char *filename)
{
  BIO    *p_bio;
  X509  *p_x509  = NULL;

  if((p_bio = BIO_new(BIO_s_file_internal())) == NULL)
    goto PEM_read_file_X509_exit;

  if(BIO_read_filename(p_bio, filename) <= 0)
    goto PEM_read_file_X509_exit;

  p_x509 = PEM_read_bio_X509(p_bio, NULL, NULL, NULL);

PEM_read_file_X509_exit:
  if(p_bio != NULL)
    BIO_free(p_bio);

  return(p_x509);
}

BOOLEAN test_ds_credential(unsigned char *p_cert_data, unsigned long cert_len, unsigned char *p_test_data, unsigned long test_len)
{
	unsigned long i;

	avoPrintf(ADSAP_ADHOC,"test_ds_credential\n");
	if(cert_len != test_len)
	{
		avoPrintf(ADSAP_WARNING,"test_ds_credential length mismatch\n");
		return(FALSE);
	}

	for(i=0; i < cert_len; i++)
	{
		if(*(p_cert_data + i) != *(p_test_data + i))
		{
			avoPrintf(ADSAP_WARNING,"test_ds_credential data mismatch\n");
			return(FALSE);
		}
	}

	return(TRUE);
}
//---------------------------------------------------------------------
//    FUNCTION: verify_ds_credentials()
//
// DESCRIPTION: Callback function to verify certificate
//
//  PARAMETERS:
//
//
//     RETURNS: none
//
//       NOTES: Compares the received credentials against the ones
//              stored in the appliance.
//
//              The data parameter is ignored in version 9.6 of
//              OpenSSL, so I had to get the socket using other means
//              -- Jamie
//---------------------------------------------------------------------
int verify_ds_credentials(void *p_store, void *data)
{
  int          ret_val = 1;

  int          adsap2_mode;

  X509        *p_x509_test;
  unsigned char    test_data[2048];
  unsigned char    *p_test_data = test_data;
  unsigned long    test_len;

  X509        *p_x509_stored;
  unsigned char    stored_data[2048];
  unsigned char    *p_stored_data = stored_data;
  unsigned long    stored_len;
  char *adsap2_mode_strings[3] = {"Trust All", "Unassigned", "Assigned"};

  avoPrintf(ADSAP_ADHOC,"verify_ds_credentials\n");
  adsap2_mode = get_ADSAP2_mode();

  if(adsap2_mode == ADSAP2_MODE_TRUSTALL)
    goto verify_ds_credentials_exit;

  // Convert certificate to DER format
  if((p_x509_test = ((X509_STORE_CTX *)p_store)->cert) == NULL)
    goto verify_ds_credentials_error;

  if((test_len = i2d_X509(p_x509_test, &p_test_data)) == 0)
    goto verify_ds_credentials_error;
  else
    p_test_data = test_data;

  if(adsap2_mode == ADSAP2_MODE_ASSIGNED)
  {
    p_stored_data  = ds_cert.p_data;
    stored_len    = ds_cert.len;
  }
  else
  {
    if((p_x509_stored = PEM_read_file_X509(GLOBAL_DSVIEW_CERTIFICATE_FILE)) == NULL)
      goto verify_ds_credentials_error;

  // Convert certificate to DER format
    if((stored_len = i2d_X509(p_x509_stored, &p_stored_data)) == 0)
      goto verify_ds_credentials_error;
    else
      p_stored_data = stored_data;
  }

  if((p_stored_data == NULL) || (stored_len == 0))
    goto verify_ds_credentials_error;

  if(test_ds_credential(p_stored_data, stored_len, p_test_data, test_len))
    goto verify_ds_credentials_exit;

verify_ds_credentials_error:
  ret_val = 0;

verify_ds_credentials_exit:
  avoPrintf(ADSAP_ADHOC,"verify_ds_credentials (%s) mode=%s\n", ((ret_val) ? "PASSED" : "FAILED"), adsap2_mode_strings[adsap2_mode]);
  return(ret_val);
}

//---------------------------------------------------------------------
//    FUNCTION: verify_preauth_credentials()
//
// DESCRIPTION: Callback function to verify certificate
//
//  PARAMETERS:
//
//
//     RETURNS: TRUE
//
//       NOTES: No verification done here; the pre-auth cert will be
//              verified after the AVSP login message is received.
//---------------------------------------------------------------------
int verify_preauth_credentials(void *p_store, void *data)
{
  avoPrintf(ADSAP_INFO,"verify_preauth_credentials\n");
  return TRUE;
}


//*****************************************************************************
//  Functions to modify for portability

//---------------------------------------------------------------------
// FUNCTION     : load_file
//
// DESCRIPTION  : Load credential from flash file into a buffer.
//
// PARAMETERS   : destination buffer, filename
//
// RETURNS      : length of file, 0 if file not found
//
// NOTES        :
//
//---------------------------------------------------------------------
int load_file( int * buffer, char * filename )
{
  FILE * fp = NULL;
  int len = 0;

  fp = fopen( filename, "r" );
  if ( fp )
  {
    /*
    ** Find the file length
    */
    fseek( fp, 0, SEEK_END ); /* end of file */
    len = ftell( fp );
    fseek( fp, 0, SEEK_SET ); /* beginning of file */
    avoPrintf(ADSAP_ADHOC, "%s file length is %d\n", filename, len);

    /*
    ** Read the file into the buffer
    */
    fread( buffer, 1, len, fp );
    fclose( fp );
  }
  else
  {
    avoPrintf(ADSAP_FAILURE ,"%s file not found\n", filename);
  }
  return len;
}

//---------------------------------------------------------------------
// FUNCTION     : ADSAP2_authorize
//
// DESCRIPTION  : Authorize a user to access a specific target
//
// PARAMETERS   : username, password, mode, address levels
//                placeholders to receive the role and preemptLevel
//
// RETURNS      : success, failed, or unavailable
//
// NOTES        : RIP Id and mode may be blank if this is an attempt
//                to authenticate a username/password only.
//
//
//---------------------------------------------------------------------
int ADSAP2_authorize(	ADSAP2_AUTHORIZE_REQ * authorizeRequest, ADSAP2_AUTHORIZE_REP * authorizeReply)
{
  void * sslClient;
  int i,j;
  int sock;
  unsigned long cred_addr;
  char IPAddrStr[INET_ADDRSTRLEN];
  WORD status;
  unsigned long	ret_val = ADSAP2_DSAUTHENTICATION_UNAVAILABLE;
  BOOLEAN finished = FALSE;

  avoPrintf(ADSAP_INFO, "ADSAP2_authorize %s\n", authorizeRequest->username);

  /*
  ** Find a DSView Server to authenticate against
  */
  for(j = 0; ((j < DS_AUTH_RETRIES) && !(finished)); j++)
  {
    avoPrintf(ADSAP_INFO,"DS User Authentication - Pass #%d\n", j+1);

    for(i = 0; ((i < DSMIB_AUTHSERVER_ADDRESS_COUNT) && !(finished)); i++)
    {
      DsmibGetAuthServerAddress( i, (IP_ADDR *)&cred_addr );

      if(cred_addr == 0)
      {
        avoPrintf(ADSAP_WARNING, "ERROR: No DS IP Address found\n");
        continue;
      }
      sprintf( IPAddrStr, "%ld.%ld.%ld.%ld",
                  (cred_addr >> 24) & 0xff,
                  (cred_addr >> 16) & 0xff,
                  (cred_addr >> 8) & 0xff,
                  cred_addr & 0xff);

    if( (sock = TCP_Connect(ADSAP2_LISTEN_PORT, IPAddrStr, DS_AUTH_TIMEOUT)) <= 0 )
      {
        continue;
      }

      if( ( sslClient = connect_ssl(sock, authorizeRequest->mode) ) )
      {
        status = send_ADSAP2_auth_request( sslClient,
                                authorizeRequest,
                                authorizeReply );

        avoPrintf(ADSAP_INFO, "send_ADSAP2_auth_request status = %d\n", status);
        if( status == ADSAP2_NOERROR )
        {
          ret_val = ADSAP2_DSAUTHENTICATION_SUCCESS;
          SSL_Close_Connection(sslClient);
          finished = TRUE; // break out of loop
        }
        else
        {
          ret_val = ADSAP2_DSAUTHENTICATION_FAILED;
          SSL_Close_Connection(sslClient);
          continue; // try the next server
        }
      }
      else
      {
        Close(sock);
      }
    } // end multiple-auth address loop
  } // end retry loop

  switch (ret_val)
  {
    case ADSAP2_DSAUTHENTICATION_SUCCESS:
      avoPrintf(ADSAP_INFO, "ADSAP2_authorize PASSED\n\n");
      break;
    case ADSAP2_DSAUTHENTICATION_FAILED:
      avoPrintf(ADSAP_INFO, "ADSAP2_authorize FAILED\n\n");
      break;
    case ADSAP2_DSAUTHENTICATION_UNAVAILABLE:
      avoPrintf(ADSAP_INFO, "ADSAP2_authorize UNAVAILABLE\n\n");
      break;
  }
  return(ret_val);
}

//---------------------------------------------------------------------
// FUNCTION     : authenticate_ds_local
//
// DESCRIPTION  : Authenticate a user against the pre-auth list
//
// PARAMETERS   :
//
// RETURNS      :
//
// NOTES        :
//
//---------------------------------------------------------------------

static unsigned long authenticate_ds_local(ADSAP2_AUTHORIZE_REQ *pAuthInfo, unsigned char *p_key_data, int key_len, ADSAP2_AUTHORIZE_REP *pAuthReply)
{
	unsigned long			ret_val = ADSAP2_DSAUTHENTICATION_UNAVAILABLE;
	P_ADSAP2_PREAUTH_CERT	p_new_cert	= NULL;
	P_ADSAP2_PREAUTH_CERT	p_list_cert	= NULL;
	PLIST_ITEM				p_item		= NULL;	
	BOOLEAN					locked		= TRUE;
	int						name_len;

	if(preauth_cert_list == NULL)
		goto authenticate_ds_local_exit;		

	if((p_new_cert = new_preauth_cert(p_key_data, (unsigned short) key_len, ADSAP2_KEY_DATA)) == NULL)
		goto authenticate_ds_local_free;

	p_new_cert->req_port = pAuthInfo->port;

	if((name_len = strlen(pAuthInfo->username)) > ADSAP2_USERNAME_STR_LEN)
		name_len = ADSAP2_USERNAME_STR_LEN;

	strncpy(p_new_cert->username, pAuthInfo->username, name_len);
	p_new_cert->username[name_len] = '\0';

	lock_unlock_list(preauth_cert_list, TRUE);

	if((p_item = find_list_item(preauth_cert_list, preauth_cert_list->list_head, (void *) p_new_cert, TRUE, ADSAP2_KEY_DATA, locked)) == NULL) {
		avoPrintf(ADSAP_INFO, "authenticate_ds_local Failed\n\n");
		goto authenticate_ds_local_free_rel_mutex;
	}

	ret_val = ADSAP2_DSAUTHENTICATION_SUCCESS;		

	if (pAuthInfo->testLocalKey)
		goto authenticate_ds_local_free_rel_mutex;

	avoPrintf(ADSAP_INFO, "authenticate_ds_local PASSED\n\n");
	p_list_cert = (P_ADSAP2_PREAUTH_CERT) p_item->p_data;

    pAuthReply->preempt= p_list_cert->preempt;
    avoPrintf(ADSAP_INFO, "authenticate_ds_local: preemption level= %d\n", pAuthReply->preempt); // Adam

    // Convert user role string to enum
    pAuthReply->role = userRole2Enum(p_list_cert->role);
	avoPrintf(ADSAP_INFO, "authenticate_ds_local: role[%s][%d]\n",p_list_cert->role, pAuthReply->role);

    pAuthReply->logdla = p_list_cert->logdla;
    avoPrintf(ADSAP_INFO, "authenticate_ds_local: logdla=%d\n", pAuthReply->logdla);

	memcpy(pAuthReply->RIP_ID, p_list_cert->RIP_ID, sizeof(pAuthReply->RIP_ID));
    avoPrintf(ADSAP_INFO, "authenticate_ds_local: RIP=%s\n", pAuthReply->RIP_ID);

	memcpy(pAuthReply->command, p_list_cert->command, sizeof(pAuthReply->command));
    avoPrintf(ADSAP_INFO, "authenticate_ds_local: CMD=%s\n", pAuthReply->command);

	del_list_item(preauth_cert_list, p_item, locked);

authenticate_ds_local_free_rel_mutex:
	lock_unlock_list(preauth_cert_list, FALSE);

authenticate_ds_local_free:
	if(p_new_cert != NULL)
		free_preauth_cert(&p_new_cert);

authenticate_ds_local_exit:
	return(ret_val);
}

//---------------------------------------------------------------------
// FUNCTION     : authenticate_ds
//
// DESCRIPTION  : Authenticate a user against the DSView server if
//                username and password was given or local list if
//                a public key (Blob) was given.
//
// PARAMETERS   :
//
// RETURNS      :
//
// NOTES        :
//
//---------------------------------------------------------------------

unsigned long authenticate_ds(ADSAP2_AUTHORIZE_REQ * pAuthReq, ADSAP2_AUTHORIZE_REP * pAuthRep)
{
	unsigned long			ret_val = ADSAP2_DSAUTHENTICATION_UNAVAILABLE;
	unsigned char			key_data[1024];
	unsigned char			*p_key_data	= NULL;
	int						key_len = 0;
	RSA						*rsa;

    avoPrintf(ADSAP_FAILURE, "###%s pAI->UN=%s pAI->mode=%d pAI->p_Blob=%p pAI->Blob_len=%d pAI->port=%d\n",
		__FUNCTION__, pAuthReq->username, pAuthReq->mode,
		pAuthReq->p_Blob, pAuthReq->Blob_len, pAuthReq->port);

	if(pAuthReq->p_Blob != NULL)
	{
		p_key_data = key_data;

		rsa = RSA_from_blob(pAuthReq->p_Blob, pAuthReq->Blob_len);

		if((key_len = i2d_RSAPublicKey(rsa, &p_key_data)) == 0)
			goto authenticate_ds_exit;

		p_key_data = key_data;	

		ret_val = authenticate_ds_local(pAuthReq, p_key_data, key_len, pAuthRep);
	}

	if(ret_val != ADSAP2_DSAUTHENTICATION_SUCCESS)
	{
		if (pAuthReq->password[0]) {
			ret_val = ADSAP2_authorize(pAuthReq, pAuthRep);	
		}
	}

authenticate_ds_exit:
    avoPrintf(ADSAP_INFO, "DS Authentication - %s\n\n",
		((ret_val == ADSAP2_DSAUTHENTICATION_SUCCESS) ? "PASSED" : "FAILED"));

	return(ret_val);
}

//---------------------------------------------------------------------
// FUNCTION     : ADSAP2_initialize_globals
//
// DESCRIPTION  : Load credentials into memory from flash file, if
//                they exist.
//
// PARAMETERS   : None
//
// RETURNS      : none
//
// NOTES        :
//
//---------------------------------------------------------------------
void ADSAP2_initialize_globals()
{
  int len = 0;
  int buffer [ 4096 ];

  avoPrintf(ADSAP_INFO, "ADSAP2_initialize_globals\n");

  /*
  ** Initialize ADSAP2 linked list
  */
  if( preauth_cert_list == NULL )
		preauth_cert_list = (PGLOBAL_LIST) ADSAP2_initialize_list();
  if ( preauth_cert_list == NULL )
    avoPrintf(ADSAP_FAILURE, "Error initializing linked list\n");

  /*
  ** Initialize global credentials
  */
  memset(& ds_cert, 0, sizeof(ds_cert) );
  memset(& ap_cert, 0, sizeof(ap_cert) );
  memset(& ap_key, 0, sizeof(ap_key) );

  /*
  ** Load stored credentials from flash file, if available
  */
  memset(buffer, 0, 4096);
  len = load_file( buffer, DSVIEW_CERTIFICATE_FILE );
  if ( len )
  {
		ds_cert.p_data = malloc(len);
		memcpy(ds_cert.p_data, buffer, len);
		ds_cert.len = len;
    avoPrintf(ADSAP_INFO, "DSView Certificate file initialized\n");
  }

  memset(buffer, 0, 4096);
  len = load_file( buffer, APPLIANCE_CERTIFICATE_FILE );
  if ( len )
  {
		ap_cert.p_data = malloc(len);
		memcpy(ap_cert.p_data, buffer, len);
		ap_cert.len = len;
    avoPrintf(ADSAP_INFO, "Appliance Certificate file initialized\n");
  }

  memset(buffer, 0, 4096);
  len = load_file( buffer, APPLIANCE_RSA_PRIVATEKEY_FILE );
  if ( len )
  {
		ap_key.p_data = malloc(len);
		memcpy(ap_key.p_data, buffer, len);
		ap_key.len = len;
    avoPrintf(ADSAP_INFO, "Appliance Private Key file initialized\n");
  }

  // Initialize session mode lookup table used to convert session
  // mode string to enum
  memset(sessionModeTbl, 0, ADSAP2_MODE_MAX*ADSAP2_USERNAME_STR_LEN);
  strcpy(sessionModeTbl[ADSAP2_NORMAL],            "NORMAL");
  strcpy(sessionModeTbl[ADSAP2_EXCLUSIVE],         "EXCLUSIVE");
  strcpy(sessionModeTbl[ADSAP2_PREEMPT],           "PREEMPT");
  strcpy(sessionModeTbl[ADSAP2_PREEMPT_EXCLUSIVE], "PREEMPT_EXCLUSIVE");
  strcpy(sessionModeTbl[ADSAP2_STEALTH],           "STEALTH");
  strcpy(sessionModeTbl[ADSAP2_SHARE],             "SHARE");
  strcpy(sessionModeTbl[ADSAP2_SHARE_FORCE],       "SHARE_FORCE");
  strcpy(sessionModeTbl[ADSAP2_SHARE_PASSIVE],     "SHARE_PASSIVE");
  strcpy(sessionModeTbl[ADSAP2_SCAN],              "SCAN");
  strcpy(sessionModeTbl[ADSAP2_SHARE_PASSIVE_FORCE], "SHARE_PASSIVE_FORCE");

  // Initialize user role lookup table used to convert user role
  // string to enum
  memset(userRoleTbl, 0, ADSAP2_USER_ROLE_MAX*ADSAP2_USER_ROLE_STR_LEN);
  strcpy(userRoleTbl[ADSAP2_APPLIANCE_ADMIN],      "APPLIANCE ADMIN");
  strcpy(userRoleTbl[ADSAP2_USER_ADMIN],           "USER ADMIN");
  strcpy(userRoleTbl[ADSAP2_USER],                 "USER");

  // Initialize user role lookup table used to convert user role
  // string to enum
  memset(preauthTypeTbl, 0, ADSAP2_PREAUTH_TYPE_MAX*ADSAP2_PREAUTH_TYPE_STR_LEN);
  strcpy(preauthTypeTbl[ADSAP2_PREAUTH_TYPE_VIDEO],         "VIDEO");
  strcpy(preauthTypeTbl[ADSAP2_PREAUTH_TYPE_VIRTUAL_MEDIA], "VIRTUAL_MEDIA");
}

static int adsap2ETHConnectSocket = 0;

//---------------------------------------------------------------------
//    FUNCTION: void * adsap2CmdThread( void *threadData )
//
// DESCRIPTION: ADSAP2 Command processing thread
//              
//							
//  PARAMETERS: 
//							
//     RETURNS: 
//
//       NOTES: 
//---------------------------------------------------------------------
static void * adsap2CmdThread( void * threadData )
{
  void * ssl = threadData;
  int adsap2Mode;

  pthreadDetach( pthread_self() );
  
  adsap2Mode = get_ADSAP2_mode();
  processADSAP2Connect(ssl, &adsap2Mode);
  SSL_Close_Connection(ssl);

  return 0;
}

static void adsap2ConnectServerExit(void *foo)
{
	close(adsap2ETHConnectSocket);
}

//---------------------------------------------------------------------
//    FUNCTION: void * adsap2ConnectServer( void *threadInfo )
//
// DESCRIPTION: ADSAP2 connection listener thread.  
//              
//							
//  PARAMETERS: 
//							
//     RETURNS: 
//
//       NOTES: 
//---------------------------------------------------------------------

void * adsap2ConnectServer( void * threadarg )
{
  int adsap2RecvSocket;
  int client_len;
  struct sockaddr_in client_address;
  pthread_t adsap2CmdThreadID;
  static char myIPAddrStr[INET_ADDRSTRLEN];
  static byte myIPAddr[4];
  void * ssl;
  struct in_addr peerAddr;
  unsigned char	dbgName[64];
  char *intf = (char *)threadarg;

  ADSAP2_initialize_globals();
  
  pthread_cleanup_push(&adsap2ConnectServerExit, NULL);

  /*
  ** Get the appliance's IP address
  */
  while (1)
  {
    if ( getIpAddress ( myIPAddr, intf ) < 0 )
    {
      avoPrintf(1, "Could not get my IP address, will try again in 10 seconds.\n");
      sleep (10);
      continue;
    }
    sprintf( myIPAddrStr, "%u.%u.%u.%u", myIPAddr[0], myIPAddr[1], myIPAddr[2], myIPAddr[3] );
    break;    
  }
  
  /*
  ** Listen for the ADSAP2 Connection
  */
  adsap2ETHConnectSocket = TCP_Listen(ADSAP2_LISTEN_PORT, (char *)myIPAddrStr);

  if (adsap2ETHConnectSocket < 0) {
    avoPrintf(ADSAP_INFO, "adsap2ConnectServer: Error opening ADSAP2 socket\n");
    kill(0, SIGTERM);
    return(0);
  }
    
  while ( 1 )
  {
    /*
    ** Wait for a ADSAP2 client to connect.
    */
    avoPrintf(ADSAP_INFO, "Waiting for ADSAP2 connection...\n");
    client_len = sizeof(client_address);
    adsap2RecvSocket = Accept( adsap2ETHConnectSocket, 
                           (struct sockaddr *)&client_address,
                           &client_len);                    

    GetPeerName(adsap2RecvSocket, & peerAddr );
    sprintf(dbgName, "Received ADSAP2 connection from %d.%d.%d.%d",
              (peerAddr.s_addr >>24) & 0xff,
              (peerAddr.s_addr >>16) & 0xff,
              (peerAddr.s_addr >>8) & 0xff,
              peerAddr.s_addr & 0xff);
    avoPrintf(ADSAP_INFO, "%s\n", dbgName);
    
    ssl = accept_ssl(adsap2RecvSocket, get_ADSAP2_mode());
    if (ssl == NULL)
    {
      avoPrintf(ADSAP_FAILURE, "Could not accept SSL \n");
      Close(adsap2RecvSocket);
      continue;
    }

    /*
    ** Spawn ADSAP2 command thread
    */
    pthreadCreate( & adsap2CmdThreadID, NULL, & adsap2CmdThread, (void *) ssl );
    usleep( 100000 );
  }

  pthread_cleanup_pop(1);

  return 0;
}

/* end */
