//------------------------------------------------------------------------
// Copyright 2000-2002, Avocent Corporation
//------------------------------------------------------------------------
// FILE        :  netwrap.c
// DESCRIPTION :  This file contains wrappers for common network 
//                programming calls
//
//------------------------------------------------------------------------
// AUTHOR      :  J. Collins
// DATE        :  
//
//------------------------------------------------------------------------
#include <stdio.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <netdb.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/utsname.h>
#include <errno.h>
#include <netinet/tcp.h>
#include <net/if_arp.h>
#include <net/if.h>
#include <signal.h>

#include "netwrap.h"
#include "types.h"
#include "avoprint.h"

#define LISTENQ  20



//---------------------------------------------------------------------
//    FUNCTION:  Socket
//
// DESCRIPTION: Create an endpoint for communication and return a
//              descriptor
//							
//  PARAMETERS: protocol family
//              type of socket
//              protocol to be used with socket
//
//     RETURNS: socket descriptor
//
//       NOTES: 
//
//---------------------------------------------------------------------
int Socket(int family, int type, int protocol)
{
  int result;

  if( ( result = socket(family, type, protocol) ) < 0 )
  {
    avoPrintf(1, "Socket: error opening socket: %s \n", strerror(errno));
    exit(0);
  }
  return (result);
}

//---------------------------------------------------------------------
//    FUNCTION: Close 
//
// DESCRIPTION: Close a file descriptor (socket)
//							
//  PARAMETERS: sockfd - file descriptor
//
//     RETURNS: 0 for success, or error code
//
//       NOTES: 
//
//---------------------------------------------------------------------
int Close(int sockfd)
{
  int result;
	struct linger	linger;

  /* Don't close stdin, stdout, or stderr */
  if ( sockfd <= 2 )
  {
    avoPrintf(1, "Close: Shouldn't close fd %d!! \n", sockfd);
    return -1;
  }
	
  /*
  ** This will cause TCP to abort the connection and discard any data
  ** still remaining in the socket send buffer.  It will send a RST to
  ** the peer.
  */
  linger.l_onoff = 1;
	linger.l_linger = 0;
	setsockopt(sockfd, SOL_SOCKET, SO_LINGER, (void *)&linger, sizeof(linger));

  if( (result = close( sockfd )) < 0 )
  {
    avoPrintf(1, "Close: error closing socket: %s \n", strerror(errno));
  }
  return (result);
}

//---------------------------------------------------------------------
//    FUNCTION: ClosePipe 
//
// DESCRIPTION: Close both pipe file descriptors
//							
//  PARAMETERS: fds
//
//     RETURNS: 0 for success, or -1 for error
//
//       NOTES: 
//
//---------------------------------------------------------------------
int ClosePipe( int fds[2] )
{
  int n;
  int retVal = 0;

  for ( n = 0; n < 2; ++n ) 
  {
    if ( fds[n] <= 2 ) 
    {
      avoPrintf(1, "ClosePipe: Shouldn't close fd %d!! \n", fds[n] );
      retVal = -1;
    }
    else 
    {
      if ( close( fds[n] ) < 0 ) 
      {
        avoPrintf(1, "ClosePipe: error closing pipe fd %d: %s \n", fds[n], strerror(errno));
        retVal = -1;
      }
    }
    fds[n] = 0;   /* RESET fd to zero after close */
  }
  return retVal;
}

//---------------------------------------------------------------------
//    FUNCTION: Setsockopt
//
// DESCRIPTION: Set a socket option
//							
//  PARAMETERS: 
//
//     RETURNS: 
//
//       NOTES: 
//
//---------------------------------------------------------------------
int Setsockopt(int sockfd, int level, int optname, const void *optval, socklen_t optlen)
{
int result;

  if( (result = setsockopt(sockfd, level, optname, optval, optlen)) < 0 )
  {
    avoPrintf(1, "Setsockopt: error setting socket option: %s \n", strerror(errno));
  }
  return (result);
}

//---------------------------------------------------------------------
//    FUNCTION: Accept
//
// DESCRIPTION: Accept a connection on a socket
//							
//  PARAMETERS:
//
//     RETURNS: Socket descriptor
//
//       NOTES: 
//
//---------------------------------------------------------------------
int Accept(int fd, struct sockaddr *sa, socklen_t *salenptr)
{
int n;

  if ( (n = accept(fd, sa, salenptr)) < 0)
  {
    avoPrintf(1, "Accept: error: %s \n", strerror(errno));
    return -1;
  }
  
  return(n);
}

//---------------------------------------------------------------------
//    FUNCTION: Bind
//
// DESCRIPTION: Bind a socket to an IP address
//							
//  PARAMETERS: 
//
//     RETURNS: -1 if bind fails; 0 otherwise
//
//       NOTES: 
//
//---------------------------------------------------------------------
int Bind(int fd, const struct sockaddr *sa, socklen_t salen)
{
  if (bind(fd, sa, salen) < 0)
  {
    avoPrintf(1, "Bind: error: %s \n", strerror(errno));
    return -1;
  }
  return 0;
}

//---------------------------------------------------------------------
//    FUNCTION: Connect
//
// DESCRIPTION: Initiate a connection on a socket
//							
//  PARAMETERS:
//
//     RETURNS: 
//
//       NOTES: 
//
//---------------------------------------------------------------------
int Connect(int fd, const struct sockaddr *sa, socklen_t salen)
{
  if (connect(fd, sa, salen) < 0)
  {
    avoPrintf(1, "Connect: error: %s \n", strerror(errno));
    return(-1);
  }
  return 0;
}

static void connectAlarm ( int signo )
{
  //avoPrintf(1, "Connect timed out!\n");
  return;
}
//---------------------------------------------------------------------
//    FUNCTION:  ConnectTimeout
//
// DESCRIPTION:  Initiate a connection on a socket, using an alarm
//               to allow the connect to timeout.
//							
//  PARAMETERS:
//
//     RETURNS: 
//
//       NOTES: 
//
//---------------------------------------------------------------------
int ConnectTimeout(int fd, const struct sockaddr *sa, socklen_t salen, int nsec)
{
  int n;
  struct sigaction act;

  /*
  ** Set up the signal handler.
  */
  act.sa_handler = connectAlarm;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;
  sigaction ( SIGALRM, &act, 0 );
  
  if ( alarm(nsec) != 0 )
  {
    avoPrintf(1, "ConnectTimeout: warning, alarm was already set\n");
  }
  
  if ( ( n = Connect ( fd, sa, salen) ) < 0 )
  {
    close ( fd );
    if ( errno == EINTR )
    {
      errno = ETIMEDOUT;
    }
    else
    {
      avoPrintf(1, "Connect error: %s \n", strerror(errno));
    }
  }
 
  alarm(0);
  return (n);
}


//---------------------------------------------------------------------
//    FUNCTION: Listen
//
// DESCRIPTION: Listen for connections on a socket
//							
//  PARAMETERS:
//
//     RETURNS: 
//
//       NOTES: 
//
//---------------------------------------------------------------------
void Listen(int fd, int backlog)
{
  char *ptr;

  /*
  ** Can override 2nd argument with environment variable
  */
  if( (ptr = getenv("LISTENQ")) != NULL)
  {
    backlog = atoi(ptr);
  }
  
  if (listen(fd, backlog) < 0)
  {
    avoPrintf(1, "Listen: error: %s \n", strerror(errno));
    exit(0);
  }
}

//---------------------------------------------------------------------
//    FUNCTION: TCP_Listen
//
// DESCRIPTION: Create a TCP socket, bind it to an IP address, and
//              listen for a connection.
//							
//  PARAMETERS:
//
//     RETURNS: The socket descriptor that is listening for a connection
//              or -1 if error
//
//       NOTES: 
//
//---------------------------------------------------------------------
int TCP_Listen ( int port, char * ip_address )
{
int listenfd;
struct sockaddr_in listen_address;
int optval = 1;

  listenfd = Socket(AF_INET, SOCK_STREAM, 0);
  
  memset(&listen_address, 0, sizeof(listen_address));
  listen_address.sin_family = AF_INET;
  
  if (!ip_address)
  {
    listen_address.sin_addr.s_addr = INADDR_ANY;
  }
  else
  {
    listen_address.sin_addr.s_addr = inet_addr(ip_address);
  }
  listen_address.sin_port = htons(port);
  
  Setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval));

  if ( Bind ( listenfd, (struct sockaddr *)&listen_address,
         sizeof(listen_address)) < 0 )
  {
    close(listenfd);
    return -1;
  }
  
  Listen(listenfd, 20);
  avoPrintf(2, "TCP Server listening at port %d ...\n", port);
 
  return (listenfd);
}

//---------------------------------------------------------------------
//    FUNCTION: TCP_Connect
//
// DESCRIPTION: Create a TCP socket and initiate a connection
//							
//  PARAMETERS:
//
//     RETURNS: Socket descriptor of the connected socket, or 
//              -1 for an error
//
//       NOTES: 
//
//---------------------------------------------------------------------
int TCP_Connect ( int port, char * ip_address, int timeout )
{
int sockfd;
struct sockaddr_in address;
//int maxseg = 128;

  avoPrintf(2, "Connecting to IP address %s, port %d.\n", ip_address, port);
  
  sockfd = Socket(AF_INET, SOCK_STREAM, 0);
  address.sin_family = AF_INET;
  address.sin_addr.s_addr = inet_addr(ip_address);
  address.sin_port = port;

  /* Jamie debug - set the maximum TCP segment size*/  
  //Setsockopt(sockfd, IPPROTO_TCP, TCP_MAXSEG, &maxseg, sizeof(maxseg));
  
  if ( ConnectTimeout(sockfd, (struct sockaddr *)&address, sizeof(address), timeout) < 0)
  {
    return (-1);
  }  
  return(sockfd);
}


//---------------------------------------------------------------------
//    FUNCTION: UNIX_Listen
//
// DESCRIPTION: Create a unix socket and listen on that socket
//							
//  PARAMETERS:
//
//     RETURNS: descriptor of listening socket
//
//       NOTES: 
//
//---------------------------------------------------------------------
int UNIX_Listen ( char * filename )
{
int listenfd;
struct sockaddr_un listen_address;

  unlink(filename);
  listenfd = Socket(AF_UNIX, SOCK_STREAM, 0);
  listen_address.sun_family = AF_UNIX;
  strcpy(listen_address.sun_path, filename);
  Bind(listenfd, (struct sockaddr *)&listen_address, sizeof(listen_address));
  Listen(listenfd, 5);

  return (listenfd);
}

//---------------------------------------------------------------------
//    FUNCTION:  UNIX_Connect
//
// DESCRIPTION: Create a UNIX socket and initate a connection
//							
//  PARAMETERS:
//
//     RETURNS: socket descriptor
//
//       NOTES: 
//
//---------------------------------------------------------------------
int UNIX_Connect ( char * filename )
{
  int sockfd;
  struct sockaddr_un address;

  sockfd = Socket(AF_UNIX, SOCK_STREAM, 0);
  address.sun_family = AF_UNIX;
  strcpy(address.sun_path, filename);
  Connect(sockfd, (struct sockaddr *)&address, sizeof(address));  
  
  return(sockfd);
}

//---------------------------------------------------------------------
//    FUNCTION: Send
//
// DESCRIPTION: Send a message from a socket
//							
//  PARAMETERS:
//
//     RETURNS: Number of bytes sent
//
//       NOTES: 
//
//---------------------------------------------------------------------
int Send(int fd, const void *ptr, size_t nbytes, int flags)
{
  int n;

  if ( (n = send(fd, ptr, nbytes, flags)) <= 0)
  {
    avoPrintf(1, "Send: error: %s \n", strerror(errno));
    return(-1);
  }
  return(n);
}

//---------------------------------------------------------------------
//    FUNCTION: Sendn
//
// DESCRIPTION: Send n bytes from a socket
//							
//  PARAMETERS:
//
//     RETURNS: Number of bytes sent
//
//       NOTES: 
//
//---------------------------------------------------------------------
ssize_t Sendn(int fd, const void *vptr, size_t n, int flags)
{
  size_t nleft;
  ssize_t nsent;
  const char *ptr;
  
  ptr = vptr;
  nleft = n;
  while (nleft > 0)
  {
    if ( (nsent = send(fd, ptr, nleft, flags)) <= 0)
    {
      if (errno == EINTR)
      {
        nsent = 0;
      }
      else
      {
        avoPrintf(1, "Sendn: error: %s \n", strerror(errno));
        return (-1);
      }
    }
    nleft -= nsent;
    ptr += nsent;
  }
  return (n);
}

//---------------------------------------------------------------------
//    FUNCTION: setRecvTimeout
//
// DESCRIPTION: Set the Recv timeout socket option:  Socket will close
//              if no data is received in the specified number of secs
//							
//  PARAMETERS: Socket descriptor
//              Number of seconds of no recv data before timeout
//
//     RETURNS: 
//
//       NOTES: 
//
//---------------------------------------------------------------------
int setRecvTimeout( int socket, int seconds )
{
  struct timeval tv;

  tv.tv_sec = seconds;
  tv.tv_usec = 0;
  if ( setsockopt( socket, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv) ) != 0 )
  {
    avoPrintf(1, "Setsockopt: error setting recv timout: %s \n", strerror(errno));
  }
  return 0;
}
   
//---------------------------------------------------------------------
//    FUNCTION: Recv
//
// DESCRIPTION: Receive data on a socket
//							
//  PARAMETERS:
//
//     RETURNS: Number of bytes received
//
//       NOTES: 
//
//---------------------------------------------------------------------
ssize_t Recv(int fd, void *ptr, size_t nbytes, int flags)
{
  ssize_t		n;

  if ( (n = recv(fd, ptr, nbytes, flags)) <= 0)
  {
    avoPrintf(1, "Recv: error: %s \n", strerror(errno));
  }
  return(n);
}

//---------------------------------------------------------------------
//    FUNCTION: Readn
//
// DESCRIPTION: Read n bytes from a descritor
//							
//  PARAMETERS:
//
//     RETURNS: Number of bytes read
//
//       NOTES: 
//
//---------------------------------------------------------------------
ssize_t Readn(int fd, void *vptr, size_t n)
{
  size_t nleft;
  ssize_t nread;
  char *ptr;

  ptr = vptr;
  nleft = n;
  
  while (nleft > 0)
  {
    if ( (nread = read(fd, ptr, nleft)) < 0)
    {
      if (errno == EINTR)
      {
        nread = 0;
      }
      else
      {
        avoPrintf(1, "Readn: error: %s \n", strerror(errno));
        return(-1);
      }
    }
    else if (nread == 0)
    {
      break;
    }
    
    nleft -= nread;
    ptr += nread;
  }
  return (n - nleft);
}

//---------------------------------------------------------------------
//    FUNCTION: Writen
//
// DESCRIPTION: Write n bytes to a descriptor
//							
//  PARAMETERS:
//
//     RETURNS: 
//
//       NOTES: 
//
//---------------------------------------------------------------------
ssize_t Writen(int fd, const void *vptr, size_t n)
{
  size_t nleft;
  ssize_t nwritten;
  const char *ptr;
  
  ptr = vptr;
  nleft = n;
  while (nleft > 0)
  {
    if ( (nwritten = write(fd, ptr, nleft)) <= 0)
    {
      if (errno == EINTR)
      {
        nwritten = 0;
      }
      else
      {
        avoPrintf(1, "Writen: error: %s \n", strerror(errno));
        return (-1);
      }
    }
    nleft -= nwritten;
    ptr += nwritten;
  }
  return (n);
}
  
//---------------------------------------------------------------------
//    FUNCTION: Select_read
//
// DESCRIPTION: Check if a socket is ready for reading
//							
//  PARAMETERS: socket
//              timeout - number of seconds before timeout, or 0 to wait
//                        indefinitely
//
//     RETURNS: FALSE if neither is ready
//              TRUE if socket is ready  
//
//       NOTES: 
//
//---------------------------------------------------------------------
int Select_read(int socket, long timeout)
{
  fd_set fd_read;
  struct timeval select_timeout;
  int ret_val;
    
  FD_ZERO(& fd_read);
  FD_SET(socket, & fd_read);
  
  if ( !timeout )
  {
    ret_val = select( socket+1, & fd_read, NULL, NULL, NULL);
  }
  else
  {
    select_timeout.tv_sec = timeout;
    select_timeout.tv_usec = 0;
    ret_val = select( socket+1, & fd_read, NULL, NULL, &select_timeout);
  }
  
  if ( ret_val == 0 )
  {
    // Select timed out
    return(FALSE);
  }

  if ( FD_ISSET(socket, & fd_read) )
  {
    // Data is available for reading
    return(TRUE);
  }

  avoPrintf(1, "ERROR: Select_read() \n");
  return(FALSE);
}


//---------------------------------------------------------------------
//    FUNCTION: Shutdown
//
// DESCRIPTION: Shutdown a socket
//							
//  PARAMETERS:
//
//     RETURNS: 
//
//       NOTES: 
//
//---------------------------------------------------------------------
void Shutdown(int fd, int how)
{
  if (shutdown(fd, how) < 0)
  {
    avoPrintf(AVODBG_ADHOC, "Shutdown: error: %s \n", strerror(errno));
  }
}

//---------------------------------------------------------------------
//    FUNCTION: Gethostname
//
// DESCRIPTION: Retrieves the current host name and returns
//              a formatted string.
//							
//  PARAMETERS: hostname_string
//              len = max len of string to return
//							
//     RETURNS: 0 for success, -1 for failure
//
//       NOTES: 
//
//---------------------------------------------------------------------
int Gethostname ( char * hostname_string, int len )
{
  if ( gethostname ( hostname_string, len ) < 0)
  {
    avoPrintf(1, "gethostname: %s \n", strerror(errno));
    return (-1);
  }  
  return (0);
}

//---------------------------------------------------------------------
//    FUNCTION: Sethostname
//
// DESCRIPTION: Sets the current host name
//              a formatted string.
//							
//  PARAMETERS: hostname_string
//              len = max len of string
//							
//     RETURNS: 0 for success, -1 for failure
//
//       NOTES: 
//
//---------------------------------------------------------------------
int Sethostname ( char * hostname_string, int len )
{
  if ( sethostname ( hostname_string, len ) < 0)
  {
    avoPrintf(1, "sethostname: %s \n", strerror(errno));
    return (-1);
  }  
  return (0);
}

//---------------------------------------------------------------------
//    FUNCTION: Getdomainname
//
// DESCRIPTION: Gets the domain name
//							
//  PARAMETERS: domainname
//              len = max len of string
//							
//     RETURNS: 0 for success, -1 for failure
//
//       NOTES: 
//
//---------------------------------------------------------------------
int Getdomainname ( char * domainname, int len )
{
  if ( getdomainname ( domainname, len ) < 0)
  {
    avoPrintf(1, "Getdomainname: %s \n", strerror(errno));
    return (-1);
  }  
  return (0);
}

//---------------------------------------------------------------------
//    FUNCTION: Setdomainname
//
// DESCRIPTION: Sets the domain name
//							
//  PARAMETERS: domainname
//              len = max len of string
//							
//     RETURNS: 0 for success, -1 for failure
//
//       NOTES: 
//
//---------------------------------------------------------------------
int Setdomainname ( char * domainname, int len )
{
  if ( setdomainname ( domainname, len ) < 0)
  {
    avoPrintf(1, "Setdomainname: %s \n", strerror(errno));
    return (-1);
  }  
  return (0);
}


//---------------------------------------------------------------------
//    FUNCTION: 
//
// DESCRIPTION: 
//							
//  PARAMETERS:
//
//     RETURNS: 
//
//       NOTES: 
//
//---------------------------------------------------------------------
// Returns the local host's IP address.
int Get_ip_address ( char * ipAddr )
{
struct hostent *hptr;
struct utsname myname;

  if ( uname(&myname) < 0)
  {
    avoPrintf(1, "Get_ip_address: uname returned error \n");
    return (-1);
  }
  
  if ( ( hptr = gethostbyname(myname.nodename)) == NULL)
  {
    avoPrintf(1, "Get_ip_address: gethostbyname returned error \n");
    return (-1);
  }  
  
  inet_ntop(AF_INET, *hptr->h_addr_list, ipAddr, INET_ADDRSTRLEN );
  if ( ipAddr == NULL )
  {
    avoPrintf(1, "Get_ip_address: inet_ntop returned error \n");
    return (-1);
  }
  
  avoPrintf(2, "My IP address is: %s.\n", ipAddr );
  
  return (0);

}

//---------------------------------------------------------------------
//    FUNCTION: GetPeerName
//
// DESCRIPTION: Get the foreign protocol address associated with
//              the socket 
//							
//  PARAMETERS:
//
//     RETURNS: 
//
//       NOTES: 
//
//---------------------------------------------------------------------
int GetPeerName ( int sockfd, struct in_addr * peerAddr )
{
  struct sockaddr_in addr;
  int addrLen;
  
  addrLen = sizeof( addr );
  if ( getpeername( sockfd, ( struct sockaddr * )& addr, & addrLen ) < 0 )
  {
    avoPrintf(AVODBG_ADHOC, "getpeername : %s \n", strerror(errno));
    return -1;
  }
  peerAddr->s_addr = addr.sin_addr.s_addr;
  return 0;
}




/* end  */





