/*
 * Author: Tatu Ylonen <ylo@cs.hut.fi>
 * Copyright (c) 1995 Tatu Ylonen <ylo@cs.hut.fi>, Espoo, Finland
 *                    All rights reserved
 * Functions for manipulating fifo buffers (that can grow if needed).
 *
 * As far as I am concerned, the code I have written for this software
 * can be used freely for any purpose.  Any derived versions of this
 * software must be clearly marked as such, and if the derived work is
 * incompatible with the protocol description in the RFC file, it must be
 * called by a name other than "ssh" or "Secure Shell".
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "avoprint.h"
#include "ssh-buffer.h"

#define ADSAP_WARNING     (AVODBG_WARNING | AVODBG_ADSAP)
#define ADSAP_FAILURE     (AVODBG_FAILURE | AVODBG_ADSAP)
#define ADSAP_INFO        (AVODBG_INFO    | AVODBG_ADSAP)
#define ADSAP_ADHOC       (AVODBG_ADHOC   | AVODBG_ADSAP)

#define GET_64BIT(cp) (((u_int64_t)(u_char)(cp)[0] << 56) | \
               ((u_int64_t)(u_char)(cp)[1] << 48) | \
               ((u_int64_t)(u_char)(cp)[2] << 40) | \
               ((u_int64_t)(u_char)(cp)[3] << 32) | \
               ((u_int64_t)(u_char)(cp)[4] << 24) | \
               ((u_int64_t)(u_char)(cp)[5] << 16) | \
               ((u_int64_t)(u_char)(cp)[6] << 8) | \
               ((u_int64_t)(u_char)(cp)[7]))

#define GET_32BIT(cp) (((u_long)(u_char)(cp)[0] << 24) | \
               ((u_long)(u_char)(cp)[1] << 16) | \
               ((u_long)(u_char)(cp)[2] << 8) | \
               ((u_long)(u_char)(cp)[3]))

#define GET_16BIT(cp) (((u_long)(u_char)(cp)[0] << 8) | \
               ((u_long)(u_char)(cp)[1]))

#define PUT_64BIT(cp, value) do { \
  (cp)[0] = (char) ((value) >> 56); \
  (cp)[1] = (char) ((value) >> 48); \
  (cp)[2] = (char) ((value) >> 40); \
  (cp)[3] = (char) ((value) >> 32); \
  (cp)[4] = (char) ((value) >> 24); \
  (cp)[5] = (char) ((value) >> 16); \
  (cp)[6] = (char) ((value) >> 8); \
  (cp)[7] = (char) ((value)); } while (0)

#define PUT_32BIT(cp, value) do { \
  (cp)[0] = (value) >> 24; \
  (cp)[1] = (value) >> 16; \
  (cp)[2] = (value) >> 8; \
  (cp)[3] = (value); } while (0)

#define PUT_16BIT(cp, value) do { \
  (cp)[0] = (value) >> 8; \
  (cp)[1] = (value); } while (0)


void ssh_buffer_init(SSH_Buffer *buffer)
{
	buffer->alloc	= 4096;
	buffer->buf		= malloc(buffer->alloc);
	buffer->offset	= 0;
	buffer->end		= 0;
}

/* Returns the number of bytes of data in the buffer. */

u_int ssh_buffer_len(SSH_Buffer *buffer)
{
	return buffer->end - buffer->offset;
}

/* Returns a pointer to the first used byte in the buffer. */

char *ssh_buffer_ptr(SSH_Buffer *buffer)
{
	return buffer->buf + buffer->offset;
}

/* Gets data from the beginning of the buffer. */

void ssh_buffer_get(SSH_Buffer *buffer, char *buf, u_int len)
{
	if (len > buffer->end - buffer->offset)
		avoPrintf(ADSAP_FAILURE, "buffer_get: trying to get more bytes %d than in buffer %d", len, buffer->end - buffer->offset);
	memcpy(buf, buffer->buf + buffer->offset, len);
	buffer->offset += len;
}

/* Consumes the given number of bytes from the beginning of the buffer. */

void ssh_buffer_consume(SSH_Buffer *buffer, u_int bytes)
{
	if(bytes > buffer->end - buffer->offset)
		avoPrintf(ADSAP_FAILURE, "ssh_buffer_consume: trying to get more bytes than in buffer");
	buffer->offset += bytes;
}

/* Frees any memory used for the buffer. */

void ssh_buffer_free(SSH_Buffer *buffer)
{
	if(buffer)
	{
		memset(buffer->buf, 0, buffer->alloc);
		free(buffer->buf);
	}
	else
		avoPrintf(ADSAP_FAILURE, "item_buffer_free: address = 0x%Lx, passed pointer is NULL", buffer);
}



/* Returns a character from the buffer (0 - 255) */
int buffer_get_char(SSH_Buffer *buffer)
{
	char ch;
	ssh_buffer_get(buffer, &ch, 1);
	return (u_char) ch;
}

/*
    * Returns an integer from the buffer (4 bytes, msb first).
    */
u_int buffer_get_int(SSH_Buffer *buffer)
{
	u_char buf[4];
	ssh_buffer_get(buffer, (char *) buf, 4);
	return GET_32BIT(buf);
}

/*
    * Returns an arbitrary binary string from the buffer.  The string cannot
    * be longer than 256k.  The returned value points to memory allocated
    * with xmalloc; it is the responsibility of the calling function to free
    * the data.  If length_ptr is non-NULL, the length of the returned data
    * will be stored there.  A null character will be automatically appended
    * to the returned string, and is not counted in length.
    */
char *buffer_get_string(SSH_Buffer *buffer, u_int *length_ptr)
{
	u_int len;
	char *value;
                /* Get the length. */
	len = buffer_get_int(buffer);
	if (len > 256 * 1024)
		avoPrintf(ADSAP_FAILURE, "Received packet with bad string length %d", len);
                /* Allocate space for the string.  Add one byte for a null character. */
	value = malloc(len + 1);
                /* Get the string. */
	ssh_buffer_get(buffer, value, len);
                /* Append a null character to make processing easier. */
	value[len] = 0;
                /* Optionally return the length of the string. */
	if (length_ptr)
		*length_ptr = len;
	return value;
}

//********************************************************************************

void ssh_buffer_append(SSH_Buffer *buffer, const char *data, u_int len)
{
	char *cp;
	ssh_buffer_append_space(buffer, &cp, len);
	memcpy(cp, data, len);
}

/*
    * Appends space to the buffer, expanding the buffer if necessary. This does
    * not actually copy the data into the buffer, but instead returns a pointer
    * to the allocated region.
    */

void ssh_buffer_append_space(SSH_Buffer *buffer, char **datap, u_int len)
{
	unsigned int	old_alloc;

                /* If the buffer is empty, start using it from the beginning. */
	if (buffer->offset == buffer->end)
	{
		buffer->offset = 0;
		buffer->end = 0;
	}
restart:
                /* If there is enough space to store all data, store it now. */
	if (buffer->end + len <= buffer->alloc)
	{
		*datap = buffer->buf + buffer->end;
		buffer->end += len;
		return;
	}
                /*
     	 	 	 * If the buffer is quite empty, but all data is at the end, move the
     	 	 	 * data to the beginning and retry.
     	 	 	 */
	if (buffer->offset > buffer->alloc / 2)
	{
		avoPrintf(ADSAP_FAILURE, "ssh_buffer_append_space: buffer moved - Old start = %d", buffer->offset);
		memmove(buffer->buf, buffer->buf + buffer->offset, buffer->end - buffer->offset);
		buffer->end -= buffer->offset;
		buffer->offset = 0;
		goto restart;
	}
                /* Increase the size of the buffer and retry. */

	old_alloc = buffer->alloc;
	buffer->alloc += SSH_BUFFER_INCREMENT;

	buffer->buf = realloc(buffer->buf, buffer->alloc);

	avoPrintf(ADSAP_FAILURE, "ssh_buffer_append_space: buffer realloc - Old size = %d, New Size - %d", old_alloc, buffer->alloc);

	goto restart;
}

void buffer_put_int(SSH_Buffer *buffer, u_int value)
{
	char buf[4];
	PUT_32BIT(buf, value);
	ssh_buffer_append(buffer, buf, 4);
}

void buffer_put_string(SSH_Buffer *buffer, const void *buf, u_int len)
{
	buffer_put_int(buffer, len);
	ssh_buffer_append(buffer, buf, len);
}

void buffer_put_cstring(SSH_Buffer *buffer, const char *s)
{
	buffer_put_string(buffer, s, strlen(s));
}

/*
    * Stores an BIGNUM in the buffer with a 2-byte msb first bit count, followed
    * by (bits+7)/8 bytes of binary data, msb first.
    */
void buffer_put_bignum(SSH_Buffer *buffer, BIGNUM *value)
{
	int bits = BN_num_bits(value);
	int bin_size = (bits + 7) / 8;
	u_char *buf = malloc(bin_size);
	int oi;
	char msg[2];

                /* Get the value of in binary */
	oi = BN_bn2bin(value, buf);
	if (oi != bin_size)
		avoPrintf(ADSAP_FAILURE, "buffer_put_bignum: BN_bn2bin() failed: oi %d != bin_size %d", oi, bin_size);

                /* Store the number of bits in the buffer in two bytes, msb first. */
	PUT_16BIT(msg, bits);
	ssh_buffer_append(buffer, msg, 2);
                /* Store the binary data. */
	ssh_buffer_append(buffer, (char *)buf, oi);

	memset(buf, 0, bin_size);
	free(buf);
}

/*
    * Retrieves an BIGNUM from the buffer.
    */
int buffer_get_bignum(SSH_Buffer *buffer, BIGNUM *value)
{
	unsigned int bits, bytes;
	u_char buf[2], *bin;

                /* Get the number for bits. */
	ssh_buffer_get(buffer, (char *) buf, 2);
	bits = GET_16BIT(buf);
                /* Compute the number of binary bytes that follow. */
	bytes = (bits + 7) / 8;
	if (ssh_buffer_len(buffer) < bytes)
		avoPrintf(ADSAP_FAILURE, "buffer_get_bignum: input buffer too small");
	bin = (u_char *) ssh_buffer_ptr(buffer);
	BN_bin2bn(bin, bytes, value);
	ssh_buffer_consume(buffer, bytes);

	return 2 + bytes;
}

void buffer_put_bignum2(SSH_Buffer *buffer, BIGNUM *value)
{
	int bytes = BN_num_bytes(value) + 1;
	u_char *buf = malloc(bytes);
	int oi;
	int hasnohigh = 0;
	buf[0] = '\0';
                /* Get the value of in binary */
	oi = BN_bn2bin(value, buf+1);
	if (oi != bytes-1)
		avoPrintf(ADSAP_FAILURE, "buffer_put_bignum: BN_bn2bin() failed: oi %d != bin_size %d", oi, bytes);
	hasnohigh = (buf[1] & 0x80) ? 0 : 1;
	if (value->neg)
	{
                                /**XXX should be two's-complement */
		int i, carry;
		u_char *uc = buf;
//		ssh_log("negativ!");
		for(i = bytes-1, carry = 1; i>=0; i--)
		{
			uc[i] ^= 0xff;
			if(carry)
				carry = !++uc[i];
		}
	}
	buffer_put_string(buffer, buf+hasnohigh, bytes-hasnohigh);
	memset(buf, 0, bytes);
	free(buf);
}

int buffer_get_bignum2(SSH_Buffer *buffer, BIGNUM *value)
{
                /**XXX should be two's-complement */
	int len;
	u_char *bin = (u_char *)buffer_get_string(buffer, (u_int *)&len);
	BN_bin2bn(bin, len, value);
	free(bin);
	return len;
}
