//------------------------------------------------------------------------
// Copyright 2002, Avocent Corporation
//------------------------------------------------------------------------
// FILE        :  ssl.c
// DESCRIPTION :  Functions which access the OpenSSL library
//
//------------------------------------------------------------------------
// AUTHOR      :  J. Collins
// DATE        :  February 2002
//
//------------------------------------------------------------------------
// HISTORY
//
// DATE           PROGRAMMER              MODIFICATIONS
// March 2006     E. Seabra               Porting to ACS
//
//------------------------------------------------------------------------
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/mman.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <netinet/tcp.h>
#include <pthread.h>
#include <termios.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "types.h"
#include "ds2mib.h"
#include "ssl.h"
#include "netwrap.h"
#include "avoprint.h"
#include "adsap2.h"

#ifdef SSL_ACCEL
#include <openssl/hardware.h>
#endif
#include <openssl/ssl.h>
#include <openssl/rsa.h>       /* SSLeay stuff */
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/err.h>

#include <openssl/lhash.h>
#include <openssl/bn.h>

#define OPENSSL_THREAD_DEFINES

#include <openssl/opensslconf.h>

#define SSL_DYNLOCK

static pthread_mutex_t sslMutex;

#ifdef SSL_DYNLOCK
struct CRYPTO_dynlock_value
{
	pthread_mutex_t mutex;
};
#endif

static void LockingCallback ( int mode, int type, const char *file, int line);
static pthread_mutex_t cryptoLocks[ CRYPTO_NUM_LOCKS ];

#define SSL_WARNING     (AVODBG_WARNING | AVODBG_SSL)
#define SSL_FAILURE     (AVODBG_FAILURE | AVODBG_SSL)
#define SSL_INFO        (AVODBG_INFO    | AVODBG_SSL)
#define SSL_ADHOC       (AVODBG_ADHOC   | AVODBG_SSL)

//---------------------------------------------------------------------
//    FUNCTION: printError
//
// DESCRIPTION: Get the last error code on the error stack and print
//              the error message
//              
//  PARAMETERS: 
//              
//     RETURNS: 
//
//       NOTES:
//---------------------------------------------------------------------
static void printError( void )
{
  int err;
  
  err = ERR_get_error();
  avoPrintf(SSL_FAILURE, "%s\n", ERR_error_string( err, 0 ));
}


#ifdef SSL_DYNLOCK
static struct CRYPTO_dynlock_value *
dyn_create_function(const char *file, int line)
{
	struct CRYPTO_dynlock_value *value;

	value = (struct CRYPTO_dynlock_value *)malloc(sizeof(struct CRYPTO_dynlock_value));
	if (!value) return NULL;

	pthread_mutex_init(&value->mutex, NULL);
	return value;
}

static void dyn_lock_function(int mode, struct CRYPTO_dynlock_value *l, 
								const char *file, int line)
{
	if ( mode & CRYPTO_LOCK ) {
		pthread_mutex_lock(&l->mutex);
	} else {
		pthread_mutex_unlock(&l->mutex);
	}
}

static void dyn_destroy_function(struct CRYPTO_dynlock_value *l, 
								const char *file, int line)
{
	pthread_mutex_destroy(&l->mutex);
	free(l);
}
#endif //SSL_DYNLOCK

//---------------------------------------------------------------------
//    FUNCTION: SSL multi-threading support functions
//
// DESCRIPTION: 
//              
//  PARAMETERS: 
//              
//     RETURNS: 
//
//       NOTES: 
//---------------------------------------------------------------------
void MultiThreadCleanup ( void )
{
  int i;
  CRYPTO_set_id_callback(NULL);
  CRYPTO_set_locking_callback ( NULL );

#ifdef SSL_DYNLOCK
  CRYPTO_set_dynlock_create_callback ( NULL );
  CRYPTO_set_dynlock_lock_callback ( NULL );
  CRYPTO_set_dynlock_destroy_callback ( NULL );
#endif //SSL_DYNLOCK

  for ( i = 0; i < CRYPTO_NUM_LOCKS; i++ )
  {
    pthread_mutex_destroy ( & cryptoLocks[ i ]);
  }
}

static unsigned long ThreadSelf(void)
{
	return ((unsigned long)pthread_self());
}

void SSL_MultiThreadSetup ( void )
{
  int i;
  
  for ( i = 0; i < CRYPTO_NUM_LOCKS; i++ )
  {
    pthread_mutex_init ( & cryptoLocks[ i ], NULL );
  }
 
  CRYPTO_set_id_callback(ThreadSelf);
  CRYPTO_set_locking_callback((void (*)(int,int,const char *,int))LockingCallback);

#ifdef SSL_DYNLOCK
  CRYPTO_set_dynlock_create_callback ( dyn_create_function );
  CRYPTO_set_dynlock_lock_callback ( dyn_lock_function );
  CRYPTO_set_dynlock_destroy_callback ( dyn_destroy_function );
#endif //SSL_DYNLOCK
}

static void LockingCallback ( int mode, int type, const char * file, int line )
{
  if ( type >= CRYPTO_NUM_LOCKS )
  {
    avoPrintf(SSL_FAILURE, "Error in SSL LockingCallback: type too big!!!!!! \n");
    return;
  }
  if ( mode & CRYPTO_LOCK )
  {
    //printf("crypto LOCK \n");
    pthread_mutex_lock ( & cryptoLocks[ type ] );
  }
  else
  {
    //printf("crypto UNLOCK \n");
    pthread_mutex_unlock ( & cryptoLocks[ type ] );
  }
}

//---------------------------------------------------------------------
//    FUNCTION: getRandomNumber
//
// DESCRIPTION: Use dev/urandom to generate a random number
//              
//  PARAMETERS: none
//              
//     RETURNS: random number between 0 and 15
//
//       NOTES: 
//---------------------------------------------------------------------
int getRandomNumber( void )
{
  int randdev;
  unsigned char randnum[1];
  int rnum;
  
  randdev = open( "/dev/urandom", O_RDONLY, 0 );
  if( randdev < 0 ) 
  {
    avoPrintf( SSL_FAILURE, "Failed opening randdev \n");
    exit(0);
  }
  read( randdev, & randnum[0], 1 );
  close(randdev);

  srand( randnum[0] );
  
  rnum = rand() % 16;
  return (rnum);
}

/*
** Canned Diffie-Hellman parameters.  Half were generated with a 
** generator of 2, and half with a generator of 5.  A set of
** parameters will be chosen at random during initializtion.
*/

// These primes were generated with a generator of 2
static unsigned char dh512_prime_g2[8][64] =
{
  { // 0
    0xDA,0x58,0x3C,0x16,0xD9,0x85,0x22,0x89,0xD0,0xE4,0xAF,0x75,
    0x6F,0x4C,0xCA,0x92,0xDD,0x4B,0xE5,0x33,0xB8,0x04,0xFB,0x0F,
    0xED,0x94,0xEF,0x9C,0x8A,0x44,0x03,0xED,0x57,0x46,0x50,0xD3,
    0x69,0x99,0xDB,0x29,0xD7,0x76,0x27,0x6B,0xA2,0xD3,0xD4,0x12,
    0xE2,0x18,0xF4,0xDD,0x1E,0x08,0x4C,0xF6,0xD8,0x00,0x3E,0x7C,
    0x47,0x74,0xE8,0x33,
  },
  { // 1
    0xFA,0x3D,0x56,0xFD,0x1C,0x28,0x2A,0xD9,0xAB,0x82,0x86,0xBE,
    0xAE,0x07,0xF9,0x10,0x71,0x17,0xA0,0x2C,0x9B,0x55,0xD2,0xB5,
    0x7E,0x1D,0x6E,0x9F,0xB7,0x25,0x6A,0x4D,0x97,0xF9,0xC5,0xB1,
    0x0F,0x18,0x5C,0xCC,0x23,0xF2,0xCC,0x42,0xA6,0x7D,0xA9,0x03,
    0xF1,0xD3,0x5C,0x4B,0xFE,0x45,0x54,0x2E,0x96,0x3E,0xE4,0xF0,
    0x26,0xEA,0x52,0xD3,
  },
  { // 2
    0xA8,0x28,0x91,0xB6,0xAB,0xB8,0x97,0x3B,0x79,0x33,0xE0,0x2E,
    0x51,0x78,0xA0,0x4F,0xDF,0x31,0xF2,0x7F,0xCA,0x74,0xB4,0x62,
    0x7B,0xD7,0x17,0xD0,0x6F,0x11,0x52,0x6F,0x8B,0x29,0x2D,0xA9,
    0x78,0x99,0xBD,0x98,0xF3,0x74,0x50,0xCD,0xE6,0xA1,0x10,0x25,
    0x0D,0x6A,0xF3,0x65,0xBB,0x57,0xB9,0x03,0xDF,0x9C,0x56,0xFE,
    0x1E,0x56,0x71,0xC3,
  },
  { // 3
    0xC4,0x9C,0x38,0x1A,0xFB,0xC8,0x99,0x93,0x1F,0x4B,0xA9,0x21,
    0x86,0x1C,0xF4,0xD7,0x31,0x18,0x6C,0x16,0x76,0x20,0xF5,0x97,
    0x53,0x4E,0xE2,0xC8,0xE3,0x22,0x81,0x92,0x71,0x46,0x02,0x4E,
    0x20,0x26,0x7D,0x12,0x1D,0x05,0xEF,0xA4,0x96,0x43,0xA5,0xB1,
    0x85,0x47,0x29,0xCA,0x9E,0x3A,0xCF,0x05,0x57,0x61,0x35,0x22,
    0xEE,0x96,0x68,0x2B,
  },
  { // 4
    0xCC,0x0D,0xF7,0x5D,0xA7,0x26,0x8C,0x6C,0x6D,0x3C,0x3E,0xB3,
    0xBC,0xDA,0xE2,0x1B,0x0B,0x20,0x58,0x6C,0x2C,0x12,0x3F,0x00,
    0xB0,0x66,0x71,0xDC,0x07,0x65,0x5C,0xC0,0xB0,0xB1,0x19,0xBE,
    0x78,0x1F,0xC5,0x13,0x54,0xBD,0xAC,0x5A,0xE1,0xAC,0x44,0xAF,
    0x99,0x1F,0x63,0xDF,0xB7,0x7B,0xFD,0x6F,0x70,0x0F,0x72,0x27,
    0xA9,0x28,0xCB,0x4B,
  },
  { // 5
    0xA6,0xD1,0x29,0x12,0x40,0xBB,0x94,0x8B,0xD7,0x10,0xBC,0x0A,
    0x2A,0x0E,0x13,0x61,0x19,0x3C,0x99,0xE7,0xE5,0x9B,0x2D,0x3B,
    0xE4,0x14,0x3B,0xBD,0xD8,0x88,0x34,0x1C,0xC0,0xB4,0x72,0x1E,
    0x61,0x4C,0x73,0xE9,0xAA,0x37,0x9F,0x8A,0xFA,0xCB,0x03,0xA6,
    0x4D,0x08,0x91,0xD2,0x7C,0xF1,0x66,0x3C,0xF5,0x93,0x03,0xA5,
    0x3C,0x10,0xAD,0xF3,
  },
  { // 6
    0x82,0x44,0xB8,0xEA,0x21,0x4A,0xB1,0xF9,0x18,0x85,0x97,0x08,
    0x2D,0x1E,0xE8,0x21,0xE3,0x13,0x43,0xD3,0x47,0xC0,0x20,0x00,
    0xE0,0x08,0x16,0x6F,0x39,0x0E,0xFD,0xCE,0x92,0xCF,0xB0,0xE6,
    0x4F,0x58,0x40,0x81,0x2B,0x54,0xAF,0x83,0x6D,0x74,0x1E,0x44,
    0x1B,0xA0,0x7C,0x9D,0x6C,0x35,0x52,0xA8,0x7C,0x2F,0xF0,0x12,
    0x01,0xAE,0x7A,0x3B,
  },
  { // 7
    0xA2,0x66,0xAF,0x74,0xFA,0x44,0xDB,0x0B,0x62,0xBC,0x18,0xC6,
    0xDA,0x2F,0x36,0x4C,0xE9,0xC6,0x50,0x42,0x0D,0x8A,0x2B,0xAE,
    0xFE,0xD8,0xED,0x8A,0xFB,0x6E,0x53,0x76,0x3B,0x0A,0xF3,0xAD,
    0xAB,0x49,0x22,0x28,0x60,0x44,0x79,0x21,0x97,0xBF,0x7E,0x81,
    0xD2,0x6A,0x16,0xC0,0x08,0xA3,0xCD,0x59,0xFC,0x6B,0x10,0x66,
    0x6B,0x48,0x9A,0x53,
  }
};
static unsigned char dh512_g2[]=
{
  0x02,
};

// These primes were generated with a generator of 5
static unsigned char dh512_prime_g5[8][64] =
{
  { // 0
    0xF0,0xAB,0x10,0xF7,0xD4,0xE2,0x8E,0x02,0xB5,0x66,0xA8,0x96,
    0x12,0x60,0x22,0x7D,0x61,0xA0,0xFB,0x36,0x10,0x1D,0xEB,0x17,
    0xA4,0x5F,0x92,0xB5,0xCD,0x68,0xD8,0x7D,0x30,0xC5,0xDA,0xED,
    0xFF,0x80,0xA8,0x4C,0x50,0xB5,0x1A,0x2C,0x50,0xDB,0xC0,0x2C,
    0xF6,0x1F,0x3D,0x86,0xB5,0xBA,0x53,0xA0,0xAA,0xAC,0x6D,0x06,
    0x65,0xC7,0x30,0x43,
  },
  { // 1
    0x91,0xDE,0x4B,0x8C,0x5F,0xFF,0xDD,0xC7,0xC8,0x70,0x05,0x9B,
    0x4E,0x18,0x84,0x53,0x21,0xC3,0x69,0x5A,0x07,0xF1,0x25,0x7E,
    0x60,0xFD,0xFE,0x6F,0x19,0x40,0xE2,0xCE,0xAF,0x4D,0xB8,0xF5,
    0xE8,0x30,0xAE,0xCE,0x10,0x60,0xC4,0xB1,0x9E,0xDA,0x90,0x16,
    0x3B,0x3F,0xCB,0x40,0x1C,0x5D,0x15,0x1D,0x85,0xCC,0x99,0x7A,
    0x4A,0xE7,0x3E,0xAB,
  },
  { // 2
    0x9B,0x70,0x57,0x00,0xA0,0x08,0xFC,0xED,0x3E,0xC4,0x8C,0x2E,
    0x65,0xA0,0x31,0xEE,0x82,0xAC,0x7C,0x35,0x78,0x67,0x66,0x70,
    0x5A,0x12,0xE1,0x42,0xF5,0x36,0xE0,0x17,0xF5,0x16,0x38,0x0F,
    0xFF,0x9E,0xDB,0x91,0xED,0x82,0x26,0x88,0xB8,0x29,0x77,0xF2,
    0xE2,0x45,0x9F,0xA8,0x15,0x13,0x49,0x8E,0xB5,0x9F,0x7D,0xC3,
    0x5C,0xE7,0xC0,0xAF,
  },
  { // 3
    0xBD,0xA4,0x0E,0xE3,0xCE,0x4F,0x6B,0x93,0xAB,0x0C,0x69,0x88,
    0x80,0xE4,0xC9,0x0D,0x27,0x1A,0x1E,0xFF,0xC7,0x34,0x2D,0x49,
    0x5E,0x8B,0x7A,0x42,0x94,0x45,0x58,0xA1,0xF8,0xD1,0x37,0xB4,
    0xB0,0xC5,0xEB,0x0E,0x86,0xB4,0xC5,0xA3,0x49,0x53,0x5C,0x25,
    0xCA,0x25,0x8E,0xF5,0xD3,0x2F,0x82,0x41,0x59,0x59,0x6D,0x3F,
    0x44,0x08,0x74,0x97,
  },
  { // 4
    0xE9,0x4E,0x68,0xC1,0xED,0x9E,0x73,0x5A,0x1E,0xCA,0x26,0x57,
    0xA8,0x18,0xAF,0x16,0x87,0x79,0x39,0x1F,0xD1,0x9E,0xCC,0x2F,
    0x88,0x6F,0x10,0xD8,0xA8,0x9D,0xBD,0x22,0x8C,0x54,0x7F,0xDE,
    0xB3,0x24,0x6C,0xE5,0x5E,0xBA,0x0F,0x40,0x45,0xD6,0x7C,0xEB,
    0x96,0x2F,0x4F,0xC4,0xC1,0x8E,0x27,0xAD,0x6C,0xA3,0x29,0x58,
    0x86,0x15,0x3F,0x57,
  },
  { // 5
    0xE9,0x4E,0x68,0xC1,0xED,0x9E,0x73,0x5A,0x1E,0xCA,0x26,0x57,
    0xA8,0x18,0xAF,0x16,0x87,0x79,0x39,0x1F,0xD1,0x9E,0xCC,0x2F,
    0x88,0x6F,0x10,0xD8,0xA8,0x9D,0xBD,0x22,0x8C,0x54,0x7F,0xDE,
    0xB3,0x24,0x6C,0xE5,0x5E,0xBA,0x0F,0x40,0x45,0xD6,0x7C,0xEB,
    0x96,0x2F,0x4F,0xC4,0xC1,0x8E,0x27,0xAD,0x6C,0xA3,0x29,0x58,
    0x86,0x15,0x3F,0x57,
  },
  { // 6
    0xFC,0xF0,0xA3,0xF1,0x78,0xC4,0x89,0x11,0xBE,0x20,0xB4,0x87,
    0x59,0x92,0xA7,0x1D,0x0B,0xD7,0xA7,0xDE,0x37,0xDE,0x63,0x71,
    0xB5,0xEE,0xA3,0x3A,0x5E,0xAC,0x6D,0x6A,0x4D,0x8A,0xD7,0xE9,
    0x76,0x4C,0x25,0x7B,0xAB,0x7B,0xAF,0x75,0xC0,0x26,0x3A,0x31,
    0x18,0xBD,0x2D,0xAC,0xE5,0x3E,0xBF,0x63,0xE1,0x76,0x80,0x59,
    0xEA,0xCF,0x94,0x13,
  },
  { // 7
    0x8F,0xD0,0x87,0x52,0x08,0x4C,0x95,0x96,0x49,0xF2,0x11,0x63,
    0x68,0x6C,0x77,0xFA,0x33,0x43,0x92,0xEB,0xE5,0x55,0x52,0xD7,
    0x34,0xC3,0x7C,0xDA,0x5D,0x31,0xBB,0x52,0x1F,0xC1,0x09,0x52,
    0xB8,0x12,0xB1,0xDC,0x1D,0xFF,0xE3,0x86,0xF8,0x7F,0xAE,0x7E,
    0xAB,0xB6,0x18,0x8C,0xF8,0x10,0xE8,0xE7,0xBE,0xA4,0xAB,0xA8,
    0x77,0xF4,0xE1,0x0B,
  }
};

static unsigned char dh512_g5[]=
{
  0x05,
};


//---------------------------------------------------------------------
//    FUNCTION: get_dh512
//
// DESCRIPTION: Randomly select a set of Diffie-Hellman parameters
//              
//  PARAMETERS: rIdx - a random index value between 0-15
//              
//     RETURNS: A DH structure containing the DH parameters
//
//       NOTES: 
//---------------------------------------------------------------------
static DH *get_dh512( int rIdx )
{
  DH * dh = NULL;

  if ( ( dh = DH_new() ) == NULL) 
  {
    return(NULL);
  }
  
  if ( rIdx < 8 )
  {
    /*
    ** Choose a prime with a generator of 2
    */
    dh->p = BN_bin2bn ( dh512_prime_g2[ rIdx ], sizeof(dh512_prime_g2[ rIdx ]), NULL);
    dh->g = BN_bin2bn ( dh512_g2,  sizeof(dh512_g2), NULL);
  }
  else
  {
    /*
    ** Choose a prime with a generator of 5
    */
    rIdx = rIdx - 8;
    dh->p = BN_bin2bn ( dh512_prime_g5[ rIdx ], sizeof(dh512_prime_g5[ rIdx ]), NULL);
    dh->g = BN_bin2bn (dh512_g5,  sizeof(dh512_g5), NULL);
  }
  if ((dh->p == NULL) || (dh->g == NULL))
  {
    return(NULL);
  }
  return(dh);
}

//---------------------------------------------------------------------
//    FUNCTION: SSL_SelectDHparms()
//
// DESCRIPTION: Select the DH parameters to be used to generate DH
//              keys 
//              
//  PARAMETERS: ctx
//              
//     RETURNS: 0 for success, -1 for failure
//
//       NOTES: 
//---------------------------------------------------------------------
static int SSL_SelectDHparms( SSL_CTX * ctx )
{
  DH * dh;
  int rIdx;
  
  rIdx = getRandomNumber();
  
  //printf("Using canned DH parameters # %d \n", rIdx);
  
  dh = get_dh512( rIdx );
  if ( dh == NULL )
  {
    avoPrintf(SSL_FAILURE, "Failed getting DH parms \n");
    printError();
    return -1;
  }
  
  if ( !SSL_CTX_set_tmp_dh( ctx, dh ) )
  {
    avoPrintf(SSL_FAILURE, "set_tmp_dh failed\n");
    printError();
    if (dh)
      DH_free(dh);
    return -1;
  }
  
  avoPrintf(SSL_INFO, "DH parameters successfully set \n");
  DH_free(dh);
  return 0;
  
  
}

//---------------------------------------------------------------------
//    FUNCTION: SSL_getCipherList()
//
// DESCRIPTION: Return the cipher list for the ssl connection based
//              on the settings in the MIB.
//              
//  PARAMETERS: 
//              
//     RETURNS: none
//
//       NOTES: 
//---------------------------------------------------------------------
void SSL_getCipherList( const DsmibEncrypTable *pEncrypTbl, char * cipherList, int context_type )
{
  int      bRC4  = FALSE;
  int      bDES  = FALSE;
  int      b3DES = FALSE;
  int      bAES  = FALSE;
  
  unsigned int  num = 0;
  unsigned int  len;
  
  cipherList[0] = '\0';
  
  if ( (context_type == SSL_CONTEXT_CERT ) ||
       (context_type == SSL_CONTEXT_ANON_OR_CERT ) )
  {
    // Use non-anonymous ciphers
    if ( pEncrypTbl->encryp128 )
    {
      strcat( cipherList, "RC4-MD5:");
      bRC4 = TRUE;
      num++;
    }
    if ( pEncrypTbl->encryp3des )
    {
      strcat( cipherList, "DES-CBC3-SHA:");
      b3DES = TRUE;
      num++;
    }
    if ( pEncrypTbl->encrypdes )
    {
      strcat( cipherList, "DES-CBC-SHA:");
      bDES = TRUE;
      num++;
    }
    if ( pEncrypTbl->encrypAes )
    {
      strcat( cipherList, "DHE-RSA-AES128-SHA:DHE-DSS-AES128-SHA:AES128-SHA:");
      bAES = TRUE;
      num++;
    }
  }
  
  if( ( context_type == SSL_CONTEXT_ANON) || 
      ( context_type == SSL_CONTEXT_ANON_OR_CERT ) )
  {
    // Use anonymous ciphers
    if ( pEncrypTbl->encryp128 )
    {
      strcat( cipherList, "ADH-RC4-MD5:");
      bRC4 = TRUE;
      num++;
    }
    if ( pEncrypTbl->encryp3des )
    {
      strcat( cipherList, "ADH-DES-CBC3-SHA:");
      b3DES = TRUE;
      num++;
    }
    if ( pEncrypTbl->encrypdes )
    {
      strcat( cipherList, "ADH-DES-CBC-SHA:");
      bDES = TRUE;
      num++;
    }
    if ( pEncrypTbl->encrypAes )
    {
      strcat( cipherList, "ADH-AES128-SHA:");
      bAES = TRUE;
      num++;
    }
  }
  
  if(num > 0)
  {
    if(bRC4)
      strcat(cipherList,"+RC4:");
    if(b3DES)
      strcat(cipherList,"+3DES:");
    if(bDES)
      strcat(cipherList,"+DES:");
    if(bAES)
      strcat(cipherList,"+AES");
  }

  if( ( len = strlen(cipherList) ) > 0)
  {
    if(cipherList[len-1] == ':')
      cipherList[len-1] = '\0';
  }
  
  avoPrintf(3, "CipherList is: %s \n", cipherList );  
}

//---------------------------------------------------------------------
//    FUNCTION: print_cipher_list()
//
// DESCRIPTION: Return the cipher list for the ssl connection based
//              on the settings in the MIB.
//              
//  PARAMETERS: none
//              
//     RETURNS: none
//
//       NOTES: 
//---------------------------------------------------------------------
void printCipherList(SSL * ssl)
{
	unsigned int	j = 0;
	const char		*p;

	avoPrintf(SSL_INFO, "SSL %d Cipher List:\n", ssl);

	while( (p = SSL_get_cipher_list(ssl, j++)) != NULL )
	{
		avoPrintf(SSL_INFO, "%s\n",p);
	}
}

//---------------------------------------------------------------------
//    FUNCTION: get_pubkey_from_cert()
//
// DESCRIPTION: Get the public key from the certificate.
//              
//  PARAMETERS: p_cert - certificate
//              
//              
//     RETURNS: public key string
//
//       NOTES: 
//---------------------------------------------------------------------
ASN1_BIT_STRING *get_pubkey_from_cert(X509 *p_cert)
{
  X509_CINF      *p_cert_info;
  X509_PUBKEY      *p_cert_key;
  ASN1_BIT_STRING    *p_cert_public_key;

  if((p_cert_info = p_cert->cert_info) == NULL)
  {
    avoPrintf(SSL_FAILURE, "get_pubkey_from_cert - Error reading certificate info\n");
    return(NULL);
  }

  if((p_cert_key = p_cert_info->key) == NULL)
  {
    avoPrintf(SSL_FAILURE, "get_pubkey_from_cert - Error reading certificate key\n");
    return(NULL);
  }

  if((p_cert_public_key = p_cert_key->public_key) == NULL)
  {
    avoPrintf(SSL_FAILURE, "get_pubkey_from_cert - Error reading certificate public key\n");
    return(NULL);
  }

  return(p_cert_public_key);
}



//---------------------------------------------------------------------
//    FUNCTION: SSL_SetDsCertCallback()
//
// DESCRIPTION: Set peer certificate verification procedure
//              
//  PARAMETERS: ctx
//              
//     RETURNS: none
//
//       NOTES: I think that the data argument is ignored in our
//              version of OpenSSL.  Need to do something about this.
//              See manpage on SSL_CTX_Set_cert_verify_callback(). 
//---------------------------------------------------------------------
void SSL_SetDsCertCallback( void * ctx, void * data) 
{
  if ( ctx )
  {
    SSL_CTX_set_cert_verify_callback((SSL_CTX *)ctx, verify_ds_credentials, (void *)data);
  }
}

//---------------------------------------------------------------------
//    FUNCTION: SSL_SetPreauthCertCallback()
//
// DESCRIPTION: Set peer certificate verification procedure
//              
//  PARAMETERS: ctx
//              
//     RETURNS: none
//
//       NOTES: I think that the data argument is ignored in our
//              version of OpenSSL.  Need to do something about this.
//              See manpage on SSL_CTX_Set_cert_verify_callback(). 
//---------------------------------------------------------------------
void SSL_SetPreauthCertCallback( void * ctx, void * data) 
{
  if ( ctx )
  {
    SSL_CTX_set_cert_verify_callback((SSL_CTX *)ctx, verify_preauth_credentials, (void *)data);
  }
}


//---------------------------------------------------------------------
//    FUNCTION: SSL_Server_Handshake()
//
// DESCRIPTION: Perform the SSL handshaking and return the SSL object
//              
//  PARAMETERS: ctx, sockfd
//              
//     RETURNS: pointer to the SSL object
//
//       NOTES: 
//---------------------------------------------------------------------
void * SSL_Server_Handshake( void * ctx, int sockfd, int context_type )
{
  SSL * ssl;
  int err;
  int handshake_timeout = 30; // seconds
	unsigned char	dbgName[64];
	struct in_addr peerAddr;
  
  pthread_mutex_lock( & sslMutex );
  
  if ( !ctx )
  {
    avoPrintf(SSL_FAILURE, "ERROR: SSL_Server_Handshake NULL CTX\n");
    pthread_mutex_unlock( & sslMutex );
    return NULL;
  }
  
	GetPeerName(sockfd, & peerAddr );
	sprintf(dbgName, "Wait for SSL Server Handshake from %d.%d.%d.%d...",
            (peerAddr.s_addr >>24) & 0xff,
						(peerAddr.s_addr >>16) & 0xff,
						(peerAddr.s_addr >>8) & 0xff,
						peerAddr.s_addr & 0xff);
  avoPrintf(SSL_INFO, "%s\n", dbgName);    
 
  /*
  ** Creates a new SSL structure which is needed to hold the data for a
  ** TLS/SSL connection.
  */    
  ssl = SSL_new ( (SSL_CTX *)ctx );
  if ( ssl == NULL )
  {
    avoPrintf(SSL_FAILURE, "Could not create SSL structure: \n");
    printError();
    pthread_mutex_unlock( & sslMutex );
    return NULL;
  }
  
  if( !init_adsap2_credentials ( ssl ) )
  {
    avoPrintf(SSL_WARNING, "SSL Server Handshake init_adsap2_credentials failed \n" );
    printError();
    SSL_Free( ssl );
    pthread_mutex_unlock( & sslMutex );
    return NULL;
  }
  
  /*
  ** Set peer certificate verification parameters.  Use the callback
  ** function already set for the ctx.
  **
  ** SSL_VERIFY_NONE means:
  ** The server will not send a client certificate request to the client,
  ** so the client will not send a certificate.
  **
  ** SSL_VERIFY_PEER means:
  ** The server certificate will be verified. If the verification process
  ** fails, the SSL handshake will be immediately terminated with an 
  ** alert message containing the reason for the verification failure. 
  ** If no server certificate is sent, because an anonymous cipher is used,
  ** SSL_VERIFY_PEER is ignored. 
  **
  ** SSL_VERIFY_FAIL_IF_NO_PEER_CERT means:
  ** If the client did not return a certificate, the TLS/SSL handshake
  ** is immediately terminated with a ``handshake failure'' alert. This
  ** flag must be used together with SSL_VERIFY_PEER.
  **  
  */  
  if ( context_type == SSL_CONTEXT_CERT ) 
  {
    SSL_set_verify(ssl, SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT, NULL);
  }
  else
  {
    SSL_set_verify(ssl, SSL_VERIFY_PEER, NULL);
  }

  //printCipherList( ssl );
  
  /*
  ** Connect the SSL object with the socket descriptor and wait for
  ** the client to initiate a handshake. Timeout if the handshake 
  ** is never initiated.
  */
  SSL_set_fd ( ssl, sockfd );
  err = Select_read(sockfd, handshake_timeout) ;
  if ( err <= 0 ) 
  { 
    avoPrintf(SSL_WARNING, "SSL_Server_Handshake SSL_read error\n");
    printError();
    SSL_Free( ssl );
    pthread_mutex_unlock( & sslMutex );
    return NULL;
  }
  else
  {
    err = SSL_accept( ssl );
    if ( err <= 0 ) 
    { 
      avoPrintf(SSL_WARNING, "SSL_accept error\n");
      printError();
      SSL_Free( ssl );
      pthread_mutex_unlock( & sslMutex );
      return NULL;
    }
  }
    
  pthread_mutex_unlock( & sslMutex );
  
  avoPrintf (SSL_INFO, "SSL connection using %s\n", SSL_get_cipher ( ssl ) );
  
  return ssl;

}

//---------------------------------------------------------------------
//    FUNCTION: SSL_Client_Handshake()
//
// DESCRIPTION: Perform the SSL handshaking and return the SSL object
//              
//  PARAMETERS: ctx, sockfd
//              
//     RETURNS: pointer to the SSL object
//
//       NOTES: 
//---------------------------------------------------------------------
void * SSL_Client_Handshake( void * ctx, int sockfd, int context_type )
{
  SSL * ssl;
  int err;
	unsigned char	dbgName[64];
	struct in_addr peerAddr;
  
  pthread_mutex_lock( & sslMutex );
  
  if ( !ctx )
  {
    avoPrintf(SSL_FAILURE, "ERROR: SSL_Client_Handshake NULL CTX\n");
    pthread_mutex_unlock( & sslMutex );
    return NULL;
  }
  
	GetPeerName(sockfd, & peerAddr );
	sprintf(dbgName, "SSL_Client_Handshake to %d.%d.%d.%d",
            (peerAddr.s_addr >>24) & 0xff,
						(peerAddr.s_addr >>16) & 0xff,
						(peerAddr.s_addr >>8) & 0xff,
						peerAddr.s_addr & 0xff);
  avoPrintf(SSL_INFO, "%s\n", dbgName);    
  
  /*
  ** Creates a new SSL structure which is needed to hold the data for a
  ** TLS/SSL connection.
  */    
  ssl = SSL_new ( (SSL_CTX *)ctx );
  if ( ssl == NULL )
  {
    avoPrintf(SSL_FAILURE, "Could not create SSL structure: \n");
    printError();
    pthread_mutex_unlock( & sslMutex );
    return NULL;
  }
  
  if( !init_adsap2_credentials ( ssl ) )
  {
    avoPrintf(SSL_WARNING, "SSL Client Handshake init_adsap2_credentials failed \n" );
    printError();
    SSL_Free( ssl );
    pthread_mutex_unlock( & sslMutex );
    return NULL;
  }
  
  /*
  ** Set peer certificate verification parameters.  Use the callback
  ** function already set for the ctx.
  ** SSL_VERIFY_PEER means:
  ** The server certificate will be verified. If the verification process
  ** fails, the SSL handshake will be immediately terminated with an 
  ** alert message containing the reason for the verification failure. 
  ** If no server certificate is sent, because an anonymous cipher is used,
  ** SSL_VERIFY_PEER is ignored. 
  */
  SSL_set_verify(ssl, SSL_VERIFY_PEER, NULL);
  
  //printCipherList( ssl );
  
  /*
  ** Connect the SSL object with the socket descriptor and connect to
  ** the server to initiate a handshake.
  */
  SSL_set_fd(ssl, sockfd);
  err = SSL_connect( ssl );
  if ( err < 0 ) 
  { 
    avoPrintf(SSL_WARNING, "SSL_connect error\n");
    printError();
    SSL_Free( ssl );
    pthread_mutex_unlock( & sslMutex );
    return NULL;
  }    
  pthread_mutex_unlock( & sslMutex );
  
  avoPrintf (SSL_INFO, "SSL connection using %s\n", SSL_get_cipher ( ssl ) );
  
  return ssl;

}

//---------------------------------------------------------------------
//    FUNCTION: init_ap_credential()
//
// DESCRIPTION: Load the certificates and private keys into the
//              SSL_CTX object. 
//              
//  PARAMETERS: SSL_CTX object ctx
//              
//     RETURNS: TRUE if successful, FALSE otherwise
//
//       NOTES: 
//              
//---------------------------------------------------------------------
static int init_ap_credential(SSL_CTX * ctx)
{
  /* 
  ** Checks the consistency of the private key with the corresponding
  ** certificate loaded into ctx.
  */
  if( !SSL_CTX_check_private_key(ctx) )
  {
    /*
    ** Loads the first certificate stored in file into ctx.
    */
    if( !SSL_CTX_use_certificate_file(ctx, "/mnt/jffs/srv_cert", SSL_FILETYPE_PEM) )
    {
      avoPrintf(SSL_FAILURE, "init_ap_credential: Could not load certificate file\n");
      return(FALSE);
    }

    /*
    ** Adds the first private key found in file to ctx.
    */
    if( !SSL_CTX_use_PrivateKey_file(ctx, "/mnt/jffs/srv_key", SSL_FILETYPE_PEM) )
    {
      avoPrintf(SSL_FAILURE, "init_ap_credential: Could not load private key file\n");
      return(FALSE);
    }
  
    /*
    ** Checks the consistency of a private key with the corresponding 
    ** certificate loaded into ctx.
    */    
    if( !SSL_CTX_check_private_key(ctx) )
    {
      avoPrintf(SSL_FAILURE, "init_ap_credential: Check Private Key failed\n");
      return(FALSE);
    }
  }

  return(TRUE);
}

//---------------------------------------------------------------------
//    FUNCTION: SSL_initialize_ctx()
//
// DESCRIPTION: Initialize a new SSL_CTX object
//              
//  PARAMETERS: context_type - determines type of SSL connection
//              server - TRUE if we are acting as the server
//                       FALSE if we are acting as the client
//              
//     RETURNS: CTX object if successful, NULL otherwise
//
//       NOTES: 
//              
//---------------------------------------------------------------------
int SSL_Init_ctx( void ** ctx, int context_type, int server, unsigned char * forceCipher )
{
  char cipherList[MAX_CIPHER_LIST];  
  int ret_val = FALSE;
  
  avoPrintf(SSL_INFO, "SSL_Init_ctx %s\n",
      (server == TRUE) ? "Server" : "Client");
	
  pthread_mutex_lock( & sslMutex );
  
  if ( *ctx == NULL )
  {
    // Create a new SSL_CTX object
    avoPrintf(SSL_INFO, "Create new SSL_CTX object\n");
    if( server )
      (SSL_CTX *)*ctx = SSL_CTX_new ( SSLv23_server_method() );
    else
      (SSL_CTX *)*ctx = SSL_CTX_new ( SSLv23_client_method() );
  }
  
  if( *ctx != NULL )
  {
    // Disable session caching 
    SSL_CTX_set_session_cache_mode((SSL_CTX *)*ctx, SSL_SESS_CACHE_OFF);
  
    if( ( context_type == SSL_CONTEXT_ANON) || 
        ( context_type == SSL_CONTEXT_ANON_OR_CERT ) )
    {
      // Set up for anonymous SSL
      if ( SSL_SelectDHparms( (SSL_CTX *)*ctx ) < 0 )
      {
        avoPrintf(SSL_FAILURE, "SSL_SelectDHparms failed\n");
        pthread_mutex_unlock( & sslMutex );
        return FALSE;
      }
      ret_val = TRUE;
    }
    else
    {
      // CERT-SSL, not anonymous
      ret_val = TRUE;
    }
  }
    
  if ( ret_val )
  {
    /*
    ** This is supposed to prevent SSL_read and SSL_write from returning
    ** with -1 to indicate that a retry is needed.
    */
		SSL_CTX_set_options( (SSL_CTX *)*ctx, SSL_MODE_AUTO_RETRY);

    /*
    ** Set the cipher
    */
		if (forceCipher)
    {
			avoPrintf(SSL_INFO, "Forced Cipher = %s\n", forceCipher);
      strcpy(cipherList,forceCipher);
    }
		else
    {
      avoPrintf(SSL_FAILURE, "No Forced Cipher!!!\n");
      pthread_mutex_unlock( & sslMutex );
      return FALSE;
    }

    if ( cipherList[0] )
    {
      if ( !SSL_CTX_set_cipher_list( (SSL_CTX *)*ctx, cipherList ) )
      {
        avoPrintf(SSL_FAILURE, "Error setting cipher list\n");
        printError();
        pthread_mutex_unlock( & sslMutex );
        return FALSE;
      }
      //avoPrintf(SSL_ADHOC, "SSL_CTX cipher list - %s\n", (SSL_CTX *)*ctx->cipher_list);
    }
  }
  pthread_mutex_unlock( & sslMutex );
  return ret_val;
}

//---------------------------------------------------------------------
//    FUNCTION: SSL_Get_Error()
//
// DESCRIPTION: Call SSL_get_error and return a code based on the
//              results
//              
//  PARAMETERS: ssl
//              err - return value of an SSL function call
//              
//     RETURNS: SSL_ERROR - indictes socket error
//              SSL_RETRY - indicated need to retry the call
//
//       NOTES: 
//---------------------------------------------------------------------
int SSL_Get_Error( void * ssl, int err )
{
  int retval;
  
  retval = SSL_get_error( (SSL *)ssl, err );
  
  switch( retval )
  {
    case SSL_ERROR_ZERO_RETURN:       /* 6 */
      avoPrintf(SSL_INFO, "SSL clean shutdown \n");
      return SSL_ERROR;
    
    case SSL_ERROR_SYSCALL:           /* 5 */
      perror("SSL_ERROR_SYSCALL");
      return SSL_ERROR;
    
    case SSL_ERROR_NONE:              /* 0 */
    case SSL_ERROR_SSL:               /* 1 */
      avoPrintf(SSL_FAILURE, "SSL_get_error returned %d \n", retval );
      return SSL_ERROR;
      
    case SSL_ERROR_WANT_READ:         /* 2 */
    case SSL_ERROR_WANT_WRITE:        /* 3 */
    case SSL_ERROR_WANT_X509_LOOKUP:  /* 4 */
    case SSL_ERROR_WANT_CONNECT:      /* 7 */
      avoPrintf(SSL_FAILURE, "SSL_get_error returned %d \n", retval );
      return SSL_RETRY_NEEDED;
  }
  return SSL_ERROR;  
}

//---------------------------------------------------------------------
//    FUNCTION: SSL_Pending()
//
// DESCRIPTION: Wrapper function for SSL_pending
//              
//  PARAMETERS: ssl
//              
//     RETURNS: Number of bytes available inside ssl for immediate read
//
//       NOTES: 
//---------------------------------------------------------------------
int SSL_Pending( void * ssl )
{
  return ( SSL_pending ( (SSL *)ssl ) );
}

//---------------------------------------------------------------------
//    FUNCTION: SSL_Read()
//
// DESCRIPTION: Wrapper function for SSL_read
//              
//  PARAMETERS: none
//              
//     RETURNS: number of bytes read, or -1 for error
//
//       NOTES: 
//---------------------------------------------------------------------
int SSL_Read( void * ssl, byte * buf, int len )
{
  return ( SSL_read ( (SSL *)ssl, buf, len ) );
}

//---------------------------------------------------------------------
//    FUNCTION: ssl_do_read_select()
//
// DESCRIPTION: See if there is SSL data waiting
//              
//  PARAMETERS: ssl
//              timeout - number of seconds to wait for data
//              
//     RETURNS: TRUE if socket is ready for reading
//
//       NOTES: 
//              
//---------------------------------------------------------------------
static int ssl_do_read_select(SSL * ssl, unsigned long lTimeout)
{
  struct in_addr peerAddr;
  int sock = SSL_get_fd( ssl );

  if( SSL_Pending( ssl ) )
    return(TRUE);

  if ( GetPeerName( sock, & peerAddr ) < 0 ) 
    return(FALSE); /* We are no longer connected */

  return( Select_read(sock, lTimeout) );
}

//---------------------------------------------------------------------
//    FUNCTION: SSL_Get_Data()
//
// DESCRIPTION: Read in data over the SSL connection
//              
//  PARAMETERS: ssl
//              buf - place to store data
//              num - number of bytes to read in
//              timeout - number of seconds to wait for data
//              
//     RETURNS: number of bytes read in
//
//       NOTES: 
//              
//---------------------------------------------------------------------
int SSL_Get_Data(void * ssl, unsigned char *buf, unsigned long num, unsigned long timeout)
{
  int  total_data = 0;
  int  data;
  int  sel;

  avoPrintf(SSL_ADHOC, "SSL_Get_Data\n");
  while( total_data < ((int) num) )
  {
    /*
    ** Check to see if data is available
    */
    if( (sel = ssl_do_read_select((SSL*)ssl, timeout)) == FALSE )
    {
      break;
    }
    
    /*
    ** Read in the data
    */
    if((data = SSL_Read(ssl, buf + total_data, num - total_data)) == 0)
    {
      return(-1);
    }

    total_data += data;
  }

  avoPrintf(SSL_ADHOC, "SSL_Get_Data returns %d bytes read\n", total_data);
  return(total_data);
}

//---------------------------------------------------------------------
//    FUNCTION: SSL_Write()
//
// DESCRIPTION: Wrapper function for SSL_write
//              
//  PARAMETERS: none
//              
//     RETURNS: number of bytes written, or -1 for error
//
//       NOTES: 
//---------------------------------------------------------------------
int SSL_Write( void * ssl, byte * buf, int len )
{
  return ( SSL_write ( (SSL *)ssl, buf, len ) );
}
 
//---------------------------------------------------------------------
//    FUNCTION: SSL_Shutdown()
//
// DESCRIPTION: Wrapper function for SSL_shutdown
//              
//  PARAMETERS: none
//              
//     RETURNS: 
//
//       NOTES: 
//---------------------------------------------------------------------
int SSL_Shutdown( void * ssl )
{
  int retval;
  
  retval = SSL_shutdown( (SSL *) ssl );
  
  if (retval < 0 )
  {
    avoPrintf(SSL_WARNING, "SSL_Shutdown error!\n");
  }
  else if ( retval == 0 )
  {
    // OpenSSL documents indicate we may need to do the shutdown call again
    if ( SSL_shutdown( (SSL *) ssl ) < 0 )
    {
      avoPrintf(SSL_WARNING, "SSL_Shutdown error!\n");
    }
  }
  return 0;
}
 
//---------------------------------------------------------------------
//    FUNCTION: SSL_Free()
//
// DESCRIPTION: Wrapper function for SSL_free
//              
//  PARAMETERS: none
//              
//     RETURNS: none
//
//       NOTES: 
//---------------------------------------------------------------------
void SSL_Free( void * ssl )
{
  if ( ssl )
  {
    SSL_free( (SSL *) ssl );
  }
  else
  {
    avoPrintf(SSL_FAILURE, "SSL_Free: ssl was null! \n");
  }
  
  /*
  ** Free the error queue for the current thread.  Doing this seems
  ** to prevent a memory leak.
  */
  ERR_remove_state(0);
}

//---------------------------------------------------------------------
//    FUNCTION: SSL_ERR_Remove_State()
//
// DESCRIPTION: Wrapper function for ERR_remove_state
//              
//  PARAMETERS: none
//              
//     RETURNS: none
//
//       NOTES: 
//---------------------------------------------------------------------
void SSL_ERR_Remove_State( void )
{
  /*
  ** Free the error queue for the current thread.  Doing this prevents
  ** a memory leak.
  */
  ERR_remove_state(0);
}
//---------------------------------------------------------------------
//    FUNCTION: SSL_Get_fd()
//
// DESCRIPTION: Wrapper function for SSL_get_fd
//              
//  PARAMETERS: ssl
//              
//     RETURNS: none
//
//       NOTES: 
//---------------------------------------------------------------------
int SSL_Get_fd( void * ssl )
{
  return SSL_get_fd((SSL *)ssl);
}

//---------------------------------------------------------------------
//    FUNCTION: SSL_Get_cipher()
//
// DESCRIPTION: Wrapper function for SSL_get_cipher
//              Same as SSL_get_current_cipher(), which returns a 
//              pointer to an SSL_CIPHER object containing the 
//              description of the actually used cipher of a connection
//              established with the ssl object.
//              
//  PARAMETERS: ssl
//              
//     RETURNS: none
//
//       NOTES: 
//---------------------------------------------------------------------
unsigned char * SSL_Get_cipher( void * ssl )
{
  return (unsigned char * )SSL_get_cipher((SSL *)ssl);
}

//---------------------------------------------------------------------
//    FUNCTION: SSL_Get_peer_certificate()
//
// DESCRIPTION: Wrapper function for SSL_get_peer_certificate
//              
//  PARAMETERS: ssl
//              
//     RETURNS: Returns a pointer to the X509 certificate the peer
//              presented. If the peer did not present a certificate,
//              NULL is returned.
//
//       NOTES: 
//---------------------------------------------------------------------
void * SSL_Get_peer_certificate( void * ssl )
{
  return (void *)SSL_get_peer_certificate((SSL *)ssl);
}


//---------------------------------------------------------------------
//    FUNCTION: SSL_CTX_Free()
//
// DESCRIPTION: Wrapper function for SSL_CTX_free
//              
//  PARAMETERS: none
//              
//     RETURNS: none
//
//       NOTES: 
//---------------------------------------------------------------------
void SSL_CTX_Free( void * ctx )
{
  SSL_CTX_free( (SSL_CTX *) ctx );
}

//---------------------------------------------------------------------
//    FUNCTION: X509_Free()
//
// DESCRIPTION: Wrapper function for X509_free
//              
//  PARAMETERS: none
//              
//     RETURNS: none
//
//       NOTES: 
//---------------------------------------------------------------------
void X509_Free( void * X509_peer_cert )
{
  X509_free( (X509 *) X509_peer_cert );
}

//---------------------------------------------------------------------
//    FUNCTION: SSL_Close_Connection()
//
// DESCRIPTION: Closes an SSL connection and frees memory
//              
//  PARAMETERS: none
//              
//     RETURNS: none
//
//       NOTES: 
//              
//---------------------------------------------------------------------
void SSL_Close_Connection(void * ssl)
{
  int  sock;
  int    ret;
  struct in_addr peerAddr;
  
  avoPrintf(SSL_ADHOC, "SSL_Close_Connection\n");
	
  if( ssl )
  {
    sock = SSL_get_fd( (SSL *)ssl );
    if ( sock >= 0 )
    {
      /*
      ** This is an SSL connection, send a shutdown message and see
      ** if  the peer has also completed shutdown.
      **
      ** I'm setting the the SIGPIPE signal to be ignored.  Otherwise if
      ** the connection has been completely closed on the other end, then
      ** the call to shutdown causes a SIGPIPE signal which causes the
      ** process to exit.  Hope this is the right thing to do. - Jamie
      */
      signal(SIGPIPE, SIG_IGN);
      
      ret = SSL_shutdown( (SSL *)ssl );
      if ( ret == 0 )
      {
        avoPrintf(SSL_ADHOC, "SSL_Close_Connection: Socket %d - SSL_shutdown did not complete(1)\n", sock);
        
        /* 
        ** The SSL_shutdown was not completed with the peer.
        ** The peer may have disappeared.  Force the socket layer to
        ** return an error to the SSL layer by disabling send and recv
        ** Shutdown the socket, disabling both send and receive operations
        */
        shutdown(sock, SHUT_RDWR);

		tcflush(sock, TCIOFLUSH);

        
        /*
        ** If we are still connected, call SSL_shutdown again to see if
        ** the SSL layer will clean up the connection.
        */
        if ( GetPeerName( sock, & peerAddr ) == 0 ) 
        {
          ret = SSL_shutdown((SSL *)ssl);
        }
        else
          ret = 2;
      }

      if(ret == 0)
        avoPrintf(SSL_WARNING, "SSL_Close_Connection: Socket %d - SSL_shutdown NOT completed\n", sock);
      if(ret == 1)
        avoPrintf(SSL_INFO, "SSL_Close_Connection: Socket %d - SSL_shutdown completed\n", sock);
      if(ret == 2)
        avoPrintf(SSL_WARNING, "SSL_Close_Connection: Socket %d already closed\n", sock);
      
      SSL_Free((SSL *)ssl);
      Close(sock);
      // Do not free the SSL_CTX struct here!
    }
    else
      avoPrintf(SSL_FAILURE, "SSL_Close_Connection: sock is <0 \n");
  }
  else
    avoPrintf(SSL_FAILURE, "SSL_Close_Connection: ssl pointer is NULL\n");
}

//---------------------------------------------------------------------
//    FUNCTION: SSL_Init()
//
// DESCRIPTION: Initialize the SSL Library
//              
//  PARAMETERS: none
//              
//     RETURNS: none
//
//       NOTES: Must be called first before any other SSL action takes
//              place
//---------------------------------------------------------------------
void SSL_Init( void )
{
  avoPrintf(SSL_INFO, "SSL Init \n");
  
  pthread_mutex_init ( & sslMutex, NULL );
  
#ifdef SSL_ACCEL
  // initialize mpc184
  ssl_hardware_init();
#else
  avoPrintf(SSL_INFO, "SSL HW Encryption disabled!!\n");
#endif
  /*
  ** Provide readable error messages
  */
  SSL_load_error_strings();
  
  /*
  ** Load all algorithms to the table (digests and ciphers) 
  */
  OpenSSL_add_ssl_algorithms();
  
  /*
  ** Set up callback functions for multi-threading
  */
  SSL_MultiThreadSetup();
}




/* end */


