//------------------------------------------------------------------------
// Copyright 2002, Avocent Corporation Corporation
//------------------------------------------------------------------------
// FILE        :  threadwrap.c
// DESCRIPTION :  This file contains wrapper functions for pthread
//                programming calls
//
//------------------------------------------------------------------------
// AUTHOR      :  J. Collins
// DATE        :  January 2000
//
//------------------------------------------------------------------------

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <signal.h>

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "types.h"
#include "avoprint.h"
#include "threadwrap.h"

/* 
** Debugging of heap memory allocations 
*/
#ifdef DMALLOC    
  #include <dmalloc.h>
#endif

//---------------------------------------------------------------------
//    FUNCTION: pthreadCreate
//
// DESCRIPTION: Create a new pthread 
//							
//  PARAMETERS: thread ID
//              thread attribute (IGNORED)
//              start routine
//              arguments to pass to the start routine
//							
//     RETURNS: none
//
//       NOTES: 
//
//---------------------------------------------------------------------
int pthreadCreate(pthread_t *tid, const pthread_attr_t *attr,
                   void * (*func)(void *), void *arg)
{
  int err;
  sigset_t addMask, oldMask;

  /*
   * mask SIGINT briefly while child thread is started
   * so it won't process the SIGINT signal. 
   * Only the main() thread is to catch this signal. 
   */
  sigemptyset( & addMask );
  sigaddset( & addMask, SIGINT );
  pthread_sigmask( SIG_BLOCK, & addMask, & oldMask );

  if ( (err = pthread_create( tid, attr, func, arg )) != 0 ) 
  {
    /* 
     * Do not use avoPrintf() here. 
     * The debug logger can call pthreadCreate(). 
     */
    printf( "pthreadCreate: error: %s \n", strerror(err) );
  }

  /*
   * restore signal mask 
   */
  pthread_sigmask( SIG_SETMASK, & oldMask, NULL );

  return err;
}

//---------------------------------------------------------------------
//    FUNCTION: pthreadDetach
//
// DESCRIPTION: Detach a pthread (place it into the detached state, 
//              which guarantees that memory resources consumed by
//              the thread will be freed immediatly when the thread
//              terminates. 
//							
//  PARAMETERS: thread ID
//							
//     RETURNS: none
//
//       NOTES: 
//
//---------------------------------------------------------------------
void pthreadDetach( pthread_t tid )
{
  pthread_detach(tid);
}

/* end  */

