/*
 *  Module:  aidp.c (Avocent Install and Discovery Protocol)
 *
 *  Author:  Marcelo Peccin <marcelo.peccin@avocent.com>
 *
 *  From the original:  Avocent Install and Discovery Protocol (aidp.c file)
 *             Author:  John M. Welch
 *
 *  Copyright:  (c) 2006 Avocent Corporation
 *
 *  Date:  May 10, 2006
 *
 *  Comments:  The Install and Discovery protocol here implemented is a port
 *             to former Cyclades products of the original Avocent Install and
 *             Discovery Protocol.
 *             It is intended to run as a daemon at appliance's first boot.
 *             This implementation is compliant with Avocent Install and Discovery
 *             Protocol Specification (Doc# 590064, Version 1.3) and responds to
 *             the following commands at UDP port 3211:
 *             1) Discover
 *             2) TestIpConfigRequest
 *             3) SetIpConfigRequest
 *             After appliance's IP address gets configured, the daemon
 *             should stop running.
 *             See also aidpaux.c (for auxiliary functions implementation).
 */

#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <linux/if_ether.h> //ETH_P_IP
#include <netinet/in.h>     //htons(), etc
#include <stdarg.h>         //va_start()
#include <sys/stat.h>
#include <fcntl.h>

#include "aidp.h"
#include "aidpaux.h"


int debug = TRUE;
static int mypid;
static int logfd = -1;
static time_t timenow;
static struct sigaction sa;


/****************************************************************/
/**********     Register caught signal to log file     **********/
/****************************************************************/
static void aidp_got_signal(int sig)
{	
	if (time(&timenow) < 0) timenow = 0;

	switch(sig) {
	case SIGHUP:
	case SIGINT:
	case SIGTERM:
		printlog("%s --> aidpd(pid %d) got signal %s\n", ctime(&timenow), mypid, strsignal(sig));
	}

	exit(0);
}

/***********************************************************/
/**********     Prepare to catch some signals     **********/
/***********************************************************/
static void prepare_signals()
{
	bzero((void*)&sa, sizeof(struct sigaction));

	sa.sa_handler = aidp_got_signal;

	sigaction(SIGHUP,  &sa, NULL);
	sigaction(SIGINT,  &sa, NULL);
	sigaction(SIGTERM, &sa, NULL);
}
/***********************************************************/


/******************************/
int main(int argc, char** argv)
/******************************/
{
	AIDP_SOCKET_MESSAGE  psocket_data;
	APP_AIDP_MESSAGE     aidp_msg;
	byte                 *prxbuf, *prxmsg, *ptxbuf, *ptxmsg, *precord_len, *pend_data;
	struct timeval       select_timeout;
	fd_set               FDRead;
	int                  arg, select_count;
	int                  sendlen=0, sentlen=0, setIPsuccess=FALSE, last_msg=-1;
	byte                 temp_ip_addr[4];
	byte                 temp_netmask[4];
	byte                 temp_gw_addr[4];

	mypid=getpid();

	if (time(&timenow) < 0) timenow = 0;
	printlog("%s --> aidpd(pid %d) started...\n", ctime(&timenow), mypid);

	for (arg=1; arg < argc; arg++) {
		if (!strcmp(argv[arg], "-debug")) {
			debug = TRUE;
		}
	}	

	if (debug)
		prepare_signals();

	// Allocate a raw socket for UDP
	if ((psocket_data.udp_socket = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_IP))) < 0)
	{
		printlog("Could not allocate raw socket: %s\n", strerror(errno));
		return 1;
	}

	// Allocate memory
	if (!(prxbuf = (byte *)malloc(RXMAX_UDPBUF)))
	{
		printlog("Could not allocate socket RX buffer\n");
		return 1;
	}
	if (!(ptxbuf = (byte *)malloc(TXMAX_UDPBUF)))
	{
		printlog("Could not allocate socket TX buffer\n");
		return 1;
	}

	// Get necessary information from Appliance
	if (initAppVars()) {
		printlog("Could not collect necessary info\n");
		return 1;
	}

	psocket_data.MsgData = prxbuf;/* where AIDP messages are received */

	for (;;)
	{
		// Must reset the time before every call to select
		select_timeout.tv_sec = 5;
		select_timeout.tv_usec = 0;

		prxmsg = prxbuf; /* update pointer to receive message */
		ptxmsg = ptxbuf; /* update pointer to build message */

		FD_ZERO(&FDRead);
		FD_SET(psocket_data.udp_socket, &FDRead);

		select_count = select(psocket_data.udp_socket+1, &FDRead, NULL, NULL, &select_timeout);

		if (select_count <= 0) {
			if (select_count < 0)
				printlog("Error reading raw socket: %s, strerror(errno)\n");
			if (setIPsuccess) {
				//the whole precess is finished
				if (kill_other_dynamicIP_daemons() < 0)
					printlog("Could not kill other dynamic IP daemons\n");

				if (time(&timenow) < 0) timenow = 0;
				printlog("%s --> aidpd(pid %d) finished.\n", ctime(&timenow), mypid);
				return 0;
			}
			sendlen = 0;
			continue;
		}

		// Message received. Process it.
		if (process_socket_message(&psocket_data))
			continue;

		// AIDP message completely received. Copy all data.
		RX_NORM_BYTE(&prxmsg, aidp_msg.MsgHdr.bySOH, TRUE);
		RX_NORM_STRING(&prxmsg, aidp_msg.MsgHdr.szUniqueID, 4, TRUE);
		RX_NORM_WORD(&prxmsg, &aidp_msg.MsgHdr.wSequenceNumber, TRUE);
		RX_NORM_BYTE(&prxmsg, aidp_msg.MsgHdr.byMsgType, TRUE);
		RX_NORM_DWORD(&prxmsg, &aidp_msg.MsgHdr.dwMsgLen, TRUE);

		if (debug) {
			if (time(&timenow) < 0) timenow = 0;
			printlog("===============\n%x %c%c%c%c %04x %x %d at %s===============\n",
					 aidp_msg.MsgHdr.bySOH,
					 aidp_msg.MsgHdr.szUniqueID[0], aidp_msg.MsgHdr.szUniqueID[1],
					 aidp_msg.MsgHdr.szUniqueID[2], aidp_msg.MsgHdr.szUniqueID[3], 
					 ntohs(aidp_msg.MsgHdr.wSequenceNumber),
					 aidp_msg.MsgHdr.byMsgType,
					 ntohl(aidp_msg.MsgHdr.dwMsgLen),
					 ctime(&timenow));
		}

		if (sendlen && (last_msg == ntohs(aidp_msg.MsgHdr.wSequenceNumber)))
			goto send_again;

		//prepare variables
		sendlen = 0;
		aidp_msg.MsgHdr.dwMsgLen = ntohl(aidp_msg.MsgHdr.dwMsgLen);
		last_msg = ntohs(aidp_msg.MsgHdr.wSequenceNumber);

		//check Record Data integrity
		if (*(prxbuf + sizeof(APP_AIDP_HEADER) + aidp_msg.MsgHdr.dwMsgLen - 1) != 0xFF)
		{
			printlog("Wrong AIDP message\n");
			continue;
		}

		// Call the correct routine to process this AIDP message number
		switch(aidp_msg.MsgHdr.byMsgType) {

		case REQ_DISCOVER:
		{
			printDEBUG("Discover Request from %d.%d.%d.%d(%d)\n",
					   psocket_data.recvfrom_ip[0],
					   psocket_data.recvfrom_ip[1],
					   psocket_data.recvfrom_ip[2],
					   psocket_data.recvfrom_ip[3],
					   ntohs(psocket_data.recvfrom_port));

			aidp_msg.MsgHdr.byMsgType = REPLY_DISCOVER;

			aidp_msg.MsgData.APPDiscover.Appliance.byFieldID = 1;
			aidp_msg.MsgData.APPDiscover.Appliance.wFieldLen = htons(2);
			aidp_msg.MsgData.APPDiscover.Appliance.wAppliance =	htons(appID);

			aidp_msg.MsgData.APPDiscover.MAC_Addr.byFieldID = 2;
			aidp_msg.MsgData.APPDiscover.MAC_Addr.wFieldLen = htons(6);
			memcpy(aidp_msg.MsgData.APPDiscover.MAC_Addr.byMacAddr, appMAC, 6);

			aidp_msg.MsgData.APPDiscover.IP_Addr.byFieldID = 3;
			aidp_msg.MsgData.APPDiscover.IP_Addr.wFieldLen = htons(4);
			memcpy(aidp_msg.MsgData.APPDiscover.IP_Addr.byIPAddr, appIP, 4);

			aidp_msg.MsgData.APPDiscover.Mask.byFieldID = 4;
			aidp_msg.MsgData.APPDiscover.Mask.wFieldLen = htons(4);
			memcpy(aidp_msg.MsgData.APPDiscover.Mask.byMask, appSubNet, 4);

			aidp_msg.MsgData.APPDiscover.GW_Addr.byFieldID = 5;
			aidp_msg.MsgData.APPDiscover.GW_Addr.wFieldLen = htons(4);
			memcpy(&aidp_msg.MsgData.APPDiscover.GW_Addr.byGWAddr, appGw, 4);

			TX_NORM_BYTE(aidp_msg.MsgHdr.bySOH, &ptxmsg, TRUE);
			TX_NORM_STRING(aidp_msg.MsgHdr.szUniqueID, &ptxmsg, 4, TRUE);
			TX_NORM_WORD(&aidp_msg.MsgHdr.wSequenceNumber, &ptxmsg, TRUE);
			TX_NORM_BYTE(aidp_msg.MsgHdr.byMsgType, &ptxmsg, TRUE);
			//save this address to load size of data record later
			precord_len = ptxmsg;
			//just increment pointer...
			TX_NORM_DWORD(&aidp_msg.MsgHdr.dwMsgLen, &ptxmsg, TRUE);

			TX_NORM_BYTE(aidp_msg.MsgData.APPDiscover.Appliance.byFieldID, &ptxmsg, TRUE);
			TX_NORM_WORD(&aidp_msg.MsgData.APPDiscover.Appliance.wFieldLen, &ptxmsg, TRUE);
			TX_NORM_WORD(&aidp_msg.MsgData.APPDiscover.Appliance.wAppliance, &ptxmsg, TRUE);

			TX_NORM_BYTE(aidp_msg.MsgData.APPDiscover.MAC_Addr.byFieldID, &ptxmsg, TRUE);
			TX_NORM_WORD(&aidp_msg.MsgData.APPDiscover.MAC_Addr.wFieldLen, &ptxmsg, TRUE);
			TX_NORM_STRING(aidp_msg.MsgData.APPDiscover.MAC_Addr.byMacAddr, &ptxmsg, 6, TRUE);

			TX_NORM_BYTE(aidp_msg.MsgData.APPDiscover.IP_Addr.byFieldID, &ptxmsg, TRUE);
			TX_NORM_WORD(&aidp_msg.MsgData.APPDiscover.IP_Addr.wFieldLen, &ptxmsg, TRUE);
			TX_NORM_STRING(aidp_msg.MsgData.APPDiscover.IP_Addr.byIPAddr, &ptxmsg, 4, TRUE);

			TX_NORM_BYTE(aidp_msg.MsgData.APPDiscover.Mask.byFieldID, &ptxmsg, TRUE);
			TX_NORM_WORD(&aidp_msg.MsgData.APPDiscover.Mask.wFieldLen, &ptxmsg, TRUE);
			TX_NORM_STRING(aidp_msg.MsgData.APPDiscover.Mask.byMask, &ptxmsg, 4, TRUE);

			TX_NORM_BYTE(aidp_msg.MsgData.APPDiscover.GW_Addr.byFieldID, &ptxmsg, TRUE);
			TX_NORM_WORD(&aidp_msg.MsgData.APPDiscover.GW_Addr.wFieldLen, &ptxmsg, TRUE);
			TX_NORM_STRING(aidp_msg.MsgData.APPDiscover.GW_Addr.byGWAddr, &ptxmsg, 4, TRUE);

			sendlen = 1; /* mark it to do the send */
		}
		break;

		case REQ_TESTIPADDR:
		{
			RX_NORM_BYTE(&prxmsg, aidp_msg.MsgData.NTTestIpAddr.MAC_Addr.byFieldID, TRUE);
			RX_NORM_WORD(&prxmsg, &aidp_msg.MsgData.NTTestIpAddr.MAC_Addr.wFieldLen, TRUE);
			RX_NORM_STRING(&prxmsg, aidp_msg.MsgData.NTTestIpAddr.MAC_Addr.byMacAddr, 6, TRUE);

			RX_NORM_BYTE(&prxmsg, aidp_msg.MsgData.NTTestIpAddr.IP_Addr.byFieldID, TRUE);
			RX_NORM_WORD(&prxmsg, &aidp_msg.MsgData.NTTestIpAddr.IP_Addr.wFieldLen, TRUE);
			RX_NORM_STRING(&prxmsg, aidp_msg.MsgData.NTTestIpAddr.IP_Addr.byIPAddr, 4, TRUE);

			RX_NORM_BYTE(&prxmsg, aidp_msg.MsgData.NTTestIpAddr.Mask.byFieldID, TRUE);
			RX_NORM_WORD(&prxmsg, &aidp_msg.MsgData.NTTestIpAddr.Mask.wFieldLen, TRUE);
			RX_NORM_STRING(&prxmsg, aidp_msg.MsgData.NTTestIpAddr.Mask.byMask, 4, TRUE);

			RX_NORM_BYTE(&prxmsg, aidp_msg.MsgData.NTTestIpAddr.GW_Addr.byFieldID, TRUE);
			RX_NORM_WORD(&prxmsg, &aidp_msg.MsgData.NTTestIpAddr.GW_Addr.wFieldLen, TRUE);
			RX_NORM_STRING(&prxmsg, aidp_msg.MsgData.NTTestIpAddr.GW_Addr.byGWAddr, 4, TRUE);

			// Test MAC Address: is this msg for me?
			if (memcmp(aidp_msg.MsgData.NTTestIpAddr.MAC_Addr.byMacAddr, appMAC, 6))
				break;

			aidp_msg.MsgHdr.byMsgType = REPLY_TESTIPADDR;

			aidp_msg.MsgData.APPTestIpAddr.APP_Status.byFieldID = 1;
			aidp_msg.MsgData.APPTestIpAddr.APP_Status.wFieldLen = htons(2);

			TX_NORM_BYTE(aidp_msg.MsgHdr.bySOH, &ptxmsg, TRUE);
			TX_NORM_STRING(aidp_msg.MsgHdr.szUniqueID, &ptxmsg, 4, TRUE);
			TX_NORM_WORD(&aidp_msg.MsgHdr.wSequenceNumber, &ptxmsg, TRUE);
			TX_NORM_BYTE(aidp_msg.MsgHdr.byMsgType, &ptxmsg, TRUE);
			//save this address to load size of data record later
			precord_len = ptxmsg;
			//just increment pointer...
			TX_NORM_DWORD(&aidp_msg.MsgHdr.dwMsgLen, &ptxmsg, TRUE);

			TX_NORM_BYTE(aidp_msg.MsgData.APPTestIpAddr.APP_Status.byFieldID, &ptxmsg, TRUE);
			TX_NORM_WORD(&aidp_msg.MsgData.APPTestIpAddr.APP_Status.wFieldLen, &ptxmsg, TRUE);

			//check Data Field order (we assumed that above...)
			if ((aidp_msg.MsgData.NTTestIpAddr.IP_Addr.byFieldID != 2) ||
				(aidp_msg.MsgData.NTTestIpAddr.Mask.byFieldID != 3) ||
				(aidp_msg.MsgData.NTTestIpAddr.GW_Addr.byFieldID != 4))
			{
				aidp_msg.MsgData.APPTestIpAddr.APP_Status.wStatus =
					htons(STAT_INV_REQ);
				TX_NORM_WORD(&aidp_msg.MsgData.APPTestIpAddr.APP_Status.wStatus, &ptxmsg, TRUE);
				sendlen = 1; /* mark it to do the send */
				break;
			}

			memcpy(temp_ip_addr, aidp_msg.MsgData.NTTestIpAddr.IP_Addr.byIPAddr, 4);
			memcpy(temp_netmask, aidp_msg.MsgData.NTTestIpAddr.Mask.byMask, 4);
			memcpy(temp_gw_addr, aidp_msg.MsgData.NTTestIpAddr.GW_Addr.byGWAddr, 4);

			printDEBUG("Test IP/subnet/Gw from %d.%d.%d.%d(%d)\n",
					   psocket_data.recvfrom_ip[0],
					   psocket_data.recvfrom_ip[1],
					   psocket_data.recvfrom_ip[2],
					   psocket_data.recvfrom_ip[3],
					   ntohs(psocket_data.recvfrom_port));

			if (set_appliance_ip(temp_ip_addr, temp_netmask, temp_gw_addr))
			{
				printlog("Could not set appliance's IP\n");
				aidp_msg.MsgData.APPTestIpAddr.APP_Status.wStatus = ntohs(STAT_APP_ERR);
			} else {
				aidp_msg.MsgData.APPTestIpAddr.APP_Status.wStatus = ntohs(STAT_SUCCESS);
			}

			TX_NORM_WORD(&aidp_msg.MsgData.APPTestIpAddr.APP_Status.wStatus, &ptxmsg, TRUE);
			sendlen = 1; /* mark it to do the send */
		}
		break;

		case REQ_SETIPADDR:
		{
			RX_NORM_BYTE(&prxmsg, aidp_msg.MsgData.NTSetIpAddr.MAC_Addr.byFieldID, TRUE);
			RX_NORM_WORD(&prxmsg, &aidp_msg.MsgData.NTSetIpAddr.MAC_Addr.wFieldLen, TRUE);
			RX_NORM_STRING(&prxmsg, aidp_msg.MsgData.NTSetIpAddr.MAC_Addr.byMacAddr, 6, TRUE);

			RX_NORM_BYTE(&prxmsg, aidp_msg.MsgData.NTSetIpAddr.IP_Addr.byFieldID, TRUE);
			RX_NORM_WORD(&prxmsg, &aidp_msg.MsgData.NTSetIpAddr.IP_Addr.wFieldLen, TRUE);
			RX_NORM_STRING(&prxmsg, aidp_msg.MsgData.NTSetIpAddr.IP_Addr.byIPAddr, 4, TRUE);

			RX_NORM_BYTE(&prxmsg, aidp_msg.MsgData.NTSetIpAddr.Mask.byFieldID, TRUE);
			RX_NORM_WORD(&prxmsg, &aidp_msg.MsgData.NTSetIpAddr.Mask.wFieldLen, TRUE);
			RX_NORM_STRING(&prxmsg, aidp_msg.MsgData.NTSetIpAddr.Mask.byMask, 4, TRUE);

			RX_NORM_BYTE(&prxmsg, aidp_msg.MsgData.NTSetIpAddr.GW_Addr.byFieldID, TRUE);
			RX_NORM_WORD(&prxmsg, &aidp_msg.MsgData.NTSetIpAddr.GW_Addr.wFieldLen, TRUE);
			RX_NORM_STRING(&prxmsg, aidp_msg.MsgData.NTSetIpAddr.GW_Addr.byGWAddr, 4, TRUE);

			// Test MAC Address: is this msg for me?
			if (memcmp(aidp_msg.MsgData.NTSetIpAddr.MAC_Addr.byMacAddr, appMAC, 6))
				break;

			aidp_msg.MsgHdr.byMsgType = REPLY_SETIPADDR;

			aidp_msg.MsgData.APPSetIpAddr.APP_Status.byFieldID = 1;
			aidp_msg.MsgData.APPSetIpAddr.APP_Status.wFieldLen = htons(2);

			TX_NORM_BYTE(aidp_msg.MsgHdr.bySOH, &ptxmsg, TRUE);
			TX_NORM_STRING(aidp_msg.MsgHdr.szUniqueID, &ptxmsg, 4, TRUE);
			TX_NORM_WORD(&aidp_msg.MsgHdr.wSequenceNumber, &ptxmsg, TRUE);
			TX_NORM_BYTE(aidp_msg.MsgHdr.byMsgType, &ptxmsg, TRUE);
			//save this address to load size of data record later
			precord_len = ptxmsg;
			//just increment pointer...
			TX_NORM_DWORD(&aidp_msg.MsgHdr.dwMsgLen, &ptxmsg, TRUE);

			TX_NORM_BYTE(aidp_msg.MsgData.APPTestIpAddr.APP_Status.byFieldID, &ptxmsg, TRUE);
			TX_NORM_WORD(&aidp_msg.MsgData.APPTestIpAddr.APP_Status.wFieldLen, &ptxmsg, TRUE);

			//check Data Field order (we assumed that above...)
			if ((aidp_msg.MsgData.NTSetIpAddr.IP_Addr.byFieldID != 2) ||
				(aidp_msg.MsgData.NTSetIpAddr.Mask.byFieldID != 3) ||
				(aidp_msg.MsgData.NTSetIpAddr.GW_Addr.byFieldID != 4))
			{
				aidp_msg.MsgData.APPSetIpAddr.APP_Status.wStatus =
					htons(STAT_INV_REQ);
				TX_NORM_WORD(&aidp_msg.MsgData.APPSetIpAddr.APP_Status.wStatus, &ptxmsg, TRUE);
				sendlen = 1; /* mark it to do the send */
				break;
			}

			memcpy(temp_ip_addr, aidp_msg.MsgData.NTSetIpAddr.IP_Addr.byIPAddr, 4);
			memcpy(temp_netmask, aidp_msg.MsgData.NTSetIpAddr.Mask.byMask, 4);
			memcpy(temp_gw_addr, aidp_msg.MsgData.NTSetIpAddr.GW_Addr.byGWAddr, 4);

			printDEBUG("Set IP/subnet/Gw from %d.%d.%d.%d(%d)\n",
					   psocket_data.recvfrom_ip[0],
					   psocket_data.recvfrom_ip[1],
					   psocket_data.recvfrom_ip[2],
					   psocket_data.recvfrom_ip[3],
					   ntohs(psocket_data.recvfrom_port));

			if (update_appliance_ip(temp_ip_addr, temp_netmask, temp_gw_addr))
			{
				printlog("Could not update appliance's IP\n");
				aidp_msg.MsgData.APPSetIpAddr.APP_Status.wStatus = ntohs(STAT_APP_ERR);
			} else {
				aidp_msg.MsgData.APPSetIpAddr.APP_Status.wStatus = htons(STAT_SUCCESS);
				setIPsuccess = TRUE;
			}

			TX_NORM_WORD(&aidp_msg.MsgData.APPSetIpAddr.APP_Status.wStatus, &ptxmsg, TRUE);
			sendlen = 1; /* mark it to do the send */
		}
		break;

		default:
		{
			printlog("Invalid MsgType %d from %d.%d.%d.%d(%d)\n",
				   aidp_msg.MsgHdr.byMsgType,
				   psocket_data.recvfrom_ip[0],
				   psocket_data.recvfrom_ip[1],
				   psocket_data.recvfrom_ip[2],
				   psocket_data.recvfrom_ip[3],
				   ntohs(psocket_data.recvfrom_port));
		}
		break;

		} //switch()

		if (sendlen > 0) //there is a reply message to send
		{
			//load byte count of data record
			pend_data = ptxmsg;
			sendlen = (pend_data - precord_len) - 3; //accounts for a DWORD in header (4)
			                                         //and Data Field Terminator byte (1)
			aidp_msg.MsgHdr.dwMsgLen = htonl(sendlen);
			TX_NORM_DWORD(&aidp_msg.MsgHdr.dwMsgLen, &precord_len, TRUE);

			// Finish AIDP message into txbuf
			*ptxmsg++ = (byte)0xFF;	/* Terminator */
			*ptxmsg++ = (byte)0x0D;	/* END FLAG */

			sendlen = ptxmsg - ptxbuf; /* entire message size for socket */

send_again:

			if (debug) {
				if (time(&timenow) < 0) timenow = 0;
				printlog("***** sending(%04x) at %s",
						 ntohs(aidp_msg.MsgHdr.wSequenceNumber),
						 ctime(&timenow));
			}

			if ((sentlen = process_send(&psocket_data, ptxbuf, sendlen)) != sendlen)	{
				printlog("Could not respond to %d.%d.%d.%d(%d)\n",
					   psocket_data.recvfrom_ip[0],
					   psocket_data.recvfrom_ip[1],
					   psocket_data.recvfrom_ip[2],
					   psocket_data.recvfrom_ip[3],
					   ntohs(psocket_data.recvfrom_port));
			}

		} //if() there is a response to send

	} //for() loop

	return 0; //should never reach this point...

}


/*********************************************/
/**********   Print to a log file   **********/
/*********************************************/
void printlog(const char *fmt, ...)
{
	va_list ap;
	char buf[1024];
//	char logfile[] = "/root/aidp.log";
	char logfile[] = "/var/log/aidp.log";

	if (logfd == -1) {
		if ((logfd = open(logfile, O_WRONLY | O_CREAT | O_TRUNC, 0600)) < 0) {
			perror("printlog");
			return;
		}
	}

	if (logfd < 0) return;

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	write(logfd, buf, strlen(buf));
	va_end(ap);
}
/**********************************************/
