/*
 * Serial interface GDB stub
 *
 * Written (hacked together) by David Grothe (dave@gcom.com)
 * Modified to allow invokation early in boot see also
 * kgdb.h for instructions by George Anzinger(george@mvista.com)
 *
 */

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/serial_reg.h>
#include <linux/config.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/mm.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/bitops.h>
#include <asm/system.h>
#include <asm/kgdb.h>

#define TEST_EXISTANCE
#ifdef PRNT
#define dbprintk(s) printk s
#else
#define dbprintk(s)
#endif 
#define TEST_INTERRUPT
#ifdef TEST_INTERRUPT
#define intprintk(s) printk s
#else
#define intprintk(s)
#endif

#define IRQ_T(info) ((info->flags & ASYNC_SHARE_IRQ) ? SA_SHIRQ : SA_INTERRUPT)


#define	GDB_BUF_SIZE	512		/* power of 2, please */

static char	gdb_buf[GDB_BUF_SIZE] ;
static int	gdb_buf_in_inx ;
static int	gdb_buf_in_cnt ;
static int	gdb_buf_out_inx ;

struct async_struct	*gdb_async_info ;
static int	gdb_async_irq ;

extern void	set_debug_traps(void) ;		/* GDB routine */
extern void	breakpoint(void) ;		/* GDB routine */

#define outb_px(a,b) outb_p(b,a)
/*
 * Get a byte from the hardware data buffer and return it
 */
static int	read_data_bfr(struct async_struct *info)
{
        
        if (inb_p(info->port + UART_LSR) & UART_LSR_DR)
                return(inb_p(info->port + UART_RX));

        return( -1 ) ;

} /* read_data_bfr */


/*
 * Get a char if available, return -1 if nothing available.
 * Empty the receive buffer first, then look at the interface hardware.
 */
static int	read_char(struct async_struct *info)
{
	int		chr ;
        unsigned long flags;
        if (gdb_buf_in_cnt != 0){	/* intr routine has q'd chars */
                chr = gdb_buf[gdb_buf_out_inx++] ;
                gdb_buf_out_inx &= (GDB_BUF_SIZE - 1) ;
                gdb_buf_in_cnt-- ;
                return(chr) ;
        }
        save_flags(flags);cli();
        chr = read_data_bfr(info);
        restore_flags(flags);
        return(chr) ;	/* read from hardware */

} /* read_char */

/*
 * Wait until the interface can accept a char, then write it.
 */
static void	write_char(struct async_struct *info, int chr)
{
    while ( !(inb_p(info->port + UART_LSR) & UART_LSR_THRE) ) ;

    outb_p(chr, info->port+UART_TX);

} /* write_char */

/*
 * This is the receiver interrupt routine for the GDB stub.
 * It will receive a limited number of characters of input
 * from the gdb  host machine and save them up in a buffer.
 *
 * When the gdb stub routine getDebugChar() is called it
 * draws characters out of the buffer until it is empty and
 * then reads directly from the serial port.
 *
 * We do not attempt to write chars from the interrupt routine
 * since the stubs do all of that via putDebugChar() which
 * writes one byte after waiting for the interface to become
 * ready.
 *
 * The debug stubs like to run with interrupts disabled since,
 * after all, they run as a consequence of a breakpoint in
 * the kernel.
 *
 * Perhaps someone who knows more about the tty driver than I
 * care to learn can make this work for any low level serial
 * driver.
 */
static void gdb_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
    int			 chr ;
    struct async_struct	*info;
    
    info = gdb_async_info ;
    if (!info || !info->tty || irq != gdb_async_irq)
	    return;

    for (;;)
    {
	chr = read_data_bfr(info) ;
        intprintk(("Debug char on int: %x hex\n",chr));
	if (chr < 0) break ;

        if (chr == 3)                   /* Ctrl-C means remote interrupt */
        {
            BREAKPOINT;
            continue ;
        }

	if (gdb_buf_in_cnt >= GDB_BUF_SIZE)
	{				/* buffer overflow, clear it */
	    gdb_buf_in_inx = 0 ;
	    gdb_buf_in_cnt = 0 ;
	    gdb_buf_out_inx = 0 ;
	    break ;
	}

	gdb_buf[gdb_buf_in_inx++] = chr ;
	gdb_buf_in_inx &= (GDB_BUF_SIZE - 1) ;
	gdb_buf_in_cnt++ ;
    }

} /* gdb_interrupt */

/*
 * Just a NULL routine for testing.
 */
void gdb_null(void)
{
} /* gdb_null */
/* These structure are filled in with values defined in asm/break.h
 */
static struct serial_state state = SB_STATE;	
static struct async_struct local_info = SB_INFO;

/*
 * Hook an IRQ for GDB.
 *
 * This routine is called from the ioctl routine of the serial driver
 * to hand the interface over to the GDB stub code.  It is most usually
 * called from putDebugChar, below.  We try to recognize the ioctl call
 * and for old time sake, throw in a breakpoint in that case.
 */
static    int		ints_disabled = 0;
int	gdb_hook_interrupt(struct async_struct *info)
{
        struct serial_state *state= info->state;
        int flags,port;
#ifdef TEST_EXISTANCE
        int scratch, scratch2;
#endif

        /* The above fails if memory managment is not set up yet. 
         * Rather than fail the set up, just keep track of the fact
         * and pick up the interrupt thing later.
         */
        gdb_async_info = info ;
        port = gdb_async_info->port;
        gdb_async_irq = state->irq ;
        info->IER = UART_IER_RDI;			/* rcvr intr only */
        printk("gdb_hook_interrupt: port =%x, IRQ=%d, divisor =%d\n",
               port,
               gdb_async_irq,
               gdb_async_info->state->custom_divisor);

        save_flags(flags);cli();
#ifdef TEST_EXISTANCE
        /* Existance test */
        /* Should not need all this, but just in case.... */

        scratch = inb_p(port + UART_IER);
        outb_px(port + UART_IER, 0);
        outb_px(0xff, 0x080);
        scratch2 = inb_p(port + UART_IER);
        outb_px(port + UART_IER, scratch);
        if (scratch2) {
                printk("gdb_hook_interrupt: Could not clear IER, not a UART!\n");
                restore_flags(flags);
                return 1;		/* We failed; there's nothing here */
        }
        scratch2 = inb_p(port + UART_LCR);
        outb_px(port + UART_LCR, 0xBF); /* set up for StarTech test */
        outb_px(port + UART_EFR, 0);	/* EFR is the same as FCR */
        outb_px(port + UART_LCR, 0);
        outb_px(port + UART_FCR, UART_FCR_ENABLE_FIFO);
        scratch = inb_p(port + UART_IIR) >> 6;
        if ( scratch == 1){
                printk("gdb_hook_interrupt: Undefined UART type!  Not a UART! \n");
                restore_flags(flags);
                return 1;
        }else{
                dbprintk(
                         ("gdb_hook_interrupt: UART type is %d where 0=16450, 2=16550 3=16550A\n",
                          scratch));
        }
        scratch = inb_p(port + UART_MCR);
        outb_px(port + UART_MCR, UART_MCR_LOOP | scratch);
        outb_px(port + UART_MCR, UART_MCR_LOOP | 0x0A);
        scratch2 = inb_p(port + UART_MSR) & 0xF0;
        outb_px(port + UART_MCR, scratch);
        if (scratch2 != 0x90) {
                printk("gdb_hook_interrupt: Loop back test failed! Not a UART!\n");
                restore_flags(flags);
                return scratch2 + 1000;  /* force 0 to fail */
        }
        
#endif /* test existance */
        (void)inb_p(port + UART_RX);
        outb_px(port + UART_IER, 0);

        (void)inb_p(port + UART_RX);      /* serial driver comments say */
        (void)inb_p(port + UART_IIR);    /* this clears the interrupt regs */
        (void)inb_p(port + UART_MSR);
        outb_px(port + UART_LCR,UART_LCR_WLEN8 | UART_LCR_DLAB);
        outb_px(port + UART_DLL, info->state->custom_divisor & 0xff); /* LS */
        outb_px(port + UART_DLM, info->state->custom_divisor >> 8);   /* MS  */
        outb_px(port + UART_MCR, info->MCR);
        outb_px(port + UART_FCR, 
                UART_FCR_ENABLE_FIFO | 
                UART_FCR_TRIGGER_1 | 
                UART_FCR_CLEAR_XMIT | 
                UART_FCR_CLEAR_RCVR); /* set fcr */
        outb_px(port + UART_LCR, UART_LCR_WLEN8);	 /* reset DLAB */
        outb_px(port + UART_FCR, 
                UART_FCR_ENABLE_FIFO | 
                UART_FCR_TRIGGER_1);/* set fcr */

        kgdb_enable_ints();
        restore_flags(flags);
    
        /*
     * Call GDB routine to setup the exception vectors for the debugger
     */
        printk("gdb_hook_interrupt: set_debug_traps()\n") ;
        if (info != &local_info){   /* must be from ioctl, do a break point */
                printk("gdb_hook_interrupt: breakpoint\n") ;
                BREAKPOINT;
        }
        return(0) ;

} /* gdb_hook_interrupt */

/*
 * getDebugChar
 *
 * This is a GDB stub routine.  It waits for a character from the
 * serial interface and then returns it.  If there is no serial
 * interface connection then it returns a bogus value which will
 * almost certainly cause the system to hang.
 */
int	getDebugChar(void)
{
        volatile int	chr ;

        dbprintk(("getDebugChar(port %x): ",gdb_async_info->port)) ;

        if (gdb_async_info == NULL)
                chr = 0xFF ;
        else
                {
                        while ( (chr = read_char(gdb_async_info)) < 0 ) ;
                }


        dbprintk(("%c\n", chr > ' ' && chr < 0x7F ? chr : ' ')) ;
        return(chr) ;

} /* getDebugChar */

/*
 * putDebugChar
 *
 * This is a GDB stub routine.  It waits until the interface is ready
 * to transmit a char and then sends it.  If there is no serial
 * interface connection then it simply returns to its caller, having
 * pretended to send the char.
 */
static int count = 3;
void kgdb_enable_ints(void)
{
        if (gdb_async_info == NULL) {
                gdb_hook_interrupt(&local_info);
    }
        free_irq(gdb_async_info->state->irq, NULL); /*each out going */
        ints_disabled = request_irq(gdb_async_info->state->irq,
                                    gdb_interrupt,
                                    IRQ_T(gdb_async_info),
                                    "GDB-stub", NULL);
        outb_px(gdb_async_info->port + UART_IER, 
                gdb_async_info->IER);	
}

void	putDebugChar(int chr)
{
        dbprintk(("putDebugChar(port %x): chr=%02x '%c', ints_on=%d\n",
                  gdb_async_info->port,
                  chr,chr > ' ' && chr < 0x7F ? chr : ' ', ints_disabled ? 0:1)) ;

        if (gdb_async_info == NULL) {
                gdb_hook_interrupt(&local_info);
        }

        write_char(gdb_async_info, chr) ;	/* this routine will wait */
        count = ( chr == '#' )? 0 : count++;
        if ( (count == 2) ){            /* try to enable after */
                if ( ints_disabled)kgdb_enable_ints();/* try to enable after */

                /* We do this a lot because, well we really want to get these
                 * interrupts.  The serial driver will clear these bits when it
                 * initializes the chip.  Every thing else it does is ok, but this.
                 */
                outb_px(gdb_async_info->port + UART_IER, 
                        gdb_async_info->IER);	
        }

} /* putDebugChar */
#if 0
/*
 * exceptionHandler
 *
 * GDB's stub code calls this routine to associate a routine with
 * an interrupt vector.  The routine, pointed to by 'handler' is
 * an assembly language routine that can have the vector pointed
 * directly at it.
 */
typedef void	(*Function)(void) ;

Function	exceptionHandler(int trapno, Function handler)
{
    if (trapno < 0x20)			/* only 32 of these */
	set_intr_gate(trapno, handler) ;

    return(NULL) ;

} /* exceptionHandler */

#endif
