
/* MPC8xx CPM I2C interface. Copyright (c) 1999 Dan Malek (dmalek@jlc.net).
 * This is a read only interface, primarily to test the microcode
 * patches.
 */

/*-----------------------------------------------------------------------------
 *
 * Copyright (C) Cyclades Corporation, 1997 - 2000
 *
 * This file :      cy_iic.c
 *
 * Description:     This file contains the routines to access for IIC protocol
 *
 * Complies with Cyclades SW coding Standard rev 1.2
 *-----------------------------------------------------------------------------
 */

 /*----------------------------------------------------------------------------
  * Change History
  *----------------------------------------------------------------------------
  * 11/8/02   Arnaldo Zimmermann     Port to linux kernel
  *----------------------------------------------------------------------------
  */

 
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/mpc8xx.h>

#include "commproc.h"
#include <linux/cy_sm.h>

static void
msdelay(uint t) 
{
	uint i;
	/* delay t miliseconds */
	for (i=0; i<t; i++) {
		udelay(1000);
	}
}

void
cpm_iic_init()
{
	uint	mem_addr, dp_addr, reloc;
	volatile cbd_t		*bdp;
	volatile cpm8xx_t	*cp;
	volatile iic_t		*iip;
	volatile i2c8xx_t	*i2c;
	volatile immap_t	*immap;

	cp = cpmp;	/* Get pointer to Communication Processor */
	immap = (immap_t *)IMAP_ADDR;	/* and to internal registers */

	iip = (iic_t *)&cp->cp_dparam[PROFF_IIC];

	/* Check for and use a microcode relocation patch.
	*/
	if (reloc = iip->iic_rpbase)
		iip = (iic_t *)&cp->cp_dpmem[iip->iic_rpbase];
		
	i2c = (i2c8xx_t *)&(immap->im_i2c);

	/* Initialize Port B IIC pins.
	*/
	cp->cp_pbpar |= 0x00000030;
	cp->cp_pbdir |= 0x00000030;
	cp->cp_pbodr |= 0x00000030;

	/* Initialize the parameter ram.
	 * We need to make sure many things are initialized to zero,
	 * especially in the case of a microcode patch.
	 */
	iip->iic_rstate = 0;
	iip->iic_rdp = 0;
	iip->iic_rbptr = 0;
	iip->iic_rbc = 0;
	iip->iic_rxtmp = 0;
	iip->iic_tstate = 0;
	iip->iic_tdp = 0;
	iip->iic_tbptr = 0;
	iip->iic_tbc = 0;
	iip->iic_txtmp = 0;

	/* Allocate space for one transmit and one receive buffer
	 * descriptor in the DP ram.
	 */
	dp_addr = m8xx_cpm_dpalloc(sizeof(cbd_t) * 2);
	if (dp_addr == CPM_DP_NOSPACE) {
		panic("Warning! Failed to allocate tx/rx buffer on DP ram for I2C");
	}

	/* Set up the IIC parameters in the parameter ram.
	*/
	iip->iic_tbase = dp_addr;
	iip->iic_rbase = dp_addr + sizeof(cbd_t);

	/*
	 *  manually init if using microcode patch
	 */
	if(reloc)
	{
		iip->iic_rbptr = iip->iic_rbase;
		iip->iic_tbptr = iip->iic_tbase;
	}

	/* Allocate space for receive buffer in the host memory.
	 * The most you can read at once from IIC is 128 bytes.
	 */
	mem_addr = m8xx_cpm_hostalloc(128);
	if (!mem_addr) {
		panic("Warning! Failed to allocate rx buffer on CPM ram for I2C");
	}

	/* Set the buffer address.
	*/
	bdp = (cbd_t *)&cp->cp_dpmem[iip->iic_rbase];
	bdp->cbd_bufaddr = __pa(mem_addr);

	mem_addr = m8xx_cpm_hostalloc(128);
	if (!mem_addr) {
		panic("Warning! Failed to allocate tx buffer on CPM ram for I2C");
	}
	
	/* Set the buffer address.
	*/
	bdp = (cbd_t *)&cp->cp_dpmem[iip->iic_tbase];
	bdp->cbd_bufaddr = __pa(mem_addr);

	iip->iic_tfcr = 0x18;	/* big-endian order */
	iip->iic_rfcr = 0x18;	/* big-endian order */

	/* Set maximum receive size.
	*/
	iip->iic_mrblr = 128;

	/* Initialize Tx/Rx parameters.
	*/
	if (reloc == 0) {
		cp->cp_cpcr =
			mk_cr_cmd(CPM_CR_CH_I2C, CPM_CR_INIT_TRX) | CPM_CR_FLG;
		while (cp->cp_cpcr & CPM_CR_FLG);
	}
	else {
		cp->cp_cpcr =
			mk_cr_cmd(CPM_CR_CH_I2C, CPM_CR_INIT_TX) | CPM_CR_FLG;
		while (cp->cp_cpcr & CPM_CR_FLG);
		cp->cp_cpcr =
			mk_cr_cmd(CPM_CR_CH_I2C, CPM_CR_INIT_RX) | CPM_CR_FLG;
		while (cp->cp_cpcr & CPM_CR_FLG);
	}

	/* Select an arbitrary address.  Just make sure it is unique.
	*/
	i2c->i2c_i2add = CY_SLAVE_ADDR;

	/* Division ration 2*(0x39+3)
	*/
	i2c->i2c_i2brg = 0x39;

	/* I2C LSB first, BRGCLK/4, clk=100kHz
	*/
	i2c->i2c_i2mod = 0x0e;

	/* I2C Master
	*/
	i2c->i2c_i2com = 0x01;

	/* Disable interrupts.
	*/
	i2c->i2c_i2cmr = 0;
	i2c->i2c_i2cer = 0xff;

	printk("I2C init done\r\n");
}

/* Read from IIC.
 * This is a two step process.  First, we send the "dummy" write
 * to set the device offset for the read.  Second, we perform
 * the read operation.
 * where buf - [0] = device address, 
 *             [1] = offset and 
 *     count - number of bytes to read, placed back at buf[0]
 */

ssize_t
cpm_iic_read(char *buf, size_t count, size_t flag)
{
	volatile iic_t *iip;
	volatile i2c8xx_t *i2c;
	volatile cbd_t	*tbdf, *rbdf;
	volatile cpm8xx_t *cp;
	u_char	*tb;
	u_char iic_device, iic_offset, retry;
	unsigned long flags;
	uint	reloc, i, status;
 	volatile sysconf8xx_t *psiu = &(((immap_t *)IMAP_ADDR)->im_siu_conf);

	cp = cpmp;	/* Get pointer to Communication Processor */
	iip = (iic_t *)&cp->cp_dparam[PROFF_IIC];

	/* Check for and use a microcode relocation patch.
	*/
	if (reloc = iip->iic_rpbase) {
		iip = (iic_t *)&cp->cp_dpmem[iip->iic_rpbase];
	}

	if (count > iip->iic_mrblr) {
		return 0;
	}

	//printk("I2C buf: ");
	//for (i=0; i<(count+1); i++) printk("%x ", (u_char)buf[i]);
	//printk("\r\nCount: %d\r\n", count);	

	iic_device = buf[0];	/* first buffer position has the device address */
	iic_offset = buf[1];	/* second buffer position has the offset */

	i2c = (i2c8xx_t *)&(((immap_t *)IMAP_ADDR)->im_i2c);

	tbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_tbase];
	rbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_rbase];
	tb = __va(tbdf->cbd_bufaddr);

	if (flag & I2C_SET_MASTER) {
		/* I2C Master */
		i2c->i2c_i2com |= 0x01;
	}
	else {
		/* I2C Slave */
		i2c->i2c_i2com &= ~0x01;
	}
	
	retry = I2C_MAX_RETRY;
	while ((flag & I2C_SET_OFFSET) && retry--) {

		/* Send a "dummy write" operation.  This is a write request with
		 * only the offset sent, followed by another start condition.
		 * This will ensure we start reading from the first location
		 * of the EEPROM.
		 */
		memset(__va(rbdf->cbd_bufaddr), 0, count);
		memset(__va(tbdf->cbd_bufaddr), 0, count);
		*tb = iic_device & 0xfe;	/* Device address, write */
		*(tb+1) = iic_offset;		/* Offset */
		tbdf->cbd_datlen = 2;		/* Length */
		tbdf->cbd_sc =
		    BD_SC_READY | BD_SC_WRAP | BD_SC_INTRPT | BD_SC_LAST;
		rbdf->cbd_datlen = 0;		/* Length */
		rbdf->cbd_sc =
		    BD_SC_EMPTY | BD_SC_WRAP | BD_SC_INTRPT;

		save_flags(flags); cli();
		i2c->i2c_i2cmr = 0;	
		i2c->i2c_i2cer = 0xff;
		i2c->i2c_i2mod |= 1;	/* Enable I2C */
		i2c->i2c_i2com |= 0x80;	/* Start transmit */

		/* Wait for IIC transfer.
		*/
		i = 10000000;
		while (!(status = i2c->i2c_i2cer & 0x16)) {
		 	psiu->sc_swsr = 0x556c;
			psiu->sc_swsr = 0xaa39;
			i--;
			if (!i) {
				break;
			}
		}

		restore_flags(flags);

		i2c->i2c_i2mod &= ~1;		/* Disable I2C */

		if (status != 0x02) {
			if (!retry) {
				printk("I2C error on writing devive. Status = 0x%04x\r\n", i2c->i2c_i2cer);
				return 0;
			}
			continue;				/* try again */
		}

		if (signal_pending(current))
			return -ERESTARTSYS;

		if (tbdf->cbd_sc & BD_SC_READY)
			printk("I2C write complete but tbuf ready\n");
		
		break;
	}
	
	/* To read, we need an empty buffer of the proper length.
	 * All that is used is the first byte for address, the remainder
	 * is just used for timing (and doesn't really have to exist).
	 */

	retry = I2C_MAX_RETRY;
	while (retry--) {

		memset(__va(rbdf->cbd_bufaddr), 0, count);
		memset(__va(tbdf->cbd_bufaddr), 0, count);
		*tb = (iic_device & 0xfe) | 1;		/* Device address, read */

		if (flag & I2C_SET_MASTER) {
			tbdf->cbd_datlen = count + 1;	/* Length */
			tbdf->cbd_sc =
			    BD_SC_READY | BD_SC_WRAP | BD_SC_INTRPT | BD_SC_LAST;
		}
		rbdf->cbd_datlen = 0;				/* Length */
		rbdf->cbd_sc =
		    BD_SC_EMPTY | BD_SC_WRAP | BD_SC_INTRPT;


		/* Chip bug, set enable here.
		*/
		save_flags(flags); cli();
		i2c->i2c_i2cer = 0xff;
		i2c->i2c_i2mod |= 1;	/* Enable I2C */
		i2c->i2c_i2com |= 0x80;	/* Start transmit/reception */

		/* Wait for IIC transfer.
		*/
		i = 20000000;

		while (!(status = i2c->i2c_i2cer & 0x15)) {
		 	psiu->sc_swsr = 0x556c;
			psiu->sc_swsr = 0xaa39;
			i--;
			if (!i) {
				break;
			}
		}

		restore_flags(flags);

		i2c->i2c_i2mod &= ~1;		/* Disable I2C */

		if (status != 0x01) {
			if (!retry) {
				printk("I2C error on reading devive. Status = 0x%04x\r\n", i2c->i2c_i2cer);
				memcpy(buf, __va(rbdf->cbd_bufaddr), count);
			    printk("I2C buf Rerr: ");
			    for (i=0; i<count; i++) printk("%02x ", (u_char)buf[i]);
			    printk("\r\n");	
				return 0;
			}
			continue;
		}

		if (signal_pending(current))
			return -ERESTARTSYS;

		if ((rbdf->cbd_datlen != count) && (flag & I2C_SET_MASTER)) {
			printk("I2C data length %d of %d\n", rbdf->cbd_datlen, count);
		}

		if (rbdf->cbd_sc & BD_SC_EMPTY)
			printk("I2C read complete but rbuf empty\n");
	
		break;
	}

	memcpy(buf, __va(rbdf->cbd_bufaddr), count);

	//printk("I2C buf R: ");
	//for (i=0; i<count; i++) printk("%02x ", (u_char)buf[i]);
	//printk("\r\n");	

	return rbdf->cbd_datlen;
}

/* Write to IIC.
 * where buf - [0] = device address, 
 *             [1] = offset and 
 *			   buf[2...] = data
 *     count - number of bytes to write + 2 bytes for header.
 */

ssize_t
cpm_iic_write(char *buf, size_t count)
{
	volatile iic_t *iip;
	volatile i2c8xx_t *i2c;
	volatile cbd_t	*tbdf, *rbdf;
	volatile cpm8xx_t *cp;
	u_char	*tb, retry;
	unsigned long flags;
	uint	reloc, i, status;
 	volatile sysconf8xx_t *psiu = &(((immap_t *)IMAP_ADDR)->im_siu_conf);

	cp = cpmp;	/* Get pointer to Communication Processor */
	iip = (iic_t *)&cp->cp_dparam[PROFF_IIC];

	/* Check for and use a microcode relocation patch.
	*/
	if (reloc = iip->iic_rpbase) {
		iip = (iic_t *)&cp->cp_dpmem[iip->iic_rpbase];
	}

	if (count > iip->iic_mrblr) {
		return 0;
	}

	//printk("I2C buf W: ");
	//for (i=0; i<count; i++) printk("%02x ", (u_char)buf[i]);
	//printk("\r\n");	

	i2c = (i2c8xx_t *)&(((immap_t *)IMAP_ADDR)->im_i2c);

	tbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_tbase];
	rbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_rbase];

	/* Send the tx buffer
	 */
	memcpy(__va(tbdf->cbd_bufaddr), buf, count);
	tb = __va(tbdf->cbd_bufaddr);
	*tb &= 0xfe;				/* Device address, write */
	tbdf->cbd_datlen = count; 	/* Length */

	retry = I2C_MAX_RETRY;
	while (retry--) {

		memset(__va(rbdf->cbd_bufaddr), 0, count);
		tbdf->cbd_sc =
		    BD_SC_READY | BD_SC_WRAP | BD_SC_INTRPT | BD_SC_LAST;
		rbdf->cbd_datlen = 0;		/* Length */
		rbdf->cbd_sc =
		    BD_SC_EMPTY | BD_SC_WRAP | BD_SC_INTRPT;

		save_flags(flags); cli();
		i2c->i2c_i2cmr = 0;	
		i2c->i2c_i2cer = 0xff;
		i2c->i2c_i2mod |= 1;	/* Enable I2C */
		i2c->i2c_i2com |= 0x81;	/* Start transmit as master*/

		/* Wait for IIC transfer.
		*/
		i = 10000000;
		while (!(status = i2c->i2c_i2cer & 0x16)) {
		 	psiu->sc_swsr = 0x556c;
			psiu->sc_swsr = 0xaa39;
			i--;
			if (!i) {
				break;
			}
		}

		restore_flags(flags);

		i2c->i2c_i2mod &= ~1;		/* Disable I2C */

		if (status != 0x02) {
			if (!retry) {
				//printk("I2C error on writing devive. Status = 0x%04x\r\n", i2c->i2c_i2cer);
				return 0;
			}
			continue;
		}

		if (signal_pending(current))
			return -ERESTARTSYS;

		if (tbdf->cbd_sc & BD_SC_READY)
			printk("I2C write complete but tbuf ready\n");

		break;
	}

	return count;
}


