/*
 *  LCD framebuffer driver for MPC823
 */

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/8xx_immap.h>
#include <asm/mpc8xx.h>

#include <linux/fb.h>
#include <video/fbcon.h>
#include <video/fbcon-mfb.h>
#include <video/fbcon-cfb4.h>
#include <video/fbcon-cfb8.h>

#include "commproc.h"
#include "lcdvideo.h"


/*
 *  Information about displays we are using.  This is for configuring
 *  the LCD controller and memory allocation.  Someone has to know what
 *  is connected, as we can't autodetect anything.
 */
#define CFG_HIGH	0	/* Pins are active high */
#define CFG_LOW		1	/* Pins are active low */

typedef struct vidinfo {
    ushort	vl_col;		/* Number of columns (i.e. 640) */
    ushort	vl_row;		/* Number of rows (i.e. 480) */
    ushort	vl_width;	/* Width of display area in millimeters */
    ushort	vl_height;	/* Height of display area in millimeters */
    
    /* LCD configuration register.
    */
    u_char	vl_clkp;	/* Clock polarity */
    u_char	vl_oep;		/* Output Enable polarity */
    u_char	vl_hsp;		/* Horizontal Sync polarity */
    u_char	vl_vsp;		/* Vertical Sync polarity */
    u_char	vl_dp;		/* Data polarity */
    u_char	vl_bpix;	/* Bits per pixel, 0 = 1, 1 = 2, 2 = 4, 3 = 8 */
    u_char	vl_lbw;		/* LCD Bus width, 0 = 4, 1 = 8 */
    u_char	vl_splt;	/* Split display, 0 = dual, 1 = single */
    u_char	vl_clor;	/* Color, 0 = mono, 1 = color */
    u_char	vl_tft;		/* 0 = passive, 1 = TFT */

    /* Horizontal control register.  Timing from data sheet.
    */
    ushort	vl_wbl;		/* Wait between lines */
    
    /* Vertical control register.
    */
    u_char	vl_vpw;		/* Vertical sync pulse width */
    u_char	vl_lcdac;	/* LCD AC timing */
    u_char	vl_wbf;		/* Wait between frames */
} vidinfo_t;

#define NEC_TFT

#ifdef NEC_TFT

/*
 *  NEC NL6648AC33-18. Active, color, single scan.
 *  Over time, we could either configure exactly one of these at build
 *  time, or create an array of supported devices and find a way to
 *  tell this driver what we have (EEPROM perhaps?).
 */
static vidinfo_t panel_info = {
    640, 480, 132, 99, CFG_HIGH, CFG_HIGH, CFG_LOW, CFG_LOW, CFG_HIGH,
    3, 0, 0, 1, 1, 144, 2, 0, 33
};
#endif

#ifdef SHARP_16x9

/* Sharp 320x240. Active, color, single scan.  It isn't 16x9, and I am
 * not sure what it is.......
 */
static vidinfo_t panel_info = {
    320, 240, 0, 0, CFG_HIGH, CFG_HIGH, CFG_HIGH, CFG_HIGH, CFG_HIGH,
    3, 0, 0, 1, 1, 15, 4, 0, 3
};
#endif

#define NBITS(bit_code)		(1 << (bit_code))
#define NCOLORS(bit_code)	(1 << NBITS(bit_code))

static struct fb_info fb_info;
static struct display display;
static int currcon = 0;

/* Frame buffer memory information */
static unsigned long lcd_fb_base;
static unsigned long lcd_fb_size;
static unsigned long lcd_fb_line_length;


/* ------------------- chipset specific functions -------------------------- */


static int
lcd823_getcolreg(
    unsigned regno,
    unsigned *red,
    unsigned *green,
    unsigned *blue,
    unsigned *transp,
    struct fb_info *info
)
{
    volatile cpm8xx_t	*cp = cpmp;

    *red = (cp->lcd_cmap[regno * 2] & 0x0f) << 12;
    *green = (cp->lcd_cmap[(regno * 2) + 1] & 0xf0) << 8;
    *blue = (cp->lcd_cmap[(regno * 2) + 1] & 0x0f) << 12;

    return 0;
}


static int
lcd823_setcolreg(
    unsigned regno,
    unsigned red,
    unsigned green,
    unsigned blue,
    unsigned transp,
    struct fb_info *info
)
{
    volatile cpm8xx_t	*cp = cpmp;

    cp->lcd_cmap[regno * 2] = (red & 0xf000) >> 12;
    cp->lcd_cmap[(regno * 2) + 1] =
	((green & 0xf000) >> 8) | ((blue & 0xf000) >> 12);

    return 0;
}


static void
lcd823_install_cmap(int con, struct fb_info *info)
{
    struct fb_cmap *cmap;

    if (fb_display[con].cmap.len != 0)
	cmap = &fb_display[con].cmap;
    else
	cmap = fb_default_cmap(NCOLORS(panel_info.vl_bpix));

    fb_set_cmap(cmap, 1, lcd823_setcolreg, info);
}


/* To determine if these occur, just cat /proc/interrupts and look
 * at the interrupt count.
 */
static void
lcd823_interrupt(int irq, void * dev, struct pt_regs * regs)
{
    volatile lcd823_t *lcdp = &(((immap_t *)IMAP_ADDR)->im_lcd);
    lcdp->lcd_lcsr = 0xff;
}


/* ------------ Hardware Independent Functions ------------ */


static int
lcd823_open(struct fb_info *info, int user)
{
    return 0;
}


static int
lcd823_release(struct fb_info *info, int user)
{
    return 0;
}


static int
lcd823_get_fix(struct fb_fix_screeninfo *fix, int con, struct fb_info *info)
{
    memset(fix, 0, sizeof(struct fb_fix_screeninfo));
    strcpy(fix->id, "MPC823");

    fix->smem_start = (char *) virt_to_phys((void *) lcd_fb_base);
    fix->smem_len = lcd_fb_size;
    fix->line_length = lcd_fb_line_length;

    fix->type = FB_TYPE_PACKED_PIXELS;

    if (panel_info.vl_bpix == 1)
	fix->visual = FB_VISUAL_MONO01;
    else
	fix->visual = FB_VISUAL_PSEUDOCOLOR;

    return 0;
}


static int
lcd823_get_var(struct fb_var_screeninfo *var, int con, struct fb_info *info)
{
    if (con == -1) {
	memset(var, 0, sizeof(struct fb_var_screeninfo));

	var->xres = var->xres_virtual = panel_info.vl_col;
	var->yres = var->yres_virtual = panel_info.vl_row;

	var->width = panel_info.vl_width;
	var->height = panel_info.vl_height;

	var->bits_per_pixel = NBITS(panel_info.vl_bpix);
	var->grayscale = (panel_info.vl_clor == 0);
	var->red.length = var->green.length = var->blue.length = 4;
    } else {
	*var = fb_display[con].var;
    }

    return 0;
}


static int
lcd823_set_var(struct fb_var_screeninfo *var, int con, struct fb_info *info)
{
    if ((var->xres != panel_info.vl_col)
	|| (var->yres != panel_info.vl_row)
	|| (var->bits_per_pixel != NBITS(panel_info.vl_bpix))
	|| (var->grayscale && panel_info.vl_clor)
	|| ((var->grayscale == 0) && (panel_info.vl_clor == 0)))
    {
	return -EINVAL;
    }

    lcd823_get_var(var, con, info);

    return 0;
}


static int
lcd823_get_cmap(struct fb_cmap *cmap, int kspc, int con, struct fb_info *info)
{
    struct fb_cmap *src_cmap;

    if (panel_info.vl_bpix == 1)
	return -EINVAL;

    if ((cmap->start + cmap->len) > NCOLORS(panel_info.vl_bpix))
	return -EINVAL;

    /* FIXME:  why can con be -1? */
    if ((con == currcon) || (con == -1))
        return fb_get_cmap(cmap, kspc, lcd823_getcolreg, info);

    if (fb_display[con].cmap.len)
	src_cmap = &fb_display[con].cmap;
    else
	src_cmap = fb_default_cmap(NCOLORS(panel_info.vl_bpix));
    fb_copy_cmap(src_cmap, cmap, kspc ? 0 : 2);

    return 0;
}


static int
lcd823_set_cmap(struct fb_cmap *cmap, int kspc, int con, struct fb_info *info)
{
    int err;

    if (panel_info.vl_bpix == 1)
	return -EINVAL;

    if ((cmap->start + cmap->len) > NCOLORS(panel_info.vl_bpix))
	return -EINVAL;

    if (fb_display[con].cmap.len == 0) {
        int size = NCOLORS(panel_info.vl_bpix);
        if ((err = fb_alloc_cmap(&fb_display[con].cmap, size, 0)))
            return err;
    }

    /* FIXME:  why can con be -1? */
    if ((con == currcon) || (con == -1))
        return fb_set_cmap(cmap, kspc, lcd823_setcolreg, info);
    else
	fb_copy_cmap(cmap, &fb_display[con].cmap, kspc ? 0 : 1);

    return 0;
}


static int
lcd823_pan_display(struct fb_var_screeninfo *var, int con, struct fb_info *info)
{
    return -EINVAL;
}


static int
lcd823_ioctl(
    struct inode *inode,
    struct file *file,
    unsigned int cmd,
    unsigned long arg,
    int con,
    struct fb_info *info
)
{
    return -EINVAL;
}


static struct fb_ops lcd823_ops = {
    lcd823_open,
    lcd823_release,
    lcd823_get_fix,
    lcd823_get_var,
    lcd823_set_var,
    lcd823_get_cmap,
    lcd823_set_cmap,
    lcd823_pan_display,
    lcd823_ioctl,
    NULL,		/* fb_mmap	*/
    NULL		/* fb_rasterimg	*/
};


/* ------------------------------------------------------------------------- */


static int
lcd823_switch(int con, struct fb_info *info)
{
    /* Do we have to save the colormap ? */
    if (fb_display[currcon].cmap.len != 0)
        fb_get_cmap(&fb_display[currcon].cmap, 1, lcd823_getcolreg, info);

    currcon = con;

    /* Install new colormap */
    lcd823_install_cmap(currcon, info);

    return 0;
}


static void
lcd823_blank(int blank, struct fb_info *info)
{
    u16 black[16];
    struct fb_cmap cmap;

    if (blank) {
        memset(black, 0, 16*sizeof(u16));
        cmap.red = black;
        cmap.green = black;
        cmap.blue = black;
        cmap.transp = NULL;
        cmap.start = 0;
        cmap.len = 16;
        fb_set_cmap(&cmap, 1, lcd823_setcolreg, info);
    } else
        lcd823_install_cmap(currcon, info);
}


static int
lcd823_updatevar(int con, struct fb_info *info)
{
    return 0;
}


/* ------------------------------------------------------------------------- */


/*
 *  This is called very early in the system initialization to
 *  easily grab physically contiguous memory pages for the LCD
 *  controller.
 */
__initfunc(void lcd_fb_setpage(unsigned long *memory_page_addr))
{
    lcd_fb_line_length = (panel_info.vl_col * NBITS(panel_info.vl_bpix)) / 8;
    lcd_fb_size = lcd_fb_line_length * panel_info.vl_row;

    /* Round up to nearest full page */
    lcd_fb_size = (lcd_fb_size + (PAGE_SIZE - 1)) & ~(PAGE_SIZE - 1);

    /* Allocate pages for the frame buffer. */
    lcd_fb_base = *memory_page_addr;
    *memory_page_addr = lcd_fb_base + lcd_fb_size;

    printk("LCD frame buffer %ld pages\n", lcd_fb_size / PAGE_SIZE);
}


__initfunc(static void lcd823_init_dev(void))
{
    volatile cpm8xx_t *cp = cpmp;
    volatile immap_t *immap = (immap_t *)IMAP_ADDR;
    volatile lcd823_t *lcdp = &immap->im_lcd;
    uint lccrtmp;

    /* Initialize the LCD control register according to the LCD
     * parameters defined.  We do everything here but enable
     * the controller.
     */
    
    lccrtmp = LCDBIT(LCCR_BNUM_BIT, 
	    (((panel_info.vl_row * panel_info.vl_col) * 8) / 128));

    lccrtmp |= LCDBIT(LCCR_CLKP_BIT, panel_info.vl_clkp) |
		    LCDBIT(LCCR_OEP_BIT, panel_info.vl_oep) |
		    LCDBIT(LCCR_HSP_BIT, panel_info.vl_hsp) |
		    LCDBIT(LCCR_VSP_BIT, panel_info.vl_vsp) |
		    LCDBIT(LCCR_DP_BIT, panel_info.vl_dp) |
		    LCDBIT(LCCR_BPIX_BIT, panel_info.vl_bpix) |
		    LCDBIT(LCCR_LBW_BIT, panel_info.vl_lbw) |
		    LCDBIT(LCCR_SPLT_BIT, panel_info.vl_splt) |
		    LCDBIT(LCCR_CLOR_BIT, panel_info.vl_clor) |
		    LCDBIT(LCCR_TFT_BIT, panel_info.vl_tft);
    
    lccrtmp |= ((SIU_LEVEL5/2) << 12);
    lccrtmp |= LCCR_EIEN;

    lcdp->lcd_lccr = lccrtmp;
    lcdp->lcd_lcsr = 0xff;		/* Clear pending interrupts */

    /* Initialize LCD controller bus priorities.
    */
    immap->im_siu_conf.sc_sdcr &= ~0x0f;	/* RAID = LAID = 0 */

    /* set SHFT/CLOCK division factor 4
     * This needs to be set based upon display type and processor
     * speed.  The TFT displays run about 20 to 30 MHz.
     * I was running 64 MHz processor speed.
     * The value for this divider must be chosen so the result is
     * an integer of the processor speed (i.e., divide by 3 with
     * 64 MHz would be bad).
     */
    immap->im_clkrst.car_sccr &= ~0x1f;
    immap->im_clkrst.car_sccr |= 8;

    /* Enable LCD on port D.
    */
    immap->im_ioport.iop_pdpar |= 0x1fff;
    immap->im_ioport.iop_pddir |= 0x1fff;

    /* Enable LCD_A/B/C on port B.
    */
    cp->cp_pbpar |= 0x00005001;
    cp->cp_pbdir |= 0x00005001;

    /* Load the physical address of the linear frame buffer
     * into the LCD controller.
     * BIG NOTE:  This has to be modified to load A and B depending
     * upon the split mode of the LCD.
     * Since we don't touch these pages, we don't care how the
     * application assigns cache mode.
     */
    lcdp->lcd_lcfaa = lcdp->lcd_lcfba = __pa(lcd_fb_base);

    /* MORE HACKS...This must be updated according to 823 manual
     * for different panels.
     */
    lcdp->lcd_lchcr = LCHCR_BO |
	    LCDBIT(LCHCR_AT_BIT, 4) |
	    LCDBIT(LCHCR_HPC_BIT, panel_info.vl_col) |
	    panel_info.vl_wbl;
    lcdp->lcd_lcvcr = LCDBIT(LCVCR_VPW_BIT, panel_info.vl_vpw) |
	    LCDBIT(LCVCR_LCD_AC_BIT, panel_info.vl_lcdac) |
	    LCDBIT(LCVCR_VPC_BIT, panel_info.vl_row) |
	    panel_info.vl_wbf;

    *((uint *)HIOX_CSR0_ADDR) &= ~(HIOX_CSR0_ENVDOCLK | HIOX_CSR0_VDORST_HL);
#if 0
    *((uint *)0xfac0000c) |= 0x08000000;
#endif

    if (request_8xxirq(SIU_LEVEL5, lcd823_interrupt, 0, "lcd823", NULL) != 0)
	panic("Could not allocate LCD IRQ!");
}


__initfunc(static void lcd823_enable(void))
{
    volatile immap_t *immap = (immap_t *) IMAP_ADDR;
    volatile lcd823_t *lcdp = &immap->im_lcd;

    /* Enable the LCD panel */
    immap->im_siu_conf.sc_sdcr |= (1 << (31 - 25));	/* LAM = 1 */
    lcdp->lcd_lccr |= LCCR_PON;
}


#if 0
static void
lcd823_disable(void)
{
    volatile immap_t *immap = (immap_t *) IMAP_ADDR;
    volatile lcd823_t *lcdp = &immap->im_lcd;

    /* Disable the LCD panel */
    lcdp->lcd_lccr &= ~LCCR_PON;
    immap->im_siu_conf.sc_sdcr &= ~(1 << (31 - 25)); /* LAM = 0 */
}
#endif


__initfunc(void lcd823_init(void))
{
    strcpy(fb_info.modename, "fixed");
    fb_info.flags = FBINFO_FLAG_DEFAULT;
    fb_info.fbops = &lcd823_ops;
    fb_info.disp = &display;
    fb_info.switch_con = &lcd823_switch;
    fb_info.blank = &lcd823_blank;
    fb_info.updatevar = &lcd823_updatevar;
    fb_info.changevar = NULL;

    lcd823_get_var(&display.var, -1, &fb_info);
    display.screen_base = (char *) lcd_fb_base;
    display.line_length = lcd_fb_line_length;
    display.type = FB_TYPE_PACKED_PIXELS;

    if (panel_info.vl_bpix == 1) {
	display.visual = FB_VISUAL_MONO01;
    } else {
	display.visual = FB_VISUAL_PSEUDOCOLOR;
	display.can_soft_blank = 1;
    }

    switch (display.var.bits_per_pixel) {
#ifdef FBCON_HAS_MFB
    case 1:
	display.dispsw = &fbcon_mfb;
	break;
#endif
#ifdef FBCON_HAS_CFB4
    case 4:
	display.dispsw = &fbcon_cfb4;
	break;
#endif
#ifdef FBCON_HAS_CFB8
    case 8:
	display.dispsw = &fbcon_cfb8;
	break;
#endif
    default:
	display.dispsw = &fbcon_dummy;
	break;
    }

    lcd823_init_dev();
    lcd823_install_cmap(0, &fb_info);

    if (register_framebuffer(&fb_info) == 0) {
	printk("fb%d:  MPC823 LCD frame buffer device\n",
	    GET_FB_IDX(fb_info.node));
	lcd823_enable();
    }
}
