/*
 * arch/ppc/kernel/k2_pci.c
 * 
 * PCI support (including autoconfiguration) for SBS K2
 *
 * Author: Matt Porter
 *         mporter@mvista.com
 *
 * Copyright 2000 MontaVista Software Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR   IMPLIED
 * WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 * USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * You should have received a copy of the  GNU General Public License along
 * with this program; if not, write  to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/malloc.h>

#include <asm/byteorder.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/machdep.h>
#include <asm/cpc710.h>
#include <asm/k2.h>
#include <asm/multi_pci.h>

#include "pci.h"

static int upper_memspc;
static int upper_iospc;
static int top_bus = 0;
extern pci_interface pci_interfaces[];

extern int multi_find_pci_interface(unsigned char);

decl_config_access_method(indirect);
decl_config_access_method(multi);

/* SBS K2 interrupt routes */
static char K2_IRQ_routes[][4] =
/*
 * 	PCI IDSEL/INTPIN->INTLINE 
 * 	A	B	C	D
 */
{
	{0,	0,	0,	0},	/* IDSEL 0 - unused */
	{0,	0,	0,	0},	/* IDSEL 1 - unused */
	{0, 	0,	0,	0},	/* IDSEL 2 - unused */
	{1, 	0,	0,	0},	/* IDSEL 3 - Ethernet */
	{5,	5,	5,	5},	/* IDSEL 4 - PMC Site 1 */
	{6,	6,	6,	6},	/* IDSEL 5 - PMC Site 2 */
	{0,     0,      0,      0},     /* IDSEL 6 - unused */
	{0,     0,      0,      0},     /* IDSEL 7 - unused */
	{0,     0,      0,      0},     /* IDSEL 8 - PCI-ISA Bridge */
	{0,     0,      0,      0},     /* IDSEL 9 - unused */
	{0,     0,      0,      0},     /* IDSEL 10 - unused */
	{0,     0,      0,      0},     /* IDSEL 11 - unused */
	{0,     0,      0,      0},     /* IDSEL 12 - unused */
	{0,     0,      0,      0},     /* IDSEL 13 - unused */
	{0,     0,      0,      0},     /* IDSEL 14 - unused */
	{0,     0,      0,      0},     /* IDSEL 15 - unused */
	{0,     0,      0,      0},     /* IDSEL 16 - unused */
	{14,     0,      0,      0},	/* IDSEL 17 - M5229 IDE (native) */
};

static unsigned char K2_CPCI_IRQ_routes[4] = {9, 10, 11, 12};

/*
 * This table represents the standard PCI swizzle defined in the
 * PCI bus specification.
 */
static unsigned char pci_intpins[4][4] =
{
	{ 1, 2, 3, 4},  /* Buses 0, 4, 8, ... */
	{ 2, 3, 4, 1},  /* Buses 1, 5, 9, ... */
	{ 3, 4, 1, 2},  /* Buses 2, 6, 10 ... */
	{ 4, 1, 2, 3},  /* Buses 3, 7, 11 ... */
};

void
K2_route_non0(struct pci_dev *dev)
{
	struct pci_bus  *pbus;          /* Parent bus structure pointer */
	struct pci_dev	*tdev = dev;	/* Temporary pci_dev pointer */
	unsigned int    devnum;         /* Accumulated device number */
	unsigned char   intpin;         /* PCI interrupt pin */
	int		pci_iface;	/* PCI interface index */

	/* Check for valid PCI dev pointer */
	if (dev == NULL) return;

	/* Set up our PCI interface index */
	pci_iface = multi_find_pci_interface(dev->bus->number);

	/* Initialize bridge IDSEL variable */
	devnum = PCI_SLOT(dev->devfn);

	/* Read the interrupt pin of the device */
	pcibios_read_config_byte(dev->bus->number, dev->devfn,
			PCI_INTERRUPT_PIN, &intpin);

	/* If device doesn't request an interrupt, return */
	if ( (intpin < 1) || (intpin > 4) )
		return;

	intpin--;

	/* Only loop if we are on a subordinate bus  */
	if (dev->bus->number != pci_interfaces[pci_iface].top_bus)
	{
		/*
		 * Walk up to the top bus, adjusting the interrupt pin for the
		 * standard PCI bus swizzle.
		 */ 
		do
		{
			intpin = (pci_intpins[devnum % 4][intpin]) - 1;
			pbus = tdev->bus;        /* up one level */
			tdev = pbus->self;
			devnum = PCI_SLOT(tdev->devfn);
		} while(tdev->bus->number != pci_interfaces[pci_iface].top_bus);
	}

	/*
	 * devnum is reduced by 6 because the CPC710
	 * addresses IDSELs differently than the PCI
	 * spec suggests.
	 */
	intpin = (pci_intpins[(devnum-6) % 4][intpin]) - 1;

	/*
	 * If we are on the first physical interface, walk the bus and
	 * note the devfn of the bridge closest to the top bus.  Use this
	 * devfn to determine interrupt routing.
	 */ 
	if (pci_iface == 0)
	{
		pbus = dev->bus;

		while (pbus->primary != 0)
			pbus = pbus->parent;

		dev->irq = K2_IRQ_routes[PCI_SLOT((pbus->self)->devfn)][intpin];
	}
	/*
	 * If we are on the second physical interface, simply use the
	 * swizzled interrupt pin value to map to a CPCI interrupt input
	 * on the PIC
	 */
	else
	{
		dev->irq = K2_CPCI_IRQ_routes[intpin];
	}

	/* Write calculated interrupt value to header and device list */
	pci_write_config_byte(dev, PCI_INTERRUPT_LINE, dev->irq);
}

void k2_pcibios_fixup(void)
{ 
	unsigned char i, intpin, index;
	struct pci_dev * dev;

	/* Force scan of CPC710 PCI64 interface (CPCI bus) */
	pci_scan_peer_bridge(pci_interfaces[1].top_bus);

        for(dev=pci_devices; dev; dev=dev->next)
	{
		/* Route PCI interrupts */
		unsigned char idsel = PCI_SLOT(dev->devfn);
		if (dev->bus->number == 0)
		{
			/* Read the interrupt pin of the device */
			pcibios_read_config_byte(dev->bus->number,
					dev->devfn,
					PCI_INTERRUPT_PIN,
					&intpin);
			if ((intpin > 0) && (intpin < 5))
			{
				/* Find interrupt value */
				intpin--;
				dev->irq = K2_IRQ_routes[idsel][intpin];
				pcibios_write_config_byte(dev->bus->number,
						dev->devfn,
						PCI_INTERRUPT_LINE,
						dev->irq);
			}
		}
		else
		{
			K2_route_non0(dev);
		}

		/*
		 * We do this since some drivers like to ioremap a raw PCI
		 * memory BAR address or use a PCI I/O BAR address assuming 
		 * there is a 1:1 CPU->PCI address mapping.  Since PCI bus
		 * addresses have no standard translation to CPU bus addresses,
		 * we perform the translation here for the K2 memory map.
		 * The values are written back to the base_address members in
		 * the pci device list.
		 */

		/* Don't fixup any host bridge BARs. */
		index = multi_find_pci_interface(dev->bus->number);
		if ( (pci_interfaces[index].top_bus == dev->bus->number) &&
			(dev->devfn == 0) )
			continue;

		for ( i = 0 ; i <= 5 ; i++ )
		{
			if ( dev->base_address[i] )
			{
				if (!(dev->base_address[i] & 0x1))
				{
					printk("K2 PCI Fixup: Old BAR = 0x%x, ",
							(unsigned int)dev->base_address[i]);
					dev->base_address[i] =
						dev->base_address[i] +
						K2_PCI32_PHY_MEM_BASE;
					printk("New BAR = 0x%x\n",
							(unsigned int)dev->base_address[i]);
				}
			}
		}
	}
}

void k2_setup_bars(int current_bus, int pci_devfn)
{
	int bar_response, bar_size, bar_value;
	int bar, addr_mask, size_mask = 0x1, bit_nr = 0, bar_nr = 0;
	int * upper_limit;
	int found_mem64 = 0;

	printk("PCI Autoconfig: Found Bus %d, Device %d, Function %d\n",
		current_bus, PCI_SLOT(pci_devfn), PCI_FUNC(pci_devfn) );

	for (bar = PCI_BASE_ADDRESS_0; bar <= PCI_BASE_ADDRESS_5; bar+=4)
	{
		/* Tickle the BAR and get the response */
		pcibios_write_config_dword(current_bus,
				pci_devfn,
				bar,
				0xffffffff);
		pcibios_read_config_dword(current_bus,
				pci_devfn,
				bar,
				&bar_response);

		/* If BAR is not implemented go to the next BAR */
		if (!bar_response)
			continue;

		/* Check the BAR type and set our "walking" mask */
		if (bar_response & PCI_BASE_ADDRESS_SPACE)
		{
			addr_mask = PCI_BASE_ADDRESS_IO_MASK;
			upper_limit = &upper_iospc;
			printk("PCI Autoconfig: BAR %d, I/O, ", bar_nr);
		}
		else
		{
			if ( (bar_response & PCI_BASE_ADDRESS_MEM_TYPE_MASK) ==
			PCI_BASE_ADDRESS_MEM_TYPE_64)
				found_mem64 = 1;

			addr_mask = PCI_BASE_ADDRESS_MEM_MASK;		
			upper_limit = &upper_memspc;
			printk("PCI Autoconfig: BAR %d, Mem, ", bar_nr);
		}

		bar_response &= addr_mask;
		/* Shift the mask left until we hit a one */
		while (bit_nr < 32)
		{
			if ( bar_response & size_mask )
				break;	
			size_mask = size_mask << 1;
			bit_nr++;
		} 

		if (bit_nr >= 32)
			printk("PCI Autoconfig: BAR %d, Fatal sizing error!\n", bar_nr);
		bar_size = 1 << bit_nr;

		bar_value = (*upper_limit - bar_size) & ~(bar_size - 1);

		pcibios_write_config_dword(current_bus,
				pci_devfn,
				bar,
				bar_value);

		*upper_limit = bar_value;

		/*
		 * If we are a 64-bit decoder then increment to the
		 * upper 32 bits of the bar and force it to locate
		 * in the lower 4GB of memory.
		 */ 
		if (found_mem64)
		{
			bar += 4;
			pcibios_write_config_dword(current_bus,
					pci_devfn,
					bar,
					0x00000000);
		}

		bar_nr++;

		printk("size=0x%x, address=0x%x\n",
			bar_size, bar_value);
	}

}

void k2_prescan_setup_bridge(int current_bus, int pci_devfn)
{
	int cmdstat;

	/* Configure bus number registers */
	pcibios_write_config_byte(current_bus,
			pci_devfn,
			PCI_PRIMARY_BUS,
			current_bus);
	pcibios_write_config_byte(current_bus,
			pci_devfn,
			PCI_SECONDARY_BUS,
			current_bus + 1);
	pcibios_write_config_byte(current_bus,
			pci_devfn,
			PCI_SUBORDINATE_BUS,
			0xff);
	/* Set up memory and I/O filter limits, assume 32-bit I/O space */
	pcibios_write_config_word(current_bus,
			pci_devfn,
			PCI_MEMORY_LIMIT,
			((upper_memspc - 1) & 0xfff00000) >> 16);
	pcibios_write_config_byte(current_bus,
			pci_devfn,
			PCI_IO_LIMIT,
			((upper_iospc - 1) & 0x0000f000) >> 8);
	pcibios_write_config_word(current_bus,
			pci_devfn,
			PCI_IO_LIMIT_UPPER16,
			((upper_iospc - 1) & 0xffff0000) >> 16);
			
	/* We don't support prefetchable memory for now, so disable */
	pcibios_write_config_word(current_bus,
			pci_devfn,
			PCI_PREF_MEMORY_BASE,
			0x1000);
	pcibios_write_config_word(current_bus,
			pci_devfn,
			PCI_PREF_MEMORY_LIMIT,
			0x0000);

	/* Enable memory and I/O accesses, enable bus master */
	pcibios_read_config_dword(current_bus,
			pci_devfn,
			PCI_COMMAND,
			&cmdstat);
	pcibios_write_config_dword(current_bus,
			pci_devfn,
			PCI_COMMAND,
			cmdstat |
			PCI_COMMAND_IO |
			PCI_COMMAND_MEMORY |
			PCI_COMMAND_MASTER);
}

void k2_postscan_setup_bridge(int current_bus, int pci_devfn, int sub_bus)
{
	/* Configure bus number registers */
	pcibios_write_config_byte(current_bus,
			pci_devfn,
			PCI_SUBORDINATE_BUS,
			sub_bus);

	/* Round memory allocator to 1MB boundary, we might waste space */
	upper_memspc &= ~(0x100000 - 1);
	pcibios_write_config_word(current_bus,
			pci_devfn,
			PCI_MEMORY_BASE,
			upper_memspc >> 16);

	/* Round I/O allocator to 4KB boundary, we might waste space */
	upper_iospc &= ~(0x1000 - 1);
	pcibios_write_config_byte(current_bus,
			pci_devfn,
			PCI_IO_BASE,
			(upper_iospc & 0x0000f000) >> 8);
	pcibios_write_config_word(current_bus,
			pci_devfn,
			PCI_IO_BASE_UPPER16,
			upper_iospc >> 16);
}

int k2_bus_scan(int current_bus)
{
	int sub_bus, pci_devfn, pci_class, cmdstat, found_multi=0;
	unsigned short vid;
	unsigned char header_type;

	sub_bus = current_bus;

	for (pci_devfn=0; pci_devfn<0xff; pci_devfn++)
	{

		if ( (current_bus == top_bus) && (pci_devfn == 0) )
			continue;

		if (PCI_FUNC(pci_devfn) && !found_multi)
			continue;

		pcibios_read_config_byte(current_bus,
				pci_devfn,
				PCI_HEADER_TYPE,
				&header_type);

		if (!PCI_FUNC(pci_devfn))
			found_multi = header_type & 0x80;

		pcibios_read_config_word(current_bus,
				pci_devfn,
				PCI_VENDOR_ID,
				&vid);

		if (vid != 0xffff)
		{
			pcibios_read_config_dword(current_bus,
					pci_devfn,
					PCI_CLASS_REVISION, &pci_class);
			if ( (pci_class >> 16) == PCI_CLASS_BRIDGE_PCI )
			{
				printk("PCI Autoconfig: Found P2P bridge, device %d\n", PCI_SLOT(pci_devfn));
				k2_prescan_setup_bridge(current_bus,
						pci_devfn);
				sub_bus = k2_bus_scan(current_bus+1);
				k2_postscan_setup_bridge(current_bus,
						pci_devfn,
						sub_bus);
			}
			else
			{
				/*
				 * Found a peripheral, enable some standard
				 * settings
				 */
				pcibios_read_config_dword(current_bus,
						pci_devfn,
						PCI_COMMAND,
						&cmdstat);
				pcibios_write_config_dword(current_bus,
						pci_devfn,
						PCI_COMMAND,
						cmdstat |
						PCI_COMMAND_IO |
						PCI_COMMAND_MEMORY |
						PCI_COMMAND_MASTER);
				pcibios_write_config_byte(current_bus,
						pci_devfn,
						PCI_LATENCY_TIMER,
						0x80);

				/* Allocate PCI I/O and/or memory space */
				k2_setup_bars(current_bus, pci_devfn);
			}
		}
	}
	return sub_bus;
}

void k2_pcibios_init(void)
{
	int sub_bus = 0, i;
	
	/* Set our PCI access method */
	set_config_access_method(indirect);

	/* Set to PCI32 config space regs */ 
	pci_config_address = (int *)K2_PCI32_CONFIG_ADDR;
	pci_config_data = (char *)K2_PCI32_CONFIG_DATA;

	/* Setup PCI32 mem-I/O space boundaries */
	upper_memspc = (int)K2_PCI32_MEM_TOP;
	upper_iospc = (int)K2_PCI32_IO_TOP;

	/* Initialize PCI32 bus registers */
	pcibios_write_config_byte(top_bus,
			PCI_DEVFN(0, 0),
			CPC710_BUS_NUMBER,
			0);
	pcibios_write_config_byte(top_bus,
			PCI_DEVFN(0, 0),
			CPC710_SUB_BUS_NUMBER,
			0xff);

	/* Scan interface A for bridges/devices */
	printk("PCI Autoconfig: Scanning interface A\n");
	sub_bus = k2_bus_scan(top_bus);

	/* Write out correct max subordinate bus number for interface A */
	pcibios_write_config_byte(top_bus,
			PCI_DEVFN(0, 0),
			CPC710_SUB_BUS_NUMBER,
			sub_bus);

	/* Setup PCI interface A */
	pci_interfaces[0].top_bus = top_bus;
	pci_interfaces[0].sub_bus = sub_bus;
	pci_interfaces[0].pci_cfga = (int *)K2_PCI32_CONFIG_ADDR;
	pci_interfaces[0].pci_cfgd = (char *)K2_PCI32_CONFIG_DATA;

	top_bus = ++sub_bus;

	/* Set to PCI64 config space regs */ 
	pci_config_address = (int *)K2_PCI64_CONFIG_ADDR;
	pci_config_data = (char *)K2_PCI64_CONFIG_DATA;

	/* Setup PCI64 mem-I/O space boundaries */
	upper_memspc = (int)K2_PCI64_MEM_TOP;
	upper_iospc = (int)K2_PCI64_IO_TOP;

	/* Initialize PCI64 bus registers */
	pcibios_write_config_byte(0,
			PCI_DEVFN(0, 0),
			CPC710_SUB_BUS_NUMBER,
			0xff);

	pcibios_write_config_byte(0,
			PCI_DEVFN(0, 0),
			CPC710_BUS_NUMBER,
			top_bus);

	/* Scan interface B for bridges/devices */
	printk("PCI Autoconfig: Scanning interface B\n");
	sub_bus = k2_bus_scan(top_bus);

	/* Write out correct max subordinate bus number for interface B */
	pcibios_write_config_byte(top_bus,
			PCI_DEVFN(0, 0),
			CPC710_SUB_BUS_NUMBER,
			sub_bus);

	/* Setup interface B */
	pci_interfaces[1].top_bus = top_bus;
	pci_interfaces[1].sub_bus = sub_bus;
	pci_interfaces[1].pci_cfga = (int *)K2_PCI64_CONFIG_ADDR;
	pci_interfaces[1].pci_cfgd = (char *)K2_PCI64_CONFIG_DATA;

	/* Fill remaining elements */
	for (i=2; i<MAX_PCI_INTERFACES; i++)
	{
		pci_interfaces[i].top_bus = 0xff;
		pci_interfaces[i].sub_bus = 0xff;
	}

	set_config_access_method(multi);
}


__init void k2_setup_pci_ptrs(void)
{
	ppc_md.pcibios_init = k2_pcibios_init;
	ppc_md.pcibios_fixup = k2_pcibios_fixup;
}
