/*
 * arch/ppc/kernel/k2_setup.c
 * 
 * Board setup routines for SBS K2
 *
 * Author: Matt Porter
 *         mporter@mvista.com
 *
 * Copyright 2000 MontaVista Software Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR   IMPLIED
 * WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 * USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * You should have received a copy of the  GNU General Public License along
 * with this program; if not, write  to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/errno.h> 
#include <linux/reboot.h>
#include <linux/pci.h>
#include <linux/kdev_t.h>
#include <linux/types.h>
#include <linux/major.h>
#include <linux/blk.h>
#include <linux/console.h>
#include <linux/delay.h>

#include <asm/system.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <asm/dma.h>
#include <asm/ide.h>
#include <asm/io.h>
#include <asm/machdep.h>
#include <asm/k2.h>

#include "time.h"

extern void k2_init_IRQ(void);
extern void k2_do_IRQ(struct pt_regs *, int, int);
extern void k2_setup_pci_ptrs(void);
extern void m48t3x_time_init(void);
extern unsigned long m48t3x_get_rtc_time(void);
extern int m48t3x_set_rtc_time(unsigned long);

extern unsigned char * m48t3x_base_address;

static unsigned int cpu_7xx[16] = {
	0, 15, 14, 0, 0, 13, 5, 9, 6, 11, 8, 10, 16, 12, 7, 0
};
static unsigned int cpu_6xx[16] = {
	0, 0, 14, 0, 0, 13, 5, 9, 6, 11, 8, 10, 0, 12, 7, 0
};

#if defined(CONFIG_BLK_DEV_IDE) || defined(CONFIG_BLK_DEV_IDE_MODULE)
/* IDE functions */
void
k2_ide_insw(ide_ioreg_t port, void *buf, int ns)
{
	ide_insw(((port)+(_IO_BASE)), buf, ns);
}

void
k2_ide_outsw(ide_ioreg_t port, void *buf, int ns)
{
	ide_outsw(((port)+_IO_BASE), buf, ns);
}

int
k2_ide_default_irq(ide_ioreg_t base)
{
	switch (base) {
		case 0x1f0: return 14;
		case 0x170: return 15;
		default:
			return 0;
	}
}

ide_ioreg_t
k2_ide_default_io_base(int index)
{
	switch (index) {
		case 0: return 0x1f0;
		case 1: return 0x170;
		default:
                        return 0;
	}
}

int
k2_ide_check_region(ide_ioreg_t from, unsigned int extent)
{
        return check_region(from, extent);
}

void
k2_ide_request_region(ide_ioreg_t from,
			unsigned int extent,
			const char *name)
{
        request_region(from, extent, name);
}

void
k2_ide_release_region(ide_ioreg_t from,
			unsigned int extent)
{
        release_region(from, extent);
}

void
k2_ide_fix_driveid(struct hd_driveid *id)
{
	ppc_generic_ide_fix_driveid(id);
}

__initfunc(void
k2_ide_init_hwif_ports (ide_ioreg_t *p, ide_ioreg_t base, int *irq))
{
	ide_ioreg_t port = base;
	int i = 8;

	while (i--)
		*p++ = port++;
	*p++ = base + 0x206;
	if (irq != NULL)
		*irq = 0;
}
#endif

static inline unsigned long _get_HID1(void)
{
	unsigned long val;

	__asm__ __volatile__("mfspr %0,1009" : "=r" (val));
	return val;
}                                                                               

int
k2_get_bus_speed(void)
{
	int bus_speed;
	unsigned char board_id;

	board_id = *(unsigned char *)K2_BOARD_ID_REG;

	switch( K2_BUS_SPD(board_id) ) {

		case 0:
		default:
			bus_speed = 100000000;
			break;

		case 1:
			bus_speed = 83333333;
			break;

		case 2:
			bus_speed = 75000000;
			break;

		case 3:
			bus_speed = 66666666;
			break;
	}
	return bus_speed;
}

int
k2_get_cpu_speed(void)
{
	unsigned long hid1;
	int cpu_speed;

	hid1 = _get_HID1() >> 28;

	if ((_get_PVR()>>16) == 8)
		hid1 = cpu_7xx[hid1];
	else
		hid1 = cpu_6xx[hid1];

	cpu_speed = k2_get_bus_speed()*hid1/2;
	return cpu_speed;
}

void __init k2_calibrate_decr(void)
{
	int freq, divisor;

	/* determine processor bus speed */
	freq = k2_get_bus_speed();
	divisor = 4;
	decrementer_count = freq / HZ / divisor;
	count_period_num = divisor;
	count_period_den = freq / 1000000;
}

void
k2_get_l2_size(char *buffer)
{
	switch ((*(unsigned char *)K2_BOARD_ID_REG) & K2_L2CACHE_MASK)
	{
		case K2_L2CACHE_512KB:
			sprintf(buffer, "512KB");
			break;

		case K2_L2CACHE_256KB:
			sprintf(buffer, "256KB");
			break;

		case K2_L2CACHE_1MB:
			sprintf(buffer, "1MB");
			break;

		case K2_L2CACHE_NONE:
		default:
			sprintf(buffer, "0KB");
			break;
	}
}

int
k2_get_cpuinfo(char *buffer)
{
	int len;
	unsigned char geo_bits, system_slot;
	char l2_size[10];

	len = sprintf( buffer, "vendor\t\t: SBS\n");

	len += sprintf( buffer+len, "machine\t\t: K2\n");

	len += sprintf( buffer+len, "cpu speed\t: %dMhz\n",
			k2_get_cpu_speed()/1000000);

	len += sprintf( buffer+len, "bus speed\t: %dMhz\n",
			k2_get_bus_speed()/1000000);

	k2_get_l2_size(l2_size);
	len += sprintf( buffer+len, "L2\t\t: %s\n", l2_size);

	len += sprintf( buffer+len, "memory type\t: SDRAM\n"); 

	geo_bits = (*(unsigned char *)K2_MSIZ_GEO_REG) & K2_GEO_ADR_MASK;
	system_slot = (*(unsigned char *)K2_MISC_REG) & K2_SYS_SLOT_MASK;
	len += sprintf( buffer+len,
		"backplane\t: %s slot board with geographical address %x\n",
		system_slot ? "Non-system" : "System",
		geo_bits);

	len += sprintf(buffer+len, "\n");

	return len;
}

extern char cmd_line[];

void __init
k2_setup_arch(unsigned long * memory_start_p, unsigned long * memory_end_p)
{
	unsigned int cpu;
	extern char cmd_line[];

	/* init to some ~sane value until calibrate_delay() runs */
        loops_per_sec = 50000000;

#ifdef CONFIG_ROOT_NFS
	/* bootable from nfsroot */
	ROOT_DEV = to_kdev_t(0x00FF);	/* /dev/nfs pseudo device */
#else
	/* bootable from IDE secondary master */
	ROOT_DEV = to_kdev_t(0x1601);	/* /dev/hdc1 */
#endif

	printk("Boot arguments: %s\n", cmd_line);

	/* Identify the system */
	printk("System Identification: SBS K2 - PowerPC 750 @ %d Mhz\n", k2_get_cpu_speed()/1000000);

	/* Identify the CPU manufacturer */
	cpu = _get_PVR();
	printk("CPU manufacturer: %s [rev=%04x]\n", (cpu & (1<<15)) ? "IBM" :
	       "Motorola", (cpu & 0xffff));
}

void
k2_restart(char *cmd)
{
	__cli();
	/* SRR0 has system reset vector, SRR1 has default MSR value */
	/* rfi restores MSR from SRR1 and sets the PC to the SRR0 value */
	__asm__ __volatile__
	("\n\
	lis	3,0xfff0
	ori	3,3,0x0100
	mtspr	26,3
	li	3,0
	mtspr	27,3
	rfi
	");
	for(;;);
}

void
k2_power_off(void)
{
	for(;;);
}

void
k2_halt(void)
{
	k2_restart(NULL);
}


void __init k2_init(unsigned long r3, unsigned long r4, unsigned long r5,
			unsigned long r6, unsigned long r7)
{

	k2_setup_pci_ptrs();

#ifdef CONFIG_BLK_DEV_INITRD
	if ( r4 )
	{
		initrd_start = r4 + KERNELBASE;
		initrd_end = r5 + KERNELBASE;
	}
#endif

	/* take care of cmd line */
	if ( r6  && (((char *) r6) != '\0'))
	{
		*(char *)(r7+KERNELBASE) = 0;
		strcpy(cmd_line, (char *)(r6+KERNELBASE));
	} 

	isa_io_base = K2_ISA_IO_BASE;
	isa_mem_base = K2_ISA_MEM_BASE;
	pci_dram_offset = K2_PCI32_SYS_MEM_BASE;
	m48t3x_base_address =
		(unsigned char *)(isa_mem_base + K2_RTC_BASE_OFFSET);

	ppc_md.setup_arch = k2_setup_arch;
	ppc_md.setup_residual = NULL;
	ppc_md.get_cpuinfo = k2_get_cpuinfo;
	ppc_md.irq_cannonicalize = NULL;
	ppc_md.init_IRQ = k2_init_IRQ;
	ppc_md.do_IRQ = k2_do_IRQ;
	ppc_md.init = NULL;

	ppc_md.restart = k2_restart;
	ppc_md.power_off = k2_power_off;
	ppc_md.halt = k2_halt;

	/* K2 has an M48T37 RTC */
	ppc_md.time_init = m48t3x_time_init;
	ppc_md.set_rtc_time = m48t3x_set_rtc_time;
	ppc_md.get_rtc_time = m48t3x_get_rtc_time;
	ppc_md.calibrate_decr = k2_calibrate_decr;

	/* K2 only has serial */
	ppc_md.kbd_setkeycode = NULL;
	ppc_md.kbd_getkeycode = NULL;
	ppc_md.kbd_translate = NULL;
	ppc_md.kbd_unexpected_up = NULL;
	ppc_md.kbd_leds = NULL;
	ppc_md.kbd_init_hw = NULL;
#ifdef CONFIG_MAGIC_SYSRQ
	ppc_md.kbd_sysrq_xlate = NULL;
#endif

#if defined(CONFIG_BLK_DEV_IDE) || defined(CONFIG_BLK_DEV_IDE_MODULE)
	ppc_ide_md.insw = k2_ide_insw;
	ppc_ide_md.outsw = k2_ide_outsw;
	ppc_ide_md.default_irq = k2_ide_default_irq;
	ppc_ide_md.default_io_base = k2_ide_default_io_base;
	ppc_ide_md.check_region = k2_ide_check_region;
	ppc_ide_md.request_region = k2_ide_request_region;
	ppc_ide_md.release_region = k2_ide_release_region;
	ppc_ide_md.fix_driveid = k2_ide_fix_driveid;
	ppc_ide_md.ide_init_hwif = k2_ide_init_hwif_ports;
	ppc_ide_md.ide_request_irq = request_irq;
	ppc_ide_md.io_base = _IO_BASE;
#endif
}
