

/* Minimal serial functions needed to send messages out the serial
 * port on the MBX console.
 *
 * The MBX uxes SMC1 for the serial port.  We reset the port and use
 * only the first BD that EPPC-Bug set up as a character FIFO.
 *
 * Later versions (at least 1.4, maybe earlier) of the MBX EPPC-Bug
 * use COM1 instead of SMC1 as the console port.  This kinda sucks
 * for the rest of the kernel, so here we force the use of SMC1 again.
 */
#include <linux/config.h>
#include <linux/types.h>
#include <asm/uaccess.h>
#include <asm/mpc8xx.h>
#include "../8xx_io/commproc.h"

#ifdef CONFIG_MBX
#define MBX_CSR1	((volatile u_char *)0xfa100000)
#define CSR1_COMEN	(u_char)0x02
#endif

/* If we are using an SCC, this will be non-zero and the address of
 * the SCC parameter ram.
 */
static uint scc_proff = 0;

static cpm8xx_t	*cpmp = (cpm8xx_t *)&(((immap_t *)IMAP_ADDR)->im_cpm);

void
serial_init(bd_t *bd)
{
	volatile smc_t		*sp;
	volatile smc_uart_t	*up;
	volatile cbd_t	*tbdf, *rbdf;
	volatile cpm8xx_t	*cp;
	uint	dpaddr, memaddr;
#ifdef CONFIG_PRxK
        volatile iop8xx_t       *iop = &(((immap_t *)IMAP_ADDR)->im_ioport);
        volatile car8xx_t       *pcar = &(((immap_t *)IMAP_ADDR)->im_clkrst);
#endif

	cp = cpmp;
	sp = (smc_t*)&(cp->cp_smc[0]);
	up = (smc_uart_t *)&cp->cp_dparam[PROFF_SMC1];

	/* Disable transmitter/receiver.
	*/
	sp->smc_smcmr &= ~(SMCMR_REN | SMCMR_TEN);

#ifndef CONFIG_MBX
	{
	/* Initialize SMC1 and use it for the console port.
	 */

	/* Enable SDMA.
	*/
	((immap_t *)IMAP_ADDR)->im_siu_conf.sc_sdcr = 1;

#ifdef CONFIG_PR3K

        cp->cp_pbpar &= ~0x20000;     /* CTS */
        cp->cp_pbdir &= ~0x20000;

        cp->cp_pbpar &= ~0x4000;      /* RTS */
        cp->cp_pbdir |= 0x4000;
        cp->cp_pbdat &= ~0x4000;

        iop->iop_pcpar &= ~0x0008;    /* DCD */
        iop->iop_pcdir &= ~0x0008;

        iop->iop_pdpar &= ~0x0400;    /* DTR */
        iop->iop_pddir |= 0x0400;
        iop->iop_pddat &= ~0x0400;

        iop->iop_pdpar &= ~0x0800;    /* DSR */
        iop->iop_pddir &= ~0x0800;

        /*
         * Enable SMTXDn and SMRRXDn
        */
        cp->cp_pbpar|= (3 << 6);
        cp->cp_pbdir&=~(3 << 6);
        cp->cp_pbodr&=~(3 << 6);

#else
	/* Use Port B for SMC1 instead of other functions.
	*/
	cp->cp_pbpar |= 0x000000c0;
	cp->cp_pbdir &= ~0x000000c0;
	cp->cp_pbodr &= ~0x000000c0;
#endif

	/* Allocate space for two buffer descriptors in the DP ram.
	 * For now, this address seems OK, but it may have to
	 * change with newer versions of the firmware.
	 */
	dpaddr = 0x0800;

	/* Grab a few bytes from the top of memory.  EPPC-Bug isn't
	 * running any more, so we can do this.
	 */
	memaddr = (bd->bi_memsize - 32) & ~15;

	/* Set the physical address of the host memory buffers in
	 * the buffer descriptors.
	 */
	rbdf = (cbd_t *)&cp->cp_dpmem[dpaddr];
	rbdf->cbd_bufaddr = memaddr;
	rbdf->cbd_sc = 0;
	tbdf = rbdf + 1;
	tbdf->cbd_bufaddr = memaddr+4;
	tbdf->cbd_sc = 0;

	/* Set up the uart parameters in the parameter ram.
	*/
	up->smc_rbase = dpaddr;
	up->smc_tbase = dpaddr+sizeof(cbd_t);
	up->smc_rfcr = SMC_EB;
	up->smc_tfcr = SMC_EB;

	/* Set UART mode, 8 bit, no parity, one stop.
	 * Enable receive and transmit.
	 */
	sp->smc_smcmr = smcr_mk_clen(9) |  SMCMR_SM_UART;

	/* Mask all interrupts and remove anything pending.
	*/
	sp->smc_smcm = 0;
	sp->smc_smce = 0xff;

	/* Set up the baud rate generator.
	 * See 8xx_io/commproc.c for details.
	 */
	cp->cp_simode = 0x10000000;
	cp->cp_brgc1 =
		((((bd->bi_intfreq * 1000000)/16) / bd->bi_baudrate) << 1) | CPM_BRG_EN;

#else /* CONFIG_MBX */
	if (*MBX_CSR1 & CSR1_COMEN) {
		/* COM1 is enabled.  Initialize SMC1 and use it for
		 * the console port.
		 */

		/* Enable SDMA.
		*/
		((immap_t *)IMAP_ADDR)->im_siu_conf.sc_sdcr = 1;

		/* Use Port B for SMC1 instead of other functions.
		*/
		cp->cp_pbpar |= 0x000000c0;
		cp->cp_pbdir &= ~0x000000c0;
		cp->cp_pbodr &= ~0x000000c0;

		/* Allocate space for two buffer descriptors in the DP ram.
		 * For now, this address seems OK, but it may have to
		 * change with newer versions of the firmware.
		 */
		dpaddr = 0x0800;

		/* Grab a few bytes from the top of memory.  EPPC-Bug isn't
		 * running any more, so we can do this.
		 */
		memaddr = (bd->bi_memsize - 32) & ~15;

		/* Set the physical address of the host memory buffers in
		 * the buffer descriptors.
		 */
		rbdf = (cbd_t *)&cp->cp_dpmem[dpaddr];
		rbdf->cbd_bufaddr = memaddr;
		rbdf->cbd_sc = 0;
		tbdf = rbdf + 1;
		tbdf->cbd_bufaddr = memaddr+4;
		tbdf->cbd_sc = 0;

		/* Set up the uart parameters in the parameter ram.
		*/
		up->smc_rbase = dpaddr;
		up->smc_tbase = dpaddr+sizeof(cbd_t);
		up->smc_rfcr = SMC_EB;
		up->smc_tfcr = SMC_EB;

		/* Set UART mode, 8 bit, no parity, one stop.
		 * Enable receive and transmit.
		 */
		sp->smc_smcmr = smcr_mk_clen(9) |  SMCMR_SM_UART;

		/* Mask all interrupts and remove anything pending.
		*/
		sp->smc_smcm = 0;
		sp->smc_smce = 0xff;

		/* Set up the baud rate generator.
		 * See 8xx_io/commproc.c for details.
		 */
		cp->cp_simode = 0x10000000;
		cp->cp_brgc1 =
			((((bd->bi_intfreq * 1000000)/16) / 9600) << 1) | CPM_BRG_EN;

		/* Enable SMC1 for console output.
		*/
		*MBX_CSR1 &= ~CSR1_COMEN;
	}
	else {
#endif /* ndef CONFIG_MBX */
		/* SMC1 is used as console port.
		*/
		tbdf = (cbd_t *)&cp->cp_dpmem[up->smc_tbase];
		rbdf = (cbd_t *)&cp->cp_dpmem[up->smc_rbase];

		/* Issue a stop transmit, and wait for it.
		*/
		cp->cp_cpcr = mk_cr_cmd(CPM_CR_CH_SMC1,
					CPM_CR_STOP_TX) | CPM_CR_FLG;
		while (cp->cp_cpcr & CPM_CR_FLG);
	}

	/* Make the first buffer the only buffer.
	*/
	tbdf->cbd_sc |= BD_SC_WRAP;
	//rbdf->cbd_sc |= BD_SC_EMPTY | BD_SC_WRAP;
	rbdf->cbd_sc |= BD_SC_WRAP;

	/* Single character receive.
	*/
	up->smc_mrblr = 1;
	up->smc_maxidl = 0;

	/* Initialize Tx/Rx parameters.
	*/
	cp->cp_cpcr = mk_cr_cmd(CPM_CR_CH_SMC1, CPM_CR_INIT_TRX) | CPM_CR_FLG;
	while (cp->cp_cpcr & CPM_CR_FLG);

	/* Enable transmitter/receiver.
	*/
	sp->smc_smcmr |= SMCMR_REN | SMCMR_TEN;
}

/* These should be passed in......
*/
#ifdef CONFIG_TS1H
#define SCC_IDX 0	/* 0-based index */
#define SCC_PROFF PROFF_SCC1
#define SCC_CHAN CPM_CR_CH_SCC1
#else
#define SCC_IDX 1	/* 0-based index */
#define SCC_PROFF PROFF_SCC2
#define SCC_CHAN CPM_CR_CH_SCC2
#endif

void
serial_scc_init(bd_t *bd)
{
	volatile scc_t		*scp;
	volatile scc_uart_t	*sup;
	volatile cbd_t	*tbdf, *rbdf;
	volatile cpm8xx_t	*cp;
	volatile immap_t	*immap;
	uint	dpaddr, memaddr, iobits, *brg;

	cp = cpmp;
	immap = (immap_t *)IMAP_ADDR;

	scp = &cp->cp_scc[SCC_IDX];
	sup = (scc_uart_t *)&cp->cp_dparam[SCC_PROFF];

	/* Disable transmitter/receiver.
	*/
	scp->scc_gsmrl &= ~(SCC_GSMRL_ENR | SCC_GSMRL_ENT);

	/* Enable SDMA.
	*/
	immap->im_siu_conf.sc_sdcr = 1;

	/* Since we don't yet do modem control, connect the port C pins
	 * as general purpose I/O.  This will assert CTS and CD for the
	 * SCC ports. Assert DTR using GP I/O.
	 */
#ifndef CONFIG_TS1H
	iobits = (0x03 << ((SCC_IDX * 2) + 4)) | (0x01 << SCC_IDX);
#else
	iobits = (0x03 << ((SCC_IDX * 2) + 4));
#endif
	immap->im_ioport.iop_pcdir |= iobits;
	immap->im_ioport.iop_pcpar &= ~iobits;
#ifndef CONFIG_TS1H
	immap->im_ioport.iop_pcdat |= 0x01 << SCC_IDX;
#endif

	/* Assert RTS using GP I/O.
	*/
	iobits = (0x01 << (SCC_IDX + 12));
	immap->im_cpm.cp_pbdir |= iobits;
	immap->im_cpm.cp_pbpar |= iobits;

#ifdef CONFIG_TS1H
	/* Assert DTR using GP I/O.
	*/
	iobits = 0x0800;
	immap->im_cpm.cp_pbdir |= iobits;
	immap->im_cpm.cp_pbpar &= ~iobits;
	immap->im_cpm.cp_pbdat |= iobits;

	/* Assert DSR using GP I/O.
	*/
	iobits = 0x0008;
	immap->im_ioport.iop_pcdir &= ~iobits;
	immap->im_ioport.iop_pcpar &= ~iobits;
#endif

	/* Connect Rx/Tx of all SCCs through port A.
	*/
	iobits = 0x03 << (SCC_IDX * 2);
	immap->im_ioport.iop_padir &= ~iobits;
	immap->im_ioport.iop_papar |= iobits;
	immap->im_ioport.iop_paodr &= ~iobits;

	/* Connect SCC to NMSI.  Connect BRG based upon SCC index.
	 */
#if 0
	cp->cp_sicr &= ~(0xff << (SCC_IDX * 8));
	cp->cp_sicr |= ((SCC_IDX << 3) | SCC_IDX) << (SCC_IDX * 8);
#else
	cp->cp_sicr = 0;
#endif


	/* Allocate space for two buffer descriptors in the DP ram.
	 * For now, this address seems OK, but it may have to
	 * change with newer versions of the firmware.
	 */
	dpaddr = 0x0800;

	/* Grab a few bytes from the top of memory.  EPPC-Bug isn't
	 * running any more, so we can do this.
	 */
	memaddr = (bd->bi_memsize - 32) & ~15;

	/* Set the physical address of the host memory buffers in
	 * the buffer descriptors.
	 */
	rbdf = (cbd_t *)&cp->cp_dpmem[dpaddr];
	rbdf->cbd_bufaddr = memaddr;
	rbdf->cbd_sc = 0;
	tbdf = rbdf + 1;
	tbdf->cbd_bufaddr = memaddr+4;
	tbdf->cbd_sc = 0;

	/* Make the first buffer the only buffer.
	*/
	tbdf->cbd_sc |= BD_SC_WRAP;
	//rbdf->cbd_sc |= BD_SC_EMPTY | BD_SC_WRAP;
	rbdf->cbd_sc |= BD_SC_WRAP;

	/* Set up the uart parameters in the parameter ram.
	*/
	sup->scc_genscc.scc_rbase = dpaddr;
	sup->scc_genscc.scc_tbase = dpaddr + sizeof(cbd_t);

	/* Set up the uart parameters in the
	 * parameter ram.
	 */
	sup->scc_genscc.scc_rfcr = SMC_EB;
	sup->scc_genscc.scc_tfcr = SMC_EB;

	/* Set this to 1 for now, so we get single
	 * character interrupts.  Using idle charater
	 * time requires some additional tuning.
	 */
	sup->scc_genscc.scc_mrblr = 1;
	sup->scc_maxidl = 0;
	sup->scc_brkcr = 1;
	sup->scc_parec = 0;
	sup->scc_frmec = 0;
	sup->scc_nosec = 0;
	sup->scc_brkec = 0;
	sup->scc_uaddr1 = 0;
	sup->scc_uaddr2 = 0;
	sup->scc_toseq = 0;
	sup->scc_char1 = 0x8000;
	sup->scc_char2 = 0x8000;
	sup->scc_char3 = 0x8000;
	sup->scc_char4 = 0x8000;
	sup->scc_char5 = 0x8000;
	sup->scc_char6 = 0x8000;
	sup->scc_char7 = 0x8000;
	sup->scc_char8 = 0x8000;
	sup->scc_rccm = 0xc0ff;

	/* Send the CPM an initialize command.
	*/
	cp->cp_cpcr = mk_cr_cmd(SCC_CHAN,
			CPM_CR_INIT_TRX) | CPM_CR_FLG;
	while (cp->cp_cpcr & CPM_CR_FLG);

	/* Set UART mode, 8 bit, no parity, one stop.
	 * Enable receive and transmit.
	 */
#ifdef CONFIG_TS1H
	scp->scc_gsmrh = SCC_GSMRH_RTSM;
#else
	scp->scc_gsmrh = 0;
#endif
	scp->scc_gsmrl = 
		(SCC_GSMRL_MODE_UART | SCC_GSMRL_TDCR_16 | SCC_GSMRL_RDCR_16);

	/* Disable all interrupts and clear all pending
	 * events.
	 */
	scp->scc_sccm = 0;
	scp->scc_scce = 0xffff;
	scp->scc_dsr = 0x7e7e;
	scp->scc_pmsr = 0x3000;

	/* Set up the baud rate generator.
	 * See 8xx_io/commproc.c for details.
	 */
#if 0
	brg = &cp->cp_brgc1;
	brg += SCC_IDX;
	*brg =
#else
	cp->cp_brgc1 =
#endif
		((((bd->bi_intfreq * 1000000)/16) / bd->bi_baudrate) << 1) | CPM_BRG_EN;

	/* Enable transmitter/receiver.
	*/
	scp->scc_gsmrl |= (SCC_GSMRL_ENR | SCC_GSMRL_ENT);

	scc_proff = 1;
#ifdef CONFIG_TS1H
	if (bd->hw_info.serial_itf)
		scc_proff = 2;
#endif
}

void
serial_putchar(const char c)
{
	volatile cbd_t		*tbdf;
	volatile char		*buf;
	volatile smc_uart_t	*up;
	volatile scc_uart_t	*sup;

	if (scc_proff) {
		if (scc_proff == 2) return;
		sup = (scc_uart_t *)&cpmp->cp_dparam[SCC_PROFF];
		tbdf = (cbd_t *)&cpmp->cp_dpmem[sup->scc_genscc.scc_tbase];
	}
	else {
		up = (smc_uart_t *)&cpmp->cp_dparam[PROFF_SMC1];
		tbdf = (cbd_t *)&cpmp->cp_dpmem[up->smc_tbase];
	}

	/* Wait for last character to go.
	*/
	buf = (char *)tbdf->cbd_bufaddr;
	while (tbdf->cbd_sc & BD_SC_READY);

	*buf = c;
	tbdf->cbd_datlen = 1;
	tbdf->cbd_sc |= BD_SC_READY;
}

char
serial_getc()
{
	volatile cbd_t		*rbdf;
	volatile char		*buf;
	volatile smc_uart_t	*up;
	volatile scc_uart_t	*sup;
	char			c;

	if (scc_proff) {
		if (scc_proff == 2) return 0;
		sup = (scc_uart_t *)&cpmp->cp_dparam[SCC_PROFF];
		rbdf = (cbd_t *)&cpmp->cp_dpmem[sup->scc_genscc.scc_rbase];
	}
	else {
		up = (smc_uart_t *)&cpmp->cp_dparam[PROFF_SMC1];
		rbdf = (cbd_t *)&cpmp->cp_dpmem[up->smc_rbase];
	}

	/* Wait for character to show up.
	*/
	buf = (char *)rbdf->cbd_bufaddr;
	while (rbdf->cbd_sc & BD_SC_EMPTY);
	c = *buf;
	rbdf->cbd_sc |= BD_SC_EMPTY;

	return(c);
}

int
serial_tstc(int st)
{
	volatile cbd_t		*rbdf;
	volatile smc_uart_t	*up;
	volatile scc_uart_t	*sup;

	if (scc_proff) {
		if (scc_proff == 2) return 0;
		sup = (scc_uart_t *)&cpmp->cp_dparam[SCC_PROFF];
		rbdf = (cbd_t *)&cpmp->cp_dpmem[sup->scc_genscc.scc_rbase];
	}
	else {
		up = (smc_uart_t *)&cpmp->cp_dparam[PROFF_SMC1];
		rbdf = (cbd_t *)&cpmp->cp_dpmem[up->smc_rbase];
	}
	if (st) {
		rbdf->cbd_sc |= BD_SC_EMPTY;
	}
	return(!(rbdf->cbd_sc & (BD_SC_EMPTY|BD_SC_FR|BD_SC_PR)));
}
