/*----------------------------------------------------------------------------
 *      Copyright (C) Cyclades Corporation, 1998.
 *      All rights reserved
 *      SM100
 *
 *      This file:      cy_sm.c
 *      Description:    Routines to handle Server Management.
 *---------------------------------------------------------------------------
 *---------------------------------------------------------------------------
 *      Change History
 *---------------------------------------------------------------------------
 *      Aug/20/2002     V 1.0.0   Arnaldo Zimmermann - Initial Implementation
 *--------------------------------------------------------------------------*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/mpc8xx.h>
#include <asm/8xx_immap.h>
#include <linux/cy_sm.h>
#include <asm-ppc/irq.h>

static int sm_atx_is_open=0;
static int sm_ipmi_is_open=0;
static int sm_sensor_is_open=0;
static u_char bmc_msg_counter=0;
extern ssize_t cpm_iic_read(char *buf, size_t count, size_t flag);
extern ssize_t cpm_iic_write(char *buf, size_t count);

static void msdelay(uint t) 
{
	uint i;
	/* delay t miliseconds */
	for (i=0; i<t; i++) {
		udelay(1000);
	}
}

static u_char ChecksumRecord(u_char *Buffer, int DataLength)
{
   u_int i;
   u_int Sum;
   
   Sum = 0;
   
   for(i=0; i < DataLength; i++)
   {
      Sum += Buffer[i];
   }
   return ((u_char)(((~Sum) + 1) & 0xFF));
}

static long long sm_llseek(struct file *file, long long offset, int origin)
{
	return -ESPIPE;
}

/*
 *	Write to Server Management Port.
 */

static ssize_t sm_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
    struct inode *inode;
	volatile immap_t *pm = (volatile immap_t *)IMAP_ADDR;
	int status;

	if (!count) {
		return -ESPIPE;
	}

    inode = file->f_dentry->d_inode;
	switch(MINOR(inode->i_rdev))
	{
		case SM_ATX_MINOR:
			if(!sm_atx_is_open) {
				return -ESPIPE;
			}
			if (buf[0] & SM100_POWER_ON) {
				if (buf[1] & SM100_POWER_ON) {
		 			pm->im_ioport.iop_padat |= SM100_POWER_ON;	
				} else {
			 		pm->im_ioport.iop_padat &= ~SM100_POWER_ON;		
				}
			}
			if (buf[0] & SM100_POWER_ENABLE) {
				if (buf[1] & SM100_POWER_ENABLE) {
		 			pm->im_ioport.iop_padat |= SM100_POWER_ENABLE;	
				} else {
			 		pm->im_ioport.iop_padat &= ~SM100_POWER_ENABLE;		
				}
			}
			if (buf[0] & MB_POWER_ENABLE) {
				if (buf[1] & MB_POWER_ENABLE) {
		 			pm->im_ioport.iop_padat |= MB_POWER_ENABLE;	
				} else {
			 		pm->im_ioport.iop_padat &= ~MB_POWER_ENABLE;		
				}
			}
			
			break;
		case SM_IPMI_MINOR:
			if(!sm_ipmi_is_open) {
				return -ESPIPE;
			}
			bmc_msg_counter = (bmc_msg_counter + 1) % BMC_MSG_MOD;
			buf[RQ_SEQ_OFFSET] |= SEQ_LUN(bmc_msg_counter, 0);
			buf[IMB_CHECK1] = ChecksumRecord(&buf[0], IMB_CHECK1);
			if (count > RQ_SA_OFFSET) {
				buf[count-1] = ChecksumRecord(&buf[RQ_SA_OFFSET], (count - 1 - RQ_SA_OFFSET));
			}
			status = cpm_iic_write(buf, count);
			return status;
		case SM_SENSOR_MINOR:
			if(!sm_sensor_is_open) {
				return -ESPIPE;
			}
			status = cpm_iic_write(buf, count);
			return status;
		default:
			return -ENODEV;
	}
	return 1;
}

/*
 *	Read to Server Management Port.
 */
 
static ssize_t sm_read(struct file *file, char *buf, size_t count, loff_t *ptr)
{
    struct inode *inode;
	volatile immap_t *pm = (volatile immap_t *)IMAP_ADDR;
	int status;

    inode = file->f_dentry->d_inode;

	if(!count) {
		return -ESPIPE;
	}

	switch(MINOR(inode->i_rdev))
	{
		case SM_ATX_MINOR:
			if(!sm_atx_is_open) {
				return -ESPIPE;
			}
			*buf = pm->im_ioport.iop_padat & 0xff;		
			return 1;
		case SM_IPMI_MINOR:
			if(!sm_ipmi_is_open) {
				return -ESPIPE;
			}
			status = cpm_iic_read(buf, count, I2C_NO_OFFSET|I2C_SET_SLAVE);
			return status;
		case SM_SENSOR_MINOR:
			if(!sm_sensor_is_open) {
				return -ESPIPE;
			}
			status = cpm_iic_read(buf, count, I2C_SET_OFFSET|I2C_SET_MASTER);
			return status;
		default:						
			return -ESPIPE;
	}

	return 1;
}

/*
 *	Open Server Management Port.
 */

static int sm_open(struct inode *inode, struct file *file)
{
	switch(MINOR(inode->i_rdev))
	{
		case SM_ATX_MINOR:
			if(sm_atx_is_open)
				return -EBUSY;
			sm_atx_is_open=1;
			break;
		case SM_IPMI_MINOR:
			if(sm_ipmi_is_open)
				return -EBUSY;
			sm_ipmi_is_open=1;
			break;
		case SM_SENSOR_MINOR:
			if(sm_sensor_is_open)
				return -EBUSY;
			sm_sensor_is_open=1;
			break;
		default:
			return -ENODEV;
	}
	return 0;
}

/*
 *	Release Server Management Port.
 */

static int sm_release(struct inode *inode, struct file *file)
{
	switch(MINOR(inode->i_rdev))
	{
		case SM_ATX_MINOR:
			if(!sm_atx_is_open)
				return -ESPIPE;
			sm_atx_is_open=0;
			break;
		case SM_IPMI_MINOR:
			if(!sm_ipmi_is_open)
				return -ESPIPE;
			sm_ipmi_is_open=0;
			break;
		case SM_SENSOR_MINOR:
			if(!sm_sensor_is_open)
				return -ESPIPE;
			sm_sensor_is_open=0;
			break;
		default:
			return -ENODEV;
	}
	return 0;
}

/*
 *	Kernel Interfaces
 */
 
 
static struct file_operations sm_fops = {
	sm_llseek,
	sm_read,
	sm_write,
	NULL,		/* No Readdir */
	NULL,		/* No Select */
	NULL,		/* No ioctl */
	NULL,		/* No mmap */
	sm_open,
	NULL,		/* flush */
	sm_release
};

static struct miscdevice sm_atx_miscdev=
{
	SM_ATX_MINOR,
	"SM_ATX",
	&sm_fops
};

static struct miscdevice sm_ipmi_miscdev=
{
	SM_IPMI_MINOR,
	"SM_IPMI",
	&sm_fops
};

static struct miscdevice sm_sensor_miscdev=
{
	SM_SENSOR_MINOR,
	"SM_SENSOR",
	&sm_fops
};



static	void
cy_sm_temperature_interrupt(int irq, void * dev_id, struct pt_regs * regs)
{
	unsigned char buf[8] = {0}, buf1[8] = {0};
	unsigned char conf_byte = 0;

	printk("Sensor Interrupt Services\r\n");

	return;

#if 0	
	disable_irq(SIU_IRQ1);

	/* first read works as broadcast for the SENSOR device */
	//buf[0] = SENSOR_SLAVE_ADDRESS;
	//cpm_iic_read(buf, 1, I2C_NO_OFFSET|I2C_SET_MASTER);

	/* now let's read the status byte 1 */
	buf[0] = SENSOR_SLAVE_ADDRESS;
	buf[1] = SENSOR_RSL_REG;
	if (cpm_iic_read(buf, 1, I2C_SET_OFFSET|I2C_SET_MASTER)) {
		printk("Sensor Status Byte1 0x%02x\r\n", buf[0]);
		if (buf[0] & SENSOR_RHIGH) {
			conf_byte |= SENSOR_MASK_RT1;
			buf1[0] = SENSOR_SLAVE_ADDRESS;
			buf1[1] = SENSOR_RRTE_REG;
			cpm_iic_read(buf1, 1, I2C_SET_OFFSET|I2C_SET_MASTER);
			printk("High temperature alert. Temperature at %d C\r\n", buf1[0]);
		}
	}

	/* read conf Byte 1 and return value on buf[0] */
	buf[0] = SENSOR_SLAVE_ADDRESS;
	buf[1] = SENSOR_RCL_REG;
	cpm_iic_read(buf, 1, I2C_SET_OFFSET|I2C_SET_MASTER);

	/* disable interrupts that already happened on Config Byte 1 */
	buf[2] = buf[0] | conf_byte;
	buf[0] = SENSOR_SLAVE_ADDRESS;
	buf[1] = SENSOR_WCA_REG;
	cpm_iic_write(buf, 3);

	conf_byte = 0;
		
	/* now let's read the status byte 2 */
	buf[0] = SENSOR_SLAVE_ADDRESS;
	buf[1] = SENSOR_RSL2_REG;
	if (cpm_iic_read(buf, 1, I2C_SET_OFFSET|I2C_SET_MASTER)) {
		printk("Sensor Status Byte2 0x%02x\r\n", buf[0]);
	}

	/* read conf Byte 2 and return value on buf[0] */
	buf[0] = SENSOR_SLAVE_ADDRESS;
	buf[1] = SENSOR_RCL2_REG;
	cpm_iic_read(buf, 1, I2C_SET_OFFSET|I2C_SET_MASTER);

	/* disable interrupts that already happened on Config Byte 2 */
	buf[2] = buf[0] | conf_byte;
	buf[0] = SENSOR_SLAVE_ADDRESS;
	buf[1] = SENSOR_WCA2_REG;
	cpm_iic_write(buf, 3);

	/* clear status Byte 1 */
	buf[0] = SENSOR_SLAVE_ADDRESS;
	buf[1] = SENSOR_RSL_REG;
	cpm_iic_read(buf, 1, I2C_SET_OFFSET|I2C_SET_MASTER);

	/* clear status Byte 2 */
	buf[0] = SENSOR_SLAVE_ADDRESS;
	buf[1] = SENSOR_RSL2_REG;
	cpm_iic_read(buf, 1, I2C_SET_OFFSET|I2C_SET_MASTER);

	enable_irq(SIU_IRQ1);
#endif

}

int cy_sm_init(void)
{
	volatile immap_t *pm = (volatile immap_t *)IMAP_ADDR;
	unsigned char buf[8] = {0};

	cy_sm_atx();
	cpm_iic_init();

	misc_register(&sm_atx_miscdev);
	misc_register(&sm_ipmi_miscdev);
	misc_register(&sm_sensor_miscdev);

#if 0
	/* set sensor thresholds for voltage/temperature */
	buf[0] = SENSOR_SLAVE_ADDRESS;
	buf[1] = SENSOR_WRHA_REG;
	buf[2] = 45;
	cpm_iic_write(buf, 3);

	/* set sensor conversion rate */
	buf[0] = SENSOR_SLAVE_ADDRESS;
	buf[1] = SENSOR_WCRW_REG;
	buf[2] = 1;
	cpm_iic_write(buf, 3);

	/* enable sensor interrupt on Config Byte 2 */
	buf[0] = SENSOR_SLAVE_ADDRESS;
	buf[1] = SENSOR_WCA2_REG;
	buf[2] = SENSOR_DIS_RT2 | SENSOR_DIS_IT | SENSOR_DIS_Vin3;
	cpm_iic_write(buf, 3);

	/* enable sensor interrupt on Config Byte 1 */
	buf[0] = SENSOR_SLAVE_ADDRESS;
	buf[1] = SENSOR_WCA_REG;
	buf[2] = SENSOR_MASK_RT2 | SENSOR_MASK_Vin3 | SENSOR_MASK_Vin2 | SENSOR_MASK_Vin1 | SENSOR_MASK_Vcc;
//	buf[2] = SENSOR_MASK_RT1 | SENSOR_MASK_Vin2 | SENSOR_MASK_Vin1 | SENSOR_MASK_Vcc;
	cpm_iic_write(buf, 3);

	/* clear status Byte 1 */
	buf[0] = SENSOR_SLAVE_ADDRESS;
	buf[1] = SENSOR_RSL_REG;
	cpm_iic_read(buf, 1, I2C_SET_OFFSET|I2C_SET_MASTER);

	/* clear status Byte 2 */
	buf[0] = SENSOR_SLAVE_ADDRESS;
	buf[1] = SENSOR_RSL2_REG;
	cpm_iic_read(buf, 1, I2C_SET_OFFSET|I2C_SET_MASTER);

	msdelay(100);

	/* Install our interrupt handler */
	if (request_8xxirq(SIU_IRQ1, cy_sm_temperature_interrupt, 0, "temp", 0) != 0)
		panic("Could not allocate TEMP IRQ!");
#endif

 	printk("Server Management init done. Power in ");
	(pm->im_ioport.iop_padat & POWER_MODE_STATUS) ? printk("IPMI mode.\r\n") : printk("ATX mode.\r\n");		

	return 0;
}

int cy_sm_atx(void)
{
	volatile immap_t *pm = (volatile immap_t *)IMAP_ADDR;

	/* enables MB, disables SM100 power */
    pm->im_ioport.iop_padat &= ~MB_POWER_ENABLE;
    pm->im_ioport.iop_padat |=  SM100_POWER_ENABLE;

	/* output signals */
    pm->im_ioport.iop_padir |= SM100_POWER_ON|SM100_POWER_ENABLE|MB_POWER_ENABLE;

	/* input signals */
    pm->im_ioport.iop_padir &= ~(POWER_MODE_STATUS | POWER_ON_MB);

	/* PA9-PA13 are general-purpose I/O */
    pm->im_ioport.iop_papar &= ~(POWER_MODE_STATUS|POWER_ON_MB|SM100_POWER_ON|SM100_POWER_ENABLE|MB_POWER_ENABLE);

	return 0;
}

