/*
 * Module: pmFwUpgrade.c
 *
 * Author:	Henrique Oliveira <henrique.oliveira@cyclades.com>
 *
 * Copyright:	(c) 1999-2003 Cyclades Corp.
 *
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terxms of the GNU General Public License
 *      as published by the Free Software Foundation; either version
 *      2 of the License, or (at your option) any later version.
 *
 *  --------------------------------------------------------------------------
 *  Description
 *  
 *  This software is a tool to upgrade the firmware of all equipments of 
 *  Cyclades AlterPath PM line. If you go to the main routine of this file
 *  you will see that this software follows the following procedure:
 *   1 - Get options passed in the command line;
 *       For more details on possible options please see the routine usage().
 *   2 - Get username from the command line
 *   3 - Get serial port number from the command line
 *   4 - Check file and load it
 *       Check if the provided file exists, load it to a intern memory region
 *       and analyse if this file can be an AlterPath-PM firmware file.
 *   5 - Get the file version
 *       Analyse the contents of the file and find out what firmware version
 *       it contains.
 *   6 - Set up serial port
 *       Open and configure the serial port that was received as parameter
 *   7 - Get PM unit's firmware version
 *       Find out what version of firmware the unit is currently running
 *   8 - Confirm Upgrade
 *       Show the user the current firmware version and the file version and
 *       ask him if he is sure about the upgrade. This step can be skip using
 *       the option -f
 *   9 - Prepare PM unit
 *       Log into the PM unit and reboot it. Chose the option "Upload a new
 *       firmware" in the boot loader menu.
 *   10 - Upgrade
 *        Send the firmware file using xmodem protocol.
 *   11 - Get unit's new version
 *        Check if the unit is able to reboot after the upgrade and try to get
 *        it's new firmware version.
 *  --------------------------------------------------------------------------
 *
 * $Log: pmFwUpgrade.c,v $
 * Revision 1.3  2004/02/02 19:02:09  regina
 * Removed the support to daisy-chain upgrade
 *
 * Revision 1.1  2003/08/27 22:51:52  henrique
 * First version of the software
 *
 */

// Linux includes
# include <stdio.h>
# include <sys/types.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <errno.h>
# include <unistd.h>
# include <termios.h>
# include <linux/ioctl.h>
# include <stdlib.h>
# include <string.h>
# include <sys/types.h>
# include <dirent.h>
# include <ctype.h>
# include <signal.h>
# include <sys/time.h>

// Version of this software
char version[] = 
"Cyclades Corporation\n"
"AlterPath-PM upgrade utility\n"
"Version 1.0";

// Basic OK or NOK
# define ERROR -1
# define SUCCESS 0

// This string will be looked for inside the fimrware file to check if the file
// is a valid firmware for AltherPath PM
# define PMFWIDSTRING "AlterPath PM"

// This type holds a firmware version
typedef int Version[3];

//The first PM version to have firmware upgrade
# define FIRSTVERSIONWITHUPGRADE {1, 1, 0}

// Maximum size for user/passwd PM string
# define PMMAXSTRLEN 8

/***************************************************************************
 * Show the usage for that customer that hasn't read the manual
 **************************************************************************/
void usage ( void ) {
        printf ("Usage:\n"
                "------\n"
                "pmFwUpgrade <options> <file> <serial port number>\n"
                "   - <options> can be:\n"
                "        -v: print this software version and exit\n"
                "        -h: show this help message and exit\n"
                "        -f: don't ask me anything. Do the upgrade\n"
                "        -p <password>: use this password to log onto the PM unit\n"
                "   - <file> is the firmware file you want to upload\n"
                "   - <serial port number> is the serial port where the PM unit is connected\n"
                "\nExample: pmFwUpgrade /tmp/pm_110.obj 3\n"
                "         Upgrade AlterPath-PM on the serial port 3 with the file /tmp/pm_110.obj\n\n"
                "         pmFwUpgrade -p cygnus /tmp/pm_109.obj 8\n"
                "         Upgrade AlterPath-PM on the serial port 8 with the file /tmp/pm_109.obj\n"
                "         Use the password cygnus to log into the PM unit\n");

}

/***************************************************************************
 * Print out the version of the software and the date and time of
 * compilation.
 **************************************************************************/
void printVersion (void) {
	printf("------------------------------------------------------------\n");
	puts(version);
	printf("Generated on %s %s\n", __DATE__, __TIME__);
	printf("------------------------------------------------------------\n");
}

/***************************************************************************
 **************************  -=-  ******************************************
 ********************** XMODEM STUFF ***************************************
 **************************  -=-  ******************************************
 **************************************************************************/
// ASCII Constants
#define      SOH  	001 
#define      EOT	004
#define      ACK  	006
#define      NAK  	025
#define	     CAN	030

// XMODEM Constants
#define      ERRORMAX  	10    // maximum errors tolerated while transferring a packet
#define      NAKMAX	120   // maximum times to wait for initial NAK when sending
#define      RETRYMAX  	5     // maximum retries to be made certain handshaking routines
#define      EOTMAX	10    // maximum times sender will send an EOT to end transfer
#define	     BBUFSIZE	128   // buffer size

/***************************************************************************
 * Set the structure <tv> to <seconds> seconds
 **************************************************************************/
void setTime ( struct timeval * tv, int seconds ) {
	tv->tv_sec = seconds;
	tv->tv_usec = 0;
}

/***************************************************************************
 * Tries to read one single byte from <fd>. If this bytes doesn't arrive in
 * <timeout> seconds returns a zero.
 **************************************************************************/
char readbyte (int fd, int timeout) {
	fd_set fdset;
	struct timeval tv;
	char ret;

	// Register fd on fdset
	FD_ZERO(&fdset);
	FD_SET(fd, &fdset);

	// Set timeout
	setTime(&tv, timeout);

	// Wait until there data or until timeout
	if ( select(fd+1, &fdset, NULL, NULL, &tv) ) {
		read(fd, &ret, 1);
		return ret;
	} else {
		return 0;
	}
}

/***************************************************************************
 * Sends one byte to fd
 **************************************************************************/
void sendbyte (int fd, char c) {
	write(fd, &c, 1);
}

/***************************************************************************
 * Send <bufSize> bytes from <buf> through <fd> using the xmodem protocol.
 * It returns ERROR if something goes wrong and SUCCESS in case it transmits
 * all bytes.
 **************************************************************************/
int xmodemSendBuffer (int fd, char *buf, int bufSize) {
	int bufctr = 0;         // array index for buf
	int sectnum;	        // packet number for packet header
	unsigned short checksum;// holds the checksum for a package
	char blockbuf[132];	// holds packet as it is constructed 
	int bufcnt = 0;         // index used while building each package
	int attempts = 0;	// number of attempts made to transmit a packet
	int sendfin = 0; 	// flag that we are sending the last packet
	int firstchar;		// first character in protocol transaction
	time_t start;		// starting time of transfer
	char sendresp;          // Stores answers received from remote side
	int i;                  // Counter
	int rate;               // Percentage of data that has been transmitted
	
	// wait for and read startup character
	printf("Waiting for AlterPath PM init character");
	do {
		while (((firstchar=readbyte(fd, 2)) != NAK) && (firstchar != CAN)) {
			if (firstchar == 0) {
				printf(".");
			}
			if (++attempts > NAKMAX) {
				printf("Remote System not responding\n");
				return ERROR;
			}

			if ((firstchar & 0x7f) == CAN) {
				printf("Transfer canceled by remote side\n");
				return ERROR;
			}
		}
	} while (firstchar != NAK);
	printf("\nInitializing transfer\n");

	// Initiate transfer
	sectnum = 1;
	attempts = 0;
	start = time((time_t *) 0);
	do {
		// Building the packet
		bufcnt = 0;
		blockbuf[bufcnt++] = SOH;        // Start of header byte
		blockbuf[bufcnt++] = sectnum;    // The packet number
		blockbuf[bufcnt++] = ~sectnum;   // The 1's complement of the packate number

		// Copying data to blockbuf and calculating checksum
		checksum = 0;
		for (i = 0; i < BBUFSIZE; i++) {
			if (bufctr < bufSize) {
				blockbuf[bufcnt++] = buf[bufctr];
				checksum = ((checksum + buf[bufctr]) & 0xff);
				bufctr++;
			} else {
				sendfin = 1;
				blockbuf[bufcnt++] = 0xff;
				checksum = ((checksum + 0xff) & 0xff);
			}
		}
		blockbuf[bufcnt++] = checksum;

		// Check if this is the last package
		if (bufctr >= bufSize) {
			sendfin = 1;
		}
			
		// Try to transmit package
		attempts = 0;
		do {
			// Send package
			write (fd, blockbuf, 132);
			attempts++;
			// Wait until it has been transmited
			tcdrain(fd);
	
			// Read the answer from remote
			sendresp = readbyte(fd, 4);
			// Check the answer
			if (sendresp != ACK) {
				if (sendresp == 0 ) {
					printf("Timeout on sector %d\n",sectnum);
				} else if (sendresp == CAN) {
					printf("ignored CAN on sector %d\n",sectnum);
				} else if (sendresp == NAK) {
					printf("NAK on sector %d\n",sectnum);
				} else {
					printf("Non-ACK on sector %d\n",sectnum);
				}
			}
		} while((sendresp != ACK) && (attempts < ERRORMAX));
    
		// Check if max allowed errors was reached
		if (attempts >= ERRORMAX) {
			sendbyte(fd, CAN); sendbyte(fd, CAN); sendbyte(fd, CAN); sendbyte(fd, CAN); sendbyte(fd, CAN);
			printf ("Too many errors in transmission\n");
			return ERROR;
		}

		// Increment sector number
		sectnum++;

		// Calculate rate and print if on the screen on a scp like format
		// "Progress: |=================================================| 100%"
		printf("\r");
		rate = (((sectnum - 1) * BBUFSIZE) * 100) / bufSize;
		printf("Progress: |");
		for (i = 0; i < 100; i += 2) {
			if (rate > i) {
				printf("=");
			} else {
				printf(" ");
			}
		}
		printf("| %d %% ", rate);
	
	} while (!sendfin);
	printf("\n");

	// Ending transmittion
	sendbyte(fd, EOT);
	attempts = 0;
	while ((readbyte(fd, 2) != ACK) && (attempts++ < EOTMAX)) {
		if (attempts > 1) {
			printf("EOT not ACKed, try %d\n", attempts);
		}
		sendbyte(fd, EOT);
	}
   
	// End didn't succed
	if (attempts >= RETRYMAX) 
		printf("Remote System Not Responding on Completion\n");

	printf("Transfer complete. %d bytes were sent in %d seconds\n", bufSize, (int)(time(NULL) - start));
	return SUCCESS;
}

/***************************************************************************
 **************************  -=-  ******************************************
 ******************  Serial Port Routines  *********************************
 **************************  -=-  ******************************************
 **************************************************************************/

/***************************************************************************
 * This routine configures the serial port pointed by fd to 9600 8n1 
 * without flow control and withou echo.
 **************************************************************************/
int configureSerialPort ( int fd ) {
	struct termios mytermio;

	// Get the configuration for the port
	if( tcgetattr(fd, &mytermio ) == -1 ) {
		perror("Could not get attributes."); 
		return ERROR;
	}

	cfmakeraw(&mytermio);

	// 8 data bits 
	mytermio.c_cflag &= ~CSIZE; 
	mytermio.c_cflag |= CS8; 

	// disable parity generation & detection
	mytermio.c_cflag &= ~(PARENB);

	// 1 stop bit
	mytermio.c_cflag &= ~(CSTOPB);

	// disable hardware flow control
	mytermio.c_cflag &= ~(CRTSCTS);

	// disable all software flow control
	mytermio.c_iflag &= ~(IXOFF | IXON | IXANY);

	// non-canonical (raw-mode)
	mytermio.c_lflag &= ~(ICANON);

	// disable extended functions
	mytermio.c_oflag &= ~(IEXTEN);

	// raw output
	mytermio.c_oflag &= ~(OPOST);

	// disable echo
	mytermio.c_lflag &= ~(ECHO);

	// Set the speed properly
	cfsetispeed (&mytermio, B9600);
	cfsetospeed (&mytermio, B9600);

	if ( tcsetattr(fd, TCSANOW, &mytermio) == -1 ) {
		perror("Could not apply configuration to serial port."); 
		return ERROR;
	}

	return SUCCESS;
}

/***************************************************************************
 * This opens the serial port, configure it to 8n1 without flow control
 * and without echo and return the file descriptor for the port.
 **************************************************************************/
int setupSerialPort ( int serialPortNumber ) {
	char devicename[12];
	int fd;

	// building device name
	sprintf(devicename, "/dev/ttyS%d", serialPortNumber);
	printf("Upgrading AlterPath PM on serial port %s\n", devicename);

	// open the device
	fd = open(devicename, O_RDWR | O_NONBLOCK);
	if ( fd == -1 ) {
		printf("error 3\n");
		perror("error opening device");
		usage();
		return ERROR;
	}

	// configure the device
	if ( configureSerialPort(fd) == ERROR ) {
		printf("error 4");
		close(fd);
		return ERROR;
	}
	
	// returning fd
	return (fd);
}


/***************************************************************************
 **************************  -=-  ******************************************
 ***************  Version management Routines  *****************************
 **************************  -=-  ******************************************
 **************************************************************************/

/***************************************************************************
 * Simple routine to print <version>
 **************************************************************************/
void printFirmwareVersion ( Version version ) {
	printf("%d.%d.%d", version[0], version[1], version[2]);
}

/***************************************************************************
 * Simple routine to clear <version>
 **************************************************************************/
void clearFirmwareVersion ( Version version ) {
	version[0] = 0;
	version[1] = 0;
	version[2] = 0;
}

/***************************************************************************
 * Simple routine to compare <version1> and <version2>.
 * It returns 1 if version1>version2
 * It returns 0 if version1=version2
 * It returns -1 if version1<version2
 **************************************************************************/
int compareFirmwareVersion (Version version1, Version version2 ) {
	if (version1[0] != version2[0]) {
		return ((version1[0] > version2[0]) ? 1:-1);
	}
	if (version1[1] != version2[1]) {
		return ((version1[1] > version2[1]) ? 1:-1);
	}
	if (version1[2] != version2[2]) {
		return ((version1[2] > version2[2]) ? 1:-1);
	}
	return 0;
}

/***************************************************************************
 * This routines prints a message on the screen telling the user what is
 * about to be done and ask him if he is sure about that. It returns
 * SUCCESS in case the user approve the upgrade and ERROR in case the user
 * cancel it.
 **************************************************************************/
int confirmUpgrade (Version file, Version unit) {
	char c;

	printf("\n");
	// Tell user what he is about to do
	if ( unit[0] == 0 && unit[1] == 0 && unit[2] == 0 ) {
		// It was not possible to detect unit version
		printf("Are you sure you want to upload the firmware version ");
		printFirmwareVersion(file);
		printf(" to your AlterPath-PM ?\n");
	} else {
		switch ( compareFirmwareVersion(file, unit) ) {
		case 1:// file version is bigger
			printf("\nAre you sure you want to upgrade from version\n");
			printFirmwareVersion(unit);
			printf(" to version ");
			printFirmwareVersion(file);
			printf(" ?\n");
			break;
		case 0: // versions are the same
			printf( "It seems you are trying to upload a firmware that is already\n"
				"running in your AlterPath-PM unit.\n"
				"Are you sure you want to proceed ?\n");
			break;
		case -1: // unit version is bigger
			printf("You are about to downgrade you AlterPath-PM unit from version\n");
			printFirmwareVersion(unit);
			printf(" to version ");
			printFirmwareVersion(file);
			printf("Are you sure you want to proceed ?\n");
			break;
		}
	}

	// Wait for user confirmation
	do {
		printf("Please answer 'y' (yes) or 'n' (no): ");
		while ( (c = (char) getchar()) == '\n' );
	} while (c != 'y' && c != 'Y' && c != 'n' && c != 'N');
	while ( ((char) getchar()) != '\n' );

	// Check user answer
	if (c == 'y' || c == 'Y') 
		return SUCCESS;
	else
		return ERROR;
}

/***************************************************************************
 **************************  -=-  ******************************************
 **  Default session. Everything that doesn't fit anywhere else  ***********
 **************************  -=-  ******************************************
 **************************************************************************/

/***************************************************************************
 * This routine searchs the buffer buf for the version string of the PM
 * firmware file. Both firmware file and output of the serial port have
 * similar formats. The format of the firmware file is:
 *       Cyclades Corporation<null>@(#)V x.y.z
 * and the format of the output is
 *       Cyclades Corporation\r\nV x.y.z
 *
 * This routine saves the version found on <buf> on <version> and returns
 * SUCCESS if it finds the firmware version, or returns ERROR in case it
 * is not able to find the firmware version 
 **************************************************************************/
int getFirmwareVersion (char *buf, int bufferSize, Version version) {
	int aux;
	char cycladesString[] = "Cyclades Corporation";

	// Looking for the string Cyclades Corporation
	for ( aux = 0; aux < bufferSize - (int)strlen(cycladesString); aux ++ ) {
		if ( !strncmp (&buf[aux], cycladesString, (int)strlen(cycladesString)) ) {
			// string cyclades found, skip it
			aux += strlen(cycladesString);

                        // Looking for 'V'
                        while ( (buf[aux] != 'V') && (aux < bufferSize - (int)strlen(cycladesString)) ) aux++;
                        if (buf[aux] == 'V') {
                                aux += 2;
                        } else {
                                return ERROR;
                        }

			// Pointing to first digit
			if ( isdigit(buf[aux]) ) {
				version[0] = (int)buf[aux] - (int)'0';
			} else {
                                return ERROR;
			}
			aux++;

			// It must be a dot
			if ( buf[aux] != '.' ) {
				return ERROR;
			}
			aux++;

			// Second digit
			if ( isdigit(buf[aux]) ) {
				version[1] = (int)buf[aux] - (int)'0';
			} else {
				return ERROR;
			}
			aux++;

			// a dot again
			if ( buf[aux] != '.' ) {
				return ERROR;
			}
			aux++;
			
			// Third digit
			if ( isdigit(buf[aux]) ) {
				version[2] = (int)buf[aux] - (int)'0';
			} else {
				return ERROR;
			}

			return SUCCESS;
		}
	}
	return ERROR;
}

/***************************************************************************
 * This routines reads the file /etc/pm.cyclades and look for the variable
 * admPasswd. If it finds this variable it copies it to <password> and
 * returns SUCCESS. If it doesn't find it it returns ERROR.
 **************************************************************************/
int getPasswordFromFile(char * password) {
	char filename[] = "/etc/pm.cyclades";
	char admPasswdId[] = "admPasswd=";
	struct stat fstat;
	char * fileBuf;
	int  fd;
	int aux;
	int i;

	// Trying to get information about the file
	if ( stat(filename, &fstat) == -1 ) {
		printf("Error getting information about file %s\n", filename);
		perror("Error on stat:");
		return ERROR;
	}

	// Opening the file for reading
	if ( (fd = open(filename, O_RDONLY)) == -1 ) {
		printf("Error opening the file %s\n", filename);
		perror("Error on open:");
		return ERROR;
	}

	// Allocating memory to load the file
	if ( (fileBuf = (char *) malloc (fstat.st_size)) == NULL ) {
		printf("Error allocating memory to read file %s. File size: %d\n", filename, (int)fstat.st_size);
		perror("Error on malloc:");
		close(fd);
		return ERROR;
	}

	// Reading the file
	printf("Loading file %s\n", filename);
	if ( (aux = read(fd, fileBuf, fstat.st_size)) != fstat.st_size ) {
		printf("Error reading file %s. Read %d bytes of %d\n", filename, aux, (int)fstat.st_size);
		perror("Error on read:");

		goto error;
	}

	// Looking for the id string. If the file does not contain this string the 
	// password can't be found.
	for ( aux = 0; aux < fstat.st_size - (int)strlen(admPasswdId); aux ++ ) {
		if ( !strncmp (&fileBuf[aux], admPasswdId, (int)strlen(admPasswdId)) ) {
			// the next char must be a '"'
			aux += strlen (admPasswdId);
			if ( fileBuf[aux++] != '"') {
				printf("Syntax error on file %s\n", filename);
				goto error;
			}
			
			for (i = 0; i < PMMAXSTRLEN; i++) {
				if ( fileBuf[aux+i] == '"' ) {
					printf("Password found !!!\n");
					password[i] = '\n';
					password[i+1] = 0;
					close(fd);
					free(fileBuf);
					return SUCCESS;
				}
				password[i] = fileBuf[aux+i];
			}
			printf("Password on file %s is too big\n", filename);
			goto error;
		}
	}

error:
	// ERROR
  	close(fd);
	free(fileBuf);
	return ERROR;
}

/***************************************************************************
 * This routine asks user to type the password of the pm unit. It stores
 * the password on the buffer pointed by <password>. Note that this buffer
 * has to be bigger the (PMMAXSTRLEN+1) because that is the maximum lenght
 * for a PM password.
 **************************************************************************/
void getPasswordFromUser(char * password) {
	char c;
	int idx = 0;

retry:
	// reset the password string
	bzero(password, PMMAXSTRLEN+1);
	idx = 0;

	// Ask user to type password
	printf("\nPlease insert the administrator password for your AlterPath-PM\n"
	       "unit and press enter: ");

	// Read user input
	do {
		// Store char
		c = getchar();
		*(password+idx++) = c;

		// Check for long passwords
		if (idx == PMMAXSTRLEN+1) {
			printf("Error: password too big\n");
			while(getchar() != '\n');
			goto retry;
		}
	} while ( c != '\n' );
}
		
/***************************************************************************
 * This routine gets the password for the PM unit. First it tries to load 
 * the password from the file pm.cyclades. If it is not possible, this 
 * routine asks the user to type the password for the unit.
 **************************************************************************/
void getPassword(char * password) {
        if ( getPasswordFromFile(password) == SUCCESS ) {
                return;
        } else {
                getPasswordFromUser(password);
                return;
        }
}

/***************************************************************************
 * This routine sends the username, the password, reboot the unit and then
 * tells the boot code that it will upload a new firmware. If it can
 * successfully tell the unit to upgrade the firmware it will return
 * SUCCESS. The caller can provided the password on the <password> field
 * and this routine will use <password> on the authentication. In case
 * <password> is not provided (*password = 0) this routine will try to get
 * the password from the file /etc/pm.cyclades and, if it fails, it will
 * ask the user for the password.
 **************************************************************************/
char username[] = "Username: ";
char passwdId[] = "Password: ";
char prompt[] = "pm>";
char menu[] = "Enter choice (1 or 2): ";
char wrongpasswd[] = "Authentication failed";
# define READINGBUFSIZE 512
# define DEFTO 10
# define MAXTO 2
int prepareUnit (int fd, char * password) {
	char c;
	char buf[READINGBUFSIZE];
	unsigned int idx;
	unsigned int i;
	int tries = 0;
        int passwdProvided = (*password) ? 1 : 0;

retry:
	// Reset the reception buffer
	bzero(buf, READINGBUFSIZE);
	idx = 0;

	// Get all bytes available. It it stays DEFTO seconds without receiving
	// one byte there is something wrong, return error.
	while ( (c = readbyte(fd,DEFTO)) ) {
		if (c == '\r' || c == '\n') {
			bzero(buf, READINGBUFSIZE);
			idx = 0;
			continue;
		}

		// store character and update idx
		buf[idx++] = c;
		// Check for buffer overflow
		if (idx >= READINGBUFSIZE) break;

		// Check for all string we would like to receive
		for (i = 0; i < idx; i++) {
			if ( (strlen(username) <= (idx - i)) && (!strcmp(&buf[i], username)) ) {
				write(fd, "admin\n", strlen("admin\n"));
				bzero(buf, READINGBUFSIZE);
				idx = 0;
			} else if ( (strlen(passwdId) <= (idx - i)) && (!strcmp(&buf[i], passwdId)) ) {
				if (password[0] == 0) {
					getPassword(password);
				} else if (!passwdProvided) {
					getPasswordFromUser(password);
				}
				write(fd, password, strlen(password));
				bzero(buf, READINGBUFSIZE);
				idx = 0;
			} else if ( (strlen(prompt) <= (idx - i)) && (!strcmp(&buf[i], prompt)) ) {
				write(fd, "reboot\n", strlen("reboot\n"));
				bzero(buf, READINGBUFSIZE);
				idx = 0;
			} else if ( (strlen(menu) <= (idx - i)) && (!strcmp(&buf[i], menu)) ) {
				write(fd, "1\n", strlen("1\n"));
				bzero(buf, READINGBUFSIZE);
				idx = 0;
				return SUCCESS;
			} else if ( (strlen(wrongpasswd) <= (idx - i)) && (!strcmp(&buf[i], wrongpasswd)) ) {
				bzero(buf, READINGBUFSIZE);
				idx = 0;
				puts(wrongpasswd);
                                if (passwdProvided) {
                                        printf("It seems you have provided a wrong password\n");
                                        return ERROR;
                                }
			}
		}
	}

	// It's been a while the PM unit doesn't send us a byte.
	// Try to excite it with a new line char
	if ( tries <= MAXTO ) {
		write(fd, "\n", strlen("\n"));
		tries++;
		goto retry;
	}
	
	printf("Could not prepare unit. Timeout\n");
	return ERROR;
}

/***************************************************************************
 * This routine tries to find out how many units in daisy chain we got
 * connected to the serial port. It look for the string:
 *       [PM]: IPDU: y
 * and then return the number y. If for some reason this routine fails to
 * detect the number of ipdu's, it returns 1 and print a warn message
 **************************************************************************/
int getNumberOfIpdu (char * buf, int bufSize) {
        int i;
        char ipdu[] = "IPDU: ";

        for (i = 0; i < bufSize - (int)strlen(ipdu); i++) {
                if (!strncmp (buf+i, ipdu, strlen(ipdu))) {
                        i += strlen(ipdu);
                        if (isdigit(buf[i])) {
                                printf("Number of units in daisy-chain: %d\n", atoi(buf+i));
                                return atoi(buf+i);
                        } else {
                                printf("[Warning]Error detecting number of units in daisy-chain. Using default number [1]\n");
                                return 1;
                        }
                }
        }
        printf("[Warning2]Error detecting number of units in daisy-chain. Using default number [1]\n");
        return 1;
}

/***************************************************************************
 * This routine tries to get the software version of the AlterPath-PM unit
 * represented by the file descriptor <fd> by analysing the header of the
 * software:
 *
 *       AlterPath PM
 *       Copyright (c) 2002-2003 Cyclades Corporation
 *       V 1.1.0m Aug 11, 2003
 *       
 *       [PM]: IPDU: y
 *       [PM]: OUT: z
 *       Username:
 *
 * It is very important that this routine does not read the "Username: "
 * string because this string will be used by the routine prepareUnit()
 * down the road.
 *
 * This routine uses the substring "OUT" on the string "[PM]: OUT:" to
 * identify the end of the header. As soon as it identifies the header
 * it sends the whole buffer to getFirmwareVersion that will return the
 * firmware version on the header.
 *
 * If this routines takes more than <timeout> to identify the header it will
 * return an error. The unit might not have an op code and might be waiting
 * for a firmware. If everything works fine this routine the number of ipdu
 * that were found in daisy-chain (the number 'y' on the text above)
 **************************************************************************/
int getVersion (int fd, Version version, int timeout) {
	char c;
	char buf[READINGBUFSIZE];
	unsigned int idx;
	time_t start = time(NULL);

        // Reset <version>
	clearFirmwareVersion(version);

	// Reset the reception buffer
	bzero(buf, READINGBUFSIZE)  ;
	idx = 0;

	// Get all bytes available. It it stays DEFTO seconds without receiving
	// one byte there is something wrong, return error.
	while ( (c = readbyte(fd,DEFTO)) ) {
		buf[idx++] = c;
		if (idx >= READINGBUFSIZE) {
			printf("Buffer overflow waiting for AlterPath-PM current version\n");
			return ERROR;
		}

		// The character 'I' means that we have received the whole header and
		// now we are receiving the string "[PM]: IPDU:...". Thus have have
		// enough information to call getFirmwareVersion.
		if (buf[idx-1] == 'T' && buf[idx-2] == 'U' && buf[idx-3] == 'O') {
			if ( getFirmwareVersion (buf, idx, version) == ERROR ) {
                                return ERROR;
                        } else {
                                return getNumberOfIpdu (buf, idx);
                        }
		}

		// Checking for timeout
		if ( (int)(time(NULL) - start) > timeout ) {
			break;
		}
	}
	return ERROR;
}

/***************************************************************************
 * This routine check the file <filename> and, if this file seems to be
 * a valid PM firmware file it reads the file into a allocated buffer
 * and returns a pointer to this buffer. If the file does not seem to be
 * a firmware file or if it fails to read the file it returns ERROR. If
 * it get to make everything it returns SUCCESS and point <*file> to the
 * allocated buffer. The variable <force> says if the routine must test
 * the firmware file <force=0> or not <force=1>. 
 **************************************************************************/
unsigned int checkFile(char *filename, char **file) {
	struct stat fstat;
	char * fileBuf;
	int  fd;
	int aux;

	// Trying to get information about the file
	if ( stat(filename, &fstat) == -1 ) {
		printf("Error getting information about file %s\n", filename);
		perror("Error on stat:");
		return ERROR;
	}

	// Opening the file for reading
	if ( (fd = open(filename, O_RDONLY)) == -1 ) {
		printf("Error opening the file %s\n", filename);
		perror("Error on open:");
		return ERROR;
	}

	// Allocating memory to load the file
	if ( (fileBuf = (char *) malloc (fstat.st_size)) == NULL ) {
		printf("Error allocating memory to read the file. File size: %d\n", (int)fstat.st_size);
		perror("Error on malloc:");
		close(fd);
		return ERROR;
	}

	// Reading the file
	printf("Loading file %s\n", filename);
	if ( (aux = read(fd, fileBuf, fstat.st_size)) != fstat.st_size ) {
		printf("Error reading file. Read %d bytes of %d\n", aux, (int)fstat.st_size);
		perror("Error on read:");
		free(fileBuf);
		close(fd);
		return ERROR;
	}

	// Looking for the id string. If the file does not contain this string it can not
	// be a firmware file.
        for ( aux = 0; aux < fstat.st_size - (int)strlen(PMFWIDSTRING); aux ++ ) {
                if ( !strncmp (&fileBuf[aux], PMFWIDSTRING, (int)strlen(PMFWIDSTRING)) ) {
                        break;
                }
        }
        if ( aux ==  fstat.st_size - (int)strlen(PMFWIDSTRING)) {
                printf("This file does not look like a AlterPath PM firmware. Aborting\n");
                free(fileBuf);
                close(fd);
                return ERROR;
        }

	// Everything went ok
	close(fd);
	*file = fileBuf;
	return fstat.st_size;
}


/***************************************************************************
 **************************  -=-  ******************************************
 ** MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN **
 **************************  -=-  ******************************************
 **************************************************************************/
extern char * optarg;
extern int optind;

int main (int argc, char ** argv) {
	int serialPortNumber;      // Number of the serial port where PM is connected
	char * filename;           // Pointer to the firmware file name
	char * file;               // Pointer to the buffer that contains the file
	unsigned int fileSize;     // Size of the buffer that contains the file
	int fd;                    // Serial port file descriptor
	Version fileVersion;       // Version of the firmware on the file
	Version unitVersion;       // See file header
        Version firstVersion = FIRSTVERSIONWITHUPGRADE;    // First Version of the pm
        int force = 0;             // force upgrade (1) or perform checks (0)
        int i;                     // index 
        int opt;                   // aux variable
        char password[PMMAXSTRLEN+1]; // holds units passwords when user provides it on the cl
        int numberOfIpdu;          // Number of ipdu's in daisy chain

        // zero the password string
        bzero (password, sizeof(password));

        // 1 - get all options
	while ((opt = getopt(argc, argv, "p:vfh")) != EOF) {
		switch (opt) {
			case 'h':           // show help and exit
                                usage();
                                return 0;
			case 'f':           // don't perform checks
				force = 1;
				break;
                        case 'v':           // show version and exit
                                printVersion();
                                return 0;
                        case 'p':           // use this password
                                for (i = 0; i < PMMAXSTRLEN; i++) {
                                        if ( *(optarg+i) == ' ') {
                                                break;
                                        }
                                        password[i] = *(optarg + i);
                                }
                                password[i] = 0;
                                break;
		}
	}

        // fixing the indexes
        argc -= optind;
        argv += optind;

	// Check if filename and serial port were provided
	if ( argc != 2 ) {
		printf("Error 1: Missing arguments\n");
		usage();
		return -1;
	}

	// 2 - Getting file location
	filename = argv[0];

	// 3 - Getting the serial port number
	serialPortNumber = atoi(argv[1]);
	if ( serialPortNumber < 1 || serialPortNumber > 48) {
		printf("Error 2: Invalid serial port number\n");
		usage();
		return -2;
	}

	// Zeroing output buffer (for fast printings) and clearing the screen
	setbuf(stdout, NULL);
	setbuf(stdin, NULL);
	printf("\033[H\033[J");

	// Show version
	printVersion();

	// 4 - Check the file and load it up
	fileSize = checkFile(filename, &file);
	if ( fileSize == (unsigned int) ERROR ) {
		return ERROR;
	}

	// 5 - Get firmware version
	clearFirmwareVersion(fileVersion);
	if ( getFirmwareVersion(file, fileSize, fileVersion) == ERROR ) {
		printf("Could not get version of the file\n");
		free(file);
		return ERROR;
	}
	printf("File %s contains firmware version:", filename);
	printFirmwareVersion(fileVersion);
	printf("\n");

	// 6 - Setting up serial port
	fd = setupSerialPort(serialPortNumber);
	if (fd == ERROR) {
		free(file);
		return ERROR;
	}

	// 7 - Try to get unit firmware version
	printf("Trying to get unit's current firmware version...\n");
	if ( (numberOfIpdu = getVersion (fd, unitVersion, DEFTO)) != ERROR ) {
		printf("Unit is running firmware version: ");
		printFirmwareVersion(unitVersion);
		printf("\n");
                if ( compareFirmwareVersion (unitVersion, firstVersion) < 0 ) {
                        printf("The AlterPath-PM firmware version ");
                        printFirmwareVersion(unitVersion);
                        printf(" does not support firmware upgrade. Aborting...\n");
                        return -1;
                } else if (numberOfIpdu > 1) {
                        printf("\n             --==== READ THIS CAREFULLY ====---\n\n");
                        printf("[IMPORTANT WARNING]This software detected that you have more than\n");
                        printf("one AlterPath-PM connected in daisy-chain. THIS SOFTWARE WILL NOT\n");
                        printf("UPGRADE THE WHOLE CHAIN. IT WILL ONLY UPGRADE THE MASTER UNIT. FOR\n");
                        printf("MORE INFORMATION ON THIS LIMITATION PLEASE CONSULT THE PRODUCT DOCUMENTATION\n");
                }
	} else {
		printf("Could not identify software version the unit is running\n");
	}

	// 8 - Ask user to confirm upgrade
        if (!force) {
                if ( confirmUpgrade(fileVersion, unitVersion) == ERROR ) {
                        printf("Aborting upgrade\n");
                        free(file);
                        close(fd);
                        return ERROR;
                }
                printf("Proceeding ...\n");
        }
	
	// 9 - Prepare AltherPath PM
	if ( prepareUnit(fd, password) == ERROR ) {
		printf("Could not complete upgrade\n");
		close(fd);
		free(file);
		return ERROR;
	}

	// 10 - Upgrade
	if ( xmodemSendBuffer(fd, file, fileSize) == ERROR ) {
	     close(fd);
	     free(file);
	     return ERROR;
	}

	// 11 - Try to get unit firmware version
	printf("\nChecking installation\n");
        printf("It might take a while...\n");
	if ( getVersion (fd, unitVersion, 30) != ERROR ) {
		printf("Unit is now running firmware version:");
		printFirmwareVersion(unitVersion);
		printf("\n");
	} else {
		printf("Could not identify software version the unit is running\n");
                printf("Something really bad might have happened.\n");
                printf ("Please try to make the upgrade again and make sure you are\n"
                        "using a valid firmware file\n");
                close(fd);
                free(file);
                return ERROR;                
	}

	// Close serial port
	close(fd);
	free(file);
	return SUCCESS;
}
