/*
 *  Module:  2way-raw.c (raw data socket client/server application)
 *
 *  Author:  Marcelo Peccin <marcelo.peccin@cyclades.com>
 *
 *  Copyright:  (c) 2004 Cyclades Corp.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 *
 *  Date:  September 16, 2004
 *
 *  Comments:  The first release of this module is intended to
 *             Telia's project. The project involves a main Application
 *             that pools and receive asynchronous alarms from
 *             a number of remote serial devices.
 *             This software is based on modem-replace.c.
 */

#include "rot_nports.h"
#include <asm/prxk.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <sys/ioctl.h>
#include "server.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include "2way-raw.h"


// Local Variables
static int     version = 1;
static int     debug = FALSE;
static int     directconn = FALSE;
static int     firstpass = TRUE;
static int     fd = -1;
static int     ffd = -1;
static int     serial_port, txdisp;
static int     fdserial, fdsock = NO_SOCK;
static uchar  *sbuffp;
static int     exitstatus = PROBLEM;
static int     sertimeout;
static time_t  lasttime;

static struct sigaction    sa;
static struct MachineType  mt;

static struct itimerval  almtimeout;


/***************************************************************************/
/**********     Print caught signal to log file and transmmit     **********/
/***************************************************************************/
static void  raw2w_got_signal(int sig)
{
	if (debug && (sig != SIGALRM))
		mpfprint("Got signal - %s !!!\n\n", strsignal(sig));
	
	switch(sig) {
	case SIGINT:
	case SIGTERM:
	case SIGHUP:
		exit(OK);
	case SIGALRM:
		if (txdisp) { //send received data to Socket Connection, if any
			if (tcpsend() == ERROR) {
				mpfprint("error writing to tcp port: %s (%d)\n", strerror(errno), errno);
				if (close(fdsock) == ERROR) exitstatus = ERRCLOSE;
				if (close(fdserial) == ERROR) exitstatus = ERRCLOSE;
				exit(exitstatus);
			}
			if (sertimeout) lasttime = time(NULL);
			if (debug) mpfprint("TX due %s...\n", strsignal(sig));
		}
	}
}


/***********************************************************/
/**********     Prepare to catch some signals     **********/
/***********************************************************/
static void  prepare_signals()
{

	bzero((void*)&sa, sizeof(struct sigaction));

	sa.sa_handler = raw2w_got_signal;

	sigaction(SIGINT,  &sa, NULL);
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGHUP,  &sa, NULL);
	sigaction(SIGALRM, &sa, NULL);

}
/***********************************************************/



/******************************/
int  main(int argc, char **argv)   // 2-way Raw Data application entry point
/******************************/
{
	raw2wconf           config;
	int                 arg;
	uchar               sdevbase[] = "/dev/ttyS";
	int                 fdlsock = NO_SOCK, flags;
	struct sockaddr_in  sock;
	int                 sal = sizeof(sock);
	fd_set              inputs, testfds;
	int                 fdready, event = RAW2W_NO_EVENT;
	struct timeval      timeout;
	uchar               dest_ip[25] = {0};
	int                 rec, connexpired;
	uchar               tcpbuff[TCPBUFFSIZE];
    uchar               serialbuff[SBUFFSIZE];


	// get machine type
	get_machine_type(&mt);
	

	// take command line arguments
	if ((argc < 2) || ((config.serial.port = atoi(argv[1])) < 1) || (config.serial.port > mt.nports)) {
		if (mt.nports > 1)
			mpfprint("2-way Raw Data must be called with a port (1..%d) as first argument.\n", mt.nports);
		else
			mpfprint("2-way Raw Data must be called with a port as first argument.\n");
		goto end_2wraw;		
	}
	serial_port = config.serial.port; //serial_port has a bigger scope
	if ((mt.board_type == BOARD_TS1H) || (mt.board_type == BOARD_TS110)) {
		// TS100/110 has its own zImage...
		config.serial.port = 0; //the shared console and serial port
		sprintf(config.serial.dev, "%s%d", sdevbase, config.serial.port);
	} else {
		sprintf(config.serial.dev, "%s%d", sdevbase, config.serial.port);
		config.serial.port--;
	}
	for (arg = 2; arg < argc; arg++) {
		if (!strcmp(argv[arg], "-debug")) {
			debug = TRUE;
  		} else if (!strcmp(argv[arg], "-connect")) {
			directconn = TRUE;
			if (++arg == argc) {
				mpfprint("-connect option must be followed by an IP address\n");
				goto end_2wraw;
			}
			strcpy(dest_ip, argv[arg]);
		}
	}	
	if (debug)
		mpfprint("\n2-way Raw Data running for serial port %02d (v%d)...\n", serial_port, version);


    // initialize Shared Memory
	if (shm_init(-1) < 0)
		goto end_2wraw;
	while (shm_get_set_val(0, &cy_shm->st_cy_ras, 0, CY_SEM_CONFIG) == 0)
		sleep(1);


	// prepare to catch some signals and load sbufp
	prepare_signals();
	sbuffp = serialbuff;


	// copy Configuration Information
	config.serial.speed    = cy_shm->line_conf[config.serial.port].speed; 
	config.serial.datasize = cy_shm->line_conf[config.serial.port].DataSize;
	config.serial.parity   = cy_shm->line_conf[config.serial.port].Parity;
	config.serial.stopbits = cy_shm->line_conf[config.serial.port].StopBits;
	config.serial.flowctrl = cy_shm->line_conf[config.serial.port].flow;
	config.serial.media	   = cy_shm->line_conf[config.serial.port].media;
	config.localip         = cy_shm->line_conf[config.serial.port].ipno;
	config.localport       = cy_shm->line_conf[config.serial.port].socket_port;
	config.remoteip        = cy_shm->line_conf[config.serial.port].host;
	config.remoteport      = cy_shm->line_conf[config.serial.port].host_port;
	config.maxidle         = cy_shm->line_conf[config.serial.port].host_idletime;
	config.tx_interval     = cy_shm->line_conf[config.serial.port].tx_interval;

	// info printing
	bzero(&sock, sizeof(sock));
	sock.sin_addr.s_addr = config.localip;
	if (debug)
	    mpfprint("\
config.serial.speed    = %d\n\
config.serial.datasize = %d\n\
config.serial.parity   = %d (1,odd,even)\n\
config.serial.stopbits = %d\n\
config.serial.flowctrl = %d (0,Hw,Sw)\n\
config.serial.media    = %d (232=0)\n\
config.localip         = %s\n\
config.localport       = %d\n",
				 config.serial.speed,
				 config.serial.datasize,
				 config.serial.parity,
				 config.serial.stopbits,
				 config.serial.flowctrl,
                 config.serial.media,
				 inet_ntoa(sock.sin_addr),
				 config.localport);
	sock.sin_addr.s_addr = config.remoteip;
	if (debug)
	    mpfprint("\
config.remoteip        = %s\n\
config.remoteport      = %d\n\
config.maxidle         = %d s\n\
config.tx_interval     = %d ms\n",
				 inet_ntoa(sock.sin_addr),
				 config.remoteport,
                 config.maxidle,
                 config.tx_interval);


    // open and configure Serial Port
	if ((fdserial = open(config.serial.dev, O_RDWR | O_NOCTTY | O_NONBLOCK)) == ERROR) {
		mpfprint("error opening serial port: %s (%d)\n", strerror(errno), errno);
		goto end_2wraw;
	}
	if (config_port(fdserial, &config.serial) == ERROR) {
		mpfprint("problems configuring serial port\n");
		goto end_closeserial;
	}
	if (debug && directconn)
		mpfprint("...serial port is ready.\n\n");


 // Main Loop
  	while (1) {
		int flag;
		sertimeout = FALSE;
		connexpired = FALSE;

		if (!directconn) {
			if (debug && !firstpass)
				mpfprint("\nentering Listen Mode\n");

			// open Listen Socket
			if ((fdlsock = socket(AF_INET, SOCK_STREAM, 0)) == ERROR) {
				mpfprint("error creating listen socket: %s (%d)\n", strerror(errno), errno);
				goto end_closeserial;
			}
			// must be non blocking
			flags = fcntl(fdlsock, F_GETFL);
			if (fcntl(fdlsock, F_SETFL, flags | O_NONBLOCK) == ERROR) {
				mpfprint("error making non blocking socket: %s (%d)\n", strerror(errno), errno);
				goto end_closelsocket;
			}
		
			bzero(&sock, sizeof(sock));
			sock.sin_family = AF_INET;
			sock.sin_addr.s_addr = config.localip;
			sock.sin_port = config.localport;
			
			// "name" Listen Socket
			flag = FALSE;
			do {
				if (bind(fdlsock, (struct sockaddr*)&sock, sal) == ERROR) {
					if (errno != EADDRINUSE) {
						mpfprint("error binding listen socket: %s (%d)\n", strerror(errno), errno);
						goto end_closelsocket;
					}
					if (debug && !flag) {
						mpfprint("waiting socket to be available...\n");
						flag = TRUE;
					}
					sleep(1);
				}
				else
					break;
			} while (1);
			
			// put Listen Socket to listen
			if (listen(fdlsock, 1) == ERROR) {
				mpfprint("error putting listen socket to listen: %s (%d)\n", strerror(errno), errno);
				goto end_closelsocket;
			}
		
			if (debug) {
				if (firstpass) {
					mpfprint("...serial port and listen socket are ready.\n\n");
					firstpass = FALSE;
				} else {
					mpfprint("listen socket ready (%s)\n", inet_ntoa(sock.sin_addr));
				}
			}

		}

		FD_ZERO(&inputs);
		FD_SET(fdserial, &inputs);

	    // wait for an EVENT to happen
		while (!event && !directconn) {

			testfds = inputs;
			timeout.tv_sec = 0;
			timeout.tv_usec = 30 * 1000;  // 30ms

			if ((fdready = select(fdserial+1, &testfds, NULL, NULL, &timeout)) == ERROR) {
				mpfprint("error in select: %s (%d)\n", strerror(errno), errno);
				goto end_closelsocket;
			}

			if (fdready) { //data received

				if ((rec = read(fdserial, serialbuff, SBUFFSIZE)) == ERROR) {
					mpfprint("error reading serial port: %s (%d)\n", strerror(errno), errno);
					goto end_closelsocket;
				}
				txdisp = rec;
				event = ALARM_RECEIVED;
				if (debug)
					mpfprint("received serial alarm\n");

			} else { //timeout
				
				if ((fdsock = accept(fdlsock, (struct sockaddr*)&sock, &sal)) == ERROR) {
					if (errno != EAGAIN) {
						mpfprint("error accepting socket connection: %s (%d)\n", strerror(errno), errno);
						goto end_closelsocket;
					}
				} else { //remote socket connection accepted

					event = SOCK_CONN_ACCEPT;
					if (debug)
						mpfprint("socket connection accepted   <<<<<<<<<<\n");

				}
			}			

		} /* while (!event) */

		if (!directconn) {
			if (close(fdlsock) == ERROR) { //stop listen
				exitstatus = ERRCLOSE;
				if (fdsock != NO_SOCK) close(fdsock);
				goto end_closeserial;
			}
			fdlsock = NO_SOCK;
		}

		if ((event == ALARM_RECEIVED) || (directconn)) {
            //serial side sent data or option -connect was issued
			
			if ((fdsock = socket(AF_INET, SOCK_STREAM, 0)) == ERROR) {
				mpfprint("error creating remote socket: %s (%d)\n", strerror(errno), errno);
				goto end_closeserial;
			}

			bzero(&sock, sizeof(sock));
			sock.sin_family = AF_INET;
			if (directconn) {
				if (!inet_aton(dest_ip, &sock.sin_addr)) {
					mpfprint("problems setting remote socket\n");
					if (directconn)
						goto end_closesocket;
					if (close(fdsock) == ERROR) {
						exitstatus = ERRCLOSE;
						goto end_closeserial;
					}
					fdsock = NO_SOCK;
					event = RAW2W_NO_EVENT; //prepare to next pass
					sleep(3/*sec*/); //before recreating the Listen Socket					
					continue;
				}
			} else {
				sock.sin_addr.s_addr = config.remoteip;
			}
			sock.sin_port = config.remoteport;

			if (connect(fdsock,(struct sockaddr*)&sock, sizeof(sock)) == ERROR) {
				mpfprint("error connecting to remote host: %s (%d)\n", strerror(errno), errno);
				if (directconn)
					goto end_closesocket;
				if (close(fdsock) == ERROR) {
					exitstatus = ERRCLOSE;
					goto end_closeserial;
				}
				fdsock = NO_SOCK;
				event = RAW2W_NO_EVENT; //prepare to next pass
				sleep(3/*sec*/); //before recreating the Listen Socket
				continue;
			}

			if (!directconn && config.maxidle) sertimeout = TRUE;
			if (debug)
				mpfprint("remote connection established   >>>>>>>>>>\n");

			if (event == ALARM_RECEIVED) { //there is an alarm to be sent immediately

				if (tcpsend() == ERROR) {
					mpfprint("error writing to tcp port: %s (%d)\n", strerror(errno), errno);
					goto end_closesocket;
				}
				lasttime = time(NULL);
			}

		}
		event = RAW2W_NO_EVENT; //prepare to next pass

		// initialize TCP transmission control
		txdisp = 0;
		if (config.tx_interval) {
			almtimeout.it_value.tv_usec = config.tx_interval * 1000;
			setitimer(ITIMER_REAL, &almtimeout, NULL);
		}


		FD_SET(fdsock, &inputs);
		/*****************************/
		/* Receive and Transmit data */
		/*****************************/
		while (!connexpired) {

			testfds = inputs;
			timeout.tv_sec = 0;
			timeout.tv_usec = config.tx_interval * 1000;  // us
			if (!timeout.tv_usec)
				timeout.tv_usec = 30 * 1000;  // 30ms

			if ((fdready = select(Max(fdserial, fdsock) + 1, &testfds, NULL, NULL, &timeout)) == ERROR) {
				if (errno == EINTR) { //Interrupted system call (due to max tx_interval)
					fdready = 0;
				} else {
					mpfprint("error in select: %s (%d)\n", strerror(errno), errno);
					goto end_closesocket;
				}
			}

			if (fdready) {

				/*****************************/
				if (FD_ISSET(fdsock, &testfds)) {   // Activity on ethernet port
				/*****************************/
			
					// read data
					if ((rec = recv(fdsock, tcpbuff, TCPBUFFSIZE, 0)) == ERROR) {
						mpfprint("error reading socket: %s (%d)\n", strerror(errno), errno);
						goto end_closesocket;
					}

					if (!rec) { //remote socket has closed
						if (debug)
							mpfprint("connection reset by peer   <<<<<<<<<<\n");
						break;
					}

					// received data is ready to be sent to Serial Port
					if (ssend(fdserial, tcpbuff, rec) == ERROR) {
						mpfprint("error writing to serial port: %s (%d)\n", strerror(errno), errno);
						goto end_closesocket;
					}
					if (sertimeout) lasttime = time(NULL);

				}

				/*******************************/
				if (FD_ISSET(fdserial, &testfds)) {   // Activity on serial port
				/*******************************/

					// read data
					if ((rec = read(fdserial, (serialbuff + txdisp), (SBUFFSIZE - txdisp))) == ERROR) {
						mpfprint("error reading serial port: %s (%d)\n", strerror(errno), errno);
						goto end_closesocket;
					}

					// check buffer limit and immediate TX
					if (((txdisp += rec) >= SBUFFLIMIT) || !config.tx_interval){

						// send received data to Socket Connection
						if (tcpsend() == ERROR) {
							mpfprint("error writing to tcp port: %s (%d)\n", strerror(errno), errno);
							goto end_closesocket;
						}
						if (sertimeout) lasttime = time(NULL);

					}

				}
			
			} else { //timeout

				if (txdisp) { //send received data to Socket Connection, if any
					if (tcpsend() == ERROR) {
						mpfprint("error writing to tcp port: %s (%d)\n", strerror(errno), errno);
						goto end_closesocket;
					}
					if (sertimeout) lasttime = time(NULL);
				}

			}

			if (sertimeout && ((config.maxidle - (time(NULL) - lasttime)) <= 0)) {
				if (debug)
					mpfprint("idle timeout   >>>>>>>>>>\n");
				connexpired = TRUE;
			}

		}


		// stop TX timer
		if (config.tx_interval) {
			almtimeout.it_value.tv_usec = 0;
			setitimer(ITIMER_REAL, &almtimeout, NULL);
		}

		// tear down connection
		if (debug)
			mpfprint("tearing down TCP connection\n");
		if (close(fdsock) == ERROR) {
			exitstatus = ERRCLOSE;
			goto end_closeserial;
		}
		fdsock = NO_SOCK;


		sleep(3/*sec*/); //before recreating the Listen Socket or trying a new connection


	} /* Main Loop */


end_closesocket:
	if (close(fdsock) == ERROR) exitstatus = ERRCLOSE;
	goto end_closeserial;

end_closelsocket:
	if (close(fdlsock) == ERROR) exitstatus = ERRCLOSE;

end_closeserial:
	if (close(fdserial) == ERROR) exitstatus = ERRCLOSE;

end_2wraw:
	if (debug) mpfprint("................................... exit status: %d\n\n", exitstatus);


	exit(exitstatus);
	

} /* main() */



/*****************************************************************/
/**********   This routine configures the Serial Port   **********/
/*****************************************************************/
static int  config_port(int fd, sconf *pconf)
{ 
	struct termios  term;
	int             aux;

	
	// Getting the port configuration
	if (tcgetattr(fd, &term) == ERROR) {
		return ERROR;
	}

	// Setting media type
	if ((mt.board_type == BOARD_TS1H) || (mt.board_type == BOARD_TS110)) {
		aux = pconf->media;
		if (ioctl(fd, TIOCSERCONFIG, &aux) == ERROR) {
			return ERROR;
		}
	}

	// Setting raw mode
	cfmakeraw(&term);

	// Setting speed
	switch (pconf->speed) {
	case 50:
		aux = B50;
		break;
	case 75:
		aux = B75;
		break;
	case 110:
		aux = B110;
		break;
	case 134:
		aux = B134;
		break;
	case 150:
		aux = B150;
		break;
	case 200:
		aux = B200;
		break;
	case 300:
		aux = B300;
		break;
	case 600:
		aux = B600;
		break;
	case 1200:
		aux = B1200;
		break;
	case 1800:
		aux = B1800;
		break;
	case 2400:
		aux = B2400;
		break;
	case 4800:
		aux = B4800;
		break;
	case 14400:
		aux = B14400;
		break;
	case 19200:
		aux = B19200;
		break;
	case 28800:
		aux = B28800;
		break;
	case 38400:
		aux = B38400;
		break;
	case 57600:
		aux = B57600;
		break;
	case 76800:
		aux = B76800;
		break;
	case 115200:
		aux = B115200;
		break;
	case 230400:
		aux = B230400;
		break;
	case 460800:
		aux = B460800;
		break;
	default:
		aux = B9600;
	}
	cfsetospeed(&term, aux);
	cfsetispeed(&term, aux);

	// Setting data size
	term.c_cflag &= ~CSIZE;
	switch (pconf->datasize) {
	case 5:
		term.c_cflag |= CS5;
		break;
	case 6:
		term.c_cflag |= CS6;
		break;
	case 7:
		term.c_cflag |= CS7;
		break;
	case 8:
	default:
		term.c_cflag |= CS8;
	}

	// Setting parity
	switch (pconf->parity) {
	case PARITY_ODD:
		term.c_cflag |= (PARENB | PARODD);
		break;
	case PARITY_EVEN:
		term.c_cflag |= (PARENB);
		term.c_cflag &= ~(PARODD);
		break;
	default:
		term.c_cflag &= ~(PARENB); //none
	}

	// Setting number of stop bits
	if (pconf->stopbits == 2)
		term.c_cflag |= (CSTOPB);
	else
		term.c_cflag &= ~(CSTOPB);

	// Setting flow control
	switch (pconf->flowctrl) {
	case FLOW_HARD:
		term.c_iflag &= ~(IXON | IXOFF);
		term.c_cflag |= (CRTSCTS);
		break;
	case FLOW_SOFT:
		term.c_iflag |= (IXON | IXOFF);
		term.c_cflag &= ~(CRTSCTS);
		break;
	default:
		term.c_iflag &= ~(IXON | IXOFF); //none
		term.c_cflag &= ~(CRTSCTS);
	}

	// Not canonical
	term.c_lflag &= ~(ICANON);

	// Disabling modem signal
	term.c_cflag |= CLOCAL;

	// Enabling modem hang up
	term.c_cflag |= HUPCL;

	// Flush the port
	tcflush(fd, TCIOFLUSH);

	// Setting the port configuration
	if (tcsetattr(fd, TCSANOW, &term) == ERROR) {
		return ERROR;
	}

	return OK;
}


/****************************************/
/**********   Print to ttyS0   **********/
/****************************************/
static void  mpprint(const char *fmt, ...)
{
	va_list   ap;
	uchar     buf[1024];

	if (fd < 0) {
		if ((fd = open("/dev/ttyS0", O_WRONLY, 0)) <= 0) {
			return;
		}
	}

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	write(fd, buf, strlen(buf));
	va_end(ap);
}


/*********************************************/
/**********   Print to a log file   **********/
/*********************************************/
static void  mpfprint(const char *fmt, ...)
{
	va_list   ap;
	uchar     buf[1024];
	uchar     filen[30];

	if (serial_port)
		sprintf(filen, "/var/log/2way-raw-%02d.log", serial_port);
	else
		sprintf(filen, "/var/log/2way-raw.log");

	if (ffd < 0) {
		if ((ffd = open(filen, O_WRONLY | O_CREAT | O_TRUNC, 0)) <= 0) {
			mpprint("error opening log file: %s !!!\r\n", strerror(errno));
			return;
		}
	}

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	write(ffd, buf, strlen(buf));
	va_end(ap);
}


/******************************************************/
/**********   Send data to the Serial Port   **********/
/******************************************************/
static int  ssend(int fds, uchar *buff, int tosend)
{
	int disp = 0, sent;

	do {
		if ((sent = write(fds, (buff + disp), (tosend - disp))) >= 0)
			disp += sent;
		else
			if (errno != EAGAIN)
				break;
		if (disp != tosend)
			usleep(1000);  // 1ms
	} while (disp < tosend);

	if (sent < 0)
		return ERROR;
	else
		return OK;

}


/************************************************************/
/**********   Send data to the Socket Connection   **********/
/************************************************************/
static int  tcpsend()
{
	int disp = 0, sent;

	do {
		if ((sent = send(fdsock, (sbuffp + disp), (txdisp - disp), 0)) >= 0)
			disp += sent; 
		else
			if (errno != EAGAIN)
				break;
		if (disp != txdisp)
			usleep(100);
	} while (disp < txdisp);

	txdisp = 0;

	if (sent < 0)
		return ERROR;
	else
		return OK;

}



/* End of 2way-raw.c */
