/*
 *  Module:  modbusd.c (Modbus/TCP daemon for TS100)
 *
 *  Author:  Marcelo Peccin <marcelo.peccin@cyclades.com>
 *
 *  Copyright:  (c) 2002 Cyclades Corp.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 *
 *  Date:  June 13, 2002
 *
 *  Comments:  
 *
 */

#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <asm/ioctls.h>
#include <sys/time.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include "server.h"
#include "modbus.h"
#include <asm/frame_mode.h>

// Sequence Number table
static ushort  seqno[256];

// Debug mode control: compiler initialized as FALSE
int         debug_mode;
static int  serial_test;
static int  bounced;

// Other control variables & timers
static int    openreq;
static int    sertimeout;
static int    serfd, slen;
static uchar  *sbuff;
static int    nstimeout;
static struct itimerval  stimeout = {{0,0},{0,0}};



/*****************************/
int main(int argc, char **argv)   // Modbus daemon entry point
/*****************************/
{
	modconf           config;
	int               arg, fdserial, exitstatus = PROBLEM;
	int               rec, tosend, disp, sent, fdready;
	sbuffstatus       status2 = EMPTY, status3 = EMPTY;
	int               move2 = 0, move3 = 0;
	ushort            remaining;
	fd_set            inputs, testfds;
	uchar             tcpbuff[MAXTCPFRAME+10];
	uchar             serialbuff1[MAXSERFRAME+10];
	uchar             *serialbuff = serialbuff1;
	uchar             serialbuff2[MAXSERFRAME+10];
	uchar             *ptr2a = serialbuff2;
	uchar             serialbuff3[MAXSERFRAME+10];
	uchar             *ptr3a = serialbuff3;
	uchar             *ptr3b = ptr3a;
	uchar             *serialbuffp;
	struct timeval    timeout;


	for (arg = 1; arg < argc; arg++) {

		if (!strcmp(argv[arg], "-debugmode")) {
			debug_mode = TRUE;
		} else if (!strcmp(argv[arg], "-serialtest")) {
			debug_mode = TRUE;
			serial_test = TRUE;
		} else if (!strcmp(argv[arg], "-retry")) {
			if (++arg == argc) {
				arg = 1000;  //missing parameter
				break;
			}
			else if (!(nstimeout = atoi(argv[arg]))) {
				arg = 1000;  //parameter error
				break;
			}
		} else {
			arg = 1000;  //wrong parameter
			break;
		}
	}
	
	if ((arg == 1000) || (serial_test && nstimeout)) {
		mpfprint("Modbus daemon usage: modbusd [ [-debugmode][-retry <chars>] | -serialtest ]\n");
		debug_mode = FALSE;
		goto closesock;
	}
	else if (debug_mode) {
		mpfprint("Modbus daemon alive: ");
		if (serial_test)
			mpfprint("(serial test");
		else {
			mpfprint("(debug mode");
			if (nstimeout)
				mpfprint(" - retrying after %d chars", nstimeout);
		}
		mpfprint(")...\n");
	}


	// check serial driver buffer size
	if (FRAMEMODE_FRAMELEN < MAXSERFRAME) {
		mpfprint("Modbus daemon: serial buffer size incompatible\n");
		debug_mode = FALSE;
		goto closesock;
	}


    // initialize Shared Memory
	if (shm_init(-1) < 0) {
		goto closesock;
	}
	while (shm_get_set_val(0, &cy_shm->st_cy_ras, 0, CY_SEM_CONFIG) == 0) {
		sleep(1);
	}


	// prepare to catch some signals
	prepare_signals();


	// copy Configuration Information
	config.serial.speed    = cy_shm->line_conf[0].speed; 
	config.serial.datasize = cy_shm->line_conf[0].DataSize;
	config.serial.parity   = cy_shm->line_conf[0].Parity;
	config.serial.stopbits = cy_shm->line_conf[0].StopBits;
	config.serial.flowctrl = cy_shm->line_conf[0].flow;
	config.serial.media	   = cy_shm->line_conf[0].media;

	config.mode        = cy_shm->line_conf[0].modbus_smode;
	config.idletimeout = cy_shm->line_conf[0].idletimeout;


	if (debug_mode)
		mpfprint("\
config.serial.speed    = %d\n\
config.serial.datasize = %d\n\
config.serial.parity   = %d (1,odd,even)\n\
config.serial.stopbits = %d\n\
config.serial.flowctrl = %d (0,Hw,Sw)\n\
config.serial.media    = %d (232=0)\n\
config.mode            = %d (RTU=1)\n\
config.idletimeout     = %d\n",
				 config.serial.speed,
				 config.serial.datasize,
				 config.serial.parity,
				 config.serial.stopbits,
				 config.serial.flowctrl,
                 config.serial.media,
				 config.mode,
				 config.idletimeout);
      

    // prepare adjusted serial timeout (to resend...)
	if (nstimeout)
		nstimeout = nstimeout * 1042 * 9600/config.serial.speed; //adjust to baud rate
	    //in fact the timeout will be exactly <nstimeout> chars only if 8N1

    // open and configure Serial Port
	if ((fdserial = open("/dev/ttyS0", O_RDWR | O_NOCTTY | O_NONBLOCK)) == ERROR) {
		if (debug_mode)
			mpfprint("problems opening serial port\n");
		goto closesock;
	}
	
	if (config_port(fdserial, &config.serial) == ERROR) {
		if (debug_mode)
			mpfprint("problems configuring serial port\n");
		goto closeserial;
	}

	if (set_frame_mode(fdserial, FRAMEMODE_ON, config.mode) == ERROR) {
		if (debug_mode)
			mpfprint("problems setting frame mode on serial port\n");
		goto closeserial;
	}

	if (debug_mode)
		mpfprint("...serial port is ready.\n");


	FD_ZERO(&inputs);
	if (!serial_test)
		FD_SET(INHSOCK, &inputs);
	FD_SET(fdserial, &inputs);


    // Main Loop
	while (1) {

		testfds = inputs;
  		timeout.tv_sec = config.idletimeout * 60;
  		timeout.tv_usec = 0;

		if (timeout.tv_sec) {
			if (((fdready = select(fdserial+1, &testfds, NULL, NULL, &timeout))
				 == ERROR) || !fdready) {
				if (errno == EINTR) { //Interrupted system call (due to resend...)
					FD_ZERO(&testfds);
				}
				else {
					if (debug_mode) {
						if (!fdready)
							mpfprint(">>>>>   Timeout! (%d min)   <<<<<\n", config.idletimeout);
						else
							mpfprint("error in select (%d)\n", errno);
					}
					break;
				}
			}
		} else {
			if (select(fdserial+1, &testfds, NULL, NULL, NULL) == ERROR) {
				if (errno == EINTR) { //Interrupted system call (due to resend...)
					FD_ZERO(&testfds);
				}
				else {
					if (debug_mode) 
						mpfprint("error in select (%d)\n", errno);
					break;
				}
			}
		}

		
		/******************************/
		if (FD_ISSET(INHSOCK, &testfds)) {   // Activity on ethernet port
		/******************************/
			
			if (debug_mode)
				mpfprint("activity on ethernet port\n");

			// check if there is a pending request (abort it)
			if (openreq) {
				openreq = FALSE;
				status2 = status3 = EMPTY;
				serialbuff = serialbuff1;
				ptr2a = serialbuff2;
				ptr3b = ptr3a = serialbuff3;
			}

			// reset serial timeout control, if needed
			if (sertimeout) {
				sertimeout = FALSE;
				stimeout.it_value.tv_usec = 0; //stop timer
				setitimer(ITIMER_REAL, &stimeout, NULL);
			}

			// reading the Header
			if ((rec = recv(INHSOCK, tcpbuff, HOFFSET, 0)) == ERROR || rec != HOFFSET) {
				if (debug_mode) mpfprint("error readind tcp header\n");
				break;
			}
			if (!rec) {
				exitstatus = OK;
				if (debug_mode) mpfprint("connection reset by peer (header)\n");				
				break;
			}

			if ((remaining = ntohs(*(ushort*)(tcpbuff + LENGTHDISP))) > MAXTCPFRAME) {
				exitstatus = ERRORMAX;
				if (debug_mode) mpfprint("tcp frame is too long\n");				
				break;
			}

			// reading the remaining packet
			if ((rec = recv(INHSOCK, tcpbuff + HOFFSET, remaining, 0)) == ERROR) {
				if (debug_mode) mpfprint("error readind tcp frame\n");
				break;
			}				
			if (!rec) {
				exitstatus = OK;
				if (debug_mode) mpfprint("connection reset by peer (frame)\n");				
				break;
			}

			if (debug_mode) {
				int i = 0;
				mpfprint("-------\n%d bytes received from ethernet port:\n", rec + HOFFSET);
				while (i < rec + HOFFSET)
					mpfprint("%02x ", tcpbuff[i++]);
				mpfprint("\n");
			}

			// received packet via TCP is ready to be processed
			if ((tosend = tcp2serial(tcpbuff, serialbuff + 2, rec + HOFFSET, config.mode)) == ERROR) {
				if (debug_mode) mpfprint("error processing tcp frame\n");				
				break;
			}

			if (tosend == ERRHEADER) {
				exitstatus = ERRHEADER;
				if (debug_mode) mpfprint("tcp header isn't consistent\n");				
				break;
			}
			if (tosend == ERRORMIN) {
				exitstatus = ERRORMIN;
				if (debug_mode) mpfprint("tcp frame is too short\n");				
				break;
			}

			// prepare frame size field
serial_send:
			tosend += 2;
			*(ushort*)serialbuff = htons(tosend);

			if (debug_mode && !serial_test) {
				int i = 0;
				mpfprint("%d bytes ready to be sent to the serial port:\n", tosend);
				while (i < tosend)
					mpfprint("%02x ", serialbuff[i++]);
				mpfprint("...\n");
			}

			// send to serial port
			disp = 0;
			do {
				int loop = 0;
				sent = write(fdserial, (serialbuff + disp), (tosend - disp));
				if (sent > 0)
					disp += sent; 
				if (disp != tosend) {
					usleep(100); loop++;
				}
				if (loop == 100) { //10 ms
					if (debug_mode) mpfprint("... ERROR!!!\n");
					break;
				}
			} while (disp < tosend);

			// set control to resend the frame in case of no response
			if (!serial_test && nstimeout) {
				sertimeout = TRUE;
				stimeout.it_value.tv_usec = nstimeout; //once
				setitimer(ITIMER_REAL, &stimeout, NULL);
				serfd = fdserial;
				sbuff = serialbuff;
				slen = tosend;
			}

			if (debug_mode) {
				if (!serial_test)
					mpfprint("... Done!\n");
				else
					mpfprint("... frame sent back!\n");
			}

			if (serial_test && bounced) {
				bounced = FALSE;
				continue;
			}

		}


		/*******************************/
		if (FD_ISSET(fdserial, &testfds)) {   // Activity on serial port
		/*******************************/

			if (debug_mode)
				mpfprint("activity on serial port\n");

			// reset serial timeout control, if needed
			if (sertimeout) {
				sertimeout = FALSE;
				stimeout.it_value.tv_usec = 0; //stop timer
				setitimer(ITIMER_REAL, &stimeout, NULL);
			}

			// reading frame size
			if ((rec = read(fdserial, serialbuff, 2)) == ERROR || rec != 2) {
				if (debug_mode) mpfprint("error readind serial header\n");
				continue;
			}
			if (!rec) {
				if (debug_mode) mpfprint("zero serial header size\n");
				continue;
			}

			remaining = ntohs(*(ushort*)serialbuff);

			// adjust remaining bytes
			remaining -= 2;
			
			if (remaining > MAXSERFRAME) {
				if (debug_mode) mpfprint("serial frame is too long\n");
				continue;
			}

			// reading the frame itself
			if (!serial_test)
				serialbuffp = serialbuff;
			else
				serialbuffp = serialbuff + 2;

            // used to test flow control
			// if (serial_test) usleep(500 * 1000);

			if ((rec = read(fdserial, serialbuffp, remaining)) == ERROR) {
				if (debug_mode) mpfprint("error readind serial frame\n");
				continue;
			}
			if (!rec) {
				if (debug_mode) mpfprint("zero serial frame size\n");
				continue;
			}

			if (debug_mode) {
				int i = 0;
				mpfprint("-------\n%d bytes received from serial port\n", rec);
				while (i < rec)
					mpfprint("%02x ", serialbuffp[i++]);
				mpfprint("\n");
			}

			// received serial frame is ready to be processed
			if (serial_test) { //bounce back received frame
				if (rec < SERIALMIN) {
					if (debug_mode) mpfprint("NOISE ???     not bounced\n");
					continue;
				}
				tosend = rec;
				bounced = TRUE;
				goto serial_send;
			}

try_again:
			if (debug_mode && openreq) {
				int i = 0;
				mpfprint("-------\ntrying this %d bytes\n", rec);
				while (i < rec)
					mpfprint("%02x ", serialbuffp[i++]);
				mpfprint("\n");
			}

			if ((tosend = serial2tcp(serialbuff, tcpbuff, rec, config.mode)) < 0) {
				if (debug_mode)
					switch (tosend) {
					case ERROR:
						mpfprint("error processing serial frame\n");				
						break;
					case ERRORMIN:
						mpfprint("tcp frame is too short\n");
						break;
					case ERRORFCS:
						mpfprint("FCS error !!!\n");
						break;
					}

				if (config.mode == RTU) { //try to correct framing problem (fragmentation)
					/****************************************************************/
					/*   In fact, we try to reassemble the last 3 invalid frames.   */
					/*   If frame A, B, C, D and E, received in sequence, are       */
					/*   invalid frames, we'll try this sequence:                   */
					/*   A, B, AB, C, BC, ABC, D, CD, BCD, E, DE, CDE               */
					/****************************************************************/
					switch(status2) {
					case EMPTY:
						openreq = TRUE;
						status2 = WAIT_1;
						memcpy(serialbuff2, serialbuff, rec);
						ptr2a += rec;
						status3 = WAIT_1;
						memcpy(serialbuff3, serialbuff, rec);
						ptr3a += rec; ptr3b = ptr3a;
						break;
					case WAIT_1:
						status3 = WAIT_2;
						memcpy(ptr3a, serialbuff, rec);
						ptr3b += rec;
						status2 = TEST;
						memcpy(ptr2a, serialbuff, rec);
						serialbuff = serialbuff2;
						move2 = rec;
						rec = (ptr2a - serialbuff2) + move2;
						goto try_again;
					case TEST:
						status2 = WAIT_2;
						memmove(serialbuff2, ptr2a, move2);
						ptr2a = serialbuff2 + move2;
						if (status3 == TEST) {
							serialbuff = serialbuff3;
							rec = (ptr3a - serialbuff3) + move3;
							goto try_again;
						}
						else
							serialbuff = serialbuff1;
						break;
					case WAIT_2:
						if (status3 == TEST) {
							status3 = WAIT_2;
							memmove(serialbuff3, ptr3a, move3);
							ptr3a = serialbuff3 + (ptr3b - ptr3a);
							ptr3b = serialbuff3 + move3;
							serialbuff = serialbuff1;
						}
						else {
							status3 = TEST;
							memcpy(ptr3b, serialbuff, rec);
							move3 = (ptr3b - ptr3a) + rec;
							status2 = TEST;
							memcpy(ptr2a, serialbuff, rec);
							serialbuff = serialbuff2;
							move2 = rec;
							rec = (ptr2a - serialbuff2) + move2;
							goto try_again;
						}
						break;
					}
				}
				continue;
			}

			// check if there were a pending request (reset control)
			if (openreq) {
				openreq = FALSE;
				status2 = status3 = EMPTY;
				serialbuff = serialbuff1;
				ptr2a = serialbuff2;
				ptr3b = ptr3a = serialbuff3;
			}

			if (debug_mode) {
				int i = 0;
				mpfprint("%d bytes ready to be sent to the ethernet port:\n", tosend);
				while (i < tosend)
					mpfprint("%02x ", tcpbuff[i++]);
				mpfprint("...\n");
			}

			// send to ethernet port
			disp = 0;
			do {
				sent = send(INHSOCK, (tcpbuff + disp), (tosend - disp), 0);
				if ( sent >= 0 )
					disp += sent; 
				else
					break;
				usleep(100);
			} while ( disp < tosend);

			if (debug_mode) mpfprint("... Done!\n");			
		}
		
	}


closeserial:	
	if (close(fdserial) == ERROR) exitstatus = ERRCLOSE;

closesock:
	if (close(INHSOCK) == ERROR) exitstatus = ERRCLOSE;
	
	if (debug_mode) mpfprint("..................................... exit status: %d\n", exitstatus);

	exit(exitstatus);
	
} /* main() */



int tcp2serial(uchar *tcpb, uchar *serialb, int len, int mode)
{
	tcpmsg  inmsg;
	int     auxlen;
	uchar  *startlrc;

	if (len < TCPMIN) return ERRORMIN; //minimum size

 	inmsg.header.inv_id   = ntohs(*((ushort*)tcpb)++);
	inmsg.header.proto_id = ntohs(*((ushort*)tcpb)++);
	inmsg.header.length   = ntohs(*((ushort*)tcpb)++);
	startlrc = tcpb;
	inmsg.header.unit_id  = *tcpb++;
	inmsg.pdu.func_code   = *tcpb++;

	// update Sequence Number table
	seqno[inmsg.header.unit_id] = inmsg.header.inv_id;

	if (inmsg.header.proto_id != MODSERVICE) return ERRHEADER;
	if (inmsg.header.length + HOFFSET != len) return ERRHEADER;

	auxlen = inmsg.header.length - 2; //minus unit_id & func_code

	switch (mode) {

	case ASCII:
		ASCIIwrite(*(ushort*)serialb, inmsg.header.unit_id); ((ushort*)serialb)++;
		ASCIIwrite(*(ushort*)serialb, inmsg.pdu.func_code);  ((ushort*)serialb)++;
		
		while (auxlen--) {
			ASCIIwrite(*(ushort*)serialb, *tcpb);
			((ushort*)serialb)++;
			tcpb++;
		}
		
		auxlen = inmsg.header.length; //byte to compute LRC
		build_LRC(startlrc, &auxlen, serialb);
		
		return auxlen;

	case RTU:
		*serialb++ = inmsg.header.unit_id;
		*serialb++ = inmsg.pdu.func_code;
		
		while (auxlen--)
			*serialb++ = *tcpb++;
		
		auxlen = inmsg.header.length;
		build_CRC(serialb - auxlen, &auxlen);
		
		return auxlen;
		
	default:
		return ERROR;
		
	}

}


int serial2tcp(const uchar *serialb, uchar *tcpb, int len, int mode)
{
	int    auxlen, index;
	uchar  *startlrc;


	switch (mode) {

	case ASCII:

		if (len < ASCIIMIN) return ERRORMIN; //minimum size

		len /= 2;         //each 2 ASCIIs lead to 1 byte
		len--;            //minus LRC
		auxlen = len - 2; //minus addr & func

		ASCIIread(*(ushort*)serialb, index);

		*((ushort*)tcpb)++ = htons(seqno[index]); //inv_id

		*((ushort*)tcpb)++ = htons(MODSERVICE); //proto_id

		*((ushort*)tcpb)++ = htons(len); //lenth

		startlrc = tcpb;

		ASCIIread(*(ushort*)serialb, *tcpb); //unit_id
		((ushort*)serialb)++;
		tcpb++;

		ASCIIread(*(ushort*)serialb, *tcpb); //func_code
		((ushort*)serialb)++;
		tcpb++;

		while (auxlen--) {
			ASCIIread(*(ushort*)serialb, *tcpb); //data
			((ushort*)serialb)++;
			tcpb++;
		}

		if (check_LRC(startlrc, &len, serialb)) return ERRORFCS;

		return (len);

	case RTU:

		if (len < RTUMIN) return ERRORMIN; //minimum size

		if (check_CRC(serialb, &len)) return ERRORFCS;

		auxlen = len - 2;

		*((ushort*)tcpb)++ = htons(seqno[*serialb]); //inv_id

		*((ushort*)tcpb)++ = htons(MODSERVICE); //proto_id

		*((ushort*)tcpb)++ = htons(len); //lenth

		*tcpb++ = *serialb++; //unit_id

		*tcpb++ = *serialb++; //func_code

		while (auxlen--)
			*tcpb++ = *serialb++; //data

		return (len + HOFFSET);

	default:
		return ERROR;

	}
	
}


/******************************************************/
/**********     Resend last serial frame     **********/
/******************************************************/
void resend(void)
{
	int disp = 0, loop = 0, sent;

	do {
		sent = write(serfd, (sbuff + disp), (slen - disp));
		if (sent > 0)
			disp += sent;
		else
			break;  //may be due to reception pending...
		if (disp != slen) {
			usleep(100); loop++;
		}
		if (loop == 100) { //10 ms
			if (debug_mode) mpfprint("... ERROR!!!\n");
			break;
		}
	} while (disp < slen);

	// set control to resend the frame in case of no response
	stimeout.it_value.tv_usec = nstimeout; //once
	setitimer(ITIMER_REAL, &stimeout, NULL);
	
	if (debug_mode)
		mpfprint("-------\n... frame resent to serial port!\n");
	
}


/* End of modbusd.c */
