/*************************************************************
ts_menu.c: This program provides a clear and easy way to 
connect to the remote devices from the serial ports. It 
creates a menu with a list of the tty configured. If the unit
is a Master unit, the first menu shown will be a list of the
Master itself and it slaves. From there you can choose to 
access a local port or a remote port (through one of the 
slaves). 

Quick run thru of the structure of this program:
In main, there is a big while loop that first determines
whether this unit is configured for clustering. If so, then 
the main menu (master TS and slaves) is shown. If not, then 
only the local ports for the TS is shown. If users choose
an incorrect port or decides to go backwards in the menu, or
if the connection closes, certain flags are set so that the
configurations can be reread and the appropriate menus will 
be shown again. 
Jennifer Chen	--	Sept 2002
*************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <termios.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/types.h>
#include <linux/ipc.h>
#include <linux/shm.h>
#include <linux/sem.h>

#include "server.h"

#define MAX_CMD 		128
#define MAX_OPT			50 
#define MAX_SLAVE		MAXLINES  //includes MasterHost
#define CHAR_CTRL_C     3
#define CHAR_CR         '\r'
#define CHAR_BS         '\b'


struct MENU_TS {
	char cmd[MAX_CMD];
	char menu_opt[MAX_OPT];
	char menu_old[MAX_OPT];
	char user[MAX_OPT];
};

/* This structure keeps a table of the master and the slaves */
struct SLAVE {
	char slave[MAX_OPT];		// slave IP
	char user[MAX_OPT];			// default or last user for a specific slave
};

struct termios raw_term = {0}, std_term;
static struct MENU_TS menu_ts[MAXLINES]={0};
static struct SLAVE slaveIP[MAX_SLAVE]={0};  

static int list_opt = 0, optTemp = 0, portUser = 0;	// user chosen options
static int max_opt = 0;			// Max options that users have for current menu. 
							    // Changes depending on what menu is next.
static int max_slave = 0;		// Max slaves in slave structure including the MasterHost.
								// Should stay constant after creating slave structure.
static int virtualFlg = 0;		// 1 = there are slaves configured. 
static int virtualFlgOff = 0;	// Control on/off of virtualFlg, 1 = virtualFlg off - ignore slaves configured
static int getSlavePorts = 0;	// 0 = do not get slave ports, 1 = get slave ports
static int virExist = 0; 		// One time flag telling that this unit has slaves configured  
static int skipNonVirtuals = 0; // 1 = skip any slaves configured. 

static char *escape_char = NULL;

static char user[MAX_OPT] = {0};
char aux[200], temp[200];

static char *ip_to_asc(char *p, int ipno)
{
	p += sprintf(p, "%d", (ipno >> 24) & 0xff);
	p += sprintf(p, ".%d", (ipno >> 16) & 0xff);
	p += sprintf(p, ".%d", (ipno >> 8) & 0xff);
	p += sprintf(p, ".%d", ipno & 0xff);
	return(p);
}

/* Wait for file descriptor to change status.
   The rfds descriptor is watched to see if
   characters become available to be read in. */
int GetKeyPressed (int flag_jmp)
{
        int ch;
        fd_set rfds;

        FD_ZERO (&rfds);
        FD_SET (0, &rfds);
        select (1, &rfds, NULL, NULL, NULL);

        if ((ch = getchar()) == CHAR_CTRL_C) {
                exit(-1);       // ctrl-C
        }

        return(ch);
}

/* Main reading of configuration to determine if unit
   is to be used for clustering. Basically, are there
   slaves configured? If so, save slaves into slave 
   structure. */
static int read_config_main(int opt)
{
	char *p, *slave_portIP=NULL;
	char slave_portIPTemp[20];
	int i, x, IPlength;
	int save = 0, masterFlag = 0;
	unsigned long addr_num;
	struct line_cfg *pconf;
	struct MENU_TS *pm;
	struct hostent *hostName;
	

	/* Clear slave structure */
	memset(&slaveIP, 0, sizeof(slaveIP));

	for (max_opt=0, i=0, pconf=&lineconf[0]; i < MAXVLINES; i++, pconf++) {

		if (pconf->this != i) {
			continue;
		}

		if (!pconf->tty || !*pconf->tty) {
			continue;
		}
	
		if (pconf->protocol != P_SOCKET_SSH &&
			pconf->protocol != P_SOCKET_SERVER &&
			pconf->protocol != P_SOCKET_SERVER_RAW &&
			//[RK]June/03/04 - SSH/Telnet to IPDU
			pconf->protocol != P_IPDU) {
			continue;
		}

		if (pconf->virtual) {
			if(virExist == 0)
				virExist = 1;	// Set one time flag to tell us that this unit has slaves configured 
		}

		/* If this unit have slaves configured, reset the max_opt so 
		   that menu_ts will take in the master and the slaves */
		if(virExist==1) {
			max_opt = 0; 		
			virtualFlg = 1;
			virExist=2;			
		}

		pm = &menu_ts[max_opt];


		if (!opt) {
			p = pm->menu_opt;
		} else {
			p = pm->menu_old;
		}

		if (pconf->virtual) {
			/* Get MasterHost just once */ 
			if(!masterFlag) {
				p = ip_to_asc(p, mainconf.ipno);
				strncpy(slaveIP[max_opt].slave, pm->menu_opt, sizeof(slaveIP[max_opt].slave));  
				addr_num = inet_addr(pm->menu_opt);
				hostName = gethostbyaddr((char *) &addr_num, sizeof(addr_num), AF_INET);
				if (hostName != NULL) {
					strncpy(pm->menu_opt, hostName->h_name, sizeof(pm->menu_opt));
				}
				masterFlag = 1;
				max_opt++;				// Increment current max_opt
				pm = &menu_ts[max_opt];
				p = pm->menu_opt;
			}
			
			/* Parse current tty to obtain IP */
			strncpy(slave_portIPTemp, pconf->tty, MAX_OPT);
			if (strchr(slave_portIPTemp, ':') != NULL) {
				slave_portIP = strtok(slave_portIPTemp, ":");
			} else {
				slave_portIP= slave_portIPTemp;
			}
			IPlength = strlen(slave_portIP);

			/* Compare current slave IP to the ones in the slave structure */
			for (x=0; x < max_opt; x++) {
				if (memcmp(slave_portIP, slaveIP[x].slave, IPlength ) == 0) { 
					save = 0;
					break;
				} else {
					save = 1;	// Not in current slave structure so must save into it
				}
			} 	//end of for loop

			if (save && max_opt >= MAX_SLAVE) {
				// Most likely, users will not exceed the total number of slaves allowed (64)
				save = 0;
				printf("\nCaution: You have exceeded the number of slaves allowed.\n");	
				sleep(2);
			}	
			if (save) { 
				/* Save current slave IP to slave structure */	
				strncpy(slaveIP[max_opt].slave, slave_portIP, sizeof(slaveIP[max_opt].slave));
				p += sprintf(p, slave_portIP);
				addr_num = inet_addr(slaveIP[max_opt].slave);
				hostName = gethostbyaddr((char *) &addr_num, sizeof(addr_num), AF_INET);
				if (hostName != NULL) {
					strncpy(pm->menu_opt, hostName->h_name, sizeof(pm->menu_opt));
				}
				max_opt++; 
			} 
		} 


		if (!opt) {
			continue;
		}

	} //end of big for loop

	if (!opt || opt > max_opt) {
		return(MAXVLINES);
	}
	pm = &menu_ts[--opt];

	if (strcmp(pm->menu_opt, pm->menu_old)) {
		return(MAXVLINES);
	}

	return(opt);
}

/* This read_config does the reading of either the 
   local ports or the ports related to the slaves */
static int read_config(int opt)
{
	char *p, *slave_portIP=NULL, *slave_portNum=NULL;
	char slave_portIPTemp[20];
	int i, IPlength, use_socketPort;
	int save = 0;
	struct line_cfg *pconf;
	struct MENU_TS *pm;
	virtualFlg = 0;
	
	for (max_opt=0, i=0, pconf=&lineconf[0]; i < MAXVLINES; i++, pconf++) {

		use_socketPort=0;

		if (pconf->this != i) {
			continue;
		}

		if (!pconf->tty || !*pconf->tty) {
			continue;
		}
	
		if (pconf->protocol != P_SOCKET_SSH &&
			pconf->protocol != P_SOCKET_SERVER &&
			pconf->protocol != P_SOCKET_SERVER_RAW &&
			pconf->protocol != P_IPDU) {
			continue;
		}

		if(skipNonVirtuals) {
			if(!pconf->virtual) {
				continue;
			}	
		}

		if (pconf->virtual) {
			virtualFlg = 1;		
			if (virtualFlgOff == 1) {
				virtualFlg = 0;
				continue;
			}
		}

		pm = &menu_ts[max_opt];


		if (!opt) {
			p = pm->menu_opt;
		} else {
			p = pm->menu_old;
		}
		switch(list_opt) {
		case 1:		// port
			if (!pconf->socket_port || pconf->virtual) {
				goto xdefault;
			} else {
				//p = ip_to_asc(p, mainconf.ipno);
				p += sprintf(p, " %u", pconf->socket_port);
				max_opt++;
			}
			break;
		case 2:		// ip local
			if (!pconf->ipno || pconf->virtual) {
				goto xdefault;
			} else {
				p = ip_to_asc(p, pconf->ipno);
				max_opt++;
			}
			break;
		default:
xdefault:
			if (!pconf->serverfarm || !*pconf->serverfarm) {
				if (pconf->virtual) {
					/* Parse current tty to obtain IP */
					strncpy(slave_portIPTemp, pconf->tty, MAX_OPT);
					if (strchr(slave_portIPTemp, ':') != NULL) {
						slave_portIP = strtok(slave_portIPTemp, ":");
						slave_portNum = strtok(NULL, ":");
					} else {
						use_socketPort = 1;
						//strcpy(slave_portIP, slave_portIPTemp);
						slave_portIP=slave_portIPTemp;
					}
					IPlength = strlen(slave_portIP);

					/* Compare if current slave is the same as the one user chose */
					if (memcmp(slave_portIP, slaveIP[optTemp-1].slave, IPlength ) == 0) { 
						save = 1;     
					} else { 
						save = 0;
						break;
					}

					if (save && max_opt >= MAXLINES) {
						printf("\nCaution: You have exceeded the number of ports allowed.\n");	
						sleep(2);
					}	
					if (save) { 
						if (use_socketPort)
							p += sprintf(p, "%d", pconf->socket_port);	
						else
							p += sprintf(p, "%d -> %s", pconf->socket_port, slave_portNum);     
					}
					max_opt++; 

				} else { 		//If current line in pslave.conf is not virtual
						strncpy(p, pconf->tty, MAX_OPT);
						max_opt++;
				}

			} else {	//If there is a serverfarm for current line
				if (pconf->virtual) {
					strncpy(slave_portIPTemp, pconf->tty, MAX_OPT);
					if (strchr(slave_portIPTemp, ':') != NULL) {
						slave_portIP = strtok(slave_portIPTemp, ":");
						slave_portNum = strtok(NULL, ":");
					} else {
						use_socketPort = 1;
						slave_portIP = slave_portIPTemp;
					}
					IPlength = strlen(slave_portIP);
                    if (memcmp(slave_portIP, slaveIP[optTemp-1].slave, IPlength ) == 0) {
                        save = 1;
                    } else { 
                    	save = 0;
                        break;
                    }

					if (save) {
						if (list_opt==1 || list_opt==2) {
							if (use_socketPort)
								p += sprintf(p, "%d", pconf->socket_port);	
							else
								p += sprintf(p, "%d -> %s", pconf->socket_port, slave_portNum);     
						} else {
							strncpy(p, pconf->serverfarm, MAX_OPT);
						}
                    }
                    max_opt++;
                } else { 	//If not virtual
						strncpy(p, pconf->serverfarm, MAX_OPT);  
						max_opt++;
                }	
			}
			break;
		}

		if (!opt) {
			continue;
		}

	/* Below begins the forming of the telnet/ssh commands */
		p = pm->cmd;

		if(pconf->virtual) {
			p += sprintf(p, "telnet ");
			if (escape_char) {
				p += sprintf(p, " -e %s", escape_char);
			}
			if (*user) {
               p += sprintf(p, " -l %s ", user);
         }
			p += sprintf(p, " %s", slave_portIP);
            if (use_socketPort)
                p += sprintf(p, " %u", pconf->socket_port);
            else
                p += sprintf(p, " %s", slave_portNum);

		} else {
 	   //[RK]June/03/04 - SSH/Telnet to IPDU
 	   	switch (pconf->protocol) {
 			case  P_IPDU :
 			p += sprintf(p,"pmCommand %d",pconf->this+1);
 			break;
 
 			case P_SOCKET_SSH : 
				p += sprintf(p, "ssh -t");
				if (escape_char) {
					p += sprintf(p, " -e %s", escape_char);
				}
				if (pconf->authtype != AUTH_NONE ||
					pconf->ipno == 0) {
					p += sprintf(p, " -l "); 
				}

				if (*user && pconf->authtype != AUTH_NONE) {
					p += sprintf(p, "%s", user);
				}

				if (pconf->ipno) {
					*p++ = ' ';
					p = ip_to_asc(p, pconf->ipno);
				} else {
					p += sprintf(p, ":%s ", pconf->tty);
					p = ip_to_asc(p, mainconf.ipno);
				}

 			break;
 
 			default :
				p += sprintf(p, "telnet ");
				if (escape_char) {
					p += sprintf(p, " -e %s", escape_char);
				}
				if (*user) {
					p += sprintf(p, " -l %s ", user);
				}
				if (pconf->ipno) {
					p = ip_to_asc(p, pconf->ipno);
				} else {
					p = ip_to_asc(p, mainconf.ipno);
					p += sprintf(p, " %u", pconf->socket_port);
				}
			}
		}
	}

	if (!opt || opt > max_opt) {
		return(MAXVLINES);
	}
	pm = &menu_ts[--opt];

	if (strcmp(pm->menu_opt, pm->menu_old)) {
		return(MAXVLINES);
	}

	return(opt);

}

static void show_menu()
{
	int max_str = 0;
	int i, j, max_lin;
	struct MENU_TS *pm;

	system("clear");

	for (i=0, pm=&menu_ts[0]; i < max_opt; i++, pm++) {
		j = strlen(pm->menu_opt);
		if (j > max_str) {
			max_str = j;
		}
	}

	if(virtualFlg && !getSlavePorts) { 
		if (max_str < 15) max_str = 30;
	} else {
		if (max_str < 15) max_str = 15;
	}

	max_lin = 80 / (max_str + 4);

	if (max_lin == 0) {
		max_lin = 1;
		max_str = 74;
	}


	if (virtualFlg==1) {
		if(!getSlavePorts) {
			printf(" Master and Slaves Console Server Connection Menu\n\n");
		} else {
			printf(" Slave Port Serial Console Server Connection Menu for Slave %s\n\n", slaveIP[optTemp-1].slave); 
		}
		for( i=0, pm=&menu_ts[0]; i<max_opt; i++, pm++){
			if ((i % max_lin) == 0) {
				printf("\n");
			} else {
				printf(" ");
			} 
			printf("%2.0d %*.*s", i+1, -max_str, max_str, pm->menu_opt);  
		}
		if(!getSlavePorts) {
			printf("\n\n Type 'q' to quit, a valid option[1-%d], or anything else to refresh : ", max_opt);
			max_slave = max_opt;
		}else{
		 	printf("\n\n Type 'q' to quit, 'b' to return to previous menu, a valid port[1-%d],\n" 
			" or anything else to refresh : ", max_opt); 
		}
	}
	else {
		printf("\n Serial Console Server Connection Menu for your Master Terminal Server\n\n");
		for (i=0, pm=&menu_ts[0]; i < max_opt; i++, pm++) {
			if ((i % max_lin) == 0) {
				printf("\n");
			} else {
				printf(" ");
			} 

			printf("%2.0d %*.*s", i+1, -max_str, max_str, pm->menu_opt);
		}
		if(max_slave!=0) {
			printf("\n\n Type 'q' to quit, 'b' to return to previous menu, a valid port[1-%d],\n"
		           " or anything else to refresh : ", max_opt); 	
		} else {
			printf("\n\n Type 'q' to quit, a valid option[1-%d], or anything else to refresh : ", max_opt);
		}
	}


}

static void show_portUser_menu(int opt)
{
        printf("\n");
		if(max_slave!=0) {
        	printf(" The current login for this port is: %s\n", slaveIP[optTemp-1].user);
		}
		else {
        	printf(" The current login for this port is: %s\n", user);
		}
        printf("\r\n Press Enter to connect to port %d with the current login\n", opt);
        printf("\r\n or enter another username : ");

}


static int get_option(void)
{
	char *p, *z, tmp[128];
	int cnt;

	while(1) {
		if ((p = z = gets(tmp)) == NULL) continue;
		cnt = 0;

		while(*z) {
			if (*z != 0x08) {
				cnt++;
				*p++ = *z++;
			} else {
				z++;
				if (cnt) {
					--cnt;
					--p;
				}
			}
		}
		*p = 0; p = tmp;


		if (!(cnt)) {
			return(0);
		}
		if (cnt == 1 && (*p == 'q' || *p == 'Q')) {
			break;
		}
		if (cnt == 1 && (*p == 'b' || *p == 'B')) {
			return(-1);
		}
		
		for (cnt=0; *p; p++) {
			if (*p <'0' || *p > '9') {
				return(0);
			}
			cnt = cnt * 10 + *p - '0';
		}
		return(cnt);
	}
	return(-2);
}

int get_param(char *f, int n, int *b)
{
        char    *r = temp;
        int     ret = 0, count = 0;

        fflush(stdout);

        while((*r = GetKeyPressed(0)) != CHAR_CR) {
                if (*r != CHAR_BS) {
                        if (*r >= 32 && *r < 127) {
                                printf ("%c", *r);
                        } else {
                                printf ("?");
                        }
                        fflush(stdout);
                        count++;
                        r++;
                } else {
                        if (count) {
                                printf ("\b \b");
                                fflush(stdout);
                                --count;
                                --r;
                        }
                }
        }

        if (count == 0) {
                return(0);
        }

        *r = 0;
        r = temp;

        if (n == 2) {
                strcpy(&b[0], r);
                ret = n;
        }

        if (ret != n) {
                return(0);
        }
	return(ret);
}

/* Set back to standard terminal upon exit */
void exit_rot(void)
{
        tcsetattr(0, TCSANOW, &std_term);
        printf("\r\n");
}

int main(int argc, char *argv[])
{
	extern char *optarg;
	extern int optind;
	int opt, firstTimeFlag = 0, i, j;
	int optPort = 0, showMainMenu=0;

	/* Parse command-line arguments. */
	while ((opt = getopt(argc, argv, "e:ipUu:h")) != EOF) {
		switch (opt) {
		case 'i':
			list_opt = 2;
			break;
		case 'p':
			list_opt = 1;
			break;
		case 'u':
			if (!optarg) {
				goto error;
			}
			strncpy(user, optarg, sizeof(user));
			break;
		case 'U':
             portUser = 1;
             firstTimeFlag = 1;
             break;
        case 'e':
            if (optarg[0] == '^' && optarg[2] == 0 &&
                (u_char) optarg[1] >= 64 &&
                (u_char) optarg[1] < 128)
                escape_char = optarg;
            else if (strlen(optarg) == 1)
                escape_char = optarg;
            else {
                goto error;
            }
            break;

		case 'h':
		default:
error:
			printf("\n\nUSAGE: %s options\n\n"
				"-p           : Display Tcp port\n"
				"-i           : Display Local Ip assigned to the serial port\n"
				"-u <name>    : Username to be used in ssh/telnet command\n"
				"-U           : Always ask for an username\n"
				"-e <[^]char> : Escape char used by telnet or ssh\n\n"
				, argv[0]);
			return(0);
		}
	}

	if (shm_init(-1) < 0) {
		return -1;
	}

	if (!*user) {
		strncpy(user, getenv("USER"), sizeof(user));
	}


	while (1) {
		if(!showMainMenu) {
			read_config_main(0);
			if (virExist == 0) { 		
				read_config(0);		// If no slaves configured, then reread to obtain local ports
			}

			show_menu();

			if ((opt = get_option()) < -1) {
				break;
			}

			if (!opt || opt > max_opt || opt == -1) {	// opt=-1 to go backwards in menu
				virtualFlgOff = 0;						// Reset flags 
				virExist = 0;
				continue;
			}
			
			showMainMenu = 1;
			optTemp = opt;
		}
		
		if (virExist == 2) {			// If there are slaves configured
			if (optTemp == 1) {			// If user chose #1 - this is the masterhost
				getSlavePorts = 0;
				virtualFlgOff = 1;		// Skip if it's a slave, so it will read only
				read_config(0);			// local ports of the masterhost
				show_menu();

				if ((opt = get_option()) < -1) {
					break;
				}

				if (opt == -1) {		// Reset flags when going to previous menu
					virExist=0;			// because configuration will be reread again 
					virtualFlgOff = 0;
					showMainMenu=0;
					continue;
				}
				else if (!opt || opt > max_opt) {
					virtualFlgOff = 0;
					continue;
				}
			}
			else {						// user chose a slave
				skipNonVirtuals = 1;	// When configuration is read, it will only read 
				getSlavePorts = 1;		// info relating to slaves
				read_config(0);
				show_menu();

				if ((opt = get_option()) < -1) {
					break;
				}

				if (opt == -1) {
					skipNonVirtuals = 0;
					getSlavePorts = 0;
					virExist = 0;
					virtualFlgOff = 0;
					showMainMenu = 0;
					continue;
				}
				else if (!opt || opt > max_opt) {
					continue;
				}
			}
			optPort = opt;
		}

 		/* Sets default login to all ports */
        if (portUser == 1 && firstTimeFlag == 1)
        {
        	for(i = 0; i < max_slave; i++)
            {
            	strcpy(slaveIP[i].user, user);
            }
            firstTimeFlag = 0;
        }
		

        if(portUser == 1)
        {
        	show_portUser_menu(opt);
            /* Set up the terminal mode. Change to raw mode. */
			tcgetattr (0, &std_term);
			raw_term.c_cflag = (std_term.c_cflag & CBAUD) | CS8 | CREAD;
			raw_term.c_iflag = IGNPAR | IGNBRK;
			atexit(exit_rot);
			tcsetattr(0, TCSADRAIN, &raw_term);
			j = get_param("%s", 2, (int *)aux);
			if (j > 0) {
				strncpy(user, aux, sizeof(user));
				if(max_slave!=0){
					strncpy(slaveIP[optTemp-1].user, user, sizeof(slaveIP[optTemp-1].user));
				}
			} else {
				if(max_slave!=0)
					 strncpy(user, slaveIP[optTemp-1].user, sizeof(user));    
			}
			exit_rot();
		}		

		/* Read configuration but this time create the command
		   to connect ports */
		opt = read_config(opt);

		if (opt < max_opt) {
			//printf("Cmd2: %s\n\n\n", menu_ts[opt].cmd);
			system(menu_ts[opt].cmd);	
			sleep(3);
		}

		showMainMenu = 0;
		if (optPort != 0){
			showMainMenu = 1;
		}

	} //end of while

	return(0);
}


