/**********************************************************************
wiz.c - Quick overview of how it works. Basically, all the parameters 
are contained in a structure. The basic wiz parameters are contained 
are in basicList and the rest are in pslaveportlist grouped according 
to their commands. The parameters have a starting point and an ending 
point. These points correlates to their position within the structure.
So for instance, authentication parameters have the starting point 
AUTH_START as 0 and AUTH_END as 8. This is because there are 7 authen-
tication parameters and the first one starts at position 0 in the stucture.  
So if you were to add new parameters, add the parameters to pslaveport-
list, and then update the starting and ending points. 

In the main program, a global flag is set according to what command 
you chose. Then executeFunc is called. executeFunc is the loop that 
this program goes through. It calls the actual function that displays
the parameters sequentially, asks if configurations are correct or if
you want to continue to configure specific ports, write configurations
to pslave.conf, activate configurations, save to flash, etc. 

The function that displays the parameters is pretty much all the same.
It'll read pslave.conf, show the current configurations, ask users if 
they want to set to defaults, displays the parameters and read in the 
values. 11-27-02
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <termios.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/types.h>
#include <linux/ipc.h>
#include <linux/shm.h>
#include <linux/sem.h>
#include <asm/cyc_config.h>
#include <asm/flash.h>
#include <asm/prxk.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <linux/if_ether.h>
#include <setjmp.h>
#include <getopt.h>
#include "server.h"
#include "rot_nports.h"
#include "wiz.h"

#define CHAR_ESC		27
#define CHAR_CTRL_C		3
#define CHAR_CR			'\r'
#define CHAR_BS			'\b'

#define SIZE_IPADD			16  
#define MAX_SERIAL_PORTS	64
#define MAX_PSLAVE_PORTS	MAXVLINES 
#define MAX_GROUP			10		
#define MAX_LENGTH			50

/* Types of entries */
#define TYPE_NUMBER		'N' /* Number */
#define TYPE_IPADDR		'A' /* IP address */
#define TYPE_IPMASK		'M' /* IP mask */
#define TYPE_STRING		'S' /* String */
#define TYPE_LIST		'L' /* List of valid entries */
#define TYPE_IPNO_INC	'I'
#define TYPE_PORT_INC	'P'

#define FLAG_READ		1 /* The parameter was read from the file */
#define FLAG_WRITTEN	4 /* The parameter was written in the file */

/* Fixed positions for the Basic List */
#define BASIC_START	0
#define HOSTNAME	0
#define DHCP		1
#define CSIP		2 
#define DOMAIN		3 //includes domain name and dns server
#define DNS			4 //includes domain name and dns server
#define GATEWAY		5
#define NTMASK		6
#define BASIC_END	7

/* Fixed positions for Pslaveport List */
#define AUTH_START		0	
#define AUTH_END		8	
#define AC_TS_START		8
#define AC_TS_END		12
#define AC_CAS_START	12 
#ifdef TS1H
	#define AC_CAS_END	22
	#define TL_START	22
	#define TL_END		27
	#define AL_START	27
	#define AL_END		28
	#define TS_START	28
	#define TS_END		31
	#define SSET_START	31
	#define SSET_END	39
	#define DB_START 	39
	#define DB_END		46
	#define SNF_START	46
	#define SNF_END		50
	#define SL_START	50	
	#define SL_END		52	
	#define PM_START	52
	#define PM_END		57
#else
	#define AC_CAS_END	21
	#define TL_START	21
	#define TL_END		26
	#define AL_START	26
	#define AL_END		27
	#define TS_START	27
	#define TS_END		30
	#define SSET_START	30
	#define SSET_END	38
	#define DB_START	38
	#define DB_END		45		
	#define SNF_START	45 
	#define SNF_END		49
	#define SL_START	49
	#define SL_END		51
	#define PM_START	51
	#define PM_END		56
#endif

#define HOSTS_FILE				"/etc/hosts"
#define HOSTNAME_FILE			"/etc/hostname"
#define RESOLV_CONF_FILE		"/etc/resolv.conf"
#define PSLAVE_CONF_FILE		"/etc/portslave/pslave.conf"
#define NTWK_ST_ROUTES_FILE	"/etc/network/st_routes"
#define NTWK_IFCFG_FILE			"/etc/network/ifcfg_eth0"
#define DHCPD_PID_FILE			"/etc/dhcpc/dhcpcd-eth0.pid"

/***************************** Routine Definition ***************************/
typedef unsigned char *(getputroutine) (unsigned char *, unsigned char *, int, int);
/****************************************************************************/

/* List of valid entries */
struct combolist {
	unsigned char **elements;
	unsigned char nelements;
};

/* Configuration parameter structure */
struct cfglist {
	unsigned char *show;		  /* name of the element as shown in the page */		 
	unsigned char *name;		  /* name of the element as shown in the config file */
	unsigned char *def_value;	 /* default value */
	unsigned char *value;		 /* value of the element */
	void *param;				  /* parameters for the element */
	getputroutine *getputrt;	  /* routine which extracts and inserts the values from fvalue */
	unsigned char *fvalue;		/* value in the file, when different */
	unsigned char type;		   /* type of element - see definitions */
	unsigned char flag;		   /* some flags READ/PRINTED/WRITTEN */
	unsigned char index;		  /* index for the element */
	unsigned short size;		  /* size allocated for the element */
	unsigned short fsize;		 /* size allocated for fvalue */
};

/* Structure for keeping track of all the groups - conf.group */
struct GROUP {
	char names[50];
	struct GROUP *next;
};

/*************************************************
 Definitions for basic configuration parameters  
*************************************************/
static struct cfglist basicList[] = {
	{"Hostname", "hostname", "CAS",
	NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 40, 0},
	{"DHCP", "dhcp_client", "2",
	NULL, NULL, NULL, NULL, 
	TYPE_NUMBER, 0, 0, 6, 0},	
	{"System IP", "eth_ip", "192.168.160.10",
	NULL, NULL, NULL, NULL,
	TYPE_IPADDR, 0, 0, SIZE_IPADD, 0},
	{"Domain name", "domain", "#",
	NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 40, 0},
	{"Primary DNS Server", "nameserver", "#",
	NULL, NULL, NULL, NULL,
	TYPE_IPADDR, 0, 0, SIZE_IPADD, 0},
	{"Gateway IP", "route", "eth0",
	NULL, NULL, NULL, NULL,
	TYPE_IPADDR, 0, 0, 40, 0},
	{"Network Mask", "eth_mask", "255.255.255.0",
	NULL, NULL, NULL, NULL,
	TYPE_IPMASK, 0, 0, SIZE_IPADD, 0}
};

/*************************************************************
 List of valid entries - List Type 
 When adding entries to list, make sure the total number of entries 
 are updated in the List Structure below 
*************************************************************/
static unsigned char *authtypelist[] = {"none", "radius", 
#ifndef OEM1
	"ldap", 
#endif
	"TacacsPlus", "local", "radius/local", "TacacsPlus/local", "local/radius", "local/TacacsPlus", "RadiusDownLocal", "TacacsPlusDownLocal"};
static unsigned char *protocollist[] = {
#ifdef TS1H
	"modbus",
#endif
	"login", "rlogin", "telnet", "ssh", "slip", "cslip", "ppp", "tcpclear", "tcplogin", "console", "socket_client", "socket_server", "socket_ssh", "raw_data", "ipdu"};
#ifdef TS1H
	static unsigned char *smodelist[] = {"ascii", "rtu"};
#endif
unsigned char *consspeedlist[] = {
"50", "75", "110", "134", "150", "200", "300", "600", "1200", "1800", 
"2400", "4800", "9600", "14400", "19200", "28800", "38400", "57600", "76800", "115200", "230400", "460800", "921600"}; //[RK]Dec/17/02
static unsigned char *numberlist[]= {"0", "1", "2", "3", "4", "5", "6", "7", "8"};
static unsigned char *paritylist[] = {"none", "odd", "even"};
static unsigned char *flowlist[] = {"none", "hard", "soft"};
static unsigned char *noDBmenulist[] = {"0", "1", "2", "3"};
static unsigned char *onezerolist[] = {"1", "0"};
static unsigned char *snifmodelist[] = {"in", "out", "i/o"};
//static unsigned char *yesnolist[] = {"yes", "no"};
static unsigned char *dbmodelist[] = {"cir", "lin"};
static unsigned char *mulsess_list[] = {"yes", "no", "RW_session", "sniff_session"};

/********************
 List Structure 
********************/
static struct combolist authtypel = {authtypelist, 12};
#ifdef TS1H
static struct combolist protocoll = {protocollist, 16};
static struct combolist smodel = {smodelist, 2};
#else
static struct combolist protocoll = {protocollist, 15};
#endif
static struct combolist consspeedl = {consspeedlist, 22}; //[RK]Dec/19/02 - excluded 921600
static struct combolist datasizel = {numberlist+4, 5};
static struct combolist stopbitl = {numberlist+1, 2};
static struct combolist parityl = {paritylist, 3};
static struct combolist flowl = {flowlist, 3};
static struct combolist noDBmenul = {noDBmenulist, 4};
static struct combolist onezerol = {onezerolist, 2};
static struct combolist snifmodel = {snifmodelist, 3};
static struct combolist facilityl = {numberlist, 8};
static struct combolist mulsesl = {mulsess_list, 4};
static struct combolist dbmodel = {dbmodelist, 2};

/**********************
 Parameter structure 
**********************/
static struct cfglist pslaveportlist[] = {
	// wiz --auth
	{"all", "authtype",
	"none", NULL, &authtypel, NULL, NULL,
	TYPE_LIST, 0, 0, 25, 0},	  
	{"all", "authhost1",
	"192.168.160.3", NULL, NULL, NULL, NULL,
	TYPE_IPADDR, 0, 0, 40, 0},
	{"all", "accthost1",
	"192.168.160.3", NULL, NULL, NULL, NULL,
	TYPE_IPADDR, 0, 0, 40, 0},
	{"all", "authhost2",
	"192.168.160.4", NULL, NULL, NULL, NULL,
	TYPE_IPADDR, 0, 0, 40, 0},
	{"all", "accthost2",
	"192.168.160.4", NULL, NULL, NULL, NULL,
	TYPE_IPADDR, 0, 0, 40, 0},
	{"all", "radtimeout",
	"3", NULL, NULL, NULL, NULL,
	TYPE_NUMBER, 0, 0, 6, 0},
	{"all", "radretries",
	"5", NULL, NULL, NULL, NULL,
	TYPE_NUMBER, 0, 0, 6, 0},
	{"all", "secret",
	"rad-secret", NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 40, 0},

	// wiz --ac ts
	{"all", "protocol",
	"rlogin", NULL, &protocoll, NULL, NULL,
	TYPE_LIST, 0, 0, 14, 0},
	{"all", "socket_port",
	"23", NULL, NULL, NULL, NULL,
	TYPE_PORT_INC, 0, 0, 40, 0},
	{"all", "telnet_client_mode",
	"0", NULL, &onezerol, NULL, NULL,
	TYPE_LIST, 0, 0, 2, 0},
	{"all", "userauto",
	"#", NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 40, 0},

	// wiz --ac cas
	{"s", "serverfarm",
	"#", NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 40, 0},
	{"all", "ipno",
	"0.0.0.0", NULL, NULL, NULL, NULL,
	TYPE_IPNO_INC, 0, 0, 40, 0},
	{"all", "socket_port",
	"7001+", NULL, NULL, NULL, NULL, 
	TYPE_PORT_INC, 0, 0, 40, 0},
	{"all", "protocol",
	"socket_server", NULL, &protocoll, NULL, NULL,
	TYPE_LIST, 0, 0, 14, 0},
#ifdef TS1H
	{"all", "modbus_smode",
	"#", NULL, &smodel, NULL, NULL,
	TYPE_LIST, 0, 0, 6, 0},
#endif
	{"all", "users",
	"#", NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 40, 0},
	{"all", "poll_interval",
	"1000", NULL, NULL, NULL, NULL,
	TYPE_NUMBER, 0, 0, 6, 0},
	{"all", "tx_interval",
	"100", NULL, NULL, NULL, NULL,
	TYPE_NUMBER, 0, 0, 6, 0},
	{"all", "idletimeout",
	"0", NULL, NULL, NULL, NULL,
	TYPE_NUMBER, 0, 0, 6, 0},
	{"conf", "group",
	"#", NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 50, 0},

	// wiz --tl
	{"all", "issue",
	"\\r\\n\\Welcome to terminal server %h port S%p \\n\\\r\n\\r\\n\\", NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 512, 0},
	{"all", "prompt",
	"%h login:", NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 40, 0},
	{"all", "lf_suppress",
	"0", NULL, &onezerol, NULL, NULL,
	TYPE_LIST, 0, 0, 2, 0},
	{"all", "auto_answer_input",
	"#", NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 40, 0},
	{"all", "auto_answer_output",
	"#", NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 40, 0},

	// wiz --al
	{"all", "alarm",
	"0", NULL, NULL, NULL, NULL,
	TYPE_NUMBER, 0, 0, 6, 0},

	// wiz --tso 
	{"all", "host",
	"192.168.160.8", NULL, NULL, NULL, NULL,
	TYPE_IPADDR, 0, 0, SIZE_IPADD, 0},
	{"all", "term",
	"vt100", NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 20, 0},
	{"conf", "locallogins",
	"0", NULL, NULL, NULL, NULL,
	TYPE_NUMBER, 0, 0, 6, 0},

	// wiz --sset
	{"all", "speed",
	"9600", NULL, &consspeedl, NULL, NULL,
	TYPE_LIST, 0, 0, 7, 0},
	{"all", "datasize",
	"8", NULL, &datasizel, NULL, NULL,
	TYPE_LIST, 0, 0, 2, 0},
	{"all", "stopbits",
	"1", NULL, &stopbitl, NULL, NULL,
	TYPE_LIST, 0, 0, 2, 0},
	{"all", "parity",
	"none", NULL, &parityl, NULL, NULL,
	TYPE_LIST, 0, 0, 5, 0},
	{"all", "flow",
	"none", NULL, &flowl, NULL, NULL,
	TYPE_LIST, 0, 0, 5, 0},
	{"all", "dcd",
	"0", NULL, &onezerol, NULL, NULL,
	TYPE_LIST, 0, 0, 2, 0},
	{"all", "DTR_reset",
	"100", NULL, NULL, NULL, NULL,
	TYPE_NUMBER, 0, 0, 6, 0},
	{"all", "sttyCmd",
	"#", NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 40, 0},

	// wiz --db
	{"conf", "nfs_data_buffering",
	"#", NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 50, 0},
	{"all", "data_buffering",
	"0", NULL, NULL, NULL, NULL,
	TYPE_NUMBER, 0, 0, 21, 0},
	{"all", "DB_mode",
	"cir", NULL, &dbmodel, NULL, NULL,
	TYPE_LIST, 0, 0, 10, 0},
	{"all", "dont_show_DBmenu",
	"0", NULL, &noDBmenul, NULL, NULL,
	TYPE_LIST, 0, 0, 2, 0},
	{"all", "DB_timestamp",
	"0", NULL, &onezerol, NULL, NULL,
	TYPE_LIST, 0, 0, 2, 0},
	{"all", "syslog_buffering",
	"0", NULL, NULL, NULL, NULL,
	TYPE_NUMBER, 0, 0, 6, 0},
	{"all", "syslog_sess",
	"0", NULL, &onezerol, NULL, NULL,
	TYPE_LIST, 0, 0, 2, 0},

	// wiz --snf
	{"all", "admin_users",
	"#", NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 40, 0},
	{"all", "sniff_mode",
	"out", NULL, &snifmodel, NULL, NULL,
	TYPE_LIST, 0, 0, 4, 0},
	{"all", "escape_char",
	"^z", NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 6, 0},
	{"all", "multiple_sessions",
	"no", NULL, &mulsesl, NULL, NULL,
	TYPE_LIST, 0, 0, 13, 0},

	// wiz --sl
	{"conf", "facility",
	"7", NULL, &facilityl, NULL, NULL,
	TYPE_LIST, 0, 0, 2, 0},
	{"conf", "DB_facility",
	"0", NULL, &facilityl, NULL, NULL,
	TYPE_LIST, 0, 0, 2, 0},

	// wiz --pm
	{"all", "protocol",
	"ipdu", NULL, &protocoll, NULL, NULL,
	TYPE_LIST, 0, 0, 14, 0},
	{"all", "pmtype",
	"cyclades", NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 40, 0},
	{"all", "pmusers",
	"#", NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 40, 0},
	{"all", "pmoutlet",
	"#", NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 40, 0},
	{"all", "pmkey",
	"^p", NULL, NULL, NULL, NULL,
	TYPE_STRING, 0, 0, 10, 0},
//[RK]Oct/21/04 - removed - not used
//	{"all", "pmNumOfOutlets",
//	"8", NULL, NULL, NULL, NULL,
//	TYPE_NUMBER, 0, 0, 6, 0}
};

/*********************
 Global Variables 
**********************/
struct termios raw_term = {0}, std_term;
struct MachineType mt;
struct GROUP *first, *current, *new; 	//pointers for conf.group link list

char aux[200], tmp[200];
char yes_no [] = {'Y', 'Y', 'N', 'N', 0xff, 0xff};
char save_op [] = {'N', 'N', 'Y', 'Y', 'Q', 'Q', 0xff, 0xff};

/** Flags **/
/*  once - controls reading of parameters either only once 
	check - activated only when checking whether user answers
			are in the valid format or not 
	grpCnt - number of groups created with conf.group 
	dhcpFlg - set if dhcp is activated 
	discard_flag - flag to discard previous port specific parameters */
static int basicOn=0, allOn=0, accessOn=0, authOn=0, tlOn=0, alarmOn=0; 
static int tsOn=0, ssetOn=0, dbOn=0, snfOn=0, slOn=0, pmOn=0;
static int portChosen=0, grpCnt=0, once=1, check=0, dhcpFlg=1, allOnOnce=1;
static int discard_flag=0;
static char *profile=NULL;

/*************************************************************************
  AllocConfig - allocates a buffer which will store all the parameters of
  the list.
**************************************************************************/
static int AllocConfig(struct cfglist *list, int start, int end)
{
	int i, size=0;
	unsigned char *buf;

	/* Defines the size of the buffer */
	for (i=start; i<end; i++) {
		size+=(list+i)->size;
		size+=(list+i)->fsize;
	}
	/* Allocates the buffer and assign the pointers for the values */
	if ((buf = malloc (size)) == NULL) return 0;
	size=0;
	for (i=start; i<end; i++) {
		(list+i)->value = buf+size;
		size+=(list+i)->size;
		(list+i)->flag=0;
		if ((list+i)->fsize) {
			(list+i)->fvalue = buf+size;
			size+=(list+i)->fsize;
		} else {
			(list+i)->fvalue = (list+i)->value;
		}
	}
	return 1;
}

/*************************************************************************
  DeallocConfig - deallocates the buffer allocated in AllocConfig.
**************************************************************************/
static void DeallocConfig(struct cfglist *list)
{
	free(list->value);
}

/************************************************************************
  DefaultConfig - Sets the parameters with the default value
*************************************************************************/
static void DefaultConfig(struct cfglist *list, int start, int end)
{
	int i,size;

	for (i=start; i<end; i++) {
		if ((list+i)->def_value) {
			if (strcmp((list+i)->show, "DHCP")==0) {
				dhcpFlg = 1;
			}
			strncpy((list+i)->value, (list+i)->def_value, (list+i)->size);
			//[RK]Oct/01/04
			if (((list+i)->type == TYPE_PORT_INC) && portChosen) {
				size = strlen((list+i)->value);
				(list+i)->value[size-2] += portChosen - 1;
				(list+i)->value[size-1] = 0x00;
			}
		} else {
			*((list+i)->value) = 0;
		}
	}
}

/**************************************************************************
  CheckConfig - Verifies whether the value is valid; if positive, copies it
  in the list. The check flag is set when users are inputting values. If 
  users entered an invalid configuration, CheckConfig will output error
  message. All configuration type accepts the # sign.  
***************************************************************************/
static int CheckConfig (struct cfglist *list, unsigned char *value)
{
	struct in_addr b;
	unsigned int j, p[6], mask, k;  
	struct combolist *combo;	
	int inc = 0;
	char a[25];
	char *c;

	switch (list->type) {
	case TYPE_STRING:
		strncpy(list->value, value, list->size);
		return 1;

	case TYPE_IPADDR:
		/* Type IP address - tries to convert the string in an IP number */
		if (strcmp(value, "#")==0 && strcmp(list->def_value, "eth0") != 0) {
	   		strcpy (list->value, value);
			return 1;
		}
		if (strcmp(value, "eth0")==0 && strcmp(list->def_value, "eth0") == 0) {
	   		strcpy (list->value, value);
			return 1;
		}
		// Make it so that it is not necessary to configure DNS. So it's ok
		// to have it empty 
		if (strcmp(list->show, "Primary DNS Server")==0 && *value == 0) {
			return 1;
		}
		if (inet_aton(value, &b)!=0) {
	   		strcpy (list->value, inet_ntoa(b));
			return 1;
		} else {
			if(check)
				printf("\r\nError: The current value, %s, is not in dot address notation.\n", value);
		}
		break;

	 case TYPE_IPMASK:
		/* Type IP mask - tries to convert the string in an IP number and
		 * checks whether it is a mask */
		if (strcmp(value, "#") == 0) {
			strcpy (list->value, value);
			return 1;
		}
		if (inet_aton(value, &b)) {
			mask=0;
			for (j=0; j<4; j++) {
				for (k=0x80; k; k/=2) {
					if (b.s_addr == mask) {
						strcpy (list->value, inet_ntoa(b));
						return 1;
					}
					*((unsigned char *)&mask + j) += k;
				}
			}
			if (b.s_addr == mask) {
				strcpy (list->value, inet_ntoa(b));
					return 1;
			}
		} else {
			if(check)
				printf("\r\nError: The current value, %s, is not in dot address notation.\n", value);
		}
		break;

	case TYPE_LIST:
		/* Type list - The value should match one of the valid entries */
		if (strcmp(value, "#")==0) {
			return 1;
		} else {
			combo = list->param;
			for (j=0; j<combo->nelements; j++) {
				if (strcmp(combo->elements[j], value)) { 
					continue;
				}
				strcpy (list->value, combo->elements[j]);
				return 1;
			}
		}
		break;

	case TYPE_NUMBER:
		/* Type number - tries to convert the string in a number and checks
		 the limits if applicable */
		if (sscanf(value,"%d", &p[0])!= 1) {
			if(strcmp(value, "#") == 0)
				return 1;
			if(check)
				printf("\r\nError: The current value, %s, is not a numeric value.\n", value);
			break;
		}
		if (list->param) {
			if (p[0] < *((unsigned int *)(list->param)) ||
				p[0] > *((unsigned int *)(list->param) + 1)) {
					break;
			}
		}
		sprintf(list->value, "%d", p[0]);
		return 1;

	case TYPE_PORT_INC:
		if (strcmp(value, "#") == 0)
			return 1;
		inc = (*(value+strlen(value)-1) == '+')?1:0;
		//If configuring for specific ports, cannot have '+'.
		if (inc == 1 && check == 1 && portChosen != 0) {
			printf("\r\nError: Please remove the '+'.");
			break;
		}
		if(inc) {
			strncpy(a, value, sizeof(a));
			c = strtok(a, "+");
			if (sscanf(c,"%d", &p[0])!= 1) {
				printf("\r\nError: The value before '+' is not a valid format.");
				printf("\r\nPlease check the appropriate format for this parameter.\n");
				break;
			}
			strncpy(list->value, value, list->size);
			return 1;
		}
		if (sscanf(value,"%d", &p[0])!= 1) {
			printf("\r\nError: The current value, %s, is not a numeric value.\n", value);
			break;
		}
		strncpy(list->value, value, list->size);
		return 1;

	case TYPE_IPNO_INC:
		if (strcmp(value, "#") == 0)
			return 1;
		inc = (*(value+strlen(value)-1) == '+')?1:0;
		//If configuring for specific ports, cannot have '+'.
		if (inc == 1 && check == 1 && portChosen != 0) {
			printf("\r\nError: Please remove the '+'.");
			break;
		}
		if(inc) {
			strncpy(a, value, sizeof(a));
			c = strtok(a, "+");
			if (inet_aton(c, &b)==0) {
				printf("\r\nError: The value before '+' is not in dot address notation.\n");
				break;
			}
	   		strcpy (list->value, inet_ntoa(b));
			strcat(list->value, "+");
			return 1;
		}
		if (inet_aton(value, &b)==0) {
			printf("\r\nError: The value before '+' is not in dot address notation.\n");
			break;
		}
	   	strcpy (list->value, inet_ntoa(b));
		return 1;
	}

	return 0;
}

/**************************************************************************
  SeekCheckConfig - Evaluates the values parsed from the file, looks for
  the parameter(s) in the structure, and CheckConfig is called to save 
  the values. Set parameter flag to READ. 
***************************************************************************/
static int SeekCheckConfig (unsigned char *name, unsigned char *value, struct cfglist *list, int start, int end)
{
	int i, j;
	unsigned char order, index;
	unsigned char *back, *pback;
	
	/* There must be a non-empty value */
	if (!(value && *value)) {
		return 0;
	}
	j=0;
	order=0;
	for (i=start; i<end; i++) {
		/* The name must match with the parameter or be NULL */
		if (name && strcmp(name, (list+i)->name)) {
			 continue;
		}
		/* The parameter should not have been read yet */
		if ((list+i)->flag & FLAG_READ) {
			continue;
		}	
		/* index will be the order of use of the parameter in a certain
		 * line. This will be useful when the same value read in the
		 * configuration will be used by more than one parameter (example:
		 * all.ipno in pslave.conf)*/
		index=(list+i)->index&0x0f;
		if (!j) {
			if (index) {
				continue;
			} else {
				/* order will be the order of use of the parameter when it
				 * is called in more than one line in the configuration.
				 * (example: nameserver, in resolv.conf) */
				order=(list+i)->index&0xf0;
			}
		}
		/* seeks only the parameters of that order... */
		if (order!=((list+i)->index&0xf0)) {
			 break;
		}
		j++;
		/* sets FLAG_READ */
		(list+i)->flag |= FLAG_READ;
	/* getputrt is a routine which will use the value read from the
		 * file (stored in fvalue) to get the value of the parameter. */
		if ((list+i)->getputrt) {
			if (!index) {
				strncpy ((list+i)->fvalue, value, (list+i)->fsize);
			}
			back = strdup(value);
			pback = (*((list+i)->getputrt)) (back, NULL, index, 0);
		} else {
			back = NULL;
			pback = value;
		}
		CheckConfig(list+i, pback);
		if (back) {
			free(back);
		}
	}
	return j;
}

/*************************************************************************
  GetRemainingParameter - This routine will send to the file all the
  parameters that weren't saved yet, because they weren't in the
  configuration file before.
**************************************************************************/
static unsigned char *GetRemainingParameter (struct cfglist *list, struct cfglist *endlist)
{
	struct cfglist *vlist;
	unsigned char order;
	if (list->flag & FLAG_WRITTEN) {
		return NULL;
	}
	if (list->index & 0x0f) return NULL;
	order=list->index & 0xf0;
	if (list->getputrt) {
		*(list->fvalue)=0;
		for (vlist=list; vlist<endlist; vlist++) {
			if (strcmp (list->name, vlist->name)) continue;
			if (order!=(vlist->index & 0xf0)) break;
			vlist->flag |= FLAG_WRITTEN;
			(*(vlist->getputrt)) (list->fvalue, vlist->value, vlist->index & 0x0f, list->fsize);
		}
	} else {
		//list->flag |= FLAG_WRITTEN;
	}
	return list->fvalue;
}

/*****************************************************************************
  Routine for parsing/read/write the most common configuration file syntax
  (host.conf, resolv.conf, cyras.conf).
 ============================================================================
  ParseLineComArg - Parses line. The syntax of a configuration line is:
  <parameter name> <parameter value>. You can define the syntax
  by specifying the number of spaces between the parameters/values.	
  The number of space tells the function whether to continue 
  parsing or not.  So in the above configuration, the number of space
  is 1 so the function will parse once to obtain the value. 
******************************************************************************/
static int ParseLineComArg(unsigned char *buf, unsigned char **argument, int numOfSpace)
{
	int i = 0;
	unsigned char *a, *s;
	s=buf;

	if (*s==0 || isspace(*s)) return 0;
	if (*s == '#') return 2;	

	while (numOfSpace != i) {
		while (*s && !isspace(*s)) {
			s++;
		}
		if (*s){
			*s++=0;
		}
		while (isspace(*s)) {
			s++;
		}
		if (*s==0) return 0;
		a=s;
		i++;
	}

	while (*s && !isspace(*s)) {
		s++;
	}
	*s = 0;

	*argument=a;
	return 1;
}

/**************************************************************************
  readComArg - Read basic configuration files (ie. hostname, hosts, etc.)
***************************************************************************/
static int readComArg(unsigned char *filename, struct cfglist *list, int start, int end, int numOfSpace)
{
	unsigned char *a, *buf, *t, *s;
	FILE *arq;
	struct stat st;
	int ret, i;

	i=start;

	/* we will first read the file in a buffer */
	if (stat(filename, &st)) return 0;
	buf = (unsigned char *)malloc(st.st_size+1);
	if (!buf) return 0;
	arq = fopen (filename, "r");
	if (!arq) {
		free(buf);
		return 0;
	}
	fread(buf, 1, st.st_size, arq);
	*(buf+st.st_size) = 0;
	fclose(arq);

	s=buf;

	while (*s) {
		t=s;
		while (*s && *s!='\n') s++;
		if (*s) *s++=0;

		ret = ParseLineComArg(t, &a, numOfSpace);
		if (ret==0) {
			i++;
			continue;
		} else if (ret == 2) {
			CheckConfig(list+i, "#");
			i++;
			continue;
		} else {
			CheckConfig(list+i, a);
			i++;
			continue;
		}
	}
	free(buf);
	return 1;
}

/**************************************************************************
  ParseLinePslave - Parses line. The syntax of a configuration line is:
 	  <word>.<command> <argument>
  Lines which start with # or empty are disregarded; lines which end with \
  will be linked with the next line.
***************************************************************************/
static int ParseLinePslave (unsigned char *buf, unsigned char **word,
							unsigned char **command, unsigned char **argument)
{
	unsigned char *p, *s, *w, *c, *a;

	p = buf;
	if (p[0] == '#' || p[0] == '\n') return 0;
	s = p + strlen(p);
	if (s > p && *(s-1) == '\\') {
		return 2; //keep adding the line
	}
	if (buf[0] == '\n' || buf[0] == 0) {
		return 0;
	}
	/* Parse Line - w=word c=command a=argument */
	s = buf;
	for (w = s; *w; w++);
	while (--w > s && *w == '\n') *w = 0;
	while isspace (*s) s++;
	w = s;
	while (*s && !isspace(*s)) s++;
	if (*w==0) return 0;
	if (*s) *s++=0;
	while (isspace(*s)) s++;
	a = s;
	if ((c = strchr(w,'.')) == NULL || c[1] == 0) return 0;
	*c++ = 0;
	*word=w;
	*command=c;
	*argument=a;
	return 1;
}

/**************************************************************************
  printLinePslave - prints the [word.command  argument] to pslave.conf.
**************************************************************************/ 
static void printLinePslave (FILE *arq, unsigned char *w, unsigned char *c, unsigned char *a, unsigned char flag)
{
	unsigned char *s;
	if (*a && !flag) {
		for (s=a; *s; s++) {
			if (*s=='\r') *s='\\';
		}
		fprintf (arq, "%s.%s\t%s\n", w, c, a);
	}
}

/**********************************************************
  ip_to_asc - change an IP in dot notation to type ascii.
**********************************************************/
static void ip_to_asc(char *p, int ipno)
{
	p += sprintf(p, "%d", (ipno >> 24) & 0xff);
	p += sprintf(p, ".%d", (ipno >> 16) & 0xff);
	p += sprintf(p, ".%d", (ipno >> 8) & 0xff);
	sprintf(p, ".%d", ipno & 0xff);
}

/* Activate dhcp or not. Returns 1 if yes, 0 if no */
static int dhcp(int y[6])
{
	int x[6];

	if (dhcpFlg && y[0] != 'Y')
		x[0] = my_getchar("\r\nDo you want to use dhcp to automatically assign an IP for\r\nyour system? (y/n)", 'y', yes_no, 1);
	else
		x[0] = my_getchar("\r\nDo you want to use dhcp to automatically assign an IP for\r\nyour system? (y/n)", 'n', yes_no, 1);

	if (x[0] == 'Y') 
		dhcpFlg = 1;
	else 
		dhcpFlg = 0;

	return (dhcpFlg);
}

/**********************************************************************************
  getIPpslave - get IP address (conf.eth_ip) from pslave.conf. 
  If conf.eth_ip is found, set dhcpFlg = 0, meaning dhcp is disabled. 
  Dhcp is by default enabled.
***********************************************************************************/
static void getIPpslave(unsigned char *filename, struct cfglist *list)
{
	FILE *arq, *fopen();
	unsigned char *buf, *s, *t, *w, *a, *c;
	struct stat st;	
	int size;
	
	if (stat(filename, &st)) exit(0);
	size=st.st_size+1;
	size*=2;
	buf = (unsigned char *)malloc(size);
	if (buf==NULL) exit(0);
	arq = fopen(filename, "r");
	if (arq==NULL) {
		free(buf);
		printf("\r\nError opening %s", filename);		
		exit(0);
	}
	fread(buf, 1, st.st_size+1, arq);
	*(buf+st.st_size)=0;
	fclose(arq);	

	t = s = buf;   
	while (*s) {
			while (*s && *s!='\n') s++;
			if (*s) *s++=0;
			if ((ParseLinePslave (t, &w, &c, &a) == 1) && (strcmp(w, "conf")==0)) { // word == conf.
				if (strcmp(c,"eth_ip") == 0) {
					strncpy((list+CSIP)->value, a, (list+CSIP)->size);
					dhcpFlg=0;
				}
			}
			t = s;
   }	
	free(buf);
}

/*******************************************************
  ReadHostName - retrieve hostname from /etc/hostname 
*******************************************************/
static void ReadHostName(struct cfglist *list)
{
	FILE *arq;
	unsigned char *s;
	char hostName[40];
	
	if ((arq = fopen(HOSTNAME_FILE, "r+")) == NULL) {
		printf("\r\nError opening file %s", HOSTNAME_FILE);
		exit(0);
	}
	
	fgets(hostName, 40L, arq);
	s=hostName;
	while(*s) {
		while (*s && !isspace(*s)) s++;
		*s = 0;
	}
	SeekCheckConfig(NULL, hostName, list, HOSTNAME, HOSTNAME+1);
	fclose(arq);
}

/************************************************************************
  ReadResolvconfig - retrieve domain name and dns from /etc/resolv.conf
************************************************************************/
static int ReadResolvConfig(struct cfglist *list)
{
	unsigned char *a, *buf, *t, *s, *c;
	FILE *arq;
	struct stat st;

	 /* we will first read the file in a buffer */
	if (stat(RESOLV_CONF_FILE, &st)) return 0;
	buf = (unsigned char *)malloc(st.st_size+1);
	if (!buf) return 0;
	arq = fopen (RESOLV_CONF_FILE, "r");
	if (!arq) {
		free(buf);
		return 0;
	}
	fread(buf, 1, st.st_size, arq);
	*(buf+st.st_size) = 0;
	fclose(arq);

	t = s = buf;
	while (*s) {
		c = t;
		while (*s && *s!='\n') s++;
		if (*s) *s++=0;

		if (t[0] == '#' || t[0] == '\n') continue;
		while (*t && !isspace(*t)) {
			t++;
		}
		if (*t){
			*t++=0;
		}
		if (strcmp(c, "domain") == 0 || strcmp(c, "search") == 0) {
			while (isspace(*t)) {
				t++;
			}
			if (*t==0) return 0;
			a=t;
			CheckConfig((list+DOMAIN), a);
		} else if (strcmp(c, "nameserver")==0) {
			while (isspace(*t)) {
				t++;
			}
			if (*t==0) return 0;
			a=t;
			CheckConfig((list+DNS), a);
		}
		t=s;
	}

	/* If parameters were commented out, values still remain NULL.
	   If that is so, then hardcode it to # sign. */
	if (*(list+DOMAIN)->value == 0)
		CheckConfig((list+DOMAIN), "#");
	if (*(list+DNS)->value == 0)
		CheckConfig((list+DNS), "#");

	return 0;
}

/********************************************
  ReadGateWayConfig - retrieve gateway IP
********************************************/
static void ReadGateWayConfig(struct cfglist *list)
{
	readComArg(NTWK_ST_ROUTES_FILE, list, GATEWAY, GATEWAY+1, 4);
}

/***************************************************************	
  ReadCSipConfig - retrieve unit's IP address from 
  /etc/portslave/pslave.conf.
***************************************************************/
static void ReadCSipConfig(struct cfglist *list)
{
	getIPpslave(PSLAVE_CONF_FILE, list);
}

/*************************************************************
  ReadNtMaskConfig - retrieves the net mask from pslave.conf.
*************************************************************/
static void ReadNtMaskConfig(struct cfglist *list)
{
	FILE *arq, *fopen();
	unsigned char *buf, *s, *t, *w, *a, *c;
	struct stat st;	
	int size, foundMask=0;

	if (stat(PSLAVE_CONF_FILE, &st)) exit(0);
	size=st.st_size+1;
	size*=2;
	buf = (unsigned char *)malloc(size);
	if (buf==NULL) exit(0);
	arq = fopen(PSLAVE_CONF_FILE, "r");
	if (arq==NULL) {
		free(buf);
		printf("\r\nError opening %s", PSLAVE_CONF_FILE);		
		exit(0);
	}
	fread(buf, 1, st.st_size+1, arq);
	*(buf+st.st_size)=0;
	fclose(arq);	

	t = s = buf;   
	while (*s) {
			while (*s && *s!='\n') s++;
			if (*s) *s++=0;
			if ((ParseLinePslave (t, &w, &c, &a) == 1) && (strcmp(w, "conf")==0)) { // word == conf.
				if (strcmp(c,"eth_mask") == 0) {
					strncpy((list+NTMASK)->value, a, (list+NTMASK)->size);
					foundMask=1;
				}
			}
			t = s;
	}	
	if (!foundMask)
		strncpy((list+NTMASK)->value, "#", (list+NTMASK)->size);
	free(buf);
}

/*****************************************************
  SaveHostName - write the hostname to /etc/hostname
*****************************************************/
static void SaveHostName(unsigned char *filename, struct cfglist *list)
{
	FILE *arq, *fopen();

	if((arq= fopen(filename, "w")) == NULL) {
			printf("\r\nError opening %s", filename);
			exit(0);
	} 
	fprintf(arq,"%s\n", (list+HOSTNAME)->value);
	fclose(arq);
}

/*******************************************************
  SaveCSipInIfcfg - Calculate the boardcast addr, and 
  save the unit's IP, netmask, and boardcast addr into
  /etc/network/ifcfg_eth0.
********************************************************/
static void SaveCSipInIfcfg(unsigned char *filename, struct cfglist *list)
{
	FILE *arq, *fopen();
	char broadcast[20];
	int temp;

	/* Calculate broadcast IP number */
	temp = inet_addr((list+CSIP)->value) & inet_addr((list+NTMASK)->value);
	temp |= ~(inet_addr((list+NTMASK)->value));
	ip_to_asc(broadcast, temp);

	arq = fopen (filename, "w");

	fprintf(arq, "/sbin/ifconfig eth0 down\n");
	fprintf(arq, "/sbin/ifconfig eth0 %s netmask %s broadcast %s mtu 1500\n", (list+CSIP)->value, (list+NTMASK)->value, broadcast);
	fprintf(arq, "/sbin/ifconfig eth0 up\n");

	fclose(arq);
}

/**********************************************************
  SaveCSipInPslave - Save the unit's IP into pslave.conf.
**********************************************************/
static void SaveCSipInPslave(unsigned char *filename, struct cfglist *list)
{
	FILE *arq, *fopen();
	unsigned char *buf, *s, *t, *w, *a, *c;
	struct stat st;	
	int size, foundDhcp=0, foundIP=0, foundMask=0,commentD=0,commentIP=0,commentM=0;
	
	if (stat(filename, &st)) exit(0);
	size = st.st_size+1;
	buf = (unsigned char *)malloc(size);
	if (buf==NULL) exit(0);
	arq = fopen(filename, "r");
	if (arq==NULL) {
		free(buf);
		printf("\r\nError opening %s", filename);		
		exit(0);
	}
	fread(buf, 1, st.st_size+1, arq);
	*(buf+st.st_size)=0;
	fclose(arq);	

	arq = fopen(PSLAVE_CONF_FILE, "w");
	if (arq==NULL) {
		printf("\r\nError writing to temporary file.");
		exit(0);
	}		

	t = s = buf;   
	while (*s) {
			while (*s && *s!='\n') s++;
			if (*s) *s++=0;
			switch (ParseLinePslave (t, &w, &c, &a)) {
				case 2:
					*(s-2) = '\r';
					*(s-1) = '\n';
					break;
				case 1: 
					if (strcmp(w, "conf")==0) { // word == conf.
						if (dhcpFlg) {	
							if (strcmp(c, (list+DHCP)->name) == 0) {
								printLinePslave(arq, w, c, a, 0);
								foundDhcp=1;	
							} else if (strcmp(c,"eth_ip") == 0) {
								if (!commentIP) 
									fprintf(arq,"#%s.%s %s\n", w, c, a);
							} else if (strcmp(c,"eth_mask") == 0) {
								if (!commentM) 
								fprintf(arq,"#%s.%s %s\n", w, c, a);
							} else {
								printLinePslave(arq, w, c, a, 0);
							}
						} else {
							if (strcmp(c, (list+DHCP)->name) == 0) {
								if (!commentD) 
									fprintf(arq,"#%s.%s %s\n", w, c, a);
							} else if (strcmp(c,"eth_ip") == 0) {
								printLinePslave(arq, w, c, (list+CSIP)->value, 0);
								foundIP=1;
							} else if (strcmp(c,"eth_mask") == 0) {
								printLinePslave(arq, w, c, (list+NTMASK)->value, 0);
								foundMask=1;
							} else {
								printLinePslave(arq, w, c, a, 0);
							}
						}
					} else {
						printLinePslave(arq, w, c, a, 0);
					}
					t=s;
					break;
				
		 		default:	
	 				if(strncmp(t,"#conf.dhcp_client",17)==0) {
	 					commentD++;
		 				if(commentD>1) {
							t = s;
		 					continue;
						}
		 			}
	 				if(strncmp(t,"#conf.eth_ip",12)==0) {
	 					commentIP++;
						if(commentIP>1){ 
							t = s;
							continue;
						}
					}
	 				if(strncmp(t,"#conf.eth_mask",14)==0) {
	 					commentM++;
						if(commentM>1){ 
							t = s;
							continue;
						}
					}
					fprintf(arq, "%s\n", t);	
					t = s;
					break;
			}
	
   	}	
	if (dhcpFlg==1 && foundDhcp==0) 
		printLinePslave(arq, "conf", (list+DHCP)->name, "2", 0);
	if (dhcpFlg==0 && foundIP==0 && strcmp((list+CSIP)->value, "#") != 0)
		printLinePslave(arq, "conf", "eth_ip", (list+CSIP)->value, 0);
	if (dhcpFlg==0 && foundMask==0 && strcmp((list+NTMASK)->value, "#") != 0)
		printLinePslave(arq, "conf", "eth_mask", (list+NTMASK)->value, 0);

	fclose(arq);
	free(buf);
}

/******************************************************
  SaveCSipInHosts - Write unit's IP out to /etc/hosts
******************************************************/
static void SaveCSipInHosts(unsigned char *filename, struct cfglist *list)
{
	FILE *arq;
	unsigned char *buf, *s, *t, *a;
	struct stat st;	
	int size;
	int flag = 0;
	
	if (stat(filename, &st)) exit(0);
	size = st.st_size+1;
	buf = (unsigned char *)malloc(size);
	if (buf==NULL) exit(0);
	arq = fopen(filename, "r");
	if (arq==NULL) {
		free(buf);
		printf("\r\nError opening %s", filename);
		exit(0);
	}
	fread(buf, 1, st.st_size+1, arq);
	*(buf+st.st_size)=0;

	fclose(arq); 

	arq = fopen (HOSTS_FILE, "w");

	s = buf;   
	while (*s) {
		if (*s == '#') { // comments
			fprintf(arq, "%s\n", s);   
			while (*s && *s!='\n') s++;
			continue;
		}			
		t=s;
		while (*s && *s!='\n') s++;
		if (*s) *s++=0;
		if (ParseLineComArg (t, &a, 1) == 1) { 
				if (!strcmp(a, (list+HOSTNAME)->value)) {
					fprintf(arq, "%s	%s\n", (list+CSIP)->value, (list+HOSTNAME)->value);
					flag = 1;
				} else {
					fprintf(arq, "%s	%s\n", t, a);	
				}
		 }
   	}
	if(!flag) {
		fprintf(arq, "%s   %s\n", (list+CSIP)->value, (list+HOSTNAME)->value); 
	}

	fclose(arq);
	free(buf);
}

/***********************************************************
  SaveResolv - Write out domain, dns to /etc/resolv.conf.
***********************************************************/
static void SaveResolv(unsigned char *filename, struct cfglist *list)
{
	FILE *arq, *fopen();

	if((arq= fopen(filename, "w")) == NULL) {
			printf("\r\nError opening %s", filename);
			exit(0);
	}   
	if (strcmp((list+DOMAIN)->value, "#")==0 || *(list+DOMAIN)->value == 0)
		fprintf(arq,"#%s %s\n", (list+DOMAIN)->name, COMPANY_DOMAIN_NAME);
	else
		fprintf(arq,"%s %s\n", (list+DOMAIN)->name, (list+DOMAIN)->value);
	if (strcmp((list+DNS)->value, "#")==0 || *(list+DNS)->value == 0) 
		fprintf(arq,"#%s %s\n", (list+DNS)->name, "127.0.0.1");
	else
		fprintf(arq,"%s %s\n", (list+DNS)->name, (list+DNS)->value);
	fclose(arq);
}

/************************************************************
  SaveGateWay - Write out gateway to /etc/network/st_routes
*************************************************************/
static void SaveGateWay(unsigned char *filename, struct cfglist *list)
{
	FILE *arq, *fopen();
	struct in_addr b; 

	if((arq= fopen(filename, "w")) == NULL) {
		printf("\r\nError opening %s", filename);
		exit(0);
	} 

	if (inet_aton((list+GATEWAY)->value, &b)!=0) {
		fprintf(arq,"%s %s\n", "route add default gw", inet_ntoa(b));
	} else { //Should only be here if user entered either "#" or "eth0"
		fprintf(arq,"%s\n", "route add default dev eth0");
	}

	fclose(arq);
}

/**********************************************************
  getDhcpPid - Get the process ID of the dhcpd process
**********************************************************/
int getDhcpPid()
{
	int pid;
	FILE *fp;

	fp = fopen( DHCPD_PID_FILE, "r" );
	if ( fp == NULL )
	{
		//printf("Error: Can't read dhcpcd pid file %s\n", DHCPD_PID_FILE);
		return -1;
	}

	if ( fscanf( fp, "%d", &pid ) != 1 )
	{
		printf("Error: Can't read dhcpcd pid from %s\n", DHCPD_PID_FILE);
		pid = -1;
	}
	fclose( fp );

	return pid;
}

/**********************************************************
  addNewGroup - add new groups (conf.group) to link list.
**********************************************************/
static void addNewGroup(char *argument)
{
	new = (struct GROUP *)malloc(sizeof(struct GROUP));

	/* See if this is the first record */
	if (first==(struct GROUP *)NULL)
		first = current = new;
	else {
		current = first; 
		while (current->next != (struct GROUP *)NULL) 	//Find last record
			current = current->next;		
		current->next = new; 							//save the address of new record
		current = new;									//make current record the new one
	}
	
	strncpy(current->names, argument, sizeof(current->names));
	
	current->next = (struct GROUP *)NULL;
}

/*****************************************************************************
  writeGroup - output the link list. Write out the groups into pslave.conf.
*****************************************************************************/
static void writeGroup(FILE *f, char *w, char *c, char *a)
{
	if (strcmp(current->names, "#")==0)
		fprintf(f, "#%s.%s\t%s\n", w, c, a);
	else {
		printLinePslave(f, w, c, current->names, 0);
	}
	current=current->next;	
}

/***************************************************
  listGroup - list the groups from the linked list.
***************************************************/
static void listGroup(struct cfglist *list)
{
	if (first == (struct GROUP *)NULL)
		printf("\r\nThere are no groups currently configured.");
	else {
		current = first;
		do {
  			printf("\r\n%s.%s : %s", (list)->show, (list)->name, current->names); 
		} while ((current = current->next)!=(struct GROUP *)NULL);
	}
}
/******************************************************
  readWritePslave - read or write to pslave.conf.
******************************************************/
static int readWritePslave (struct cfglist *list, int start, int end, int *active, int flag)
{
	FILE *arq, *fopen();
	struct cfglist *pslavelist2;
	unsigned char *buf, *s, *t, *w, *a, *c, *v;
	unsigned char group[8];
	struct stat st;
	int size, i, found, grpCntTmp=0, tmp=0;
	unsigned char *inactive = "#";  

	/* Read in pslave.conf */
	if (stat(PSLAVE_CONF_FILE, &st)) exit(0);
	size = st.st_size+1;
	buf = (unsigned char *)malloc(size);
	if (buf==NULL) exit(0);
	arq = fopen(PSLAVE_CONF_FILE, "r");
	if (arq==NULL) {
		free(buf);
		printf("\r\nError opening %s", PSLAVE_CONF_FILE);
		exit(0);
	}
	fread(buf, 1, st.st_size+1, arq);
	*(buf+st.st_size)=0;
	fclose(arq);

	if(portChosen) {
		sprintf(group, "s%d", portChosen);
	 } else {
		 strcpy(group, "all");
	}

	if (portChosen > 0 && portChosen <=MAX_PSLAVE_PORTS && flag==1) {
		/* pslavelist2: backup of pslaveportlist, where the configuration
		 * for "all" will be stored. This will be used to compare the port
		 * configuration with the "all" configuration, and only the
		 * parameters that are different will be written */
		 pslavelist2=malloc (sizeof(pslaveportlist));
		 if (pslavelist2==NULL) {
			free(buf);
			return(0);
		 }
		 memcpy (pslavelist2, pslaveportlist, sizeof(pslaveportlist));
		 if (!AllocConfig(pslavelist2, start, end)) {
			free(pslavelist2);
			free(buf);
			return(0);
		 }
	}

	if(flag) {	//WRITE only
		arq = fopen(PSLAVE_CONF_FILE, "w");
		if (arq==NULL) {
			printf("\r\nError opening %s", PSLAVE_CONF_FILE);
			exit(0);
		}

		/* Set up the pointers for the link list */
		current = first; 		//pointers for the conf.group list structure
								//set the current pointer to the first				
	} else { 	//READ only
		/* Initialize/clear pointers to the link list when reading */
		first = (struct GROUP *)NULL;	
		current = (struct GROUP *)NULL;	
		new = (struct GROUP *)NULL;	
	
		/* Clear FLAG_READ */
		for(i=start; i<end; i++) {
			if ((list+i)->flag & FLAG_READ) {
				(list+i)->flag &= ~FLAG_READ;
			}
		}	
	}

	t = s = buf;
	while (*s) {
		found = 0;
		while (*s && *s!='\n') s++;
		if (*s) *s++=0;
		switch (ParseLinePslave (t, &w, &c, &a)) {
		case 2:
			*(s-2) = '\r';
			*(s-1) = '\n';
			break;
		case 1:
			/***** Takes care of all. and conf. parameters *****/
			if (!portChosen) {
				if (strcmp(w, "all")==0) { // word == all.
					for (i=start; i<end; i++) {
						if (strcmp((list+i)->name, c) == 0) {
							if(flag) {
								if(strcmp((list+i)->value, inactive)==0) {		//If value entered is #
									fprintf(arq, "#%s.%s\t%s\n", w, c, a);
									found = 1;
								} else {
									printLinePslave(arq, w, c, (list+i)->value, 0);	 //Write out whatever value in list
									found = 1;
								}
							} else {
								SeekCheckConfig(c, a, list, start, end);	  //Flag = 0; Read in argument into list
							}
						}
					}
				} else if (strcmp(w, "conf")==0) {  
					for (i=start; i<end; i++) {
						if (strcmp((list+i)->name, c) == 0) {
							if(flag) {
								if (strcmp(c, "group")==0) {
									writeGroup(arq, w, c, a);
									grpCntTmp++;
									found = 1;
								} else if (strcmp((list+i)->value, inactive)==0) {		//If value entered is #
									fprintf(arq, "#%s.%s\t%s\n", w, c, a);
									found = 1;
								} else {
									printLinePslave(arq, w, c, (list+i)->value, 0);	 //Write out whatever value in list
									found = 1;
								}
							} else {
								if (strcmp(c, "group")==0) {
									addNewGroup(a);	
									SeekCheckConfig(c, a, list, start, end);	//Flag = 0; Read in argument into list
									grpCntTmp++;
								} else {
									SeekCheckConfig(c, a, list, start, end);	  //Flag = 0; Read in argument into list
								}
							}
						}
					}
				} else if (*w == 's')  { //Takes care of discarding previous port specific parameters
                    for (i=start; i<end; i++) {
                        if (strcmp((list+i)->name, c) == 0) {
                            if (flag) {
                                printLinePslave(arq, w, c, a, discard_flag);
                                found = 1;
                            }
                        }
                    }
				} else {	//If word is not "all", then just write it out
					if (flag) {
						printLinePslave(arq, w, c, a, 0);
						found = 1;
					}
				}
				if (flag==1 && found==0) {
					if (*w == 's' && tmp==0) {
               	for (i=start; i<end; i++) {
                  	if (!(list+i)->flag & FLAG_READ) {
                     	if (strcmp((list+i)->value, inactive) !=0 ) {//If param is activated, print new value out
									printLinePslave(arq, (list+i)->show, (list+i)->name, (list+i)->value, 0);
                        }
                     }
                  }
                  tmp=1;
              	}
					printLinePslave(arq, w, c, a, 0);
				}
			} else {  /***** Takes care of port specific parameters *****/
		
				/* Either list or pslavelist2 contains the parameter of the "all" parameters */
				if (strcmp(w, "all")==0) { // word == all and *group==s
					if (flag) { 
						/*Save all. params into pslavelist2 when flag is WRITE*/
						SeekCheckConfig(c, a, pslavelist2, start, end);  
					} else {
						/*Save all. params into list when flag is READ*/
						SeekCheckConfig(c, a, list, start, end);		  
					}
				} //end of w=all, group=s

				/* For Read only. Checks to see if port is activated. */
				if(!flag) {
					if (*w =='s' && strcmp(c, "tty")==0 && strcmp(w, group)==0)
						*active = 1;
				}

				/* Look for s# that matches the port the user chose */
				if (strcmp(w, group) == 0) {
					if(flag) {
						for (i=start; i<end; i++) {
							if (!(pslavelist2+i)->flag & FLAG_READ) {
								strncpy((pslavelist2+i)->value, inactive, (pslavelist2+i)->size);
							}
						}

						/* For s# that matches, now look for the similar commands */
						for(i=start; i<end;i++) {
							if (strcmp((list+i)->name, c) == 0) {
								if(strcmp(a, (list+i)->value) != 0) {
									if(strcmp((list+i)->value, inactive)==0) {
										fprintf(arq, "#%s.%s\t%s\n", group, c, a);
										strncpy( (pslavelist2+i)->value, (list+i)->value,  (list+i)->size);
									} 
								} else { //if argument and list value is the same
									printLinePslave(arq, w, c, a, 0);
									strncpy( (pslavelist2+i)->value, (list+i)->value,  (list+i)->size);
								}
								found = 1;
							}
						} //end of for loop
					} else {  //If READ only, if found command for the S#, then copy the argument into the value
						for(i=start; i<end;i++) {
							if (strcmp((list+i)->name, c) == 0) {
								strncpy((list+i)->value, a, (list+i)->size);
								SeekCheckConfig(c, a, list, start, end);
								//(list+i)->flag | FLAG_READ;
							}
						} //end of for loop
					} // end of read only and s# == user chosen port

					if (flag==1 && found==0) { //Print param current command doesn't match the listcommand
						printLinePslave(arq, w, c, a, 0);
					}

				} else { //Print all other parameters that isn't current word != all or s#
					if (flag) {
						printLinePslave(arq, w, c, a, 0);
					}
				}
			} //end of portChosen != 0
			t=s;
			break;

			default:
				if(flag) {
					fprintf(arq, "%s\n", t);
				}
				t = s;
				break;
		} //end of switch
	}


	/* By now, list contains the original "all" parameters, pslavelist2 "all" configurations
		and also the ones that were already there for the ports. So if there are new port configurations
		that are not the same as the original "all" parameters, print it out */
	if(flag) {
		if(*group == 's') {
			for(i=start; i<end;i++) {
				if((v=GetRemainingParameter(list+i, list+end))== NULL)
					continue;
				if(strcmp(v, (pslavelist2+i)->value)==0 || strcmp(v, inactive)==0) {
					continue;
				}
				printLinePslave(arq, group, (pslavelist2+i)->name, v, 0);
			} //end of for loop
		}
	}

	/* The following takes care of (when flag=0) copying '#' into the 
	   variable if the corresponding command did not exist in pslave.conf.
	   When flag =1 to write, the following writes out the value if it is
	   has changed from '#' to something else. The last part takes care of 
	   writing new groups to pslave.conf. */ 
	if (!portChosen) {
		for (i=start; i<end; i++) {
			if (!(list+i)->flag & FLAG_READ) {
				if(!flag) {
					if (strcmp((list+i)->name, "group")==0) {
						addNewGroup(inactive);
						grpCntTmp++;
					} 
					else {
						strncpy((list+i)->value, inactive, (list+i)->size);
					}
				} 
			}
		} //end of for loop

		/* This takes care of the new groups created */
		if(flag) {
			while(grpCnt > grpCntTmp) {
				if(strcmp(current->names, inactive)!=0) {
					printLinePslave(arq, "conf", "group", current->names, 0);
				}
				current=current->next;
				grpCntTmp++;
			}
		}
		grpCnt = grpCntTmp;
	}

	if(portChosen) {
		int size;
		for (i=start; i<end; i++) {
			//[RK]Oct/01/04
			if ((list+i)->type == TYPE_PORT_INC) { 
			    	size = strlen((list+i)->value);
				if ((list+i)->value[size-1] == '+') {
					(list+i)->value[size-2] += portChosen - 1;
					(list+i)->value[size-1] = 0x00;
				}
			}
			if (!(list+i)->flag & FLAG_READ) {
				if(!flag) {
					strncpy((list+i)->value, inactive, (list+i)->size);
				}
			} 
		}
	}

	if (flag) {
		if(portChosen)
			free(pslavelist2);
		fclose(arq);
		discard_flag = 0;
	}

	free(buf);
	return(1);

}

/**********************************************
  lowerCase - convert string to all lowercase.
**********************************************/
void lowerCase(char *z) {

	while(*z != '\0') {
		*z = tolower(*z);
		++z;
	}
}

/****************************************************************
  exit_rot - Set terminal back to standard terminal upon exit 
****************************************************************/
void exit_rot(void)
{
	tcsetattr(0, TCSANOW, &std_term);
	printf("\r\n");
}

/*****************************************
  setRawMode - set terminal to raw mode.
*****************************************/
void setRawMode(void)
{
	/* Set up the terminal mode. Change to raw mode. */
	tcgetattr (0, &std_term);
	raw_term.c_cflag = (std_term.c_cflag & CBAUD) | CS8 | CREAD;
	raw_term.c_iflag = IGNPAR | IGNBRK;
	atexit(exit_rot);
	tcsetattr(0, TCSADRAIN, &raw_term);
}

/**************************************************************
  GetKeyPressed - obtain key pressed. If CTRL-C, exit program.
**************************************************************/
int GetKeyPressed (int flag_jmp)
{
	int ch;
	fd_set rfds;

	FD_ZERO (&rfds);
	FD_SET (0, &rfds);
	select (1, &rfds, NULL, NULL, NULL);

	if ((ch = getchar()) == CHAR_CTRL_C) {
			exit(-1);	   // ctrl-C
	}

	return(ch);
}

/**********************************************************
  my_getchar - Output the prompt, and get the character.
**********************************************************/
char my_getchar(char *fmt, char cur, char *op, int flag_any)
{
	char *p;
	int cnt;

	while(1) {
		printf("\r\n%s [%c] : ", fmt, cur);
//		fflush(NULL);
		fflush(stdout);
		p = tmp;
		cnt = 0;

		while((*p = GetKeyPressed(flag_any&2)) != CHAR_CR) {
			if (*p != CHAR_BS) {
				if (*p >= 32 && *p < 127) {
					printf ("%c", *p);
				} else {
					printf ("?");
				}
				fflush(stdout);
				cnt++;
				p++;
			} else {
				if (cnt) {
					printf ("\b \b");
					fflush(stdout);
					--cnt;
					--p;
				}
			}
		}
		*p = 0;

		if (!(cnt)) {
			cur &= 0xdf;
			return(cur);
		}
		if (cnt > 1) {
			printf("\r\nERROR : Only one char should be typed...");
			continue;
		}
		if (*tmp >= 'a' && *tmp <= 'z')
			*tmp &= 0xdf;

		for(p = op; *p != 0xff; p+=2) {
			if (*tmp == *p) {
				return(*tmp);
			}
		}
		if (flag_any & 1) {
			return(cur);
		}
		printf ("\r\nERROR : Only [");
		for (p = op+2; *p != 0xff; p+=2)
			if (p == op+2)
				printf("%c", *(p-2));
			else
				printf(", %c", *(p-2));
			printf (" or %c] are accepted!", *(p-2));
	}
}

/*************************************************
  get_option - obtains the option user chose and
  execute appropriately.
*************************************************/
static int get_option(void)
{
	char *p, *z, temp[128];
	int cnt;

	while(1) {
		p = z = gets(temp);
		cnt = 0;

		while(*z) {
			if (*z != 0x08) {
				cnt++;
				*p++ = *z++;
			} else {
				z++;
				if (cnt) {
					--cnt;
					--p;
				}
			}
		}
		*p = 0; p = temp;

		if (!(cnt)) {
			return(0);
		}
		if (cnt == 1 && (*p == 'q' || *p == 'Q')) {
			break;
		} else if (cnt == 1 && (*p == 'c' || *p == 'C')) {
			return(-1);
		}
		for (cnt=0; *p; p++) {
			if (*p <'0' || *p > '9') {
				return(0);
			}
			cnt = cnt * 10 + *p - '0';
		}
		return(cnt);
	} //end of outer while loop
	return(-2);
}

/*******************************************************
  get_param - used in getValue. Takes in user's input.
*******************************************************/
int get_param(char *f, int n, int *p)
{
	char	*r = tmp;
	char	local[3];
	int	 i, ret = 0, cnt = 0;

	fflush(stdout);

	while((*r = GetKeyPressed(0)) != CHAR_CR) {
		if (*r != CHAR_BS) {
			if (*r >= 32 && *r < 127) {
				printf ("%c", *r);
			} else if (*r == CHAR_ESC) {
				return(-1);
			}
			else {
				printf ("?");
			}
			fflush(stdout);
			cnt++;
			r++;
		} else {
			if (cnt) {
				printf ("\b \b");
				fflush(stdout);
				--cnt;
				--r;
			}
		}
	} //end of while loop

	if (cnt == 0) {
		return(0);
	}

	*r = 0;
	r = tmp;

	if (n == 6) { // MAC address
		if (r[2] == ':') { // using format NN:NN:NN:NN:NN:NN
			ret = sscanf(r, f, &p[0], &p[1], &p[2], &p[3], &p[4], &p[5]);
		}
		else { // using format NNNNNNNNNNNN
			for (i = 0; i < n; i++) {
				local[0] = r[i*2];
				local[1] = r[i*2+1];
				local[2] = 0;
				ret = ret + sscanf(local, "%x", &p[i]);
			}
		}
	}
	if (n == 4) { // IP address
		ret = sscanf(r, f, &p[0], &p[1], &p[2], &p[3]);
	}
	if (n == 1) { // string or speed
		ret = sscanf(r, f, &p[0]);
	}
	if (n == 2) {
		strcpy(&p[0], r);
		ret = n;
	}

	if (ret != n) {
		return(0);
	}

	return(ret);
}

/******************************************************
  getValue - get user's input and send to CheckConfig 
  to be checked if it's a valid input.
******************************************************/
static int getValue(struct cfglist *list, int i)
{
	int j, ret=0;

	while(ret==0) {
		if(basicOn)
			printf("\r\n\r\n%s[%s] : ", list->show, list->value);
		else {
			if (portChosen != 0) {
				printf("\r\n\r\ns%d.%s[%s] : ", portChosen, list->name, list->value);
			} else {
				printf("\r\n\r\n%s.%s[%s] : ", list->show, list->name, list->value);
			}
		}
		j = get_param("%s", 2, (int *)aux);
		if (j > 0) {
			check = 1;
			ret = CheckConfig(list, aux);
		} else if (j == 0) {
			ret = CheckConfig(list, list->value);
		} else if (j == -1) {
			ret = CheckConfig(list, list->value);
			if (ret == 1) {
				return(j);
			}
		}
	} //end of while loop 
	check = 0;

	i++;
	return(i);
}

/*****************************************************
  wizardHeader - sets the appropriate wizard Header. 
*****************************************************/
void wizardHeader(int option)
{
	if (option == 0) {
		system("clear");
		printf("***************************************************************"
		"\r\n*********** C O N F I G U R A T I O N   W I Z A R D ***********"
		"\r\n***************************************************************\r\n\r\n");
	} else if (option == 1) {
		system("clear");
		printf("***************************************************************"
		"\r\n*********** C O N F I G U R A T I O N   W I Z A R D ***********"
		"\r\n***************************************************************\r\n\r\n");
		printf("INSTRUCTIONS for using the Wizard:"
		"\r\nYou can:"
		"\r\n   1) Enter the appropriate information for your system"
		"\r\n   and press ENTER. Enter '#' if you want to "
		"\r\n   deactivate that parameter or"
		"\r\n   2) Press ENTER if you are satisfied with the value"
		"\r\n   within the brackets [ ] and want to go on to the "
		"\r\n   next parameter or"
		"\r\n   3) Press ESC if you want to exit.");
		printf("\r\n\r\nNOTE: For some parameters, if there is nothing within"
		"\r\nthe brackets, it will continue to ask for a value."
		"\r\nIn that case, you must enter a valid value or # if you"
		"\r\ndo not wish to configure the value.\r\n\r\n");
		exit_rot();
		printf("Press ENTER to continue...");
		getchar();
		setRawMode();

	} else if (option == 2) {
		system("clear");
		printf("***********************************************************"
		"\r\n********* C O N F I G U R A T I O N   W I Z A R D *********"
		"\r\n***********************************************************\r\n\r\n");
		printf("INSTRUCTIONS for using the Wizard:"
		"\r\nYou can:"
		"\r\n   1) Enter the appropriate information for your system"
		"\r\n   and press ENTER or"
		"\r\n   2) Press ENTER if you are satisfied with the value"
		"\r\n   within the brackets [ ] and want to go on to the "
		"\r\n   next parameter or"
		"\r\n   3) Press ESC if you want to exit.");
		printf("\r\n\r\nNOTE: For some parameters, if there is nothing within"
		"\r\nthe brackets, it will continue to ask for a value."
		"\r\nIn that case, you must enter a valid value or # if you"
		"\r\ndo not wish to configure the value.\r\n\r\n");
		exit_rot();
		printf("Press ENTER to continue...");
		getchar();
		setRawMode();
	}
}

/********************************************
  showConfig - show current configurations.
********************************************/
void showConfig(struct cfglist *list, int start, int end, int portFlag)
{
	int i;
	
	if (portFlag) {
		printf("Here are port %d's configurations: "
		"\r\n""(The ones with the '#' means it's not activated.)\r\n", portChosen);
		for (i=start; i<end; i++) {
			if (*(list+i)->show != 'c') {
				printf("\r\ns%d.%s : %s", portChosen, (list+i)->name, (list+i)->value);
			}
		}
	} else if (basicOn) {
		printf("Current configuration:\r\n");
		for(i=start; i<end; i++) {
			if (!strcmp((list+i)->show, "DHCP")) {
				if (dhcpFlg) { 
					printf("\r\n%s : enabled",  (list+i)->show);
					i++;
					continue;
				} else {
					printf("\r\n%s : disabled",  (list+i)->show);
					continue;
				}
			}
			if (!strcmp((list+i)->show, "Network Mask") && dhcpFlg==1) continue; 
			printf("\r\n%s : %s",  (list+i)->show, (list+i)->value);
		}
	} else {
		printf("Current configuration: "
			   "\r\n""(The ones with the '#' means it's not activated.)\r\n");

		for(i=start; i<end; i++) {
			if(*(list+i)->show != 's') {		   
				if(*(list+i)->show == 'c' && strcmp((list+i)->name, "group")==0) {
					listGroup(list+i);
				} else {
					printf("\r\n%s.%s : %s", (list+i)->show, (list+i)->name, (list+i)->value);
				}
			}
		}
	} 
	printf("\r\n\r\n\r\n");
}

/************************************************************
  formatSpacing - format spacings between parameters shown.
  After two parameters shown, a new screen should be shown.
************************************************************/
static void formatSpacing(int *a) {

	if (!((*a)++%2)) {
		wizardHeader(0);
	}
	else
		printf("\r\n\r\n\r\n");

}

/***********************************************
  basicInfoFunc - takes care of 'wiz' command.
***********************************************/
void basicInfoFunc(struct cfglist *list, int start, int end)
{
	int x[6];
	int i = start;

	if(once) {				//Read only once.
		ReadHostName(list);
		ReadCSipConfig(list);
		ReadResolvConfig(list);
		ReadGateWayConfig(list);
		ReadNtMaskConfig(list);
	}

	wizardHeader(0);	
	showConfig(list, start, end, portChosen);
	x[0] = my_getchar("\r\nSet to defaults? (y/n)", 'n', yes_no, 1);
	if (x[0] == 'Y') {
		DefaultConfig(basicList, start, end); 
	} 

	wizardHeader(0);	
	printf("\r\nHOSTNAME - An alias for your system."
		   "\r\nThis way you can always refer to the system by this name"
		   "\r\nrather than its IP address.");
	i = getValue(list+i, i);
	if (i == -1) {
		return;
	}

	if (!dhcp(x)) {
		i++;
		wizardHeader(0);
		printf("\r\nIP - The IP address of your system"" (on its Ethernet "
			   "\r\ninterface.) This is the address of the system within your" 
			   "\r\nnetwork. See you network administrator to obtain a valid"
			   "\r\nIP address for the system.");
		i = getValue(list+i, i);
		if (i == -1) {
			return;
		}
	} else {
		i+=2;
	}

	wizardHeader(0);
	printf("\r\nDOMAIN NAME - A name that locates or identifies your"
		   "\r\norganization within the Internet.");
	i = getValue(list+i, i);
	if (i == -1) {
		return;
	}

	wizardHeader(0);
	printf("\r\nDOMAIN NAME SERVER - The IP address of the server that"
		   "\r\nresolves domain names. Your domain name is alphabetic so"
		   "\r\nthat it is easier to remember. Everytime you see the domain"
		   "\r\nname, it is actually being translated into an IP address by"
		   "\r\nthe domain name server. See your network administrator to"
		   "\r\nobtain this IP address for the domain name server.");
	i = getValue(list+i, i);
	if (i == -1) {
		return;
	}

	wizardHeader(0);
	printf("\r\nGATEWAY - A node on a network that serves as an entrance"
		   "\r\npoint into another network. See your network administrator"
		   "\r\nto find out your organization's gateway address.");
	i = getValue(list+i, i);
	if (i == -1) {
		return;
	}

	if (!dhcpFlg) { //skip if dhcp is active
		wizardHeader(0);
		printf("\r\nNETMASK - A string of 0's and 1's that mask or screen out"
			   "\r\nthe host part of an IP address so that only the network"
			   "\r\npart of the address remains.");
		i = getValue(list+i, i);
		if (i == -1) {
				return;
		}
	}
} // end of basicInfoFunc

/****************************************************************
  accessFunc - takes care of 'wiz --ac <cas> or <ts>' commmand.
****************************************************************/
static int accessFunc(struct cfglist *list, int start, int end)
{
	int x[6], a = 1, i = 0, j, portFlag = 0;
	char port[4];
	char *ALL = "ALL";
	int active = 0;
	FILE *arq;

	if (once) {		
		readWritePslave(list, start, end, &active, 0);
	}
	if (portChosen != 0) {
		portFlag = 1;
		sprintf(port, "S%d", portChosen);
		if (!active && once) {
			printf("\r\nPort %d is currently not active.", portChosen);
			x[0] = my_getchar("Would you like to activate it? (y/n)", 'n', yes_no, 1);
			if (x[0] == 'N') {
				return 0;
			} else {
				arq = fopen(PSLAVE_CONF_FILE, "a");
				if(!arq) {
					printf("\r\nError opening temporary file.");
					exit(0);
				}
				fprintf (arq, "s%d.tty\tttyS%d\n", portChosen, portChosen);
				fclose(arq);
			}
		}
	}

	if (!allOn) {
		wizardHeader(0);
		showConfig(list, start, end, portFlag);
		x[0] = my_getchar("Set to defaults? (y/n)", 'n', yes_no, 1);
		if (x[0] == 'Y') {
			DefaultConfig(list , start, end);
		} 
	}
	printf("\r\n\r\n");

	i = start;
	wizardHeader(0);
	if (strcmp(profile, "ts")==0) {
		printf("%s.PROTOCOL - Users can access the servers through the"
			   "\r\nserial port using ssh, ssh2, telnet, login, rlogin,"
			   "\r\nor socket_client."
			   "\r\n""(e.g. login -requests username and password, rlogin -"	
			   "\r\nreceives username from the system and requests a password," 
			   "\r\netc.)", portFlag ? port : ALL); 

		i = getValue(list+i, i);
		if (i == -1)
			return 1;
		formatSpacing(&a);

		printf("%s.SOCKET_PORT - This defines the port(s) to be used by"
			   "\r\nthe protocols telnet and socket_client. For these two"
			   "\r\nprotocols a default value of 23 is used when no value"
			   "\r\nis configured.", portFlag ? port : ALL);
		i = getValue(list+i, i);
		if (i == -1)
			return 1;
		formatSpacing(&a);

		printf("%s.TELNET_CLIENT_MODE - This parameter only applies if"
				"\r\nthe current protocol configured is telnet. Configuring as"
				"\r\nbinary"" (1) causes an attempt to negotiate the TELNET"
				"\r\nBINARY option on both input and output with the Telnet"
				"\r\nserver. Thus, it puts the telnet client in binary mode."
				"\r\nThe default is 0 which represents text mode.", portFlag ? port : ALL); 
		i = getValue(list+i, i);
		if (i == -1)
			return 1;
		formatSpacing(&a);

		printf("%s.USERAUTO - Username used when connected to a Unix"
			   "\r\nserver from the user's serial terminal.", portFlag ? port : ALL);
		i = getValue(list+i, i);
		if (i == -1)
			return 1;

		return 1;
	}

	/******END OF ACCESS PARAMS FOR TS PROFILE*****/
	
	if (portFlag) {
		printf("%s.SERVERFARM - Alias for a specific port.", port);
		i = getValue(list+i, i);
		if (i == -1)
			return 1;
		formatSpacing(&a);

	} else {
		i++;
	}
	printf("%s.IPNO - This is the default IP address of the system's "
		   "\r\nserial ports. If configured as 192.168.1.101+, the '+'" 
			"\r\nindicates that the first port should be addressed as "
			"\r\n192.168.1.101 and the following ports should have "
			"\r\nconsecutive values. Any host can access a port using "
			"\r\nits IP address as long as a path to the address exists"
			"\r\nin the host's routing table.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.SOCKET_PORT - This defines an alternative labeling"
			   "\r\nsystem for the system ports. The '+' after the numerical"
			   "\r\nvalue causes the interfaces"" (or ports) to be numbered "
			   "\r\nconsecutively."
			   "\r\n""(e.g. interface 1 of your system is assigned port 7001, "
			   "\r\ninterface 2 has the value 7002, etc.)", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	//For now, all boards besides ts1h won't have modbus as a choice because it hasn't been
	//implemented. 
	if ((mt.board_type == BOARD_TS1H) || (mt.board_type == BOARD_TS110)) {
		printf("%s.PROTOCOL - The possible protocols are telnet,"
			   "\r\nssh1/ssh2, raw data, or modbus."
			   "\r\n""(e.g. socket_server -telnet protocol, socket_ssh -ssh1/ssh2"
			   "\r\nprotocol, raw_data -used to exchange data in transparent"
			   "\r\nmode; similar to socket_server mode but without telnet "
			   "\r\nnegotiation breaks to serial ports, modbus -an application"
			   "\r\nlayer messaging protocol for client/server communication"
			   "\r\nwidely used for industrial automation, etc.)", portFlag ? port : ALL);
	} else {
		printf("%s.PROTOCOL - The possible protocols are telnet,"
				   "\r\nssh1/ssh2, or raw data."
				   "\r\n""(e.g. socket_server -telnet protocol, socket_ssh -ssh1/ssh2"
				   "\r\nprotocol, raw_data -used to exchange data in transparent"
				   "\r\nmode; similar to socket_server mode but without telnet "
				   "\r\nnegotiation breaks to serial ports.)", portFlag ? port : ALL);	  
	}	
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	if ((mt.board_type == BOARD_TS1H) || (mt.board_type == BOARD_TS110)) {
			printf("%s.MODBUS_SMODE - Communication mode through the serial "
				   "\r\nports. This parameter is valid only if the protocol"
					"\r\nconfigured is modbus. If it is and this parameter is "
					"\r\nnot configured, ASCII mode will be assumed."
				   "\r\n""(e.g. ascii -normal TX/RX mode, rtu -Remote Transmission"
				   "\r\nmode where some time contraints are observed between"
				   "\r\ncharacters while transmitting a frame)", portFlag ? port : ALL);

			i = getValue(list+i, i);
			if (i == -1)
				return 1;
			formatSpacing(&a);
	} 

	printf("%s.USERS - Restricts access to ports by user name. Only"
		   "\r\nthe users listed can access the port, or using a '!',"
		   "\r\nall but the users listed can access the port."
		   "\r\nA single comma and spaces/tabs may be used between names."
		   "\r\nA comma may NOT appear between the '!' and the first user"
		   "\r\nname. The users may be local, Radius or TacacsPlus. User "
		   "\r\ngroups"" (defined with the parameter conf.group) can be"
		   "\r\nused in combination with user names in the parameter list."
		   "\r\nNotice that these are common users, not administrators."
		   "\r\n""(e.g. !joe, mark, grp1 -the users, Joe, Mark, and members"
		   "\r\nof grp1, cannot access the port.)", portFlag ? port : ALL);  
		

	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.POLL_INTERVAL - Valid for protocols socket_server and"
		   "\r\nraw_data. When not set to 0, this parameter sets the wait"
		   "\r\nfor a TCP connection keep-alive timer. If no traffic passes"
		   "\r\nthrough the system for this period of time, the system will "
		   "\r\nsend a line status message to the remote device to see if"
		   "\r\nthe connection is still up. If not configured, default is "
		   "\r\n1000ms. If set to 0, line status messages will not be sent "
		   "\r\nto the socket client.", portFlag ? port : ALL);

	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.TX_INTERVAL - Valid for protocols socket_server and"
		   "\r\nraw_data. This parameter defines the delay"" (in milli-"
		   "\r\nseconds) before transmission to the Ethernet of data "
		   "\r\nreceived through a serial port. If not configured, 100ms"
		   "\r\nis assumed. If set to 0 or a value above 1000, no buffering "
		   "\r\nwill take place.", portFlag ? port : ALL);

	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.IDLETIMEOUT - This parameter specifies how long"" (in"
		   "\r\nminutes) a connection can remain inactive before it is cut"
		   "\r\noff. If set to 0"" (the default), the connection will not"
		   "\r\ntime out.", portFlag ? port : ALL);

	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	if (!portFlag) {
		printf("CONF.GROUP - Used to combine users into a group. This"
			   "\r\nsimplifies the parameter, all.users. You can define more"
			   "\r\nthan one group."
			   "\r\n""(e.g. groupName: user1, user2)\r\n");
			if (first==(struct GROUP *)NULL)
				printf("\r\nThere are no groups. Should not be here");
			else {
				current = first;
				do {
					printf("\r\n%s.%s[%s] : ", (list+i)->show, (list+i)->name, current->names);	
					j = get_param("%s", 2, (int *)aux);
					if (j > 0) {
						strncpy(current->names, aux, sizeof(current->names));
					} else if (j == -1) {
						return 1;
					}	
				} while ((current=current->next) != (struct GROUP *)NULL);
			}

L2:
		x[0] = my_getchar("\r\nWould you like to create another group? (y/n)", 'n', yes_no, 1);
		if (x[0] == 'Y') {
			 printf("\r\nPlease follow this format"
				   "\r\n	groupName: user1, user2, ..."
				   "\r\nafter the following prompt to create a new group.");
			printf("\r\n\r\n%s.%s : ", (list+i)->show, (list+i)->name);
			j = get_param("%s", 2, (int *)aux);
			if (j > 0) {
				addNewGroup(aux);
				grpCnt++;
			} else if (j == -1) {
				return 1;
			}
			goto L2;
		}
	} //end of !portFlag if statement

	return 1;
} //end of accessFunc

/****************************************************************
  authFunc - takes care of 'wiz --auth' commmand.
****************************************************************/
static int authFunc(struct cfglist *list, int start, int end)
{
	int x[6], a=1, i=0, portFlag=0;
	int radius=0, active=0;
	char port[4];
	char *ALL = "ALL";
	FILE *arq;

	if(once)
		readWritePslave(list, start, end, &active, 0);
	if (portChosen != 0) {
		portFlag = 1;
		sprintf(port, "S%d", portChosen);
		if (!active && once) {
			printf("\r\nPort %d is currently not active.", portChosen);
			x[0] = my_getchar("Would you like to activate it? (y/n)", 'n', yes_no, 1);
			if (x[0] == 'N') {
				return 0;
			} else {
				arq = fopen(PSLAVE_CONF_FILE, "a");
				if(!arq) {
					printf("\r\nError opening temporary file.");
					exit(0);
				}
				fprintf (arq, "s%d.tty\tttyS%d\n", portChosen, portChosen);
				fclose(arq);
			}
		}
	}

	if (!allOn) {
		wizardHeader(0);
		showConfig(list, start, end, portFlag);
		x[0] = my_getchar("Set to defaults? (y/n)", 'n', yes_no, 1);
		if (x[0] == 'Y') {
			DefaultConfig(list , start, end);
		}  
	} 
	printf("\r\n\r\n");

	i = start;
	wizardHeader(0);
	printf("%s.AUTHTYPE - This parameter controls the authentication"
		   "\r\nrequired by the system. Users' access to the server"
		   "\r\nthrough the serial port is granted through the check of"
		   "\r\nusername and password locally or remotely."
		   "\r\n""(e.g. none, local, TacacsPlus"" (note the capital 'T'"
#ifndef OEM1
	"\r\nin TacacsPlus), radius, ldap, etc.)", portFlag ? port : ALL);
#else 
	"\r\nin TacacsPlus), radius, etc.)", portFlag ? port : ALL);
#endif
   	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);
	
	if (portFlag) {
		--i;
		//Can't just check if it's Radius because signal_ras hup is case sensitive. 
		//Capital Radius only works if used as RadiusDownLocal 
		if (strstr((list+i)->value, "radius") != NULL || strstr((list+i)->value, "RadiusDownLocal") != NULL || strstr((list+i)->value, "TacacsPlus")!= NULL)
		{
			radius = 1;
			i++;
		}
	} else {
		--i;
		if (strstr((list+i)->value, "radius") != NULL || strstr((list+i)->value, "RadiusDownLocal") != NULL || strstr((list+i)->value, "TacacsPlus") != NULL) {
			radius = 1;
			i++;
		}
	}
	if (radius) {
		printf("%s.AUTHHOST1 - This IP address indicates where the"
			   "\r\nRadius or TacacsPlus authentication server is located.", portFlag ? port : ALL);
   		i = getValue(list+i, i);
		if (i == -1)
			return 1;
		formatSpacing(&a);

		printf("%s.ACCTHOST1 - This IP address indicates where the Radius"
			   "\r\nor TacacsPlus accounting server is located. The accounting"
			   "\r\nserver can be used to track how long users are connected"
			   "\r\nafter being authorized by the authentication server.", portFlag ? port : ALL);
		i = getValue(list+i, i);
		if (i == -1)
			return 1;
		formatSpacing(&a);

		printf("%s.AUTHHOST2 - This IP address indicates where the SECOND"
			   "\r\nRadius or TacacsPlus authentication server is located.", portFlag ? port : ALL);
		i = getValue(list+i, i);
		if (i == -1)
			return 1;
		formatSpacing(&a);

		printf("%s.ACCTHOST2 - This IP address indicates where the SECOND"
			   "\r\nRadius or TacacsPlus accounting server is located.", portFlag ? port : ALL);
			i = getValue(list+i, i);
			if (i == -1)
				return 1;
		formatSpacing(&a);

		printf("%s.RADTIMEOUT- This is the timeout (in seconds) for a"
			   "\r\nRadius or TacacsPlus authentication query to be answered.", portFlag ? port : ALL);
		i = getValue(list+i, i);
		if (i == -1)
				return 1;
		formatSpacing(&a);

		printf("%s.RADRETRIES - This defines the number of times each"
			   "\r\nRadius or TacacsPlus server is tried before another is"
	 		   "\r\ncontacted.", portFlag ? port : ALL);
		i = getValue(list+i, i);
		if (i == -1)
			return 1;
		formatSpacing(&a);

		printf("%s.SECRET - This is the shared secret necessary for "
			   "\r\ncommunication between the system and the Radius or "
			   "\r\nTacacsPlus servers.", portFlag ? port : ALL);   
		i = getValue(list+i, i);
	} 

	return 1;

} //end of authFunc		

/****************************************************************
  termFunc - takes care of 'wiz --tl' commmand.
****************************************************************/
static int termFunc(struct cfglist *list, int start, int end)
{
	int x[6], a = 1, i = 0, portFlag = 0;
	char port[4];
	char *ALL = "ALL";
	int active = 0;
	FILE *arq;

	if (once)
		readWritePslave(list, start, end, &active, 0);
	if (portChosen != 0) {
		portFlag = 1;
		sprintf(port, "S%d", portChosen);
		if (!active && once) {
			printf("\r\nPort %d is currently not active.", portChosen);
			x[0] = my_getchar("Would you like to activate it? (y/n)", 'n', yes_no, 1);
			if (x[0] == 'N') {
				return 0;
			} else {
				arq = fopen(PSLAVE_CONF_FILE, "a");
				if(!arq) {
					printf("\r\nError opening temporary file.");
					exit(0);
				}
				fprintf (arq, "s%d.tty\tttyS%d\n", portChosen, portChosen);
				fclose(arq);
			}
		}
	}

	if (!allOn) {
		wizardHeader(0);
		showConfig(list, start, end, portFlag);
		x[0] = my_getchar("Set to defaults? (y/n)", 'n', yes_no, 1);
		if (x[0] == 'Y') {
			DefaultConfig(list , start, end);
		} 
	}
	printf("\r\n\r\n");

	i = start;
	wizardHeader(0);
	printf("%s.ISSUE - This text determines the format of the login"
		"\r\nbanner that is issued when a connection is made to the"
		"\r\nsystem. \\n represents a new line and \\r respresents a"
		"\r\ncarriage return.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.PROMPT - This text defines the format of the login"
		"\r\nprompt.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.LF_SUPPRESS - This activates line feed suppression. "
		"\r\nWhen configured as 0, line feed suppression will not be"
		"\r\nperformed. When 1, extra line feed will be suppressed.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.AUTO_ANSWER_INPUT - This parameter is used in conjunc-"
		"\r\ntion with the next parameter, auto_answer_output. Please"
		"\r\nrefer to the manual for more info."
		"\r\nIf configured and if there is no session established to"
		"\r\nthe port, this parameter will constantly be compared and"
		"\r\nmatched up to the string of bytes coming in remotely from"
		"\r\nthe server. If a match is found, the string configured in"
		"\r\nauto_answer_output is sent back to the server. To repre-"
		"\r\nsent the ESC character as part of this string, use the"
		"\r\ncontrol character, ^[.", portFlag ? port : ALL);
		i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.AUTO_ANSWER_OUTPUT - This parameter is used in conjunc-"
		"\r\ntion with the previous parameter, auto_answer_input. Please"
		"\r\nrefer to the manual for more info."
		"\r\nIf configured, and if there is no session established to"
		"\r\nthe port, this parameter is sent back to the server"
		"\r\nwhen there is a match between the incoming data and"
		"\r\nauto_answer_input. To represent the ESC character as part"
		"\r\nof this string, use the control character, ^[.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;

	return(1);
} //end of termFunc

/****************************************************************
  alarmFunc - takes care of 'wiz --al' commmand.
****************************************************************/
static int alarmFunc(struct cfglist *list, int start, int end)
{
	int x[6], i = 0, portFlag = 0;
	char port[4];
	char *ALL = "ALL";
	int active = 0;
	FILE *arq;

	if (once)
		readWritePslave(list, start, end, &active, 0);
	if (portChosen != 0) {
		portFlag = 1;
		sprintf(port, "S%d", portChosen);
		if (!active && once) {
			printf("\r\nPort %d is currently not active.", portChosen);
			x[0] = my_getchar("Would you like to activate it? (y/n)", 'n', yes_no, 1);
			if (x[0] == 'N') {
				return 0;
			} else {
				arq = fopen(PSLAVE_CONF_FILE, "a");
				if(!arq) {
					printf("\r\nError opening temporary file.");
					exit(0);
				}
				fprintf (arq, "s%d.tty\tttyS%d\n", portChosen, portChosen);
				fclose(arq);
			}
		}
	}

	if (!allOn) {
		wizardHeader(0);
		showConfig(list, start, end, portFlag);
		x[0] = my_getchar("Set to defaults? (y/n)", 'n', yes_no, 1);
		if (x[0] == 'Y') {
			DefaultConfig(list , start, end);
		} 
	}
	printf("\r\n\r\n");

	i = start;
	wizardHeader(0);
	printf("%s.ALARM - When non zero, all data received from the port"
           "\r\nare captured and sent to syslog-ng with INFO level and"
           "\r\nLOCAL[0+conf.DB_facility] facility. The syslog-ng.conf"
           "\r\nfile should be set accordingly, for the syslog-ng to take"
           "\r\nsome action."
           "\r\n""(Please see the 'Syslog-ng Configuration to use with Alarm"
           "\r\nFeature' section under Generating Alarms in Chapter 3 of"
           "\r\nthe system's manual for the syslog-ng configuration file.)", portFlag ? port : ALL);

	i = getValue(list+i, i);
	if (i == -1)
		return 1;

	return(1);

} //end of alarmFunc

/****************************************************************
  tsFunc - takes care of 'wiz --tso' commmand.
****************************************************************/
static int tsFunc(struct cfglist *list, int start, int end)
{
	int x[6], a = 1, i = 0, portFlag = 0;
	char port[4];
	char *ALL = "ALL";
	int active = 0;
	FILE *arq;

	if (once)
		readWritePslave(list, start, end, &active, 0);
	if (portChosen != 0) {
		portFlag = 1;
		sprintf(port, "S%d", portChosen);
		if (!active && once) {
			printf("\r\nPort %d is currently not active.", portChosen);
			x[0] = my_getchar("Would you like to activate it? (y/n)", 'n', yes_no, 1);
			if (x[0] == 'N') {
				return(0);
			} else {
				arq = fopen(PSLAVE_CONF_FILE, "a");
				if(!arq) {
					printf("\r\nError opening temporary file.");
					exit(0);
				}
				fprintf (arq, "s%d.tty\tttyS%d\n", portChosen, portChosen);
				fclose(arq);
			}
		}
	}

	if (!allOn) {
		wizardHeader(0);
		showConfig(list, start, end, portFlag);
		x[0] = my_getchar("Set to defaults? (y/n)", 'n', yes_no, 1);
		if (x[0] == 'Y') {
			DefaultConfig(list , start, end);
		} 
	}
	printf("\r\n\r\n");

	i = start;
	wizardHeader(0);
	printf("%s.HOST - The IP address of the host to which the"
		   "\r\nterminals will connect.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.TERM - This parameter defines the terminal type assumed"
		   "\r\nwhen performing rlogin or telnet to other hosts.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;

	if (!portFlag) {
		formatSpacing(&a);
		printf("CONF.LOCALLOGINS - This parameter is only necessary when "
			   "\r\nauthentication is being performed for a port. When set to"
			   "\r\n1, it is possible to log into the system directly by "
			   "\r\nplacing a '!' before users' login name, then using their"
			   "\r\nnormal password. This is useful if the Radius authentica-"
			   "\r\ntion server is down.");
		i = getValue(list+i, i);
		if (i == -1)
			return 1;
	}

	return(1);
} //end of tsFunc

/****************************************************************
  ssetFunc - takes care of 'wiz --sset' commmand.
****************************************************************/
static int ssetFunc(struct cfglist *list, int start, int end)
{
	int x[6], a = 1, i = 0, portFlag = 0;
	char port[4];
	char *ALL = "ALL";
	int active = 0;
	FILE *arq;

	if (once)
		readWritePslave(list, start, end, &active, 0);
	if (portChosen != 0) {
		portFlag = 1;
		sprintf(port, "S%d", portChosen);
		if (!active && once) {
			printf("\r\nPort %d is currently not active.", portChosen);
			x[0] = my_getchar("Would you like to activate it? (y/n)", 'n', yes_no, 1);
			if (x[0] == 'N') {
				return(0);
			} else {
				arq = fopen(PSLAVE_CONF_FILE, "a");
				if(!arq) {
					printf("\r\nError opening temporary file.");
					exit(0);
				}
				fprintf (arq, "s%d.tty\tttyS%d\n", portChosen, portChosen);
				fclose(arq);
			}
		}
	}

	if (!allOn) {
		wizardHeader(0);
		showConfig(list, start, end, portFlag);
		x[0] = my_getchar("Set to defaults? (y/n)", 'n', yes_no, 1);
		if (x[0] == 'Y') {
			DefaultConfig(list , start, end);
		} 
	}
	printf("\r\n\r\n");

	i = start;
	wizardHeader(0);
	printf("%s.SPEED - The data speed in bits per second"" (bps) of"
		   "\r\nall ports.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.DATASIZE - The data size in bits per character of" 
		   "\r\nall ports.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.STOPBITS - The number of stop bits for all ports.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.PARITY - The parity for all ports."
		   "\r\n""(e.g. none, odd, even)", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.FLOW - This sets the flow control to hardware,"
		   "\r\nsoftware, or none."" (e.g. hard, soft, none)", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.DCD - DCD signal"" (sets the tty parameter CLOCAL)."
		   "\r\nValid values are 0 or 1. In a socket session, if" 
		   "\r\nall.dcd=0, a connection request (telnet or ssh) will be "
		   "\r\naccepted regardless of the DCD signal and the connection "
		   "\r\nwill not be closed if the DCD signal is set to DOWN. In a"
		   "\r\nsocket connection, if all.dcd=1 a connection request will"
		   "\r\nbe accepted only if the DCD signal is UP and the connection"
		   "\r\n(telnet or ssh) will be closed if the DCD signal is set to"
		   "\r\nDOWN.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;

	if (strcmp(profile, "cas") == 0) {
		formatSpacing(&a);
		printf("%s.DTR_RESET - This parameter specifies the behavior of"
			"\r\nthe DTR signal in the serial port. If set to 0, the DTR"
			"\r\nsignal will be ON if there is a connection to the serial"
			"\r\nport, otherwise it will be OFF. If set from 1 - 99, the DTR"
			"\r\nsignal will always be ON. A value greater than or equal to"
			"\r\n100 specifies how long ""(in milliseconds) the DTR signal "
			"\r\nwill be turned off before it is turned back on again when a"
			"\r\nconnection to the serial port is closed.", portFlag ? port : ALL);
		i = getValue(list+i, i);
		if (i == -1)
			return 1;
		formatSpacing(&a);

		printf("%s.STTYCMD - Tty settings after a socket connection to"
			   "\r\nthat serial port is established. The tty is programmed to"
			   "\r\nwork as a CAS profile and this user specific configuration"
			   "\r\nis applied over that serial port. Parameters must be "
			   "\r\nseparated by space."
			   "\r\n""(e.g. all.sttyCmd -igncr -onlcr opost -icrnl)"
			   "\r\n-igncr tells the terminal not to ignore the carriage-return"
			   "\r\non input, -onlcr means do not map newline character to a"
			   "\r\ncarriage return/newline character sequence on output, opost"
			   "\r\nrepresents post-process output, -icrnl means do not map"
			   "\r\ncarriage-return to a newline character on input.", portFlag ? port : ALL);
		i = getValue(list+i, i);
		if (i == -1)
			return 1;
	}

	return(1);

} //end of ssetFunc

/****************************************************************
  dbFunc - takes care of 'wiz --db' commmand.
****************************************************************/
static int dbFunc(struct cfglist *list, int start, int end)
{
	int x[6], a = 1, i = 0, portFlag = 0;
	char port[4];
	char *ALL = "ALL";
	int active = 0;
	FILE *arq;

	if (once) 
		readWritePslave(list, start, end, &active, 0);
	if (portChosen != 0) {
		portFlag = 1;
		sprintf(port, "S%d", portChosen);
		if (!active && once) {
			printf("\r\nPort %d is currently not active.", portChosen);
			x[0] = my_getchar("Would you like to activate it? (y/n)", 'n', yes_no, 1);
			if (x[0] == 'N') {
				return(0);
			} else {
				arq = fopen(PSLAVE_CONF_FILE, "a");
				if(!arq) {
					printf("\r\nError opening temporary file.");
					exit(0);
				}
				fprintf (arq, "s%d.tty\tttyS%d\n", portChosen, portChosen);
				fclose(arq);
			}
		}
	}

	if (!allOn) {
		wizardHeader(0);
		showConfig(list, start, end, portFlag);
		x[0] = my_getchar("Set to defaults? (y/n)", 'n', yes_no, 1);
		if (x[0] == 'Y') {
			DefaultConfig(list , start, end);
		} 
	}
	printf("\r\n\r\n");

	i = start;
	wizardHeader(0);
	if (!portFlag) {
		printf("CONF.NFS_DATA_BUFFERING - This parameter applies only if"
				"\r\nusers choose to remotely buffer data. This is the remote"
				"\r\ndirectory name where data buffering will be written to"
				"\r\ninstead of the default directory '/var/run'. If deactiva-"
				"\r\nvated, data buffering will be done locally.");
		i = getValue(list+i, i);
		if (i == -1)
			return 1;
		formatSpacing(&a);
	} else {
		i++;
	}

	printf("%s.DATA_BUFFERING - For local data buffering, this para-"
		   "\r\nmeter represents the maximum file size in bytes allowed to"
		   "\r\nbe captured before it is discarded for new space. If re-"
		   "\r\nmote this parameter is just a flag to either activate"" (any "
		   "\r\nvalue greater than 0) or deactivate data buffering.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.DB_MODE - For local data buffering, this parameter"
		   "\r\nallow users to buffer data in either a circular or linear"
		   "\r\nfashion. Circular format"" (cir) is a revolving buffer "
		   "\r\nfile that is overwritten whenever the limit of the buffer "
		   "\r\nsize"" (set by all.data_buffering) is reached. In linear "
		   "\r\nformat"" (lin), data transmission between the remote device"
		   "\r\nand the serial port ceases once the 4k bytes Rx buffer in "
		   "\r\nthe kernel is reached. Then if a session is established to"
		   "\r\nthe serial port, the data in the buffer is shown"
		   "\r\n""(dont_show_DBmenu must be 2), cleared, and data transmi-"
		   "\r\nssion is resumed. Linear buffering is impossible if flow "
		   "\r\ncontrol is set to none. Default is cir.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.DONT_SHOW_DB_MENU - When 0, a menu with data "
			"\r\nbuffering options is shown when a non-empty data"
			"\r\nbuffering file is found. When 1, the data buffering"
			"\r\nmenu is not shown. When 2, the data buffering menu is"
			"\r\nnot shown but the data buffering file is shown if not "
			"\r\nempty. When 3, the data buffering menu is shown, but "
			"\r\nwithout the 'erase and show' and 'erase' options.", portFlag ? port : ALL);

	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	 printf("%s.DB_TIMESTAMP - Records the time stamp in the data"
			"\r\nbuffering file"" (1) or not"" (0). In case it is configured as"
			"\r\n1, the software will accumulate input characters until it"
			"\r\nreceives a CR and LF from the serial port, or the accumu-"
			"\r\nlated data reaches 256 characters. Either way, the accumu-"
			"\r\nlated data will be recorded in the data buffering file "
			"\r\nalong with the current time. The parameter, all.data_buf-"
			"\r\nfering, has to be nonzero in order for this parameter to "
			"\r\nwork.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.SYSLOG_BUFFERING - This parameter is another option to "
		   "\r\ndata buffering. Users can also have syslog perform this"
		   "\r\nfunction along with data buffering into files. When "
		   "\r\nnonzero, the contents of the data buffer are sent to the"
		   "\r\nsyslog-ng every time a quantity of data equal to this "
		   "\r\nparameter is collected. The syslog level for data buffering"
		   "\r\nis hard coded to level 5"" (notice) and facility "
		   "\r\nconf.DB_facility. The file /etc/syslog-ng/syslog-ng.conf"
		   "\r\nshould be set accordingly for the syslog-ng to take some"
		   "\r\naction."
		   "\r\n""(Please see the 'Syslog-ng Configuration to use with"
		   "\r\nSyslog Buffering Feature' section under Generating Alarms"
		   "\r\nin Chapter 3 of the system's manual for the syslog-ng"
		   "\r\nconfiguration file.)", portFlag ? port : ALL);

	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.SYSLOG_SESS - In order for this parameter to function,"
		"\r\nmake sure syslog buffering is activate. When set as 0,"
		"\r\nsyslog messages are always generated whether or not there"
		"\r\nis a connection to the port that is sending data to your"
		"\r\nunit. When set to 1, syslog messages are NOT generated when"
		"\r\nthere IS a connection to the port that is sending data. It"
		"\r\nis only generated when there isn't a session to the port"
		"\r\nthat is sending data to your unit." , portFlag ? port : ALL);

	i = getValue(list+i, i);
	if (i == -1)
		return 1;

	return(1);

} //end of dbFunc

/****************************************************************
  snfFunc - takes care of 'wiz --snf' commmand.
****************************************************************/
static int snfFunc(struct cfglist *list, int start, int end)
{
	int x[6], a = 1, i = 0, portFlag = 0;
	char port[4];
	char *ALL = "ALL";
	int active = 0;
	FILE *arq;

	if (once)
		readWritePslave(list, start, end, &active, 0);
	if (portChosen != 0) {
		portFlag = 1;
		sprintf(port, "S%d", portChosen);
		if (!active && once) {
			printf("\r\nPort %d is currently not active.", portChosen);
			x[0] = my_getchar("Would you like to activate it? (y/n)", 'n', yes_no, 1);
			if (x[0] == 'N') {
				return(0);
			} else {
				arq = fopen(PSLAVE_CONF_FILE, "a");
				if(!arq) {
					printf("\r\nError opening temporary file.");
					exit(0);
				}
				fprintf (arq, "s%d.tty\tttyS%d\n", portChosen, portChosen);
				fclose(arq);
			}
		}
	}

	if (!allOn) {
		wizardHeader(0);
		showConfig(list, start, end, portFlag);
		x[0] = my_getchar("Set to defaults? (y/n)", 'n', yes_no, 1);
		if (x[0] == 'Y') {
			DefaultConfig(list , start, end);
		} 
	}
	printf("\r\n\r\n");

	i = start;
	wizardHeader(0);
	printf("%s.ADMIN_USERS - This parameter determines which users"
		   "\r\ncan open a sniff session, which is where other users"
		   "\r\nconnected to the very same port can see everything that"
		   "\r\nthe first user is doing. The other users connected to the"
		   "\r\nvery same port can also cancel the first user's session"
		   "\r\n""(and take over). If the parameter, all.multiple_sessions,"
		   "\r\nis configured as 'no', then only two users can connect to"
		   "\r\nthe same port simultaneously. If it is configured as 'yes',"
		   "\r\nmore simultaneous users can can sniff the session or have"
		   "\r\nread/write permissions." 
		   "\r\n""(Please see details in Session Sniffing in Chapter 3 of"
		   "\r\nthe system's manual.)", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.SNIFF_MODE - This parameter determines what other "
		   "\r\nusers connected to the very same port can see of the"
		   "\r\nsession of the first connected user"" (main session). The"
		   "\r\nsecond session is called a sniff session and this "
		   "\r\nfeature is activated whenever the protocol is set to"
		   "\r\nsocket_ssh or socket_server."
		   "\r\n""(e.g. in -shows data written to the port, out -shows data"
		   "\r\nreceived from the port, i/o -shows both streams.)", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.ESCAPE_CHAR - This parameter determines which "
		   "\r\ncharacter must be typed to make the session enter into"
		   "\r\n\"menu mode.\" The possible values are <CRTL-a> to <CRTL-z>,"
		   "\r\nand this is only valid when the port protocol is "
		   "\r\nsocket_server or socket_ssh. Represent the CRTL"
		   "\r\ncharacter with '^'. Default value is ^z.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.MULTIPLE_SESSIONS - Allows users to open multiple"
			"\r\ncommon and sniff sessions on the same port. The options"
			"\r\nare 'yes', 'no', 'RW_session', or 'sniff_session'." 
			"\r\nDefault is set to 'no'.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	return(1);

} //end of snfFunc	

/****************************************************************
  slFunc - takes care of 'wiz --sl' commmand.
****************************************************************/
static int slFunc(struct cfglist *list, int start, int end)
{
	int x[6], a = 1, i = 0, portFlag = 0;
	char port[4];
	int active = 0;
	FILE *arq;

	if (once) 
		readWritePslave(list, start, end, &active, 0);
	if (portChosen != 0) {
		portFlag = 1;
		sprintf(port, "S%d", portChosen);
		if (!active && once) {
			printf("\r\nPort %d is currently not active.", portChosen);
			x[0] = my_getchar("Would you like to activate it? (y/n)", 'n', yes_no, 1);
			if (x[0] == 'N') {
				return(0);
			} else {
				arq = fopen(PSLAVE_CONF_FILE, "a");
				if(!arq) {
					printf("\r\nError opening temporary file.");
					exit(0);
				}
				fprintf (arq, "s%d.tty\tttyS%d\n", portChosen, portChosen);
				fclose(arq);
			}
		}
	}

	if (!allOn) {
		wizardHeader(0);
		showConfig(list, start, end, portFlag);
		x[0] = my_getchar("Set to defaults? (y/n)", 'n', yes_no, 1);
		if (x[0] == 'Y') {
			DefaultConfig(list , start, end);
		}
	}
	printf("\r\n\r\n");

	i = start;
	wizardHeader(0);
	if (!portFlag) {
		printf("CONF.FACILITY - This value (0-7) is the Local facility sent"
			   "\r\nto the syslog. The file /etc/syslog-ng/syslog-ng.conf "
			   "\r\ncontains a mapping between the facility number and the"
			   "\r\naction."
			   "\r\n""(Please see the 'Syslog-ng Configuration to use with Syslog"
			   "\r\nBuffering Feature' section under Generating Alarms in "
			   "\r\nChapter 3 the system's manual for the syslog-ng "
			   "\r\nconfiguration file.)");
		i = getValue(list+i, i);
		if (i == -1)
			return 1;
		formatSpacing(&a);

		printf("CONF.DB_FACILITY - This value (0-7) is the Local facility"
			   "\r\nsent to the syslog with the data when syslog_buffering is"
			   "\r\nactive. The file /etc/syslog-ng/syslog-ng.conf contains a"
			   "\r\nmapping between the facility number and the action." 
			   "\r\n""(Please see the 'Syslog-ng Configuration to use with Syslog"
			   "\r\nBuffering Feature' section under Generating Alarms in "
			   "\r\nChapter 3 the system's manual for the syslog-ng "
			   "\r\nconfiguration file.)");
		i = getValue(list+i, i);
		if (i == -1)
			return 1;
	}

	return(1);

} //end of slFunc	 

/****************************************************************
  pmFunc - takes care of 'wiz --pm' commmand.
****************************************************************/
static int pmFunc(struct cfglist *list, int start, int end)
{
	int x[6], a = 1, i = 0, portFlag = 0;
	char port[4];
	char *ALL = "ALL";
	int active = 0;
	FILE *arq;

	if (once) 
		readWritePslave(list, start, end, &active, 0);
	if (portChosen != 0) {
		portFlag = 1;
		sprintf(port, "S%d", portChosen);
		if (!active && once) {
			printf("\r\nPort %d is currently not active.", portChosen);
			x[0] = my_getchar("Would you like to activate it? (y/n)", 'n', yes_no, 1);
			if (x[0] == 'N') {
				return(0);
			} else {
				arq = fopen(PSLAVE_CONF_FILE, "a");
				if(!arq) {
					printf("\r\nError opening temporary file.");
					exit(0);
				}
				fprintf (arq, "s%d.tty\tttyS%d\n", portChosen, portChosen);
				fclose(arq);
			}
		}
	}

	if (!allOn) {
		wizardHeader(0);
		showConfig(list, start, end, portFlag);
		x[0] = my_getchar("Set to defaults? (y/n)", 'n', yes_no, 1);
		if (x[0] == 'Y') {
			DefaultConfig(list , start, end);
		} 
	}
	printf("\r\n\r\n");

	i = start;
	wizardHeader(0);
	printf("%s.PROTOCOL - The possible protocols are telnet"" (socket_"
			"\r\nserver), ssh1/ssh2"" (socket_ssh), raw data"" (raw_data),"
			"\r\nor integrated power distributed unit"" (ipdu).", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.PMTYPE - Name of the IPDU manufacturer.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.PMUSERS - List of the outlets each user can access."
			"\r\n""(e.g. Joe: 1-3; Jane: 4,5,6)", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.PMOUTLET - The number of the outlet where the server"
			"\r\nis plugged.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

	printf("%s.PMKEY - The hot-key that identifies the power "
			"\r\nmanagement command.", portFlag ? port : ALL);
	i = getValue(list+i, i);
	if (i == -1)
		return 1;
	formatSpacing(&a);

//[RK]Oct/21/04 - removed - not used
//	printf("%s.PMNUMOFOUTLETS - The number of outlets you have on the"
//			"\r\nAlterPath-PM.", portFlag ? port : ALL);
//	i = getValue(list+i, i);
//	if (i == -1)
//		return 1;
	
	return(1);

} //end of pmFunc	 

/****************************************************************
  executeBasicFunc - executes entire 'wiz' commmand.
****************************************************************/
void executeBasicFunc(struct cfglist *list, int start, int end)
{
	int x[6], opt, portFlag=0, dhcpPid;
	
	wizardHeader(2);
	while(1) {
		basicInfoFunc(list, start, end);
		
		wizardHeader(0);
		showConfig(list, start, end, portFlag);

		x[0] = my_getchar("Are all these parameters correct? (y/n)", 'n', yes_no, 1);

		if (x[0] == 'Y') {
			wizardHeader(0);
			SaveHostName(HOSTNAME_FILE, basicList);
			SaveGateWay(NTWK_ST_ROUTES_FILE, basicList);
			// Takes care of killing dhcp so that the right
         // info gets written to resolv.conf and the other files.
         if (!dhcpFlg) {
				if ((dhcpPid = getDhcpPid()) != -1) {
					kill(dhcpPid, SIGKILL);
					while(1) {
						if ((kill(dhcpPid, 0)) == -1 && errno == ESRCH) {
							break;
						}
					}
				}
				SaveCSipInHosts(HOSTS_FILE, basicList);
				SaveCSipInIfcfg(NTWK_IFCFG_FILE, basicList);
         }
			SaveResolv(RESOLV_CONF_FILE, basicList);
			SaveCSipInPslave(PSLAVE_CONF_FILE, basicList);
			printf("\r\nYou can now use the browser to finish your system configu-"
				   "\r\nrations, but before that, please read below.\r\n\r\n");
			exit_rot();
			return;
		} else {
			exit_rot();
			printf("\nType 'c' to go back and CORRECT the current configuration"
				   "\nparameters or 'q' to QUIT : ");

			if ((opt = get_option()) < -1) {
				exit(0);
			} else {
				setRawMode();
				once = 0;		//set to 0, so that when func loops back it wouldn't reread parameters
				continue;
			}
		}
	} //end of while loop	
}

/****************************************************************
  executeFunc - executes all other commands besides 'wiz'.
****************************************************************/
void executeFunc(int start, int end, int type)
{
	int opt, x[6], portFlag=0, ret=0, active;
	struct cfglist *list = pslaveportlist;

	// Show instruction header only once at the beginning of a wiz command
	if(allOnOnce) {
		wizardHeader(1);
		if(allOn)
			allOnOnce = 0;
	}
	while(1){
		while(1) {
			switch(type) {
				case 0:
					ret = accessFunc(list, start, end);
					break;
				case 1:
					ret = authFunc(list, start, end);
					break;
				case 2:
					ret = termFunc(list, start, end);
					break;
				case 3: 
					ret = alarmFunc(list, start, end);
					break;
				case 4: 
					ret = tsFunc(list, start, end);
					break;
				case 5:
					ret = ssetFunc(list, start, end);
					break;
				case 6:
					ret = dbFunc(list, start, end);
					break;
				case 7:
					ret = snfFunc(list, start, end);
					break;
				case 8:
					ret = slFunc(list, start, end);
					break;
				case 9:
					ret = pmFunc(list, start, end);
					break;
			}
			wizardHeader(0);
			if(portChosen != 0)
				portFlag = 1;

			if(ret == 0) {
				goto L1;
			}
			showConfig(list, start, end, portFlag);

			x[0] = my_getchar("Are these configuration(s) all correct? (y/n)", 'n', yes_no, 1);

			if (x[0] == 'Y') {
				if (!portChosen && type != 8) {
					x[0] = my_getchar("\r\nDiscard previous port-specific parameters? (y/n)", 'n', yes_no, 1);
					if (x[0] == 'Y')
						discard_flag = 1;
				}
				exit_rot();
				// Read in configurations
				readWritePslave (list, start, end, &active, 1);
			
				// Type = 8 is for syslog configurations which has only conf. parameters
				// So need special case so that it won't ask users to configure them for specific ports.
				if (type != 8) { 
					printf("\nType 'c' to CONTINUE to set these parameters for "
					   "\nspecific ports or 'q' to QUIT : ");
					if ((opt = get_option()) < -1) {
						return;
					} else {
						once = 1;
						setRawMode();
						break;
					}
				} else {		//If type = 8, just ask to activate/save to flash.
					return;	
				}
			} else {
				exit_rot();
				printf("\nType 'c' to go back and CORRECT these parameters"
					   "\nor 'q' to QUIT : ");
				if ((opt = get_option()) < -1) {
					exit(0);
				} else {
					setRawMode();
					once = 0; 		//Set to 0 so that when this loops back it won't reread pslave.conf 
					continue;		//to obtain the parameters. The parameters will be whatever was 
				}					//previously entered.				
			}
		} //end of while loop

		while(1) {
			wizardHeader(0);
L1:
			exit_rot();
			printf("You have %d available ports on this system.", mt.nports);
			printf("\n\nType 'q' to quit, a valid port number[1-%d], or anything"
				   "\nelse to refresh : ", mt.nports);
			if ((opt = get_option()) < -1) {
				return;
			}
			if (!opt || opt > mt.nports) {
				continue;
			}
			break;
		} //end of choosePort while loop

		portChosen = opt;
		setRawMode();
		continue;

	}  //end of outer most while loop
}

/*************************************************************
  myExit - exit function. Calls signal_ras hup and saveconf.
*************************************************************/
void myExit(int showHeader) {
	int x[6];

	setRawMode();
	
	if (showHeader)
		wizardHeader(0);
	printf("""(Note: If you are NOT connected to this unit through a"
		"\r\nconsole, and you have just reconfigured the IP of this"
		"\r\nunit, activating the new configurations may cause you to"
		"\r\nlose connection. In that case, please reconnect to the "
		"\r\nunit by the new IP address, and manually issue a saveconf"
		"\r\nto save your configurations to flash.)\r\n");
	x[0] = my_getchar("Do you want to activate your configurations now? (y/n)", 'y', yes_no, 1);

	if (x[0] == 'Y') {
		system("signal_ras hup");
	}
	wizardHeader(0);
	printf("Flash refers to a type of memory that can be erased and"
		   "\r\nreprogrammed in units of memory known as blocks rather than"
		   "\r\none byte at a time; thus, making updating to memory easier.\r\n"
		   "\r\nIf you choose to save to flash, your configurations thus"
		   "\r\nfar will still be in the memory of the system even after "
		   "\r\nyou reboot it. If you don't save to flash and if you were "
		   "\r\nto reboot the system, all your new configurations will be "
		   "\r\nlost and you will have to reconfigure the system.\r\n");
	x[0] = my_getchar("Do you want to save your configurations to flash? (y/n)", 'n', yes_no, 1);

	if (x[0] == 'Y') {
		printf("\r\nThis will only take a moment...\r\n");
		system("saveconf");
	} 

	if (basicOn) {
		DeallocConfig(basicList);
	} else {
		DeallocConfig(pslaveportlist);		
	}
	printf("\r\n\r\nThanks for using the Configuration Wizard! Good-bye!\r\n");
}

int main (argc, argv)
	int argc;
	char **argv;
{
	int c, portOn=0;

	while (1) {
		int option_index = 0;
		static struct option long_options[] = {
			{"ac", 1, 0, 0},  
			{"auth", 0, 0, 1},
			{"tl", 0, 0, 2},  
			{"al", 0, 0, 3}, 
			{"tso", 0, 0, 4},	
			{"sset", 1, 0, 5},	
			{"db", 0, 0, 6},	
			{"snf", 0, 0, 7},
			{"sl", 0, 0, 8},
			{"pm", 0, 0, 9},
			{"all", 1, 0, 10},
			{"port", 1, 0, 11},
			{"help", 0, 0, 'h'},
			{0, 0, 0, 0}
		};

		get_machine_type(&mt);

		c = getopt_long (argc, argv, "h", long_options, &option_index);
		if (c == -1)
			break;

		switch (c) {
			case 0:
				if (optarg) {
					//printf ("arg %s", optarg);
					profile = optarg;
					lowerCase(profile);
					if(strcmp(profile, "ts")!=0 && strcmp(profile, "cas")!=0) {
						printf("Error: Please provide a valid profile"" ('cas' or 'ts').\n\n");
						return(0);
					}
				}
				accessOn = 1;
				break;
			case 1:
				authOn = 1;
				break;
			case 2:
				tlOn = 1;
				break;
			case 3:
				alarmOn = 1;
				break;
			case 4:
				tsOn = 1;
				break;
			case 5:
				if (optarg) {
					profile = optarg;
					lowerCase(profile);
					if(strcmp(profile, "ts")!=0 && strcmp(profile, "cas")!=0) {
						printf("Error: Please provide a valid profile"" ('cas' or 'ts').\n\n");
						return(0);
					}
				}
				ssetOn = 1;
				break;
			case 6: 
				dbOn = 1;
				break;
			case 7:
				snfOn = 1;
				break;
			case 8:
				slOn = 1;
				break;
			case 9:
				pmOn = 1;
				break;
			case 10:
				if (optarg) {
					profile = optarg;
					lowerCase(profile);
					if(strcmp(profile, "ts")!=0 && strcmp(profile, "cas")!=0) {
						printf("Error: Please provide a valid profile"" ('cas' or 'ts').\n\n");
						return(0);
					}
				}
				allOn = 1;
				accessOn = 1;
				authOn = 1;
				tlOn = 1;
				alarmOn = 1;
				ssetOn = 1;
				dbOn = 1;
				snfOn = 1;
				slOn = 1;
				pmOn = 1;
				if (!strcmp(profile, "ts")) {
					tsOn = 1;
					dbOn = 0;
					snfOn = 0;
					slOn = 0;
				}
				break;
			case 11:
				portOn = 1;
				portChosen =  atoi(optarg);
				break;
			case 'h':
			case '?':
				printf("\nSYNOPSIS: %s [--OPTIONS] [--port <port number>]"
					   "\n\nGENERAL OPTIONS\n\n"
					" auth              : Configuration of authentication parameters\n"
					" tl                : Configuration of terminal login display parameters\n"
					" al                : Configuration of alarm parameter\n"
					" db                : Configuration of data buffering parameters\n"
					" snf               : Configuration of sniffing parameters\n"
					" sl                : Configuration of syslog parameters\n"
					" tso               : Configuration of other parameters specific\n" 
					"                     to the TS profile\n"
					" pm                : Configuration of power management parameters\n"
					" ac   <cas or ts>  : Configuration of access method parameters\n"
					" sset <cas or ts>  : Configuration of serial setting parameters\n"
					" all  <cas or ts>  : Configuration of all parameters\n"
					" help              : Print this help message\n\n"
						, argv[0]);
				printf("To directly configure a feature for a specific serial port\n"
					   "use the '--port <port number>' option after 'wiz --[option]'.\n\n");
				return(0);
				break;
			default:
				printf ("?? getopt returned character code 0%o ??\n", c);
		} //end of switch
	} //end of while

	if (argc == 1) {	// for basic wiz parameters
		setRawMode();
		basicOn=1;
		if(!AllocConfig(basicList, BASIC_START, BASIC_END)) {
			free(basicList);
			printf("Error: Was unable to allocate space for configurations.\n\n");
			return(0);
		}
		executeBasicFunc(basicList, BASIC_START, BASIC_END);
	}

	if (accessOn) {
		/* Initialize pointers for group structure */
		setRawMode();
		once = 1;
		if (strcmp(profile, "ts")==0) {
			if(!AllocConfig(pslaveportlist, AC_TS_START, AC_TS_END)) {
				free(pslaveportlist);
				printf("Error: Was unable to allocate space for configurations.\n\n");
				return(0);
			}
			executeFunc(AC_TS_START, AC_TS_END, 0);
		} else if (strcmp(profile, "cas")==0) {
			if(!AllocConfig(pslaveportlist, AC_CAS_START, AC_CAS_END)) {
				free(pslaveportlist);
				printf("Error: Was unable to allocate space for configurations.\n\n");
				return(0);
			}
			executeFunc(AC_CAS_START, AC_CAS_END, 0);
		}
		if(!portOn)
			portChosen = 0;
	}
	if (authOn) {
		setRawMode();
		once = 1;
		if(!AllocConfig(pslaveportlist, AUTH_START, AUTH_END)) {
			free(pslaveportlist);
			printf("Error: Was unable to allocate space for configurations.\n\n");
			return(0);
		}
		executeFunc(AUTH_START, AUTH_END, 1);
		if(!portOn)
			portChosen = 0;
	}
	if (tlOn) {
		setRawMode();
		once = 1;
		if(!AllocConfig(pslaveportlist, TL_START, TL_END)) {
			free(pslaveportlist);
			printf("Error: Was unable to allocate space for configurations.\n\n");
			return(0);
		}
		executeFunc(TL_START, TL_END, 2);
		if(!portOn)
			portChosen = 0;
	}
	if (alarmOn) {
		setRawMode();	
		once = 1;
		if(!AllocConfig(pslaveportlist, AL_START, AL_END)) {
			free(pslaveportlist);
			printf("Error: Was unable to allocate space for configurations.\n\n");
			return(0);
		}
		executeFunc(AL_START, AL_END, 3);
		if(!portOn)
			portChosen = 0;
	}
	if (tsOn) {
		setRawMode();
		once = 1;
		if(!AllocConfig(pslaveportlist, TS_START, TS_END)) {
			free(pslaveportlist);
			printf("Error: Was unable to allocate space for configurations.\n\n");
			return(0);
		}
		executeFunc(TS_START, TS_END, 4);		
		if(!portOn)
			portChosen = 0;
	}
	if(ssetOn) {
		setRawMode();
		once = 1;
		if (strcmp(profile, "ts")==0) {
			if(!AllocConfig(pslaveportlist, SSET_START, SSET_END-2)) {
				free(pslaveportlist);
				printf("Error: Was unable to allocate space for configurations.\n\n");
				return(0);
			}
			executeFunc(SSET_START, SSET_END-2, 5);
		} else if (strcmp(profile, "cas")==0) {
			if(!AllocConfig(pslaveportlist, SSET_START, SSET_END)) {
				free(pslaveportlist);
				printf("Error: Was unable to allocate space for configurations.\n\n");
				return(0);
			}
			executeFunc(SSET_START, SSET_END, 5);
		}
		if(!portOn)
			portChosen = 0;
	}
	if(dbOn) {
		setRawMode();
		once = 1;
		if(!AllocConfig(pslaveportlist, DB_START, DB_END)) {
			free(pslaveportlist);
			printf("Error: Was unable to allocate space for configurations.\n\n");
			return(0);
		}
		executeFunc(DB_START, DB_END, 6);		
		if(!portOn)
			portChosen = 0;
	}
	if (snfOn) {
		setRawMode();
		once = 1;
		if(!AllocConfig(pslaveportlist, SNF_START, SNF_END)) {
			free(pslaveportlist);
			printf("Error: Was unable to allocate space for configurations.\n\n");
			return(0);
		}
		executeFunc(SNF_START, SNF_END, 7);
		if(!portOn)
			portChosen = 0;
	}
	if (slOn) {
		setRawMode();
		once = 1;
		if(!AllocConfig(pslaveportlist, SL_START, SL_END)) {
			free(pslaveportlist);
			printf("Error: Was unable to allocate space for configurations.\n\n");
			return(0);
		}
		executeFunc(SL_START, SL_END, 8);
		if(!portOn)
			portChosen = 0;
	}

	if (pmOn) {
		setRawMode();
		once = 1;
		if(!AllocConfig(pslaveportlist, PM_START, PM_END)) {
			free(pslaveportlist);
			printf("Error: Was unable to allocate space for configurations.\n\n");
			return(0);
		}
		executeFunc(PM_START, PM_END, 9);
		if(!portOn)
			portChosen = 0;
	}

	if (argc == 1) {
		myExit(0);
	} else
		myExit(1);

	exit_rot();
	return(0);
}


