/*
 * appl1.c -- Routines for the web pages (LINUX version)
 *
 */

/******************************** Description *********************************/

/*
 *	Routines for the web pages (asp/form/cgi)
 *	This module manages the web pages for configuration tables. The tables
 *	available for web configuration at this moment are the host table (hosts),
 *	static routes (network/st_routes) and the ipchains entries (network/
 *	firewall.conf). A common structure is used for all the tables.
 *	Include new table : users group used in pslave.conf.
 */

/********************************* Includes ***********************************/

#include	"appl.h"
#include	<signal.h>
#include	<unistd.h>
#include	<sys/types.h>
#include	<sys/wait.h>

/******************************* Definitions **********************************/

#define SIZE_IPADD           16

/*********** Entry Types **********/
#define TYPE_NUMBER          'N' /* Number */
#define TYPE_IPADDR          'I' /* IP Address format <nnn.nnn.nnn.nnn>*/
#define TYPE_IPMASK          'M' /* IP Mask */
#define TYPE_STRING          'S' /* Any type of string */
#define TYPE_LIST            'L' /* List of valid entries */
#define TYPE_YN              'Y' /* Yes/No */
#define TYPE_STRING_NOT_NULL 's' /* String which cannot be null */
#define TYPE_RADIO1          '1' /* First option */
#define TYPE_RADIO2          '2' /* Second, third ... option */
#define TYPE_NBLIST          'B' /* List of valid entries or a number */
#define TYPE_VARLIST         'V' /* Dynamic List */
#define TYPE_DEF_IP          'D' /* 'Default' or IP Address format */

/************** Flags *************/
#define FLAG_READ      1 /* Parameter already read  */
#define FLAG_PRINTED   2 /* Parameter already printed */
#define FLAG_WRITTEN   4 /* Parameter already written */
#define FLAG_ERROR     8 /* Error in the parameter or entry */
#define FLAG_INACTIVE 16 /* Parameter inactive (not read, not shown) */
#define FLAG_INSERT   32 /* Entry is to be inserted */
#define FLAG_DELETE   64 /* Entry is to be deleted */

/************* Configuration *************/
#define CONFIG_ORDER           1 /* Table order is important */
#define CONFIG_NAME            2 /* Entry is identified by name, and not by number */
#define CONFIG_FILE            4 /* File contains more than one table */
#define CONFIG_READ_ONLY	   8  /* table is read only */

#define CONFIG_DISC            1 /* The parameter can be discarded */
#define CONFIG_TABLE           2 /* The parameter will NOT be shown */
#define CONFIG_DONT_OPEN_CELL  4
#define CONFIG_DONT_CLOSE_CELL 8

/********** Routine types to be defined ************/
typedef int (parseroutine) (unsigned char *);
typedef void (writeroutine) (FILE *, unsigned char **);
typedef char_t *(verifyroutine) (unsigned char **);
typedef void *(getlistroutine) ();
typedef int (printroutine) (webs_t , unsigned char *, unsigned char **);

/************ Structures to be defined ***********/
/* List of valid entries */
struct strlist {
    unsigned char **filelist; /* List of strings as shown in the file */
    unsigned char **showlist; /* List of strings as shown in the web page */
    unsigned char nlist;      /* Number of elements on the list */
};
/* Range of numbers */
struct nmbrange {
    unsigned int minrange;
    unsigned int maxrange;
};
/* Valid Entries + range of numbers */
struct listrange {
    unsigned int minrange;
    unsigned int maxrange;
    unsigned char **filelist;
    unsigned char **showlist;
    unsigned char nlist;
};
/* Structure for a parameter */
struct paramentry {
    unsigned char *name;      /* Name of the parameter (written in the page) */
    unsigned char *defvalue;  /* Default value, if applicable */
    unsigned char *value;     /* Current value of the parameter */
    unsigned char *value2;    /* Value read from the file or from the form */
    void          *list;      /* List of valid entries, if applicable */
    printroutine  *print;     /* Routine to print the parameter in the table */
    unsigned char flag;       /* Set of flags (see #define's) */
    unsigned char type;	      /* Type of the paramter (see #define's) */
    unsigned short size;      /* Size occupied by the parameter */
    unsigned char field_size; /* Size of the input entry (in the page) */
    unsigned char config;
    unsigned char filler[2];
};
/* Structure for the table */
struct entrylist {
    unsigned char *filename;     /* File name with the path */
    unsigned char *name;         /* Name of the table */
    unsigned char *entryname;    /* Name of the entry */
    unsigned char *title;        /* Title for the page */
    unsigned char *msg;          /* Messages */
    struct paramentry *param;    /* Parameter List */
    unsigned char **value;       /* Values for all parameters/entries */
    unsigned char **permentries; /* Permanent Entries */
    struct entrylist *nexttable; /* Next Table linked with the current one */
    struct entrylist *prevtable; /* Previous Table linked with the current one */
    unsigned char *nextvalue;    /* Value used as reference for the next table */
    unsigned char *getentries;   /* Query string for the page form */
    parseroutine *parse;         /* Routine to parse the entry */
    writeroutine *write;         /* Routine to write the entry in a file */
    verifyroutine *verify;	 /* Routine to verify if the entries are valid */
    unsigned char *buf;          /* Buffer where the file is copied to */
    unsigned char *buf2;         /* Buffer used to write in the file */
    unsigned char **backup;      /* Backup for the entries, for future use */
    unsigned short maxentries;   /* Maximum number of entries */
    unsigned short nentries;     /* Number of entries */
    unsigned int currentry;      /* Number of the current entry */
    unsigned char nparams;       /* Number of parameters */
    unsigned char config;        /* Set of configuration bits (see #define's) */
    unsigned char flag;          /* Set of flag bits (see #define's) */
    unsigned char radio;         /* Parameter with the first option for RADIO type */
    unsigned char radioindex;    /* Number of the option for the RADIO type */
};

/*****************************************************************************
 * Definitions for Host Table
 */

/* Routines */
static int ParseHostEntry (unsigned char *s);
static void WriteHostEntry (FILE *arq, unsigned char **value);
/* Parameters */
struct paramentry hostparams[] = {
    {"IP Address", NULL, NULL, NULL, NULL, NULL, 0, TYPE_IPADDR, SIZE_IPADD, SIZE_IPADD, 0},
    {"Name",       NULL, NULL, NULL, NULL, NULL, 0, TYPE_STRING, 30,         30,         0}
/*  {"Aliases",    NULL, NULL, NULL, NULL, NULL, 0, TYPE_STRING, 50,         50,         CONFIG_DISC}*/
};

/* Table structure */
struct entrylist hosttable = {
    "/etc/hosts", "Host", "host", "Host", NULL,
    hostparams, NULL, NULL, NULL, NULL, NULL, NULL,
    ParseHostEntry, WriteHostEntry, NULL, NULL, NULL, NULL,
    256, 0, 0, sizeof(hostparams)/sizeof(struct paramentry),
    CONFIG_ORDER, 0, 0, 0
};

/*****************************************************************************
 * Definitions for Static Route Table
 */

/* List of valid entries */
unsigned char *gwitfs[] = {"gw", "dev"};
struct strlist gwitflist = {gwitfs, gwitfs, 2};

/* Routines */
static int ParseRouteEntry (unsigned char *s);
static void WriteRouteEntry (FILE *arq, unsigned char **value);

/* Parameters */
struct paramentry routeparams[] = {
    {"Destination IP",   NULL, NULL, NULL, NULL,               NULL, 0, TYPE_DEF_IP, SIZE_IPADD, SIZE_IPADD, 0},
    {"Destination Mask", NULL, NULL, NULL, NULL,               NULL, 0, TYPE_IPMASK, SIZE_IPADD, SIZE_IPADD, CONFIG_DISC},
    {"Route to",         "gw", NULL, NULL, (void *)&gwitflist, NULL, 0, TYPE_RADIO1, 4,          0,          CONFIG_TABLE | CONFIG_DONT_CLOSE_CELL},
    {"Gateway",          NULL, NULL, NULL, NULL,               NULL, 0, TYPE_IPADDR, SIZE_IPADD, SIZE_IPADD, CONFIG_DISC | CONFIG_DONT_OPEN_CELL | CONFIG_DONT_CLOSE_CELL},
    {NULL,               NULL, NULL, NULL, (void *)&gwitflist, NULL, 0, TYPE_RADIO2, 4,          0,          CONFIG_TABLE | CONFIG_DISC | CONFIG_DONT_OPEN_CELL | CONFIG_DONT_CLOSE_CELL},
    {"Interface",        NULL, NULL, NULL, NULL,               NULL, 0, TYPE_STRING, 8,          8,          CONFIG_DISC | CONFIG_DONT_OPEN_CELL}
};

/* Table structure */
struct entrylist routetable = {
    "/etc/network/st_routes", "Route", "route", "Static Route", NULL,
    routeparams, NULL, NULL, NULL, NULL, NULL, NULL,
    ParseRouteEntry, WriteRouteEntry, NULL, NULL, NULL, NULL,
    256, 0, 0, sizeof(routeparams)/sizeof(struct paramentry),
    0, 0, 0, 0
};

/*****************************************************************************
 * Definitions for Filter Rules Table
 */

/* List of valid entries */
unsigned char *scopes[70]; /* This is for a dynamic list */
struct strlist scopelist  = {scopes, scopes, 5};

unsigned char *invstr[] = {"!", ""};
struct strlist invlist = {invstr, invstr, 2};

struct nmbrange socketrange = {0, 65535};
unsigned char *protshow[] = {"Any", "ICMP", "TCP", "UDP"};
unsigned char *protstr[] = {"0", "1", "6", "17"};
struct listrange protlist = {0, 255, protstr, protshow, 4};

/* List of permanent entries */
#ifdef OEM1
unsigned char *chainpentries[] = {"input", "ACCEPT", "forward", "DENY", "output", "ACCEPT", NULL};
#else
unsigned char *chainpentries[] = {"input", "ACCEPT", "forward", "ACCEPT", "output", "ACCEPT", NULL};
#endif

/* Routines */
static void *GetChainList2 ();
static int ParseRuleEntry (unsigned char *s);
static void WriteRuleEntry (FILE *arq, unsigned char **value);
static char_t *VerifyRuleEntry (unsigned char **value);
static int PrintIPRule(webs_t wp, unsigned char *str, unsigned char **value);
static int PrintSockRule(webs_t wp, unsigned char *str, unsigned char **value);
static int PrintProtRule(webs_t wp, unsigned char *str, unsigned char **value);

/* Parameters */
struct paramentry ruleparams[] = {
    {"Target",             "ACCEPT",  NULL, NULL, (void *)GetChainList2, NULL,          0, TYPE_VARLIST,16,         0,          0},
    {"Source IP",          "0.0.0.0", NULL, NULL, NULL,          PrintIPRule,   0, TYPE_IPADDR, SIZE_IPADD, SIZE_IPADD, CONFIG_DONT_CLOSE_CELL},
    {"Mask",               "0.0.0.0", NULL, NULL, NULL,                  NULL,          0, TYPE_IPMASK, SIZE_IPADD, SIZE_IPADD, CONFIG_TABLE | CONFIG_DONT_OPEN_CELL | CONFIG_DONT_CLOSE_CELL},
    {"Inverted",           NULL,      NULL, NULL, (void *)&invlist,      NULL,          0, TYPE_YN,     2,          0,          CONFIG_DISC | CONFIG_TABLE | CONFIG_DONT_OPEN_CELL},
    {"Source Port",        NULL,      NULL, NULL, (void *)&socketrange,  PrintSockRule, 0, TYPE_NUMBER, 8,          8,          CONFIG_DISC | CONFIG_DONT_CLOSE_CELL},
    {"To",                 NULL,      NULL, NULL, (void *)&socketrange,  NULL,          0, TYPE_NUMBER, 8,          8,          CONFIG_DISC | CONFIG_TABLE | CONFIG_DONT_CLOSE_CELL | CONFIG_DONT_OPEN_CELL},
    {"Inverted",           NULL,      NULL, NULL, (void *)&invlist,      NULL,          0, TYPE_YN,     2,          0,          CONFIG_DISC | CONFIG_TABLE | CONFIG_DONT_OPEN_CELL},
    {"Destination IP",     "0.0.0.0", NULL, NULL, NULL,                  PrintIPRule,   0, TYPE_IPADDR, SIZE_IPADD, SIZE_IPADD, CONFIG_DONT_CLOSE_CELL},
    {"Mask",               "0.0.0.0", NULL, NULL, NULL,                  NULL,          0, TYPE_IPMASK, SIZE_IPADD, SIZE_IPADD, CONFIG_TABLE | CONFIG_DONT_OPEN_CELL | CONFIG_DONT_CLOSE_CELL},
    {"Inverted",           NULL,      NULL, NULL, (void *)&invlist,      NULL,          0, TYPE_YN,     2,          0,          CONFIG_DISC | CONFIG_TABLE | CONFIG_DONT_OPEN_CELL},
    {"Destination Port",   NULL,      NULL, NULL, (void *)&socketrange,  PrintSockRule, 0, TYPE_NUMBER, 8,          8,          CONFIG_DISC | CONFIG_DONT_CLOSE_CELL},
    {"To",                 NULL,      NULL, NULL, (void *)&socketrange,  NULL,          0, TYPE_NUMBER, 8,          8,          CONFIG_DISC | CONFIG_TABLE| CONFIG_DONT_OPEN_CELL | CONFIG_DONT_CLOSE_CELL},
    {"Inverted",           NULL,      NULL, NULL, (void *)&invlist,      NULL,          0, TYPE_YN,     2,          0,          CONFIG_DISC | CONFIG_TABLE | CONFIG_DONT_OPEN_CELL},
    {"Protocol",           "0",       NULL, NULL, (void *)&protlist,     PrintProtRule, 0, TYPE_NBLIST, 8,          8,          CONFIG_DISC | CONFIG_DONT_CLOSE_CELL},
    {"Inverted",           NULL,      NULL, NULL, (void *)&invlist,      NULL,          0, TYPE_YN,     2,          0,          CONFIG_DISC | CONFIG_TABLE | CONFIG_DONT_OPEN_CELL},
    {"Interface",          NULL,      NULL, NULL, NULL,                  PrintProtRule, 0, TYPE_STRING, 8,          8,          CONFIG_DISC | CONFIG_DONT_CLOSE_CELL},
    {"Inverted",           NULL,      NULL, NULL, (void *)&invlist,      NULL,          0, TYPE_YN,     2,          0,          CONFIG_DISC | CONFIG_TABLE | CONFIG_DONT_OPEN_CELL},
};

/* Table structure*/
struct entrylist ruletable = {
    "/etc/network/firewall", "Rule", "rule", "Rule", NULL,
    ruleparams, NULL, NULL, NULL, NULL, NULL, NULL,
    ParseRuleEntry, WriteRuleEntry, VerifyRuleEntry, NULL, NULL, NULL,
    256, 0, 0, sizeof(ruleparams)/sizeof(struct paramentry),
    CONFIG_ORDER|CONFIG_FILE, 0, 0, 0
};

/*****************************************************************************
 * Definitions for IP Chains Table
 */

/* List of valid entries */
unsigned char *defscopes[] = {"-", "ACCEPT", "DENY", "REJECT", "RETURN", "MASQ"};
struct strlist scopelistc = {defscopes,   defscopes,   1};
struct strlist scopelisti = {defscopes+1, defscopes+1, 4};
struct strlist scopelistf = {defscopes+1, defscopes+1, 5};

/* Backup */
static unsigned char *ipchains[64];

/* Routines */
static void *GetChainList ();
static int ParseChainEntry (unsigned char *s);
static void WriteChainEntry (FILE *arq, unsigned char **value);

/* Parameters */
struct paramentry chainparams[] = {
    {"Name"          , NULL, NULL, NULL, NULL,                 NULL, 0, TYPE_STRING,  30, 10, CONFIG_DONT_OPEN_CELL | CONFIG_DONT_CLOSE_CELL},
    {"Default Target", NULL, NULL, NULL, (void *)GetChainList, NULL, 0, TYPE_VARLIST, 16, 0,  0}
};

/* Table structure */
struct entrylist chaintable = {
    "/etc/network/firewall", "Chain", "chain", "Filter Chain", NULL,
    chainparams, NULL, chainpentries, &ruletable, NULL, NULL, NULL, 
    ParseChainEntry, WriteChainEntry, NULL, NULL, NULL, ipchains,
    64, 0, 0, sizeof(chainparams)/sizeof(struct paramentry), CONFIG_NAME|CONFIG_FILE, 0, 0, 0
};

/*****************************************************************************
 * Definitions for Users Group - Pslave Table
 */
static unsigned char hostname[40], SPUG_file[40],conf_incl,conf_find_host;

/* Routines */
static int ParseSPUGroupEntry (unsigned char *s);
static void WriteSPUGroupEntry (FILE *arq, unsigned char **value);

/* Parameters */
struct paramentry SPUGroupparams[] = {
    {"Group Name",  NULL, NULL, NULL, NULL, NULL, 0, TYPE_STRING, 20, 20, 0},
    {"Users",       NULL, NULL, NULL, NULL, NULL, 0, TYPE_STRING, 40, 40, 0}
};

/* Table structure */
struct entrylist SPUGrouptable = {
    "/etc/portslave/pslave.conf", "SPUGroup", "group", 
	"Serial Ports - Users Group", NULL,
    SPUGroupparams, NULL, NULL, NULL, NULL, NULL, NULL,
    ParseSPUGroupEntry, WriteSPUGroupEntry, NULL, NULL, NULL, NULL,
    256, 0, 0, sizeof(SPUGroupparams)/sizeof(struct paramentry),
    CONFIG_FILE, 0, 0, 0
};

/*****************************************************************************
 * Messages
 */
unsigned char msginvalidentry[] = "Invalid values entered. See parameters in red";
unsigned char msgchangedconfig[] = "Configuration changed since last page";
unsigned char msginvalidvariable[] = "Variable not found";
unsigned char msgnoresource[] = "Lack of resources";
unsigned char msgreadfileerror[] = "Invalid entries read from the file";
unsigned char msgtoomanyentries[] = "Number of entries reached the limit";
unsigned char msgentryalreadyexists[] = "Entry already exists";
unsigned char msginvaliddelete[] = "This entry cannot be deleted";
unsigned char msgnoentries[] = "No entry";
unsigned char msgnoentryselected[] = "No entry selected";
unsigned char msgporterr[] = "IPchains can only specify ports for icmp, tcp or udp";

/*****************************************************************************
 * AllocEntryAddresses - This routine allocates the buffers to read/write in
 * the file.
 */
static int AllocEntryAddresses (struct entrylist *table)
{
    FILE *arq;
    struct stat st;
    int size;
    table->buf=NULL;
    table->buf2=NULL;
    table->flag=0;
    /* Allocates the space for the pointers to the values of the parameters
     * of all the entries */
    table->value=(unsigned char **)malloc(table->maxentries*table->nparams * sizeof(unsigned char *));
    if (!table->value) return 0;
    /* Get file size, allocate a buffer and copies the contents of the file
     * to that buffer*/
    if (stat(table->filename, &st)) return 1;
    size=st.st_size+1;
    if (table->config & CONFIG_FILE) size*=2;
    if (!(table->buf=malloc(size))) return 0;
    if (!(arq = fopen(table->filename, "r"))) return 0;
    fread(table->buf, 1, st.st_size, arq);
    *(table->buf+st.st_size)=0;
    fclose(arq);
    /* If there are more than one table in the file, we need to use another
     * buffer for writing */
    if (table->config & CONFIG_FILE) {
        table->buf2 = table->buf+st.st_size+1;
        memcpy (table->buf2, table->buf, st.st_size+1);
    }
    return 1;
}

/*****************************************************************************
 * DeallocEntryAddresses - This routine deallocates the buffers allocated in
 * the routine above.
 */
static void DeallocEntryAddresses (struct entrylist *table)
{
    if (table->value) {
		sync();
        free(table->value);
        table->value=0;
    }
    if (table->buf) {
       free(table->buf);
        table->buf=0;
    }
}

/*****************************************************************************
 * GetStringList - This routine returns the structure with the list of strings.
 */
static struct strlist *GetStringList (struct paramentry *param)
{
    getlistroutine *listrt;
    switch (param->type) {
    case TYPE_VARLIST:
        /* Dynamic list - Call the routine which returns the list */
        listrt = (getlistroutine *)(param->list);
        return (struct strlist *)((*listrt) ());
    case TYPE_NBLIST:
        /* List or number - The pointer must be shifted in 2 integers */
        return (struct strlist *)((unsigned int *)(param->list) + 2);
    default:
        return (struct strlist *)(param->list);
    }
}

/*****************************************************************************
 * VerifyEntryParamValue - This routine checks whether the parameter read is
 * valid.
 */
static int VerifyEntryParamValue (struct entrylist *table, int index)
{
    struct in_addr b;
    unsigned int mask;
    int j;
    unsigned char k;
    unsigned char *value=(table->param+index)->value2;
    unsigned char type = (table->param+index)->type;
    struct strlist *tbllist = GetStringList(table->param+index);

    /* Consistency for null or empty values */
    if (!value || !(*value)) {
        if ((table->param+index)->config & CONFIG_DISC) {
            return 1;
        } else {
            return 0;
        }
    }

    switch (type) {
    case TYPE_STRING:
        /* Type string - any value is OK.*/
        return 1;

    case TYPE_IPADDR:
        /* Type IP Address - tries to convert the string in an IP address */
        if (inet_aton(value, &b)) return 1;
        break;

    case TYPE_IPMASK:
        /* Type IP mask - After converting in an IP address, verifies whether
         * is is a mask */
        if (inet_aton(value, &b)) {
            mask=0;
            for (j=0; j<4; j++) {
                for (k=0x80; k; k/=2) {
                    if (b.s_addr == mask) return 1;
                    *((unsigned char *)&mask + j) += k;
                }
            }
            if (b.s_addr == mask) return 1;
        }
        break;

    case TYPE_RADIO2:
        /* Type RADIO for the option 2, 3 .. - it should be null */
        break;

    case TYPE_RADIO1:
    case TYPE_LIST:
    case TYPE_NBLIST:
    case TYPE_VARLIST:
        /* Type List - Checks whether the value matches one of the valid
         * entries*/
        for (j=0; j<tbllist->nlist; j++) {
            if (strcmp(tbllist->filelist[j], value)) continue;
            return 1;
        }
        if (type != TYPE_NBLIST) break;

    case TYPE_NUMBER:
        /* Type number - Tries to convert the string into a number */
        if (!sscanf (value, "%d", &mask)) return 0;
        if (mask >= *(int *)((table->param+index)->list) &&
            mask <= *((int *)((table->param+index)->list)+1)) return 1;
        break;

    case TYPE_YN:
        /* Type yes/no - The value should be either null(no) or the entry for
         "yes" */
        if (strcmp(tbllist->filelist[0], value) == 0) {
            return 1;
        }

    case TYPE_DEF_IP: //[RK] default
	if (strcasecmp(value,"default") == 0) {
		return 1;
	}
       	/* Type IP Address - tries to convert the string in an IP address */
       	if (inet_aton(value, &b)) return 1;
        break;

    default:
        break;
    }
    return 0;
}

/*****************************************************************************
 * VerifyEntryParamValue - This routine checks whether the parameter read is
 * valid.
 */
static void ReadEntryTableFromFile (struct entrylist *table)
{
    unsigned int nparam;
    unsigned char *s, *t, **p, **v;
    int i, j;

    t=table->buf;
    table->nentries=0;
    /* Tries to use the permanent values as a beginning; otherwise, consider
     * the table empty */
    if (table->permentries) {
        nparam=0;
        p = table->permentries;
        v = table->value;
        while (*p) {
            *v++=*p++;
            if (++nparam == table->nparams) {
                if (table->backup) {
                    /* This part is to store the entries for the future */
                    if (*(table->backup+table->nentries)) {
                        free (*(table->backup+table->nentries));
                    }
                    *(table->backup+table->nentries) = strdup (*(v-table->nparams));
                }
                table->nentries++;
                nparam=0;
            }
        }
    }
    if (!t) return;
    while (*t) {
        /* Gets one line, which could be an entry. */
        s=t;
        while (*t && *t != '\n') t++;
        if (*t) *t++=0;
	/* Skips the blank lines */
	if (*s == 0) continue;
        /* Parses the line and gets the values for each parameter */
        if ((*(table->parse))(s) != 1) continue;
        /* Now check the validity of each parameter. If they are OK, consider
         * it as one entry */
        for (i=0; i<table->nparams; i++) {
            if (VerifyEntryParamValue (table, i)) {
                table->value[table->nentries*table->nparams+i]=(table->param+i)->value2;
            } else {
                table->flag |= FLAG_ERROR;
                table->msg = msgreadfileerror;
                break;
            }
        }
        if (i!=table->nparams) continue;
        /* If the entry is identified by name, checks whether the entry already exists.
         * If so, set up the entry. */
        if (table->config & CONFIG_NAME) {
            for (j=0; j<table->nentries; j++) {
                if (strcmp(table->value[j*table->nparams], table->param->value2) == 0) {
                    for (i=1; i<table->nparams; i++) {
                        table->value[j*table->nparams+i]=(table->param+i)->value2;
                    }
                    break;
                }
            }
            if (j!=table->nentries) {
                continue;
            }
        }
        /* This part is to store the entries for the future */
        if (table->backup) {
            if (*(table->backup+table->nentries)) {
                free (*(table->backup+table->nentries));
            }
            *(table->backup+table->nentries) = strdup ((table->param)->value2);
        }
        /* Check the number of entries */
        table->nentries++;
        if (table->nentries==table->maxentries) {
            table->msg = msgtoomanyentries;
            break;
        }
    }
}

/*****************************************************************************
 * SaveEntryTableToFile - This routine saves the entries in the memory to
 * the file.
 */
static void SaveEntryTableToFile (struct entrylist *table)
{
    FILE *arq;
    int i, nentries, flag=0;
    unsigned char *s, *t, line[256];
    arq = fopen(table->filename, "w");
	rewind(arq);
    if (table->config & CONFIG_FILE) {
        /* If the file contains more than one table, we should modify only
         * the lines with the current table. The other lines should be just
         * copied */
        t=table->buf2;
        nentries=0;
        while (t && *t) {
            s=t;
            while (*t && *t != '\n') t++;
            /* We don't want the buf2 to be used for parsing, because
             * we could use it to write in the file. So, we will copy the
             * line to another buffer. */
            strncpy (line, s, 255);
            if (t-s<255 && *t) {
                line[t-s] = 0;
                t++;
            }

            switch ((*(table->parse))(line)) {
            case 1:
                /* This line is from this table -  ignore it and writes the
                 * table, if it is the first occurence. */
                if (!flag) {
                    for (i=0; i<table->nentries; i++) {
                        (*(table->write))(arq, table->value+i*table->nparams);
                    }
                    flag=1;
                }
                break;
            case 2:
                /* This line is not from this table, but is related to the
                 * reference value. If the command is to delete the entry,
                 * don't write it on the file */
                if (table->flag & FLAG_DELETE) break;
            case 0:
                /* This line has nothing  to do with the table: just writes it.
                 */
                fwrite(s, 1, t-s, arq);
                break;
            }
        }
    }
    /* If the table was not found, let's write it in the end. */
    /* If the file has only this table, we just overwrite the table
     * from the buffer */
    if (!flag) {
        for (i=0; i<table->nentries; i++) {
            (*(table->write))(arq, table->value+i*table->nparams);
        }
    }
	fflush(arq);
    fclose(arq);
}

/*****************************************************************************
 * PrintEntryParam - This routine prints in the page the inputs for a
 * parameter.
 */
static void PrintEntryParam (webs_t wp, struct entrylist *table, int i)
{
    struct paramentry *param = table->param + i;
    struct strlist *tbllist = GetStringList(param);

    unsigned char *s;
    int j, idx, selected;

    /* Specific case of RADIO2 type - use the first option as index */
    if (param->type == TYPE_RADIO2) {
        idx = table->radio;
    } else {
        idx = i;
    }
    table->flag &= ~FLAG_ERROR;

    /* Converts a null value to an empty value (""). */
    if (!param->value) param->value="";

    /* Writes the "space" before writing the entry. If it is null, use the
     * default value. */
//    if (param->space) {
//       websWrite(wp, param->space);
//    } else {
//        if (i!=0) {
//            websWrite(wp, "<br>");
//        }
//    }

    if (!(param->config & CONFIG_DONT_OPEN_CELL)) {
    	websWrite(wp, "<tr><td>");
    } else if (param->type == TYPE_RADIO2) {
    	websWrite(wp, "<br>");
    }
    if (param->type != TYPE_RADIO2 && param->type != TYPE_YN && param->name) {
        websWrite(wp, T("<b%s>%s:</b>&nbsp;"), (param->flag & FLAG_ERROR)?" class=\"err\"": "", param->name);
    }
    if (!(param->config & CONFIG_DONT_OPEN_CELL)) {
    	websWrite(wp, "</td><td>");
    }

    switch (param->type) {

    case TYPE_VARLIST:
    case TYPE_LIST:
    case TYPE_NBLIST:
        /* For lists: select */
        websWrite(wp, T("<select name=\"%02x\">"), idx);
        selected = 0;
        for (j=0; j<tbllist->nlist; j++) {
			if (strcmp(param->value, tbllist->filelist[j]) == 0) {
                s = " selected";
                selected = 1;
            } else {
            	s = "";
            }
            websWrite(wp, T("<option value=\"%s\"%s>%s"), tbllist->filelist[j],
            	s, tbllist->showlist[j]);
        }
        if (param->type != TYPE_NBLIST) {
            websWrite(wp, T("</select>"));
            break;
        }
        websWrite(wp, T("<option value=\"%s\""), "other");
        s = (selected)?"":" selected";
        websWrite(wp, T("<option value=\"%s\"%s>%s"), "other",
            	s, "other");
        websWrite(wp, T("</select>"));
        websWrite(wp, T("<input class=\"tx\" type=\"text\" name=\"%02xa\" value=\"%s\" size=\"%d\">"),
                  idx, param->value, param->field_size);
        break;

    case TYPE_IPADDR:
    case TYPE_DEF_IP:
    case TYPE_IPMASK:
    case TYPE_STRING:
    case TYPE_NUMBER:
        /* For most of the types : input type text */
        websWrite(wp, T("<input class=\"tx\" type=\"text\" name=\"%02x\" value=\"%s\" size=\"%d\">"),
                  idx, param->value, param->field_size);
        break;


    case TYPE_RADIO1:
        /* For radio types - input type radio. */
        table->radio=i;
        table->radioindex=0;
    case TYPE_RADIO2:
        tbllist=param->list;
        websWrite(wp, T("<input type=\"radio\" name=\"%02x\" value=\"%s\"%s>"),
                  idx, tbllist->filelist[table->radioindex],
                  (strcmp((table->param+table->radio)->value, tbllist->filelist[table->radioindex]))?"":" checked");
        table->radioindex++;
        break;

    case TYPE_YN:
        /* For yes/no type - input type checkbox */
        tbllist=param->list;
        websWrite(wp, T("<input type=\"checkbox\" name=\"%02x\" value=\"%s\"%s>"),
                  idx, tbllist->filelist[0], (strcmp(param->value, tbllist->filelist[0]))?"":" checked");
        break;
    }
    if ((param->type == TYPE_RADIO2 || param->type == TYPE_YN) && (param->name)) {
        websWrite(wp, T("&nbsp;<b%s>%s</b>"), (param->flag & FLAG_ERROR)?" class=\"err\"": "", param->name);
    }
    if (!(param->config & CONFIG_DONT_CLOSE_CELL)) {
	    websWrite(wp, "</td></tr>");
    } else if (param->type == TYPE_RADIO1 || param->type == TYPE_RADIO2) {
	    websWrite(wp, "&nbsp;");
    } else {
	    websWrite(wp, "&nbsp;&nbsp;&nbsp;&nbsp;");
    }
    param->flag &= ~FLAG_ERROR;
}

/*****************************************************************************
 * PrintEntry - This routine creates the web page to input the entries.
 */
static void PrintEntry (webs_t wp, struct entrylist *table)
{
    int i;
    struct entrylist *tableaux;
    unsigned char title[128], links[256], form[64], addheader[128], father[64];
	struct cyc_webs_header hdr;
	
    sprintf(title, "%s %s Table Entry", (table->flag & FLAG_INSERT)?"Add":"Edit", table->title);

    links[0]=0;
    for (tableaux=table; tableaux->prevtable!=NULL; tableaux=tableaux->prevtable);
    sprintf(father, "/goform/List%sTable", tableaux->name);
    for (;; tableaux=tableaux->nexttable) {
        sprintf(links+strlen(links), "&nbsp;<a href=\"/goform/");
        if (tableaux->prevtable) {
            sprintf(links+strlen(links), "cfg/Write%sTable?%s\"", tableaux->prevtable->name, tableaux->prevtable->getentries);
        } else {
            sprintf(links+strlen(links), "List%sTable\"", tableaux->name);
        }
        sprintf(links+strlen(links), ">[%s Table]</a>", tableaux->title);
        if (tableaux==table) break;
    }

    sprintf(form, "cfg/Set%sEntry", table->name);

    addheader[0] = 0;
    if (table->prevtable) {
        sprintf(addheader, "<br>%s - %s", (table->prevtable)->title, (table->prevtable)->nextvalue);
    }

	websCycHdrInit(&hdr);
	hdr.need_header = 1;
	hdr.title = title;
	hdr.form = form;
	hdr.links = links;
	hdr.father = father;
	websPageStart(wp, &hdr);

    /* Error message */
    if (table->msg) {
        websWrite(wp, T("<center><b class=err>Error - %s</b></center><br>"), table->msg);
        table->msg=0;
    }
    /* Opens the window for the entries */
    websWrite(wp, T("<center><table border align=center cols=1><tr><th class=\"tt\" colspan=\"2\">%s "),
    	title);
    if (table->config & CONFIG_NAME) {
        /* If the table is indexed by name, don't use it as input. Just shows it*/
        (table->param)->value2 = (table->param)->value;
        websWrite(wp, T("(%s: %s)<input type=\"hidden\" name=\"entry\" value=\"%s\">"),
        	(table->param)->name, (table->param)->value, (table->param)->value);
        i=1;
    } else {
       	/* If the table is indexed by number, shows it */
       	websWrite(wp, T("#%d<input type=\"hidden\" name=\"entry\" value=\"%d\">"),
			table->currentry+1, table->currentry);
       	i=0;
    }
   	websWrite(wp, T("</th></tr><input type=\"hidden\" name=\"insert\" value=\"%c\">"),
    		(table->flag & FLAG_INSERT)?'a':'e');
	for (tableaux=table->prevtable; tableaux; tableaux=tableaux->prevtable) {
		websWrite(wp, T("<input type=\"hidden\" name=\"chain%s\" value=\"%s\">"), tableaux->name, tableaux->nextvalue);
	}

    /* Print the inputs for each parameter */
    for (;i<table->nparams; i++) {
       	PrintEntryParam(wp, table, i);
    }
    /* Closes the window and inserts the submit button */
    websWrite(wp, T("</table></center><br><center>"
    	"<input type=\"submit\" name=\"opt\" value=\"Submit\">"
    	"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"
    	"<input type=\"submit\" name=\"opt\" value=\"Cancel\">"
    	"</center>"));
	websPageEnd(wp, &hdr);
}

/*****************************************************************************
 * PrintTable - This routine creates a web page with the table and the options
 * to add/edit/delete.
 */
static int PrintTable (webs_t wp, struct entrylist *table)
{
    struct entrylist *tableaux;
    unsigned char title[128], links[256], form[64], addheader[128], father[64];
    int i, j, k;
    unsigned char aux=0, *value, *message = table->msg;
    struct cyc_webs_header hdr;
    table->msg = 0;
    /* Allocates the buffers and read the entries */
    if (!AllocEntryAddresses (table)) {
        DeallocEntryAddresses (table);
        return 0;
    }
    ReadEntryTableFromFile (table);

    /* If there are no entries, don't show the table: go straight to the
     entry page as if adding the entry. */
    if (!table->nentries) {
        table->msg = msgnoentries;
        //RK table->flag |= FLAG_INSERT;
        //RK table->currentry = 0;
        //RK for (i=0; i<table->nparams; i++) {
        //RK     (table->param+i)->value=(table->param+i)->defvalue;
       //RK  }
        //RK PrintEntry (wp, table);
        //RK DeallocEntryAddresses (table);
       //RK  return 1;
    }

    sprintf(title, T("%s Table"), table->title);
    links[0]=0;
	for (tableaux=table; tableaux->prevtable!=NULL; tableaux=tableaux->prevtable);
    sprintf(father, "/goform/List%sTable", tableaux->name);
    for (; tableaux!=table; tableaux=tableaux->nexttable) {
	    sprintf(links+strlen(links), "&nbsp;<a href=\"/goform/");
        if (tableaux->prevtable) {
            sprintf(links+strlen(links), T("cfg/Write%sTable?%s\""), tableaux->prevtable->name, tableaux->prevtable->getentries);
        } else {
            sprintf(links+strlen(links), T("List%sTable\""), tableaux->name);
        }
        sprintf(links+strlen(links), T(">[%s Table]</a>"), tableaux->title);
    }
    sprintf(form, "%sWrite%sTable", (table->nexttable)?"":"cfg/", table->name);
    addheader[0]=0;
    if (table->prevtable) {
        sprintf(addheader, "%s - %s", (table->prevtable)->title, (table->prevtable)->nextvalue);
    }
	
        websCycHdrInit(&hdr);
        hdr.need_header = 1;
        hdr.title = title;
        hdr.form = form;
        hdr.links = links;
        hdr.father = father;
        websPageStart(wp, &hdr);

    /* Message */
	if (!table->msg) table->msg=message;
    if (table->msg) {
        websWrite(wp, T("<center><b class=err>Error - %s</b></center><br>"), table->msg);
        table->msg=0;
    }

	if (table->config & CONFIG_READ_ONLY) {
    	if (table->nentries) {
	   	 	/* Opens the window */
   	 		websWrite(wp, T("<center><table border align=center>"));
    		websWrite(wp, T("<tr><th class=\"tt\" colspan=\"%d\">%s</th></tr>")
							, 3, title);
           	websWrite(wp, T("<th class=\"st\">&nbsp</th>"));
    		/* Writes the table headers */
    		for (i=0; i<table->nparams; i++) {
            	websWrite(wp, T("<th class=\"st\">%s</th>"), (table->param+i)->name);
			}
    		websWrite(wp, T("</tr>"));
    		for (i=0; i<table->nentries; i++) {
        		websWrite(wp, T("<tr><td>%d</td>"),i+1);
        		for (j=0; j<table->nparams; j++) {
            		value = table->value[i*table->nparams+j];
            		if (!value) value=&aux;
            		websWrite(wp, T("<td>%s</td>"), value);
        		}
        		websWrite(wp, T("</tr>"));
			}
    		/* Close the table */
    		websWrite(wp, T("</table></center>"));
		}
		websPageEnd(wp, &hdr);
    	DeallocEntryAddresses (table);
    	return 1;
	}

   	websWrite(wp, T("<input type=\"hidden\" name=\"chain\" value=\""));
	for (tableaux=table->prevtable; tableaux; tableaux=tableaux->prevtable) {
		websWrite(wp, T("%s\\"), tableaux->nextvalue);
	}
	websWrite(wp, T("\">"), i+1);
    if (table->nentries) {
	    /* Opens the window */
   	 	websWrite(wp, T("<center><table border align=center>"));
    	/* Writes the table headers */
    	j=1;
    	for (i=0; i<table->nparams; i++) {
        	if (!((table->param+i)->config & CONFIG_TABLE))
        		j++;
    	}
    	websWrite(wp, T("<tr><th class=\"tt\" colspan=\"%d\">%s</th></tr>"), j, title);
    	websWrite(wp, T("<tr><th class=\"st\">Entry</th>"));
    	for (i=0; i<table->nparams; i++) {
        	if (!((table->param+i)->config & CONFIG_TABLE))
            	websWrite(wp, T("<th class=\"st\">%s</th>"), (table->param+i)->name);
    	}
    	websWrite(wp, T("</tr>"));
    	/* Writes the entries */
    	for (i=0; i<table->nentries; i++) {
        	websWrite(wp, T("<tr>"));
		if ((websGetPriority(wp) == PRIV_FULL && 
			((table->config & CONFIG_READ_ONLY) == 0)) ||
				table->nexttable) {
	        	websWrite(wp, T("<td><input type=\"radio\" name=\"entry\" value=\"%d\\"), i);
        		for (j=0; j<table->nparams; j++) {
            			value = table->value[i*table->nparams+j];
		            	if (!value) value=&aux;
            			websWrite(wp, T("%s\\"), value);
        		}
//        		websWrite (wp, T("\""));
//        		if (i==0) websWrite (wp, T(" checked"));
        		websWrite(wp, T("\">&nbsp;%d</td>"), i+1);
		} else {
			websWrite(wp, T("<td>%d</td>"), i+1);
		}
        	for (j=0; j<table->nparams; j++) {
            	if (!((table->param+j)->config & CONFIG_TABLE)) {
                	int type;
                	value = table->value[i*table->nparams+j];
                	if (!value) value=&aux;
                	/* For parameters type list, try to show the value in "showlist"*/
                	type = (table->param+j)->type;
                	if (type == TYPE_LIST || type == TYPE_VARLIST || type == TYPE_NBLIST) {
                    	struct strlist *tbllist = GetStringList(table->param+j);
                    	int k;
                    	for (k=0; k<tbllist->nlist; k++) {
                        	if (strcmp(value, tbllist->filelist[k])==0) {
                            	value = tbllist->showlist[k];
                            	break;
                        	}
                    	}
                	}
                	websWrite(wp, T("<td nowrap>"));
                	if ((table->param+j)->print) {
                    	k = (*((table->param+j)->print)) (wp, value, table->value+i*table->nparams+j);
                	} else {
                		k = websWrite(wp, value);
                	}
                	if (k==0) websWrite(wp, "&nbsp;");
                	websWrite(wp, T("</td>"));
            	}
        	}
        	websWrite(wp, T("</tr>"));
    	}
    	/* Close the table */
    	websWrite(wp, T("</table></center>"));
	}
    /* What to do with the entry selected (Add/Edit/Delete...) */
    websWrite(wp, T("<center><br>"));
    if (table->nexttable) {
        websWrite(wp, T("&nbsp;&nbsp;&nbsp;&nbsp;"
        	"<input type=\"submit\" name=\"opt\" value=\""));
        websWrite(wp, T("List %ss\">"), (table->nexttable)->entryname);
    }
    if (websGetPriority(wp) == PRIV_FULL) {
	if (table->nentries) {
	    websWrite(wp, T("&nbsp;&nbsp;&nbsp;&nbsp;"
		"<input type=\"submit\" name=\"opt\" value=\""));
	    websWrite(wp, T("Edit %s\" checked>"), table->entryname);
	    websWrite(wp, T("&nbsp;&nbsp;&nbsp;&nbsp;"
		"<input type=\"submit\" name=\"opt\" value=\""));
	    websWrite(wp, T("Delete %s\">"), table->entryname);
	}
	websWrite(wp, "&nbsp;&nbsp;&nbsp;&nbsp;"
	    "<input type=\"submit\" name=\"opt\" value=\"");
	websWrite(wp, T("Add %s"), table->entryname);
	if (table->config & CONFIG_ORDER) {
	    websWrite(wp, T(" in line\">&nbsp;<select name=\"line\">"));
	    for (i=0; i<table->nentries+1; i++) {
		websWrite(wp, T("<option value=\"%d\""), i);
		if (i==table->nentries) websWrite(wp, T(" selected"));
		websWrite(wp, T(">%d"), i+1);
	    }
	    websWrite(wp, T("</select>"));
	} else {
	    websWrite(wp, T("\">"));
	}
	if (table->config & CONFIG_NAME) {
	    websWrite(wp, T("/"));
	    table->flag |= FLAG_INSERT;
	    table->flag &= ~FLAG_ERROR;
	    (table->param)->value = NULL;
	    PrintEntryParam(wp, table, 0);
	}
    }
    /* Submit button */
    websWrite(wp, T("</center>"));
	websPageEnd(wp, &hdr);
    //DeallocEntryAddresses (&hosttable);
    DeallocEntryAddresses (table);
    return 1;
}

/*****************************************************************************
 * ListTableAction - This routine is called when the "table page" entries
 * are submitted.
 */
static void ListTableAction(webs_t wp, struct entrylist *table, unsigned char *query)
{
    unsigned char *option, *entry, *line, *name, *chain, *s;
    unsigned char buf[512], buf2[256];
    int i, j;
    struct entrylist *tableaux;
    /* copies the query. It will be useful to write in the links */
    if (table->nexttable) {
        if (table->getentries) {
            free(table->getentries);
        }
        table->getentries = strdup(query);
    }
    option = websGetVar(wp, T("opt"), NULL);
    entry = websGetVar(wp, T("entry"), NULL);
    line = websGetVar(wp, T("line"), NULL);
    name = websGetVar(wp, T("00"), NULL);
    chain = websGetVar(wp, T("chain"), NULL);

    /* The variables option and chain are mandatory */
    if (!option || !chain) {
        table->msg = msginvalidvariable;
        PrintTable (wp, table);
        return;
    }

    /* Copy the entry in a buffer */
    strncpy(buf2, chain, 256);

   	/* Parses the buffer and assign values for the parameters */
    s=buf2;
    for (tableaux=table->prevtable; tableaux; tableaux=tableaux->prevtable) {
   	    tableaux->nextvalue = s;
       	while (*s && *s!='\\') s++;
        if (*s) *s++=0;
   	}

    /* For options which are not Add the variable entry is mandatory */
    if (!entry && *option != 'A') {
        table->msg = msgnoentryselected;
   	    PrintTable (wp, table);
       	return;
	}

	if (*option == 'A') {
	    for (i=0; i<table->nparams; i++)
	    	(table->param+i)->value = (table->param+i)->defvalue;
	} else {
	    /* Copy the entry in a buffer */
	    strncpy(buf, entry, 512);
    	/* Parses the buffer and assign values for the parameters */
	    s=buf;
    	while (*s && *s!='\\') s++;
	    if (*s) *s++=0;
    	sscanf(buf, "%d", &(table->currentry));
//	    for (tableaux=table->prevtable; tableaux; tableaux=tableaux->prevtable) {
//   	    tableaux->nextvalue = s;
//      	while (*s && *s!='\\') s++;
//	        if (*s) *s++=0;
//   	}
	    for (i=0; i<table->nparams; i++) {
    	    (table->param+i)->value = s;
        	while (*s && *s!='\\') s++;
	        if (*s) *s++=0;
    	}
 	}

    switch (*option) {
    case 'L':
        /* Option to go to the next table */
        if (!table->nexttable) {
            table->msg=msginvalidvariable;
            PrintTable(wp, table);
            return;
        }
        /* Copies the name of the entry */
        table->nextvalue = (table->param)->value;
        /* Links the next table with the current one */
        (table->nexttable)->prevtable = table;
        /* Switches the the next table */
        PrintTable(wp, table->nexttable);
        return;

    case 'A':
        /* Add entry */
        table->flag |= FLAG_INSERT;
        /* If order is important, we need to know the number of line in
         which the entry will be inserted; otherwise we will append it */
        if (table->config & CONFIG_ORDER) {
            if (!line || !sscanf(line, "%d", &(table->currentry))) return;
        } else {
            table->currentry=table->nentries;
        }
        /* If indexed on name, we need to get the name of the entry */
        if (table->config & CONFIG_NAME) {
            if (name && *name) {
                (table->param)->value = name;
            } else {
                (table->param)->flag |= FLAG_ERROR;
                table->msg = msginvalidentry;
                PrintTable(wp, table);
                return;
            }
        }
        PrintEntry (wp, table);
        return;

    case 'E':
        /* Edit entry */
        table->flag &= ~FLAG_INSERT;
        PrintEntry (wp, table);
        return;

    case 'D':
        /* Delete Entry */
        /* First Allocate and read the entries */
        if (!AllocEntryAddresses (table)) {
            table->msg = msgnoresource;
            DeallocEntryAddresses (table);
            PrintTable(wp, table);
            return;
        }
        ReadEntryTableFromFile (table);

        /* Verifies if the entry is still in the file. */
        for (i=0; i<table->nparams; i++) {
            if ( !(*((table->param+i)->value)) &&
                table->value[table->currentry*table->nparams+i] == NULL) continue;
            if (strcmp(table->value[table->currentry*table->nparams+i],
                       (table->param+i)->value) == 0) continue;
            table->msg = msgchangedconfig;
            DeallocEntryAddresses (table);
            PrintTable(wp, table);
            return;
        }
        /* The permanent entries cannot be deleted */
        if (table->permentries) {
            unsigned char **p = table->permentries, nentries = 0, nparam = 0;
            while (*p) {
                if (nparam==0) {
                    if (strcmp((table->param)->value, *p) == 0) {
                        table->msg = msginvaliddelete;
                        DeallocEntryAddresses (table);
                        PrintTable(wp, table);
                        return;
                    }
                }
                if (++nparam == table->nparams) {
                    nparam=0;
                    nentries++;
                }
                p++;
            }
        }
        /* Deletes the entry and saves it in the file */
        for (j=table->currentry+1; j<table->nentries; j++) {
            for (i=0; i<table->nparams; i++) {
                table->value[(j-1)*table->nparams+i] = table->value[j*table->nparams+i];
            }
        }
        table->nentries--;
        table->flag |= FLAG_DELETE;
        SaveEntryTableToFile (table);
        table->flag &= ~FLAG_DELETE;
        DeallocEntryAddresses (table);
        PrintTable(wp, table);
        return;
    }
}

/*****************************************************************************
 * EntryTableAction - This routine is called when the "entry page" entries
 * are submitted
 */
static void EntryTableAction (webs_t wp, struct entrylist *table, unsigned char *query)
{
    int i, j, idx;
    unsigned char *value, flag, entry[4], entryname[32], *msg;
    struct entrylist *tableaux;

	value = websGetVar(wp, "opt", "C");
	if (*value == 'C') {
		PrintTable(wp, table);
		return;
	}
    value = websGetVar(wp, "insert", NULL);
    if (value && *value == 'a') {
        table->flag |= FLAG_INSERT;
    } else if (value && *value == 'e') {
        table->flag &= ~FLAG_INSERT;
    } else {
        table->msg=msginvalidvariable;
        PrintTable(wp, table);
        return;
    }

    for (tableaux=table->prevtable; tableaux; tableaux=tableaux->prevtable) {
    	sprintf(entryname, "chain%s", tableaux->name);
	    tableaux->nextvalue = websGetVar(wp, entryname, "");
    }

    value=websGetVar(wp, "entry", "");
    if (table->config & CONFIG_NAME) {
        (table->param)->value = (table->param)->value2 = value;
        idx=1;
    } else {
        if (sscanf (value, "%d", &(table->currentry))) {
	    	idx=0;
	    } else {
            table->msg = msginvalidvariable;
            PrintTable(wp, table);
            return;
	    }
    }

    /* Gets the value of each parameter, written by the user. Checks the
     * validity of them. The parameter with invalid value will be marked. */
    for (i=idx; i<table->nparams; i++) {
        sprintf(entry, "%02x", i);
        (table->param+i)->value2 = websGetVar(wp, entry, "");
        /* When the type is List or Number, the option "other" will cause
         the value to be the one written in the second input. */
        if ((table->param+i)->type==TYPE_NBLIST &&
            strcmp((table->param+i)->value2, "other") == 0) {
            sprintf(entry, "%02xa", i);
            (table->param+i)->value2 = websGetVar(wp, entry, "");
        }
        (table->param+i)->value = (table->param+i)->value2;
        if (!VerifyEntryParamValue (table, i)) {
            (table->param+i)->flag |= FLAG_ERROR;
            table->flag |= FLAG_ERROR;
            continue;
        }
    }
    if (table->flag & FLAG_ERROR) {
        table->msg=msginvalidentry;
        PrintEntry(wp, table);
        return;
    }
    if (table->verify) {
        if ((msg = (*(table->verify))(&(table->param->value)) ) != NULL) {
            table->msg=msg;
            PrintEntry(wp, table);
            return;
        }
    }
    /* As the allocation resets the flag, save it before and load it later */
    flag=table->flag;
    /* Allocate buffers and read the file */
    if (!AllocEntryAddresses (table)) {
        table->msg = msgnoresource;
        DeallocEntryAddresses (table);
        PrintTable(wp, table);
        return;
    }
    ReadEntryTableFromFile (table);
    table->flag=flag;

    /* Checks whether we cannot insert more entries. */
    if ((table->flag & FLAG_INSERT) && table->nentries >= table->maxentries) {
        table->msg = msgtoomanyentries;
        DeallocEntryAddresses (table);
        PrintTable(wp, table);
        return;
    }
    /* If the table is indexed by name, the entry name cannot exist in the
     * file, if it is for "add" and must be there if it is for "edit" */
    if (table->config & CONFIG_NAME) {
        for (j=0; j<table->nentries; j++) {
            if (strcmp(table->value[j*table->nparams],(table->param)->value)==0) {
                break;
            }
        }
        table->currentry = j;
        if ((j==table->nentries && !(table->flag & FLAG_INSERT)) ||
            (j<table->nentries && (table->flag & FLAG_INSERT))) {
            table->msg=msgchangedconfig;
            DeallocEntryAddresses (table);
            PrintTable(wp, table);
            return;
        }
    }
    /* Shift the entries if we are adding an entry. */
    if (table->flag & FLAG_INSERT) {
        for (j=table->nentries; j>table->currentry; j--) {
            for (i=0; i<table->nparams; i++) {
                table->value[j*table->nparams+i] = table->value[(j-1)*table->nparams+i];
            }
        }
        table->nentries++;
    }
    /* Set the entry in the list */
    for (i=0; i<table->nparams; i++) {
        table->value[table->currentry*table->nparams+i] = (table->param+i)->value;
    }
    /* Save the entries in the file */
    SaveEntryTableToFile (table);
    DeallocEntryAddresses (table);
    PrintTable(wp, table);
}

/*============================================================================
 * Routines related to Host Table
 *============================================================================
 */

/*****************************************************************************
 * ParseHostEntry - Parse a host entry. The syntax of a host entry line is:
 *          <ip address> <host name> <aliases>
 */
static int ParseHostEntry (unsigned char *s)
{
    struct paramentry *param=hosttable.param;

    while (isspace(*s)) s++;
    param->value2=s;
    while (*s && !isspace(*s)) s++;
    if (*s) *s++=0;
    while (isspace(*s)) s++;
    (param+1)->value2=s;
    while (*s && !isspace(*s)) s++;
    if (*s) *s++=0;
//  while (isspace(*s)) s++;
//  (param+2)->value2=s;
    return 1;
}

/*****************************************************************************
 * WriteHostEntry - Writes a host entry.
 */
static void WriteHostEntry (FILE *arq, unsigned char **value)
{
//  fprintf (arq, "%s\t%s\t%s\n", *value, *(value+1), *(value+2));
    fprintf (arq, "%s\t%s\n", *value, *(value+1));
}

/*****************************************************************************
 * ListHostTable - Prints the host table in the web page
 */
static void ListHostTable(webs_t wp, char_t *path, char_t *query)
{
    PrintTable (wp, &hosttable);
}

/*****************************************************************************
 * WriteHostTable - Answer to the submission of the host table page
 */
static void WriteHostTable(webs_t wp, char_t *path, char_t *query)
{
    ListTableAction(wp, &hosttable, query);
}

/*****************************************************************************
 * SetHostEntry - Answer to the submission of the host table entry page
 */
static void SetHostEntry(webs_t wp, char_t *path, char_t *query)
{
    EntryTableAction (wp, &hosttable, query);
}

/*============================================================================
 * Routines related to Static Route Table
 *============================================================================
 */

/*****************************************************************************
 * ParseRouteEntry - Parse a host entry. The syntax of a host entry line is:
 *          net <Destination ip> mask <Destination mask> gw <gateway IP>
 * or
 *          net <Destination ip> mask <Destination mask> dev <interface>
 */
static int ParseRouteEntry (unsigned char *s)
{
    int i;
    unsigned char *word, gw;
    struct paramentry *param=routetable.param;

    for (i=0; i<6; i++) (param+i)->value2=NULL;
    for (i=0; i<8; i++) {
        while (isspace(*s)) s++;
        word=s;
        while (*s && !isspace(*s)) s++;
        if (*s) *s++=0;
        switch (i) {
        case 2:
        	if (strcmp(word,"default")==0) {
        		param->value2=word;
        		(param+1)->value2=NULL;
        		i=5;
        	}
        	break;
        case 3:
            param->value2=word;
            break;
        case 5:
            (param+1)->value2=word;
            break;
        case 6:
            if (strcmp(word,"gw")==0) {
                gw=1;
            } else {
                gw=0;
            }
            (param+2)->value2=word;
            (param+4)->value2=NULL;
            break;
        case 7:
            if (gw) {
                (param+3)->value2=word;
                (param+5)->value2=word+strlen(word);
            } else {
                (param+3)->value2=word+strlen(word);
                (param+5)->value2=word;
            }
            break;
        }
    }
    return 1;
}

/*****************************************************************************
 * WriteRouteEntry - Writes a route entry.
 */
static void WriteRouteEntry (FILE *arq, unsigned char **value)
{
	if (strcasecmp(*value,"default") == 0) {
		fprintf(arq,"route add default ");
	} else {
    fprintf (arq, "route add -net %s netmask %s", *value, *(value+1));
	}
    if (strcmp(*(value+2), "gw") == 0) {
        fprintf (arq, " gw %s\n", *(value+3));
    } else {
        fprintf (arq, " dev %s\n", *(value+5));
    }
}

/*****************************************************************************
 * ListRouteTable - Prints the route table.
 */
static void ListRouteTable(webs_t wp, char_t *path, char_t *query)
{
    PrintTable (wp, &routetable);
}

/*****************************************************************************
 * WriteRouteTable - Answer to the submission of the route table page.
 */
static void WriteRouteTable(webs_t wp, char_t *path, char_t *query)
{
    ListTableAction(wp, &routetable, query);
}

/*****************************************************************************
 * SetRouteEntry - Answer to the submission of the route table entry page.
 */
static void SetRouteEntry(webs_t wp, char_t *path, char_t *query)
{
    EntryTableAction (wp, &routetable, query);
}

/*============================================================================
 * Routines related to IP Chains Table
 *============================================================================
 */
/*****************************************************************************
 * GetChainList - Returns the list of scope to be used for that entry
 */
static void *GetChainList ()
{
    unsigned char *s=(chaintable.param)->value2;
    if (s==0) {
        return (void *)&scopelistc;
    } else if ((strcmp(s, "input") == 0) || (strcmp(s, "output") == 0)) {
        return (void *)&scopelisti;
    } else if (strcmp(s, "forward") == 0) {
        return (void *)&scopelistf;
    } else {
        return (void *)&scopelistc;
    }
}

/*****************************************************************************
 * ParseChainEntry - Parse a ip chain entry. The syntax of a ip chain entry
 * line is:
 *          :<chain> <default scope>
 */
static int ParseChainEntry (unsigned char *s)
{
    struct paramentry *param=chaintable.param;
    unsigned char *value;

    if (*s==':') {
        param->value2=++s;
        while (*s && !isspace(*s)) s++;
        if (*s) *s++=0;
        while (isspace(*s)) s++;
        (param+1)->value2 = s;
        while (*s && !isspace(*s)) s++;
        if (*s) *s++=0;
        return 1;
    }
    if ((chaintable.param)->value) {
        while (*s && !isspace(*s)) s++;
        while (isspace(*s)) s++;
        value = s;
        while (*s && !isspace(*s)) s++;
        if (*s) *s++=0;
        if (strcmp (value, (chaintable.param)->value) == 0) {
            return 2;
        }
    }
    return 0;
}

/*****************************************************************************
 * WriteChainEntry - Write a ip chain entry.
 */
static void WriteChainEntry (FILE *arq, unsigned char **value)
{
    fprintf (arq, ":%s %s\n", *value, *(value+1));
}

/*****************************************************************************
 * ListChainTable - Prints the ip chain table.
 */
static void ListChainTable(webs_t wp, char_t *path, char_t *query)
{
    struct stat st;
    if (stat(chaintable.filename, &st) || st.st_size == 0) {
        system("/sbin/fwset");
    }
    PrintTable (wp, &chaintable);
}

/*****************************************************************************
 * WriteChainTable - Answer to the submission of the ip chain table page.
 */
static void WriteChainTable(webs_t wp, char_t *path, char_t *query)
{
    ListTableAction(wp, &chaintable, query);
    system("/bin/fwset restore");
}

/*****************************************************************************
 * SetChainEntry - Answer to the submission of the ip chain table entry page.
 */
static void SetChainEntry(webs_t wp, char_t *path, char_t *query)
{
    EntryTableAction (wp, &chaintable, query);
    system("/bin/fwset restore");
}

/*============================================================================
 * Routines related to IP Chains Rules Table
 *============================================================================
 */
static int PrintIPRule(webs_t wp, unsigned char *str, unsigned char **value)
{
    struct in_addr a;
    unsigned char j;
    int i, k=0;
    unsigned int mask;
    mask = 0;
    inet_aton(*(value+1), &a);
    for (i=0; ; i++) {
        j=0x80>>(i%8);
        if (a.s_addr==mask) {
            k+=websWrite (wp, T("%s/%d"), str, i);
            if (*(value+2) && *(*(value+2))) {
                k+=websWrite (wp, T("[inv]"));
            }
            break;
        }
        if (i==32) break;
        *((unsigned char *)&mask + i/8) += j;
    }
    return k;
}

static int PrintSockRule(webs_t wp, unsigned char *str, unsigned char **value)
{
	int k=0;
    if (str && *str) {
        k+=websWrite (wp, str);
    }
    if (*(value+1) && *(*(value+1))) {
        k+=websWrite (wp, ":%s", *(value+1));
    }
    if (*(value+2) && *(*(value+2))) {
        k+=websWrite (wp, T("[inv]"));
    }
    return k;
}

static int PrintProtRule(webs_t wp, unsigned char *str, unsigned char **value)
{
	int k=0;
    if (str && *str) {
        k+=websWrite (wp, str);
    }
    if (*(value+1) && *(*(value+1))) {
        k+=websWrite (wp, T("[inv]"));
    }
    return k;
}

/*****************************************************************************
 * GetChainList2 - Returns the list of scopes to be used for that entry
 */
static void *GetChainList2 ()
{
    unsigned char *s=chaintable.nextvalue, *t;
    int i, n;
    if (!s || (strcmp(s, "input") && strcmp(s, "output"))) {
        n=5;
    } else {
        n=4;
    }
    for (i=0; i<n; i++) {
        scopes[i] = defscopes[i+1];
    }
    for (i=0; i<chaintable.nentries; i++) {
        t = *(chaintable.backup+i);
        if (strcmp(t, "input") && strcmp(t, "output") &&
            strcmp(t, "forward") && strcmp(t, s)) {
            scopes[n++] = t;
        }
    }
    scopelist.nlist = n;
    return (void *)&scopelist;
}

/*****************************************************************************
 * ParseRuleEntry - Parse a rule entry. The syntax of a ip chain entry
 * line is:
 *          -A <chain> -s [!] <IP Address>[/<mask>] [!] [<socket_init>[:<socket_end>]]
 *                     -d [!] <IP Address>[/<mask>] [!] [<socket_init>[:<socket_end>]]
 *                     [-p [!] <protocol>] [-i [!] <interface>] -j <target>
 */
static int ParseRuleEntry (unsigned char *s)
{
    struct paramentry *param=ruletable.param;
    unsigned char *value, *t;
    int idx, subidx;
    for (idx=0; idx<17; idx++) {
        (param+idx)->value2=(param+idx)->defvalue;
    }
    if (*s==':') {
        return 0;
    }
    if ((ruletable.prevtable)->nextvalue) {
        while (*s && !isspace(*s)) s++;
        while (isspace(*s)) s++;
        value = s;
        while (*s && !isspace(*s)) s++;
        if (*s) *s++=0;
        if (strcmp (value, (ruletable.prevtable)->nextvalue)) {
            return 0;
        }
        idx=0;
        while (*s) {
            while (isspace(*s)) s++;
            value = s;
            while (*s && !isspace(*s)) s++;
            if (*s) *s++=0;
            if (*value == '-') {
                subidx=0;
                switch (*(value+1)) {
                case 's':
                    idx = 1;
                    break;
                case 'd':
                    idx = 2;
                    break;
                case 'p':
                    idx = 3;
                    break;
                case 'i':
                    idx = 4;
                    break;
                case 'j':
                    idx = 5;
                    break;
                default:
                    return 0;
                }
                continue;
            }
            if (*value == '!') {
                switch (idx) {
                case 1:
                case 2:
                    if (subidx==0) {
                        (param+idx*6-3)->value2 = value;
                    } else {
                        (param+idx*6)->value2 = value;
                    }
                    break;
                case 3:
                case 4:
                    (param+idx*2+8)->value2 = value;
                    break;
                default:
                    return 0;
                }
                continue;
            }
            t=value;
            while (*t && *t!='/' && *t!=':') t++;
            switch (*t) {
            case '/':
                if ((idx!=1 && idx!=2) || subidx) return 0;
                *t++=0;
                (param+idx*6-5)->value2 = value;
                (param+idx*6-4)->value2 = t;
                subidx=1;
                break;
            case ':':
                if ((idx!=1 && idx!=2) || subidx!=1) return 0;
                *t++=0;
                (param+idx*6-2)->value2 = value;
                (param+idx*6-1)->value2 = t;
                break;
            case 0:
                if (idx==3) {
                    (param+13)->value2 = value;
                } else if (idx==4) {
                    (param+15)->value2 = value;
                } else if (idx==5) {
                    param->value2 = value;
                } else {
                    return 0;
                }
                break;
            }
        }
    }
    return 1;
}

/*****************************************************************************
 * WriteRuleEntry - Write a ip rule entry.
 */
static void WriteRuleEntry (FILE *arq, unsigned char **value)
{
    int i, j;
    unsigned char **value2;

    for (i=0; i<2; i++) {
        value2=value+i*6;
        if (i==0) {
            fprintf (arq, "-A %s -s", (ruletable.prevtable)->nextvalue);
        } else {
            fprintf (arq, " -d");
        }
        if (*(value2+3) && *(*(value2+3))) {
            fprintf (arq, " %s", *(value2+3));
        }
        for (j=0; j<2; j++) {
            if (j==0) {
                fprintf (arq, " ");
            } else {
                fprintf (arq, "/");
            }
            if (*(value2+1+j) && *(*(value2+1+j))) {
                fprintf(arq, "%s", *(value2+1+j));
            } else {
                fprintf(arq, "0.0.0.0");
            }
        }
        if (*(value2+4) && *(*(value2+4))) {
            if (*(value2+6) && *(*(value2+6))) {
                fprintf (arq, " %s", *(value2+6));
            }
            fprintf (arq, " %s", *(value2+4));
            if (*(value2+5) && *(*(value2+5))) {
                fprintf (arq, ":%s", *(value2+5));
            } else {
                fprintf (arq, ":%s", *(value2+4));
            }
        }
    }
    for (i=0; i<2; i++) {
        value2=value+13+i*2;
        if (*value2 && *(*value2)) {
            if (i==0) {
                fprintf (arq, " -p");
            } else {
                fprintf (arq, " -i");
            }
            if (*(value2+1) && *(*(value2+1))) {
                fprintf (arq, " %s", *(value2+1));
            }
            fprintf (arq, " %s", *value2);
        }
    }
    fprintf (arq, " -j %s\n", *value);
}

/*****************************************************************************
 * VerifyRuleEntry - Verify if the rule entry is OK.
 */
static char_t *VerifyRuleEntry (unsigned char **value)
{
    int parm_size = sizeof(struct paramentry);
    unsigned char *srcportstr   = *(unsigned char **)((int)value+parm_size* 4);
    unsigned char *dstportstr   = *(unsigned char **)((int)value+parm_size*10);
    unsigned char *protostr     = *(unsigned char **)((int)value+parm_size*13);
    unsigned char *protoinvstr  = *(unsigned char **)((int)value+parm_size*14);
    int protocol = 0;
    /* The source/destination sockets cannot be set when the protocol is not
    tcp, udp or icmp */
    if (protostr && *protostr) {
        protocol = atoi (protostr);
    }
    if ((protocol != 1 && protocol != 6 && protocol != 17) ||
        (protoinvstr && *protoinvstr)) {
        if ((srcportstr && *srcportstr) || (dstportstr && *dstportstr)) {
            return msgporterr;
	}
    }
    return NULL;
}

/*****************************************************************************
 * WriteRuleTable - Answer to the submission of the rule table page.
 */
static void WriteRuleTable(webs_t wp, char_t *path, char_t *query)
{
    ListTableAction(wp, &ruletable, query);
    system("/bin/fwset restore");
}

/*****************************************************************************
 * SetRuleEntry - Answer to the submission of the rule table entry page.
 */
static void SetRuleEntry(webs_t wp, char_t *path, char_t *query)
{
    EntryTableAction (wp, &ruletable, query);
    system("/bin/fwset restore");
}

/*============================================================================
 * Routines related to Serial Ports Users Group Table
 *============================================================================
 */
static int ParseSPUGroup(unsigned char *s, unsigned char **com, 
				unsigned char **a1, unsigned char **a2)
{
	unsigned char *c,*aux;
	
	if ((*s==0) || (*s=='#') || (*s=='\n')) return 0;
	
	c = s;
    while (*s && !isspace(*s)) s++;
	if (*s) *s++ = 0;
	
	if ((aux = strchr(c,'.')) == NULL) return 0;
	*aux++=0;
	
	if (strcmp(c,"conf")) {
		return 0;
	}
	*com = aux;
	
	*a1 = s;
	
	if (!strcmp(aux,"group")) {
    	while (*s && *s != ':' && !isspace(*s)) s++;
		*s++ = 0;
    	while (*s && isspace(*s)) s++;
    	*a2=s;
	    while (*s && (*s != '\n')) s++;
    	if (*s) *s++=0;
	} else {
    	while (*s && !isspace(*s)) s++;
		*s++ = 0;
		if (strcmp(*com,"host_config")) {
			if (strcmp(*com,"include")) {
				if (strcmp(*com,"hostname")) {
					return 0;
				}
			}
		}
	}
	return 1;
}

static int ParseSPUGroupEntry (unsigned	char *s)
{
    struct paramentry *param=SPUGrouptable.param;
	unsigned char *c, *a1, *a2;

	if (ParseSPUGroup(s,&c,&a1,&a2) != 1) {
		return 0;
	}	
	
	if (conf_incl) {
		if (!strcmp(c,"host_config")) {
			if (conf_find_host) {
				return 4;
			} 
			if (!strcmp(a1,hostname)) {
				conf_find_host = 1;
				return 0;
			}
		}	
		if (!conf_find_host) return 0;
	}
	param->value2 = a1;
    (param+1)->value2=a2;
    return 1;
}

/*****************************************************************************
 * WriteSPUGroupEntry - Writes a host entry.
 */
static void WriteSPUGroupEntry (FILE *arq, unsigned char **value)
{
	if (!conf_incl) {
    	fprintf (arq, "conf.group %s: %s\n", *value, *(value+1));
	}
}

/*****************************************************************************
 * ListSPUGroupTable - Prints the host table in the web page
 */
static void ListSPUGroupTable(webs_t wp, char_t *path, char_t *query)
{
    FILE *arq;
    struct stat st;
    int size;
	unsigned char *buf, *s, *t, *c, *a1, *a2;
	unsigned char psname[40]="/etc/portslave/pslave.conf";

	/* verify conf.include */
	conf_incl = 0;
	hostname[0] = 0;
    if (!stat(psname, &st)) {
    	size=st.st_size+1;
    	if (SPUGrouptable.config & CONFIG_FILE) size*=2;
    	if (arq = fopen(psname, "r")){
			buf = (unsigned char *) malloc(size);
			if (buf){
    			fread(buf, 1, st.st_size, arq);
    			*(buf+st.st_size)=0;
			}
   			fclose(arq);
			s = t = buf;
			while (*s) {
				while (*s && *s!='\n') s++;
				if (*s) *s++=0;
				if (ParseSPUGroup(t, &c, &a1, &a2) == 1) {
					if (!strcmp(c,"include")) {
						strcpy(SPUG_file,a1);
						conf_incl = 1;
					} else {
						if (!strcmp(c, "hostname")) {
							strcpy(hostname,a1);
						}
					}	
				}
				t = s;
			}
			if (buf) free(buf);
		}
	}
	if (conf_incl) {
		conf_find_host = 0;
		SPUGrouptable.config &= ~CONFIG_FILE;
		SPUGrouptable.config |= CONFIG_READ_ONLY;
		if (hostname[0] == 0) {
			if ((arq = fopen("/etc/hostname","r")) == NULL) return;
			fgets(hostname, 40L,arq);
			fclose(arq);
		}
	}else{
		SPUGrouptable.config |= CONFIG_FILE;
		strcpy(SPUG_file, psname);
	}
	SPUGrouptable.filename = SPUG_file;
    PrintTable (wp, &SPUGrouptable);
}

/*****************************************************************************
 * WriteSPUGroupTable - Answer to the submission of the host table page
 */
static void WriteSPUGroupTable(webs_t wp, char_t *path, char_t *query)
{
    ListTableAction(wp, &SPUGrouptable, query);
}

/*****************************************************************************
 * SetSPUGroupEntry - Answer to the submission of the host table entry page
 */
static void SetSPUGroupEntry(webs_t wp, char_t *path, char_t *query)
{
    EntryTableAction (wp, &SPUGrouptable, query);
}

/*============================================================================
 * DefineWebAspform1 - This is the only routine which is not static. It's
 * called in main.c.
 *============================================================================
 */
void DefineWebAspForm1() {
    int i;
    for (i=0; i<64; i++) ipchains[i] = NULL;
    websFormDefine(T("ListChainTable"), ListChainTable);
    websFormDefine(T("WriteChainTable"), WriteChainTable);
    websFormDefine(T("cfg/SetChainEntry"), SetChainEntry);
    websFormDefine(T("cfg/WriteRuleTable"), WriteRuleTable);
    websFormDefine(T("cfg/SetRuleEntry"), SetRuleEntry);
    websFormDefine(T("ListHostTable"), ListHostTable);
    websFormDefine(T("cfg/WriteHostTable"), WriteHostTable);
    websFormDefine(T("cfg/SetHostEntry"), SetHostEntry);
    websFormDefine(T("ListRouteTable"), ListRouteTable);
    websFormDefine(T("cfg/WriteRouteTable"), WriteRouteTable);
    websFormDefine(T("cfg/SetRouteEntry"), SetRouteEntry);
    websFormDefine(T("ListSPUGroupTable"), ListSPUGroupTable);
    websFormDefine(T("cfg/WriteSPUGroupTable"), WriteSPUGroupTable);
    websFormDefine(T("cfg/SetSPUGroupEntry"), SetSPUGroupEntry);
}

