/*
 * security.c -- Security handler
 *
 * Copyright (c) GoAhead Software Inc., 1995-2000. All Rights Reserved.
 *
 * See the file "license.txt" for usage and redistribution license requirements
 *
 * $Id: security.c,v 1.7 2006/08/18 23:36:23 edson Exp $
 */

/******************************** Description *********************************/

/*
 *	This module provides a basic security policy.
 */

/********************************* Includes ***********************************/

#include	"wsIntrn.h"
#include	"um.h"
#ifdef DIGEST_ACCESS_SUPPORT
#include	"websda.h"
#endif

/********************************** Defines ***********************************/
/*
 *	The following #defines change the behaviour of security in the absence 
 *	of User Management.
 *	Note that use of User management functions require prior calling of
 *	umInit() to behave correctly
 */

#ifndef USER_MANAGEMENT_SUPPORT
#define umGetAccessMethodForURL(url) AM_FULL
#define umUserExists(userid) 0
#define umUserCanAccessURL(userid, url) 1
#define umGetUserPassword(userid) websGetPassword()
#define umGetAccessLimitSecure(accessLimit) 0
#define umGetAccessLimit(url) NULL
#endif

/******************************** Local Data **********************************/

static char_t	websPassword[WEBS_MAX_PASS];	/* Access password (decoded) */
#ifdef _DEBUG
static int		debugSecurity = 1;
#else
static int		debugSecurity = 0;
#endif

//Helio Fujimoto - 08/13/2001
extern void websLoginPage (webs_t, char_t *, char_t *, char_t *, unsigned char);
static void websGetUserPassFromCookie(webs_t);

/*********************************** Code *************************************/
/*
 *	Determine if this request should be honored
 */

int websSecurityHandler(webs_t wp, char_t *urlPrefix, char_t *webDir, int arg, 
						char_t *url, char_t *path, char_t *query)
{
	char_t			*type, *userid, *password, *accessLimit;
	int				flags, nRet;
	accessMeth_t	am;

	a_assert(websValid(wp));
	a_assert(url && *url);
	a_assert(path && *path);
/*
 *	Get the critical request details
 */
	websGetUserPassFromCookie(wp); //Helio Fujimoto - 08/10/2001 - Try to get the username/password from the cookie
	
	type = websGetRequestType(wp);
	password = websGetRequestPassword(wp);
	userid = websGetRequestUserName(wp);
	flags = websGetRequestFlags(wp);
	
/*
 *	Get the access limit for the URL.  Exit if none found.
 */
	accessLimit = umGetAccessLimit(path);
	if (accessLimit == NULL) {
		return 0;
	}
/*
 *	Check to see if URL must be encrypted
 */
#ifdef WEBS_SSL_SUPPORT
	nRet = umGetAccessLimitSecure(accessLimit);
	if (nRet && ((flags & WEBS_SECURE) == 0)) {
		websStats.access++;
// Jul/25/2001 Cyclades TSxK Livio: Call redirect to start https conection.
#ifndef CYCLADES_TSxK
		websError(wp, 405, T("Access Denied\nSecure access is required."));
		trace(3, T("SEC: Non-secure access attempted on <%s>\n"), path);
      /* bugfix 5/24/02 -- we were leaking the memory pointed to by
       * 'accessLimit'. Thanks to Simon Byholm.
       */
      bfree(B_L, accessLimit);
		return 1;
#else
		websRedirect(wp, url, NULL);
		trace(3, T("SEC: Non-secure access attempted on <%s>. Redirected.\n"), path);
		/* bugfix 5/24/02 -- we were leaking the memory pointed to by
		* 'accessLimit'. Thanks to Simon Byholm.
		*/
		bfree(B_L, accessLimit);
		return 1;
#endif
	}
#endif

/*
 *	Get the access limit for the URL
 */
	am = umGetAccessMethodForURL(accessLimit);

	nRet = 0;
	if ((flags & WEBS_LOCAL_REQUEST) && (debugSecurity == 0)) {
/*
 *		Local access is always allowed (defeat when debugging)
 */
	} else if (am == AM_NONE) {
/*
 *		URL is supposed to be hidden!  Make like it wasn't found.
 */
		websStats.access++;
		websError(wp, 404, T("Page Not Found"));
		nRet = 1;
#ifdef CYCLADES_TSxK
	} else if (am == AM_FULL) {
		/* If access is full, don't require a right username/password */
#endif
	} else 	if (userid && *userid) {
		if (!umUserExists(userid)) {
			websStats.access++;
// Jul/25/2001 Cyclades TSxK Livio: Send code to try login again.
#ifndef CYCLADES_TSxK
			websError(wp, 401, T("Access Denied\nUnknown User"));
#else
			if (am == AM_COOKIE) {
				websLoginPage(wp, url, path, "Access Denied: Unknown User", 1);
			} else {
				websError(wp, 401, T("Access Denied: Unknown User"));
			}
#endif
			trace(3, T("SEC: Unknown user <%s> attempted to access <%s>\n"), 
				userid, path);
			nRet = 1;
		} else if (!umUserCanAccessURL(userid, accessLimit)) {
			websStats.access++;
// Jul/25/2001 Cyclades TSxK Livio: Send code to try login again.
#ifndef CYCLADES_TSxK
			websError(wp, 403, T("Access Denied\nProhibited User"));
#else
			if (am == AM_COOKIE) {
				websLoginPage(wp, url, path, "Access Denied\nProhibited User", 1);
			} else {
				websError(wp, 401, T("Access Denied\nProhibited User"));
			}
#endif			
			nRet = 1;
		} else if (password && * password) {
			char_t * userpass = umGetUserPassword(userid);
			if (userpass) {
				if (gstrcmp(password, userpass) != 0) {
					websStats.access++;
// Jul/25/2001 Cyclades TSxK Livio: Send code to try login again.
#ifndef CYCLADES_TSxK
					websError(wp, 401, T("Access Denied\nWrong Password"));
#else
					if (am == AM_COOKIE) {
						websLoginPage(wp, url, path, "Access Denied: Wrong Password", 1);
					} else {
						websError(wp, 401, T("Access Denied\nWrong Password"));
					}
#endif					
					trace(3, T("SEC: Password fail for user <%s>")
								T("attempt to access <%s>\n"), userid, path);
					nRet = 1;
				} else {
/*
 *					User and password check out.
 */
					setenv("LOGNAME", userid, 1);
				}

				bfree (B_L, userpass);
			}
#ifdef DIGEST_ACCESS_SUPPORT
		} else if (flags & WEBS_AUTH_DIGEST) {

			char_t *digestCalc;

/*
 *			Check digest for equivalence
 */
			wp->password = umGetUserPassword(userid);

			a_assert(wp->digest);
			a_assert(wp->nonce);
			a_assert(wp->password);
							 
			digestCalc = websCalcDigest(wp);
			a_assert(digestCalc);

			if (gstrcmp(wp->digest, digestCalc) != 0) {
				websStats.access++;
// Jul/25/2001 Cyclades TSxK Livio: Send code to try login again.
#ifdef CYCLADES_TSxK
				if (am == AM_COOKIE)
					websLoginPage(wp, url, path,
						"Access Denied: Wrong Password", 1);
				else
#endif
            /* 16 Jun 03 -- error code changed from 405 to 401 -- thanks to
             * Jay Chalfant.
             */
				websError(wp, 401, T("Access Denied\nWrong Password"));
				nRet = 1;
			}

			bfree (B_L, digestCalc);
#endif
		} else {
/*
 *			No password has been specified
 */
#ifdef DIGEST_ACCESS_SUPPORT
			if (am == AM_DIGEST) {
				wp->flags |= WEBS_AUTH_DIGEST;
			}
#endif
			websStats.errors++;
#ifndef CYCLADES_TSxK
			websError(wp, 401, 
				T("Access to this document requires a password"));
#else
			if (am == AM_COOKIE) {
				websLoginPage(wp, url, path, "Access to this document requires a password", 1);
			} else {
				websError(wp, 401,
					T("Access to this document requires a password"));
			}
#endif
			nRet = 1;
		}
#ifndef CYCLADES_TSxK
	} else if (am != AM_FULL) {
#else
	} else {	// Helio Fujimoto - 08/15/01 - The consistency for AM_FULL is in the top of the routine
#endif
/*
 *		This will cause the browser to display a password / username
 *		dialog
 */
#ifdef DIGEST_ACCESS_SUPPORT
		if (am == AM_DIGEST) {
			wp->flags |= WEBS_AUTH_DIGEST;
		}
#endif
		websStats.errors++;
#ifndef CYCLADES_TSxK
		websError(wp, 401, T("Access to this document requires a User ID"));
#else
		if (am == AM_COOKIE) {
			websLoginPage(wp, url, path, "You must enter username/password to access this document.", 0);
		} else {
		websError(wp, 401, T("Access to this document requires a User ID"));
		}
#endif
		nRet = 1;
	}

	bfree(B_L, accessLimit);

	return nRet;
}

/******************************************************************************/
/*
 *	Delete the default security handler
 */

void websSecurityDelete()
{
	websUrlHandlerDelete(websSecurityHandler);
}

/******************************************************************************/
/*
 *	Store the new password, expect a decoded password. Store in websPassword in 
 *	the decoded form.
 */

void websSetPassword(char_t *password)
{
	a_assert(password);

	gstrncpy(websPassword, password, TSZ(websPassword));
}

/******************************************************************************/
/*
 *	Get password, return the decoded form
 */

char_t *websGetPassword()
{
	return bstrdup(B_L, websPassword);
}

/******************************************************************************/

/* For now, the format of the cookie SessionId will be <username>:<password>.
Later it must be encrypted for password protection */
static void websGetUserPassFromCookie(webs_t wp)
{
	unsigned char *s, *t;
	char_t aux[256];
	if ((wp->flags & WEBS_COOKIE) && (s = strstr(wp->cookie, "sessionId="))) {
		wp->flags &= ~(WEBS_AUTH_BASIC | WEBS_AUTH_DIGEST);
		if (wp->userName) bfree(B_L, wp->userName);
		if (wp->password) bfree(B_L, wp->password);
		s += strlen("sessionId=");
		if ((t = strchr(s, ';')) != NULL) *t=0;
		websDecode64(aux, s, 254);
		s = aux;
		if ((t = strchr(s, ':')) != NULL) {
			*t++=0;
			wp->userName = bstrdup(B_L, s);
			wp->password = bstrdup(B_L, t);
		} else {
			wp->userName = bstrdup(B_L, "");
			wp->password = bstrdup(B_L, "");
		}
	}
}

char_t *SetCookieFromUserPass(char_t *login, char_t *password)
{
	char_t sessionID[256], *cookie;
	cookie = balloc(B_L, 512);
	snprintf(cookie, 512, "%s:%s", login, password);
	websEncode64(sessionID, cookie, 254);
	sessionID[254] = 0;
	snprintf(cookie, 512, "sessionId=%s; path=/", sessionID);
	return cookie;
}


